//**************************************************************
//*   filename: output.cc                                      *
//*                                                            *
//**************************************************************
//* programmed by: Thomas Wagner                               *
//* last change:  (XXX: not updated)                           *
//**************************************************************
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>

#include "output.h"
#include "connect.h"
#include "regel.h"
#include "onofficon.h"
#include "init.h"
#include "quicksort.h"

#include "include_bmps/makerules.bmp"
#include "include_bmps/ghost.bmp"
#include "include_bmps/load.bmp"
#include "include_bmps/save.bmp"
#include "include_bmps/go.bmp"
#include "include_bmps/classify.bmp"
#include "include_bmps/reset.bmp"

#include "include_bmps/d2.bmp"
#include "include_bmps/d3.bmp"
#include "include_bmps/singlemembership.bmp"
#include "include_bmps/pointtyppoint.bmp"
#include "include_bmps/pointtypcross.bmp"
#include "include_bmps/roll.bmp"
#include "include_bmps/move.bmp"
#include "include_bmps/fitscale.bmp"
#include "include_bmps/movepoint.bmp"
#include "include_bmps/selectpoint.bmp"
#include "include_bmps/deletepoint.bmp"
#include "include_bmps/createpoint.bmp"
#include "include_bmps/drawonlyoriginal.bmp"
#include "include_bmps/drawonlyresult.bmp"
#include "include_bmps/drawboth.bmp"
#include "include_bmps/d2cluster.bmp"
#include "include_bmps/d3cluster.bmp"
#include "include_bmps/d2membertyp1.bmp"
#include "include_bmps/d3membertyp1.bmp"
#include "include_bmps/d2membertyp2.bmp"
#include "include_bmps/d3membertyp2.bmp"
#include "include_bmps/d2guete.bmp"
#include "include_bmps/clustertypellipsis.bmp"
#include "include_bmps/clustertypsymbol.bmp"
#include "include_bmps/clustertypstar.bmp"
#include "include_bmps/clustertypcolor.bmp"
#include "include_bmps/createcluster.bmp"
#include "include_bmps/movecluster.bmp"
#include "include_bmps/deletecluster.bmp"
#include "include_bmps/selectcluster.bmp"

//including from Olli:
#include "verfahren.h"
#include "matrix.h"
#include "regel.h"

extern unsigned long buttonpix, blackpix, whitepix, shadowpix;
extern unsigned long lightpix, redpix;
extern Cursor moveconnectcursor, moveaxescursor, delconnectcursor, maincursor;
extern Cursor waitcursor, movepointcursor, createpointcursor, rotatecursor;



/*
 * dump the given pixmap to a data-file (xbm-format)
 */

static void
dump_bitmap (Display *display, Pixmap pixmap, int width, int height)
{
  char *fname = "dump.xbm";

  fprintf (stderr, "dumping bitmap to ``%s'' ... ", fname);
  fflush (stderr);

  if (BitmapSuccess != XWriteBitmapFile(display, fname, pixmap,
					width, height, -1, -1)) {
    fprintf (stderr, "error - not done.\n");
  } else {
    fprintf (stderr, "done.\n");
  }
}



void
Output::DumpBitmap ()
{
  dump_bitmap (display, pixmap, width, height);
}



int 
Getvektorindex (DVektor * vector, DVektorArray * array)
{
  short counter, foundindex = 0;
  for (counter = 0; counter < array->Lese_Groesse (); counter++)
    if (&(array->Lese_Vektor (counter)) == vector) {
      foundindex = counter;
      break;			// for(...)

    }
  return foundindex;
}

RuleResultInfo::RuleResultInfo (Datensatz * Dataset)
{
  dataset = Dataset;
  result = new DVektorArray (Dataset->Lese_Daten ().Lese_Dim (),
			     Dataset->Lese_Daten ().Lese_Groesse (),
			     NULL);
}

RuleResultInfo::~RuleResultInfo ()
{
  delete result;
}

void 
RuleResultInfo::operator= (RuleResultInfo & source)
{
  result = source.result;
  dataset = source.dataset;
}

Output::Output (Display * initdisplay,
	GC initgc,
	MoveableIcon * initOutputicon,
	XFontStruct * fontstruct,
	int initx, int inity,
	int initwidth, int initheight,
	char *text):
BigWindow (initdisplay,
	   initgc,
	   DefaultRootWindow (initdisplay),
	   fontstruct,
	   initx, inity,
	   initwidth, initheight,
	   text,
	   DEFAULT_BDWIDTH)
{
  unsigned int depth;
  Dataset = NULL;
  myicon = initOutputicon;
  depth = DefaultDepth (display, DefaultScreen (display));
  pixmap = XCreatePixmap (display, window, width, height, depth);
  XSetForeground (display, gc, whitepix);
  XFillRectangle (display, pixmap, gc, 0, 0, width, height);

  ghost = new ActionIcon (display,
			  gc,
			  this,
			  fontstruct,
			  0, 0, 16, 16,
			  "", ghost_bits,
			  ghost_width, ghost_height,
			  OUTPUT_GHOST,
			  ICON_SELECTABLE);
  ghost->MapMe ();

  SetMinSize (width, height);
}

Output::~Output ()
{
  XFreePixmap (display, pixmap);
  delete ghost;
}

void 
Output::ClearPixmap ()
{
  if (pixmap != 0) {
    XSetForeground (display, gc, whitepix);
    XFillRectangle (display, pixmap, gc, 0, 0, width, height);
  }
}

void 
Output::MapMe ()
{
  XMapWindow (display, window);
  RaiseMe ();
}


void 
Output::DrawSymbol (short xpos,
		    short ypos,
		    Drawable drawable,
		    int symbolindex)
{
  if (symbolindex < MAXINT) {
    switch (symbolindex % (SYMBOLNUMBER - 1)) {
    case 0:
      {
//   |
	// --|--
	//   |
	XSegment lines[2];
	lines[0].x1 = lines[0].x2 = xpos;
	lines[0].y1 = ypos - SYMBOLHALFWIDTH;
	lines[0].y2 = ypos + SYMBOLHALFWIDTH;

	lines[1].x1 = xpos - SYMBOLHALFWIDTH;
	lines[1].x2 = xpos + SYMBOLHALFWIDTH;
	lines[1].y1 = lines[1].y2 = ypos;
	XDrawSegments (display, drawable, gc, lines, 2);
      }
      break;

    case 1:
      {
//   /\
	//  /  \
	//  ----
	XPoint points[4];
	points[3].x = points[0].x = xpos;
	points[3].y = points[0].y = ypos - SYMBOLHALFWIDTH;
	points[2].y = points[1].y = ypos + SYMBOLHALFWIDTH;
	points[1].x = xpos + SYMBOLHALFWIDTH;
	points[2].x = xpos - SYMBOLHALFWIDTH;
	XDrawLines (display, drawable, gc, points, 4, CoordModeOrigin);
      }
      break;

    case 2:
      {
//   ___
	//  |   |
	//  |   |
	//   ---
	XDrawRectangle (display,
			drawable,
			gc,
			xpos - SYMBOLHALFWIDTH, ypos - SYMBOLHALFWIDTH,
			2 * SYMBOLHALFWIDTH, 2 * SYMBOLHALFWIDTH);
      }
      break;

    case 3:
      {
//    ___
	//   /   \
	//  |     |
	//   \___/
	XDrawArc (display,
		  drawable,
		  gc,
		  xpos - SYMBOLHALFWIDTH, ypos - SYMBOLHALFWIDTH,
		  SYMBOLHALFWIDTH * 2, SYMBOLHALFWIDTH * 2,
		  0, 23040);
      }
      break;

    case 4:
      {
//   /\
	//  /xx\
	//  ----
	XPoint points[3];
	points[0].x = xpos;
	points[0].y = ypos - SYMBOLHALFWIDTH;
	points[2].y = points[1].y = ypos + SYMBOLHALFWIDTH;
	points[1].x = xpos + SYMBOLHALFWIDTH;
	points[2].x = xpos - SYMBOLHALFWIDTH;
	XFillPolygon (display, drawable, gc, points, 3, Convex, CoordModeOrigin);
      }
      break;

    case 5:
      {
//   ___
	//  |xxx|
	//  |xxx|
	//   ---
	XFillRectangle (display,
			drawable,
			gc,
			xpos - SYMBOLHALFWIDTH, ypos - SYMBOLHALFWIDTH,
			2 * SYMBOLHALFWIDTH, 2 * SYMBOLHALFWIDTH);
      }
      break;

    case 6:
      {
//    ___
	//   /xxx\
	//  |xxxxx|
	//   \___/
	XFillArc (display,
		  drawable,
		  gc,
		  xpos - SYMBOLHALFWIDTH, ypos - SYMBOLHALFWIDTH,
		  SYMBOLHALFWIDTH * 2, SYMBOLHALFWIDTH * 2,
		  0, 23040);
      }
      break;

    case 7:
      {
//  ----
	//  \  /
	//   \/
	XPoint points[4];
	points[3].x = points[0].x = xpos;
	points[3].y = points[0].y = ypos + SYMBOLHALFWIDTH;
	points[2].y = points[1].y = ypos - SYMBOLHALFWIDTH;
	points[1].x = xpos + SYMBOLHALFWIDTH;
	points[2].x = xpos - SYMBOLHALFWIDTH;
	XDrawLines (display, drawable, gc, points, 4, CoordModeOrigin);
      }
      break;

    case 8:
      {
//   /\
	//  /  \
	//  \  /
	//   \/
	XPoint points[5];
	points[4].x = points[0].x = points[2].x = xpos;
	points[1].y = points[3].y = ypos;
	points[4].y = points[0].y = ypos - SYMBOLHALFWIDTH;
	points[2].y = ypos + SYMBOLHALFWIDTH;
	points[1].x = xpos + SYMBOLHALFWIDTH;
	points[3].x = xpos - SYMBOLHALFWIDTH;
	XDrawLines (display, drawable, gc, points, 5, CoordModeOrigin);
      }
      break;

    case 9:
      {
//  ----
	//  \**/
	//   \/
	XPoint points[3];
	points[0].x = xpos;
	points[0].y = ypos + SYMBOLHALFWIDTH;
	points[2].y = points[1].y = ypos - SYMBOLHALFWIDTH;
	points[1].x = xpos + SYMBOLHALFWIDTH;
	points[2].x = xpos - SYMBOLHALFWIDTH;
	XFillPolygon (display, drawable, gc, points, 3, Convex, CoordModeOrigin);
      }
      break;

    case 10:
      {
//   /\
	//  /**\
	//  \**/
	//   \/
	XPoint points[4];
	points[0].x = points[2].x = xpos;
	points[1].y = points[3].y = ypos;
	points[0].y = ypos - SYMBOLHALFWIDTH;
	points[2].y = ypos + SYMBOLHALFWIDTH;
	points[1].x = xpos + SYMBOLHALFWIDTH;
	points[3].x = xpos - SYMBOLHALFWIDTH;
	XFillPolygon (display, drawable, gc, points, 4, Convex, CoordModeOrigin);
      }
      break;
    }
  } else {
//  \  /
    //   \/
    //   /\
    //  /  \

#if 1
#endif
    XSegment lines[2];
    lines[0].x1 = lines[1].x2 = xpos - SYMBOLHALFWIDTH;
    lines[0].x2 = lines[1].x1 = xpos + SYMBOLHALFWIDTH;
    lines[0].y1 = lines[1].y1 = ypos - SYMBOLHALFWIDTH;
    lines[0].y2 = lines[1].y2 = ypos + SYMBOLHALFWIDTH;
    XDrawSegments (display, drawable, gc, lines, 2);
  }
}


void 
Output::DrawColorbox (short xpos,
		      short ypos,
		      Drawable drawable,
		      int symbolindex)
{
  if (symbolindex < MAXINT)
    XSetForeground (display, gc, indexcolor[symbolindex % (SYMBOLNUMBER - 1)]);
  else
    XSetForeground (display, gc, buttonpix);

  XFillRectangle (display,
		  drawable,
		  gc,
		  xpos - SYMBOLHALFWIDTH, ypos - SYMBOLHALFWIDTH,
		  2 * SYMBOLHALFWIDTH, 2 * SYMBOLHALFWIDTH);
}

void 
Output::UnDrawBox (short xpos,
		   short ypos,
		   Drawable drawable)
{
  XSetForeground (display, gc, whitepix);
  XFillRectangle (display,
		  drawable,
		  gc,
		  xpos - SYMBOLHALFWIDTH, ypos - SYMBOLHALFWIDTH,
		  2 * SYMBOLHALFWIDTH, 2 * SYMBOLHALFWIDTH);
}

void 
Output::DrawBox (short xpos,
		 short ypos,
		 Drawable drawable)
{
  XSetForeground (display, gc, blackpix);

  XDrawRectangle (display,
		  drawable,
		  gc,
		  xpos - SYMBOLHALFWIDTH, ypos - SYMBOLHALFWIDTH,
		  2 * SYMBOLHALFWIDTH, 2 * SYMBOLHALFWIDTH);
}


void 
Output::UpdateOutput ()
{
  XCopyArea (display, pixmap, window, gc, 0, 0, width, height, 0, 0);
}

void 
Output::Resized (int newwidth, int newheight)
{
  int depth;
  width = newwidth;
  height = newheight;
  XFreePixmap (display, pixmap);
  depth = DefaultDepth (display, DefaultScreen (display));
  pixmap = XCreatePixmap (display, window, width, height, depth);
  XResizeWindow (display, window, width, height);
}


Output2d_3d::Output2d_3d (Display * initdisplay,
	     GC initgc,
	     MoveableIcon * initOutputicon,
	     XFontStruct * fontstruct,
	     int initx, int inity, int initwidth, int initheight,
	     char *text):
Output (initdisplay,
	initgc,
	initOutputicon,
	fontstruct,
	initx, inity, initwidth, initheight,
	text)
{
  xindex = DEFAULTXINDEX;
  yindex = DEFAULTYINDEX;
  zindex = DEFAULTZINDEX;

  float_to_pixel = 1.0;
}

void 
Output2d_3d::Calculate_scale_2d (DVektorArray * inputData)
{
  if (inputData->Lese_Dim () > 1)
    Calculate_scale_2d (MAX (inputData->Lese_Max (xindex), -inputData->Lese_Min (xindex)),
	 MAX (inputData->Lese_Max (yindex), -inputData->Lese_Min (yindex)));
  else
    Calculate_scale_2d (MAX (inputData->Lese_Max (xindex), -inputData->Lese_Min (xindex)),
	 MAX (inputData->Lese_Max (xindex), -inputData->Lese_Min (xindex)));
}

void 
Output2d_3d::Calculate_scale_2d (double xsize)
{
  if (xsize == 0.0)
    xscale = 1.0;
  else
    xscale = 1.0 / xsize;

  yscale = 1.0;
}

void 
Output2d_3d::Calculate_scale_2d (double xsize, double ysize)
{
  if (fitscaleicon->IsSelected ()) {
    if (xsize == 0.0)
      xscale = 1.0;
    else
      xscale = 1.0 / xsize;
    if (ysize == 0.0)
      yscale = 1.0;
    else
      yscale = 1.0 / ysize;
  } else {
    xsize = MAX (xsize, ysize);
    if (xsize == 0.0)
      xsize = 1.0;
    xscale = 1.0 / xsize;
    yscale = 1.0 / xsize;
  }
}

void 
Output2d_3d::Calculate_scale_3d (double xsize, double ysize)
{
  if (fitscaleicon->IsSelected ()) {
    if (xsize == 0.0)
      xscale = 1.0;
    else
      xscale = 1.0 / xsize;
    if (ysize == 0.0)
      yscale = 1.0;
    else
      yscale = 1.0 / ysize;
  } else {
    xsize = MAX (xsize, ysize);
    if (xsize == 0.0)
      xsize = 1.0;
    xscale = 1.0 / xsize;
    yscale = 1.0 / xsize;
  }
  zscale = 1.0;
}

void 
Output2d_3d::Calculate_scale_3d (double xsize, double ysize, double zsize)
{
  if (fitscaleicon->IsSelected ()) {
    if (xsize == 0.0)
      xscale = 1.0;
    else
      xscale = 1.0 / xsize;
    if (ysize == 0.0)
      yscale = 1.0;
    else
      yscale = 1.0 / ysize;
    if (zsize == 0.0)
      zscale = 1.0;
    else
      zscale = 1.0 / zsize;
  } else {
    xsize = MAX (xsize, ysize);
    xsize = MAX (xsize, zsize);
    if (xsize == 0.0)
      xscale = 1.0;
    xscale = 1.0 / xsize;
    yscale = 1.0 / xsize;
    zscale = 1.0 / xsize;
  }
}


void 
Output2d_3d::Calculate2dparameter (DVektorArray * inputData)
{
  if (inputData->Lese_Dim () > 1)
    Calculate2dparameter (inputData->Lese_Min (xindex),
			  inputData->Lese_Min (yindex),
			  inputData->Lese_Max (xindex),
			  inputData->Lese_Max (yindex));
  else
    Calculate2dparameter (inputData->Lese_Min (xindex),
			  inputData->Lese_Min (xindex),
			  inputData->Lese_Max (xindex),
			  inputData->Lese_Max (xindex));
}

void 
Output2d_3d::Calculate2dparameter (double xmin, double ymin, double xmax, double ymax)
{
  double xdifference, ydifference;
  double float_to_pixelx, float_to_pixely;

  xmin = MIN (xmin, 0.0);
  ymin = MIN (ymin, 0.0);

// *** NEW ****
  xdifference = (xmax - xmin) * xscale;
  ydifference = (ymax - ymin) * yscale;

  if (xdifference == 0.0)
    xdifference = 1.0;
  if (ydifference == 0.0)
    ydifference = 1.0;

  float_to_pixelx = (width - 2 * OUTPUTBORDERWIDTH) / xdifference;
  float_to_pixely = (height - 2 * OUTPUTBORDERWIDTH) / ydifference;
  float_to_pixel = MIN (float_to_pixelx, float_to_pixely);

  centerx = OUTPUTBORDERWIDTH - (short) (xmin * xscale * float_to_pixel);
  centery = height - OUTPUTBORDERWIDTH + (short) (ymin * yscale * float_to_pixel);
}

void 
Output2d_3d::Calculate3dparameter (char options)
{
  if (options & RESET_AXES) {
/*  axisframe[0][0]=axisframe[1][1]=axisframe[2][2] = 1.0;
   axisframe[0][1]=axisframe[0][2] = 0.0;
   axisframe[1][0]=axisframe[1][2] = 0.0;
   axisframe[2][0]=axisframe[2][1] = 0.0;
 */
    axisframe[0][1] = axisframe[1][2] = axisframe[2][0] = 1.0;
    axisframe[0][0] = axisframe[0][2] = 0.0;
    axisframe[1][0] = axisframe[1][1] = 0.0;
    axisframe[2][1] = axisframe[2][2] = 0.0;
  }
  if ((options & RESET_AXES) ||
      (options & CENTER_AXES)) {
    axisframe[0][3] = -0.5;
    axisframe[1][3] = -0.5;
    axisframe[2][3] = -1.0;
  }
  scalefactor = (height - 2 * OUTPUTBORDERWIDTH);
  eyedistance = 2.0;
  xcenter = width / 2;
  ycenter = height / 2;
}

void 
Output2d_3d::Resized (int newwidth, int newheight)
{
  ((Output *) this)->Resized (newwidth, newheight);
  if (Dataset != NULL) {
    Calculate2dparameter (&(Dataset->Lese_Daten ()));
    if (Dataset->Lese_Daten ().Lese_Dim () > 2)
      Calculate3dparameter (0);
  }
}

void 
Output2d_3d::DrawCoordinates (short xpos, short ypos)
{
  double xoutput, youtput;
  int xnumber;
  char coord[30];
  char box[30] = "                    ";
  short textheight = fontstruct->max_bounds.ascent + fontstruct->max_bounds.descent;

  if (Dataset->Lese_Daten ().Lese_Dim () > 1) {
    PixeltoWorld2d (xpos, ypos, &xoutput, &youtput);
    sprintf (coord, "%f %f", xoutput, youtput);
  } else {
    PixeltoWorld1d (xpos, ypos, &xnumber, &youtput);
    sprintf (coord, "%d %f", xnumber, youtput);
  }

  XSetForeground (display, gc, whitepix);
  XFillRectangle (display,
		  window,
		  gc,
		  1, height - textheight,
		  XTextWidth (fontstruct, box, strlen (box)),
		  textheight);

  XSetForeground (display, gc, blackpix);
  XDrawString (display, window, gc,
	       1,
	       height - 1,
	       coord, strlen (coord));
}

void 
Output2d_3d::PixeltoWorld1d (short xpos, short ypos, int *number, double *xis)
{
  *number = (int) (Dataset->Lese_Daten ().Lese_Groesse () * (xpos - centerx) / float_to_pixel);
  *xis = (centery - ypos) / (xscale * float_to_pixel);
}

void 
Output2d_3d::WorldtoPixel1d (short number, double xpos, short *xis, short *yis)
{
  *xis = centerx + (short) (number 
			    / (double) Dataset->Lese_Daten ().Lese_Groesse ()
			    * float_to_pixel);
  *yis = centery - (short) (xpos * float_to_pixel * xscale);
}

void 
Output2d_3d::PixeltoWorld2d (short xpos, short ypos, double *xis, double *yis)
{
  *xis = (xpos - centerx) / (xscale * float_to_pixel);
  *yis = (centery - ypos) / (yscale * float_to_pixel);
}

void 
Output2d_3d::WorldtoPixel2d (double xpos, double ypos, short *xis, short *yis)
{
  xpos *= xscale;
  ypos *= yscale;

  *xis = centerx + (short) (xpos * float_to_pixel);
  *yis = centery - (short) (ypos * float_to_pixel);
}

void 
Output2d_3d::WorldtoPixel3d (double xpos,
			     double ypos,
			     double zpos,
			     short *xis, short *yis)
{
  double viewx, viewy, viewz, x_2d, y_2d;

  xpos *= xscale;
  ypos *= yscale;
  zpos *= zscale;

  viewx = xpos * axisframe[0][0] +
    ypos * axisframe[0][1] +
    zpos * axisframe[0][2] +
    axisframe[0][3];
  viewy = xpos * axisframe[1][0] +
    ypos * axisframe[1][1] +
    zpos * axisframe[1][2] +
    axisframe[1][3];
  viewz = xpos * axisframe[2][0] +
    ypos * axisframe[2][1] +
    zpos * axisframe[2][2] +
    axisframe[2][3];

  y_2d = (viewy * eyedistance) / (eyedistance - viewz);
  x_2d = (viewx * eyedistance) / (eyedistance - viewz);

  *xis = xcenter + (short) (x_2d * scalefactor);
  *yis = ycenter - (short) (y_2d * scalefactor);
}

void 
Output2d_3d::WorldtoPixel3d (double xpos,
			     double ypos,
			     double zpos,
			     double zworld,
			     short *xis, short *yis)
{
  xpos *= xscale;
  ypos *= yscale;
  zpos *= zscale;
  double viewx, viewy, x_2d, y_2d;

  viewx = xpos * axisframe[0][0] +
    ypos * axisframe[0][1] +
    zpos * axisframe[0][2] +
    axisframe[0][3];
  viewy = xpos * axisframe[1][0] +
    ypos * axisframe[1][1] +
    zpos * axisframe[1][2] +
    axisframe[1][3];

  y_2d = (viewy * eyedistance) / (eyedistance - zworld);
  x_2d = (viewx * eyedistance) / (eyedistance - zworld);

  *xis = xcenter + (short) (x_2d * scalefactor);
  *yis = ycenter - (short) (y_2d * scalefactor);
}

void 
Output2d_3d::WorldtoZvalue (double xpos,
			    double ypos,
			    double zpos,
			    double *zis)
{
  xpos *= xscale;
  ypos *= yscale;
  zpos *= zscale;
  *zis = xpos * axisframe[2][0] +
    ypos * axisframe[2][1] +
    zpos * axisframe[2][2] +
    axisframe[2][3];
}


void 
Output2d_3d::AxisRotx (double angle)
{
  double sinus = sin (angle), cosinus = cos (angle);
  double temp[4];		// temp. save the rowes of the matrix

// first row
  temp[1] = axisframe[0][1] * cosinus +
    axisframe[0][2] * sinus;
  temp[2] = -axisframe[0][1] * sinus +
    axisframe[0][2] * cosinus;
  axisframe[0][1] = temp[1];
  axisframe[0][2] = temp[2];

// second row
  temp[1] = axisframe[1][1] * cosinus +
    axisframe[1][2] * sinus;
  temp[2] = -axisframe[1][1] * sinus +
    axisframe[1][2] * cosinus;
  axisframe[1][1] = temp[1];
  axisframe[1][2] = temp[2];

// third row
  temp[1] = axisframe[2][1] * cosinus +
    axisframe[2][2] * sinus;
  temp[2] = -axisframe[2][1] * sinus +
    axisframe[2][2] * cosinus;
  axisframe[2][1] = temp[1];
  axisframe[2][2] = temp[2];
}

void 
Output2d_3d::AxisRoty (double angle)
{
  double sinus = sin (angle), cosinus = cos (angle);
  double temp[4];		// temp. save the rowes of the matrix

// first row
  temp[0] = axisframe[0][0] * cosinus -
    axisframe[0][2] * sinus;
  temp[2] = axisframe[0][0] * sinus +
    axisframe[0][2] * cosinus;
  axisframe[0][0] = temp[0];
  axisframe[0][2] = temp[2];

// second row
  temp[0] = axisframe[1][0] * cosinus -
    axisframe[1][2] * sinus;
  temp[2] = axisframe[1][0] * sinus +
    axisframe[1][2] * cosinus;
  axisframe[1][0] = temp[0];
  axisframe[1][2] = temp[2];

// third row
  temp[0] = axisframe[2][0] * cosinus -
    axisframe[2][2] * sinus;
  temp[2] = axisframe[2][0] * sinus +
    axisframe[2][2] * cosinus;
  axisframe[2][0] = temp[0];
  axisframe[2][2] = temp[2];
}

void 
Output2d_3d::AxisRotz (double angle)
{
  double sinus = sin (angle), cosinus = cos (angle);
  double temp[4];		// temp. save the rowes of the matrix

// first row
  temp[0] = axisframe[0][0] * cosinus +
    axisframe[0][1] * sinus;
  temp[1] = -axisframe[0][0] * sinus +
    axisframe[0][1] * cosinus;
  axisframe[0][0] = temp[0];
  axisframe[0][1] = temp[1];

// second row
  temp[0] = axisframe[1][0] * cosinus +
    axisframe[1][1] * sinus;
  temp[1] = -axisframe[1][0] * sinus +
    axisframe[1][1] * cosinus;
  axisframe[1][0] = temp[0];
  axisframe[1][1] = temp[1];

// third row
  temp[0] = axisframe[2][0] * cosinus +
    axisframe[2][1] * sinus;
  temp[1] = -axisframe[2][0] * sinus +
    axisframe[2][1] * cosinus;
  axisframe[2][0] = temp[0];
  axisframe[2][1] = temp[1];
}

DVektor *
Output2d_3d::Getnearestpoint1d (short xpos, short ypos, int *index, short *xis, short *yis)
{
  DVektorArray *inputData = &(Dataset->Lese_Daten ());
  short tempx, tempy;
  int counter, nearestcounter = 0;
  int xoutput;
  double youtput;
  double distance, mindistance = MAXDOUBLE;	// distance high 2

  PixeltoWorld1d (xpos, ypos, &xoutput, &youtput);
  for (counter = 0; counter < inputData->Lese_Groesse (); counter++) {
    WorldtoPixel1d (counter,
		    inputData->Lese_Vektor (counter, xindex),
		    &tempx, &tempy);
    distance = ((xpos - tempx) * (xpos - tempx)) +
      ((ypos - tempy) * (ypos - tempy));
    if (distance < mindistance) {
      mindistance = distance;
      nearestcounter = counter;
    }
  }
  *index = nearestcounter;
  WorldtoPixel1d (nearestcounter,
		  inputData->Lese_Vektor (nearestcounter, xindex),
		  xis, yis);

  return &(inputData->Lese_Vektor (nearestcounter));
}

DVektor *
Output2d_3d::Getnearestpoint2d (short xpos, short ypos, int *index, short *xis, short *yis)
{
  DVektorArray *inputData = &(Dataset->Lese_Daten ());
  short tempx, tempy;
  int counter, nearestcounter = 0;
  double xoutput, youtput;
  double distance, mindistance = MAXDOUBLE;	// distance high 2

  PixeltoWorld2d (xpos, ypos, &xoutput, &youtput);
  for (counter = 0; counter < inputData->Lese_Groesse (); counter++) {
    WorldtoPixel2d (inputData->Lese_Vektor (counter, xindex),
		    inputData->Lese_Vektor (counter, yindex),
		    &tempx, &tempy);
    distance = ((inputData->Lese_Vektor (counter, xindex) - xoutput) * (inputData->Lese_Vektor (counter, xindex) - xoutput)) +
      ((inputData->Lese_Vektor (counter, yindex) - youtput) * (inputData->Lese_Vektor (counter, yindex) - youtput));
    if (distance < mindistance) {
      mindistance = distance;
      nearestcounter = counter;
    }
  }
  *index = nearestcounter;
  WorldtoPixel2d (inputData->Lese_Vektor (nearestcounter, xindex),
		  inputData->Lese_Vektor (nearestcounter, yindex),
		  xis, yis);

  return &(inputData->Lese_Vektor (nearestcounter));
}

DVektor *
Output2d_3d::Getnearestpoint3d (short xpos, short ypos, int *index, short *xis, short *yis)
{
  DVektorArray *inputData = &(Dataset->Lese_Daten ());
  short tempx, tempy;
  int counter, nearestcounter = 0;
  double distance, mindistance = MAXDOUBLE;	// distance * distance

  for (counter = 0; counter < inputData->Lese_Groesse (); counter++) {
    WorldtoPixel3d (inputData->Lese_Vektor (counter, xindex),
		    inputData->Lese_Vektor (counter, yindex),
		    inputData->Lese_Vektor (counter, zindex),
		    &tempx, &tempy);

    distance = ((xpos - tempx) * (xpos - tempx)) +
      ((ypos - tempy) * (ypos - tempy));
    if (distance < mindistance) {
      *xis = tempx;
      *yis = tempy;
      mindistance = distance;
      nearestcounter = counter;
    }
  }
  *index = nearestcounter;
  return &(inputData->Lese_Vektor (nearestcounter));
}

void 
Output2d_3d::DrawPoint (short xpos, short ypos, int options, int number, short pointtyp)
{
  Drawable drawable;
  if (options & DRAWINWINDOW)
    drawable = window;
  else
    drawable = pixmap;

  if (options & DRAWMARKED)
    XSetForeground (display, gc, redpix);
  else
    XSetForeground (display, gc, blackpix);

  if (options & DRAWTEMP)
    XSetFunction (display, gc, GXxor);

  switch (pointtyp) {
  case OUTPUT_POINTTYPPOINT:
    if (options & DRAWMARKED) {
      XSegment lines[2];
      lines[0].y1 = lines[0].y2 = ypos;
      lines[0].x1 = xpos - CROSSWIDTH;
      lines[0].x2 = xpos + CROSSWIDTH;

      lines[1].x1 = lines[1].x2 = xpos;
      lines[1].y1 = ypos - CROSSWIDTH;
      lines[1].y2 = ypos + CROSSWIDTH;
      XDrawSegments (display, drawable, gc, lines, 2);
    } else
      XDrawPoint (display, drawable, gc, xpos, ypos);
    break;			// case OUTPUT_POINTTYPPOINT

  case OUTPUT_POINTTYPCROSS:
    XSegment lines[2];
    if (options & DRAWMARKED) {
      lines[0].y1 = lines[0].y2 = ypos;
      lines[0].x1 = xpos - 2 * CROSSWIDTH;
      lines[0].x2 = xpos + 2 * CROSSWIDTH;

      lines[1].x1 = lines[1].x2 = xpos;
      lines[1].y1 = ypos - 2 * CROSSWIDTH;
      lines[1].y2 = ypos + 2 * CROSSWIDTH;
    } else {
      lines[0].y1 = lines[0].y2 = ypos;
      lines[0].x1 = xpos - CROSSWIDTH;
      lines[0].x2 = xpos + CROSSWIDTH;

      lines[1].x1 = lines[1].x2 = xpos;
      lines[1].y1 = ypos - CROSSWIDTH;
      lines[1].y2 = ypos + CROSSWIDTH;
    }
    XDrawSegments (display, drawable, gc, lines, 2);
    break;			// case OUTPUT_POINTTYPCROSS

  case OUTPUT_POINTTYPCLASSES:
    DrawSymbol (xpos, ypos,
		drawable, number);
    break;

  case OUTPUT_POINTTYPCOLOR:
    DrawBox (xpos, ypos, drawable);
    break;
  }
  XSetFunction (display, gc, GXcopy);
}

void 
Output2d_3d::DrawPoint2d (DVektor * me, int options, short pointtyp)
{
  if (Dataset->Lese_Daten ().Lese_Dim () > 1) {
    short tempx, tempy;
    WorldtoPixel2d (me->Lese_i (xindex), me->Lese_i (yindex),
		    &tempx, &tempy);
    DrawPoint (tempx, tempy, options, me->Lese_Nummer (), pointtyp);
  } else {
    short xpixel, ypixel;
    WorldtoPixel1d (Getvektorindex (me, &(Dataset->Lese_Daten ())),
		    me->Lese_i (xindex),
		    &xpixel, &ypixel);
    DrawPoint (xpixel, ypixel, options, me->Lese_Nummer (), pointtyp);
  }
}

void 
Output2d_3d::DrawPoint3d (DVektor * me, int options, short pointtyp)
{
  short xpixel, ypixel;
  WorldtoPixel3d (me->Lese_i (xindex),
		  me->Lese_i (yindex),
		  me->Lese_i (zindex),
		  &xpixel, &ypixel);
  DrawPoint (xpixel, ypixel,
	     options,
	     me->Lese_Nummer (),
	     pointtyp);
}

void 
Output2d_3d::DrawAxes2d (char options, char *text1, char *text2)
{
  short axis[3][2];
  double xscalebackup = xscale, yscalebackup = yscale;
  xscale = yscale = 1.0;

  double arrowwidth, arrowlength;
  short textoffset, textheight = fontstruct->max_bounds.ascent + fontstruct->max_bounds.descent;

  arrowwidth = ARROWWIDTHMULT;
  arrowlength = ARROWLENGTHMULT;

  XPoint arrow1[3], arrow2[3];
  Drawable drawable;
  if (options & DRAWINWINDOW)
    drawable = window;
  else
    drawable = pixmap;

  XSetForeground (display, gc, blackpix);

//       arrow2[]
  //        / \ axis[2][.]
  //         |
  //         |
  //         |
  //         |
  //         ----------> axis[1][.]
  //    axis[0][.]   arrow1[]

  WorldtoPixel2d (0.0, 0.0, &(axis[0][0]), &(axis[0][1]));
  WorldtoPixel2d (1.0, 0.0, &(axis[1][0]), &(axis[1][1]));
  WorldtoPixel2d (0.0, 1.0, &(axis[2][0]), &(axis[2][1]));

  WorldtoPixel2d (1.0, 0, &(arrow1[0].x), &(arrow1[0].y));
  WorldtoPixel2d (1.0 - arrowlength, arrowwidth, &(arrow1[1].x), &(arrow1[1].y));
  WorldtoPixel2d (1.0 - arrowlength, -arrowwidth, &(arrow1[2].x), &(arrow1[2].y));

  WorldtoPixel2d (0, 1.0, &(arrow2[0].x), &(arrow2[0].y));
  WorldtoPixel2d (arrowwidth, 1.0 - arrowlength, &(arrow2[1].x), &(arrow2[1].y));
  WorldtoPixel2d (-arrowwidth, 1.0 - arrowlength, &(arrow2[2].x), &(arrow2[2].y));

  XSegment lines[] =
  {
    {axis[0][0], axis[0][1], axis[1][0], axis[1][1]},
    {axis[0][0], axis[0][1], axis[2][0], axis[2][1]}
  };

  if (text1 != 0) {
    textoffset = XTextWidth (fontstruct, text1, strlen (text1)) / 2;
    XDrawString (display, drawable, gc,
		 axis[1][0] - textoffset, axis[1][1] + textheight,
		 text1, strlen (text1));
  }
  if (text2 != 0) {
    textoffset = XTextWidth (fontstruct, text2, strlen (text2)) / 2;
    XDrawString (display, drawable, gc,
		 axis[2][0] - textoffset, axis[2][1],
		 text2, strlen (text2));
  }
  XDrawSegments (display, drawable, gc, lines, 2);
  XFillPolygon (display, drawable, gc, arrow1, 3, Convex, CoordModeOrigin);
  XFillPolygon (display, drawable, gc, arrow2, 3, Convex, CoordModeOrigin);
  xscale = xscalebackup;
  yscale = yscalebackup;
}

void 
Output2d_3d::DrawAxes3d (char options, char *text1, char *text2, char *text3)
{
  int axis[7][2];
  double x, y, z;
  XPoint points[3];
  XPoint points2[4];
  double arrowwidth, arrowlength;
  short textoffset;

// select arrowlength from shortesr axis:
  arrowwidth = ARROWWIDTHMULT;
  arrowlength = ARROWLENGTHMULT;

  Drawable drawable;

  if (options & DRAWINWINDOW)
    drawable = window;
  else
    drawable = pixmap;
  XSetForeground (display, gc, blackpix);

// (0,0,0)
  // transformation: axescoordinates -> worldcoordinates
  x = axisframe[0][3];
  y = axisframe[1][3];
  z = axisframe[2][3];
// projection: 3d -> 2d
  y = (y * eyedistance) / (eyedistance - z);
  x = (x * eyedistance) / (eyedistance - z);
// 2d -> pixelvalues
  axis[0][0] = xcenter + (int) (x * scalefactor);
  axis[0][1] = ycenter - (int) (y * scalefactor);

// (1.0,0,0)
  x = axisframe[0][0] + axisframe[0][3];
  y = axisframe[1][0] + axisframe[1][3];
  z = axisframe[2][0] + axisframe[2][3];
  y = (y * eyedistance) / (eyedistance - z);
  x = (x * eyedistance) / (eyedistance - z);
  axis[1][0] = xcenter + (int) (x * scalefactor);
  axis[1][1] = ycenter - (int) (y * scalefactor);

  if (text1 != 0) {
    textoffset = XTextWidth (fontstruct, text1, strlen (text1)) / 2;
    XDrawString (display, drawable, gc, axis[1][0] - textoffset, axis[1][1], text1, strlen (text1));
  }
// (0,1.0,0)
  x = axisframe[0][1] + axisframe[0][3];
  y = axisframe[1][1] + axisframe[1][3];
  z = axisframe[2][1] + axisframe[2][3];
  y = (y * eyedistance) / (eyedistance - z);
  x = (x * eyedistance) / (eyedistance - z);
  axis[2][0] = xcenter + (int) (x * scalefactor);
  axis[2][1] = ycenter - (int) (y * scalefactor);

  if (text2 != 0) {
    textoffset = XTextWidth (fontstruct, text2, strlen (text2)) / 2;
    XDrawString (display, drawable, gc, axis[2][0] - textoffset, axis[2][1], text2, strlen (text2));
  }
// (0,0,1.0)
  x = axisframe[0][2] + axisframe[0][3];
  y = axisframe[1][2] + axisframe[1][3];
  z = axisframe[2][2] + axisframe[2][3];
  y = (y * eyedistance) / (eyedistance - z);
  x = (x * eyedistance) / (eyedistance - z);
  axis[3][0] = xcenter + (int) (x * scalefactor);
  axis[3][1] = ycenter - (int) (y * scalefactor);

  if (text3 != 0) {
    textoffset = XTextWidth (fontstruct, text3, strlen (text3)) / 2;
    XDrawString (display, drawable, gc, axis[3][0] - textoffset, axis[3][1], text3, strlen (text3));
  }
  XSegment linesright[] =
  {
    {axis[0][0], axis[0][1], axis[1][0], axis[1][1]},
    {axis[0][0], axis[0][1], axis[2][0], axis[2][1]},
    {axis[0][0], axis[0][1], axis[3][0], axis[3][1]}
  };

  XDrawSegments (display, drawable, gc, linesright, 3);

// next draw arrows on end of axes
  double localx, localy, localz;

// first arrows on X-axis
  //local point (1.0, 0, 0)
  points[0].x = axis[1][0];
  points[0].y = axis[1][1];

//local point (1.0-ARROWLENGTHMULT,
  //             ARROWWIDTHMULT,
  //             0)
  localx = 1.0 - arrowlength;
  localy = arrowwidth;
  x = axisframe[0][0] * localx +
    axisframe[0][1] * localy +
    axisframe[0][3];
  y = axisframe[1][0] * localx +
    axisframe[1][1] * localy +
    axisframe[1][3];
  z = axisframe[2][0] * localx +
    axisframe[2][1] * localy +
    axisframe[2][3];
  y = (y * eyedistance) / (eyedistance - z);
  x = (x * eyedistance) / (eyedistance - z);
  points2[0].x = points[1].x = xcenter + (int) (x * scalefactor);
  points2[0].y = points[1].y = ycenter - (int) (y * scalefactor);

//local point (1-ARROWLENGTHMULT,
  //             -ARROWWIDTHMULT,
  //             0)
  x = axisframe[0][0] * localx -
    axisframe[0][1] * localy +
    axisframe[0][3];
  y = axisframe[1][0] * localx -
    axisframe[1][1] * localy +
    axisframe[1][3];
  z = axisframe[2][0] * localx -
    axisframe[2][1] * localy +
    axisframe[2][3];
  y = (y * eyedistance) / (eyedistance - z);
  x = (x * eyedistance) / (eyedistance - z);
  points2[2].x = points[2].x = xcenter + (int) (x * scalefactor);
  points2[2].y = points[2].y = ycenter - (int) (y * scalefactor);
  XFillPolygon (display, drawable, gc, points, 3, Convex, CoordModeOrigin);

// local point (1-ARROWLENGTHMULT,
  //              0,
  //              ARROWWIDTHMULT)
  localx = 1.0 - arrowlength;
  localz = arrowwidth;
  x = axisframe[0][0] * localx +
    axisframe[0][2] * localz +
    axisframe[0][3];
  y = axisframe[1][0] * localx +
    axisframe[1][2] * localz +
    axisframe[1][3];
  z = axisframe[2][0] * localx +
    axisframe[2][2] * localz +
    axisframe[2][3];
  y = (y * eyedistance) / (eyedistance - z);
  x = (x * eyedistance) / (eyedistance - z);
  points2[1].x = points[1].x = xcenter + (int) (x * scalefactor);
  points2[1].y = points[1].y = ycenter - (int) (y * scalefactor);

// local point (1-ARROWLENGTHMULT,
  //              0,
  //              -ARROWWIDTHMULT)
  x = axisframe[0][0] * localx -
    axisframe[0][2] * localz +
    axisframe[0][3];
  y = axisframe[1][0] * localx -
    axisframe[1][2] * localz +
    axisframe[1][3];
  z = axisframe[2][0] * localx -
    axisframe[2][2] * localz +
    axisframe[2][3];
  y = (y * eyedistance) / (eyedistance - z);
  x = (x * eyedistance) / (eyedistance - z);
  points2[3].x = points[2].x = xcenter + (int) (x * scalefactor);
  points2[3].y = points[2].y = ycenter - (int) (y * scalefactor);
  XFillPolygon (display, drawable, gc, points, 3, Convex, CoordModeOrigin);
  XFillPolygon (display, drawable, gc, points2, 4, Convex, CoordModeOrigin);


// second arrows on Y-axis
  //local point (0, 1.0, 0)
  points[0].x = axis[2][0];
  points[0].y = axis[2][1];

//local point (0,
  //             1-ARROWLENGTHMULT,
  //             ARROWWIDTHMULT)
  localy = 1.0 - arrowlength;
  localz = arrowwidth;
  x = axisframe[0][1] * localy +
    axisframe[0][2] * localz +
    axisframe[0][3];
  y = axisframe[1][1] * localy +
    axisframe[1][2] * localz +
    axisframe[1][3];
  z = axisframe[2][1] * localy +
    axisframe[2][2] * localz +
    axisframe[2][3];
  y = (y * eyedistance) / (eyedistance - z);
  x = (x * eyedistance) / (eyedistance - z);
  points2[0].x = points[1].x = xcenter + (int) (x * scalefactor);
  points2[0].y = points[1].y = ycenter - (int) (y * scalefactor);

//local point (0,
  //             1-ARROWLENGTHMULT,
  //             -ARROWWIDTHMULT)
  x = axisframe[0][1] * localy -
    axisframe[0][2] * localz +
    axisframe[0][3];
  y = axisframe[1][1] * localy -
    axisframe[1][2] * localz +
    axisframe[1][3];
  z = axisframe[2][1] * localy -
    axisframe[2][2] * localz +
    axisframe[2][3];
  y = (y * eyedistance) / (eyedistance - z);
  x = (x * eyedistance) / (eyedistance - z);
  points2[2].x = points[2].x = xcenter + (int) (x * scalefactor);
  points2[2].y = points[2].y = ycenter - (int) (y * scalefactor);
  XFillPolygon (display, drawable, gc, points, 3, Convex, CoordModeOrigin);

// local point (ARROWWIDTHHMULT,
  //              1-ARROWLENGTHMULT,
  //              0)
  localx = arrowwidth;
  localy = 1.0 - arrowlength;
  x = axisframe[0][0] * localx +
    axisframe[0][1] * localy +
    axisframe[0][3];
  y = axisframe[1][0] * localx +
    axisframe[1][1] * localy +
    axisframe[1][3];
  z = axisframe[2][0] * localx +
    axisframe[2][1] * localy +
    axisframe[2][3];
  y = (y * eyedistance) / (eyedistance - z);
  x = (x * eyedistance) / (eyedistance - z);
  points2[1].x = points[1].x = xcenter + (int) (x * scalefactor);
  points2[1].y = points[1].y = ycenter - (int) (y * scalefactor);

// local point (-ARROWWIDTHHMULT,
  //              1-ARROWLENGTHMULT,
  //              0)
  x = -axisframe[0][0] * localx +
    axisframe[0][1] * localy +
    axisframe[0][3];
  y = -axisframe[1][0] * localx +
    axisframe[1][1] * localy +
    axisframe[1][3];
  z = -axisframe[2][0] * localx +
    axisframe[2][1] * localy +
    axisframe[2][3];
  y = (y * eyedistance) / (eyedistance - z);
  x = (x * eyedistance) / (eyedistance - z);
  points2[3].x = points[2].x = xcenter + (int) (x * scalefactor);
  points2[3].y = points[2].y = ycenter - (int) (y * scalefactor);
  XFillPolygon (display, drawable, gc, points, 3, Convex, CoordModeOrigin);
  XFillPolygon (display, drawable, gc, points2, 4, Convex, CoordModeOrigin);


// last the arrows on Z-axis
  //local point (0, 0, 1.0)
  points[0].x = axis[3][0];
  points[0].y = axis[3][1];

//local point (ARROWWIDTHMULT,
  //             0,
  //             1-ARROWLENGTHMULT)
  localx = arrowwidth;
  localz = 1.0 - arrowlength;
  x = axisframe[0][0] * localx +
    axisframe[0][2] * localz +
    axisframe[0][3];
  y = axisframe[1][0] * localx +
    axisframe[1][2] * localz +
    axisframe[1][3];
  z = axisframe[2][0] * localx +
    axisframe[2][2] * localz +
    axisframe[2][3];
  y = (y * eyedistance) / (eyedistance - z);
  x = (x * eyedistance) / (eyedistance - z);
  points2[0].x = points[1].x = xcenter + (int) (x * scalefactor);
  points2[0].y = points[1].y = ycenter - (int) (y * scalefactor);

//local point (-ARROWWIDTHMULT,
  //             0,
  //             1-ARROWLENGTHMULT)
  x = -axisframe[0][0] * localx +
    axisframe[0][2] * localz +
    axisframe[0][3];
  y = -axisframe[1][0] * localx +
    axisframe[1][2] * localz +
    axisframe[1][3];
  z = -axisframe[2][0] * localx +
    axisframe[2][2] * localz +
    axisframe[2][3];
  y = (y * eyedistance) / (eyedistance - z);
  x = (x * eyedistance) / (eyedistance - z);
  points2[2].x = points[2].x = xcenter + (int) (x * scalefactor);
  points2[2].y = points[2].y = ycenter - (int) (y * scalefactor);
  XFillPolygon (display, drawable, gc, points, 3, Convex, CoordModeOrigin);

// local point (0
  //              ARROWWIDTHHMULT,
  //              1-ARROWLENGTHMULT)
  localy = arrowwidth;
  localz = 1.0 - arrowlength;
  x = axisframe[0][1] * localy +
    axisframe[0][2] * localz +
    axisframe[0][3];
  y = axisframe[1][1] * localy +
    axisframe[1][2] * localz +
    axisframe[1][3];
  z = axisframe[2][1] * localy +
    axisframe[2][2] * localz +
    axisframe[2][3];
  y = (y * eyedistance) / (eyedistance - z);
  x = (x * eyedistance) / (eyedistance - z);
  points2[1].x = points[1].x = xcenter + (int) (x * scalefactor);
  points2[1].y = points[1].y = ycenter - (int) (y * scalefactor);

// local point (0
  //              -ARROWWIDTHHMULT,
  //              1-ARROWLENGTHMULT)
  x = -axisframe[0][1] * localy +
    axisframe[0][2] * localz +
    axisframe[0][3];
  y = -axisframe[1][1] * localy +
    axisframe[1][2] * localz +
    axisframe[1][3];
  z = -axisframe[2][1] * localy +
    axisframe[2][2] * localz +
    axisframe[2][3];
  y = (y * eyedistance) / (eyedistance - z);
  x = (x * eyedistance) / (eyedistance - z);
  points2[3].x = points[2].x = xcenter + (int) (x * scalefactor);
  points2[3].y = points[2].y = ycenter - (int) (y * scalefactor);
  XFillPolygon (display, drawable, gc, points, 3, Convex, CoordModeOrigin);
  XFillPolygon (display, drawable, gc, points2, 4, Convex, CoordModeOrigin);
}

void 
Output2d_3d::DrawDVektorArray1d (int options, char pointtyp,
			      DVektorArray * inputData, BVektor * eindeutig)
{
  Drawable drawable;

  if (options & DRAWINWINDOW)
    drawable = window;
  else
    drawable = pixmap;

  if (options & DRAWGREY)
    XSetForeground (display, gc, shadowpix);
  else
    XSetForeground (display, gc, blackpix);

  if (Dataset != NULL) {
    int counter, number = inputData->Lese_Groesse ();
    switch (pointtyp) {
    case OUTPUT_POINTTYPPOINT:
      XPoint * Points;

      if ((Points = (XPoint *) malloc (sizeof (XPoint) * number)) == NULL) {
	printf ("Out of memory!\n");
	exit (1);
      }
      for (counter = 0; counter < number; counter++)
	FillPoint (&(Points[counter]),
		   counter,
		   inputData->Lese_Vektor (counter, xindex));

      XDrawPoints (display, drawable, gc, Points, number, CoordModeOrigin);
      free (Points);
      break;			// OUTPUT_POINTTYPPOINT

    case OUTPUT_POINTTYPCLASSES:
      {
	short xpos, ypos;
	for (counter = 0; counter < number; counter++) {
	  if (eindeutig != NULL) {
	    if (eindeutig->Lese_i (counter))
	      XSetForeground (display, gc, blackpix);
	    else
	      XSetForeground (display, gc, lightpix);
	  }
	  WorldtoPixel1d (counter, inputData->Lese_Vektor (counter, xindex),
			  &xpos, &ypos);
	  DrawSymbol (xpos, ypos,
		      drawable,
		      inputData->Lese_Vektor (counter).Lese_Nummer ());
	}
      }
      break;			// OUTPUT_POINTTYPCLASSES

    case OUTPUT_POINTTYPCOLOR:
      {
	short xpos, ypos;
	for (counter = 0; counter < number; counter++) {
	  WorldtoPixel1d (counter, inputData->Lese_Vektor (counter, xindex),
			  &xpos, &ypos);
	  if (eindeutig != NULL) {
	    if (eindeutig->Lese_i (counter))
	      DrawColorbox (xpos, ypos, drawable, inputData->Lese_Vektor (counter).Lese_Nummer ());
	    else
	      DrawBox (xpos, ypos, drawable);
	  } else
	    DrawColorbox (xpos, ypos, drawable, inputData->Lese_Vektor (counter).Lese_Nummer ());
	}
      }
      break;			// OUTPUT_POINTTYPCOLOR

    case OUTPUT_POINTTYPCROSS:
      XSegment * lines = NULL;
      int index = 0;
      short xpixel, ypixel;
      if ((lines = (XSegment *) malloc (sizeof (XSegment) * 2 * number)) == NULL) {
	printf ("Out of memory!\n");
	exit (1);
      }
      for (counter = 0; counter < 2 * number; counter += 2) {
	FillLines (&(lines[counter]),
		   index,
		   inputData->Lese_Vektor (index, xindex),
		   &xpixel, &ypixel);

	index++;
      }
      XDrawSegments (display, drawable, gc, lines, number * 2);
      free (lines);
      break;			// case OUTPUT_POINTTYPCROSS

    }
  }
}

void 
Output2d_3d::DrawDVektorArray2d (int options, char pointtyp,
			      DVektorArray * inputData, BVektor * eindeutig)
{
  Drawable drawable;

  if (options & DRAWINWINDOW)
    drawable = window;
  else
    drawable = pixmap;

// draw points/crosses
  if (options & DRAWGREY)
    XSetForeground (display, gc, shadowpix);
  else
    XSetForeground (display, gc, blackpix);

  if (inputData != NULL) {
    int counter, number = inputData->Lese_Groesse ();
    switch (pointtyp) {
    case OUTPUT_POINTTYPPOINT:
      XPoint * Points;

      if ((Points = (XPoint *) malloc (sizeof (XPoint) * number)) == NULL) {
	printf ("Out of memory!\n");
	exit (1);
      }
      for (counter = 0; counter < number; counter++)
	FillPoint (&(Points[counter]),
		   inputData->Lese_Vektor (counter, xindex),
		   inputData->Lese_Vektor (counter, yindex));

      XDrawPoints (display, drawable, gc, Points, number, CoordModeOrigin);
      free (Points);
      break;			// OUTPUT_POINTTYPPOINT

    case OUTPUT_POINTTYPCLASSES:
      {
	short xpos, ypos;
	for (counter = 0; counter < number; counter++) {
	  if (eindeutig != NULL) {
	    if (eindeutig->Lese_i (counter))
	      XSetForeground (display, gc, blackpix);
	    else
	      XSetForeground (display, gc, lightpix);
	  }
	  WorldtoPixel2d (inputData->Lese_Vektor (counter, xindex),
			  inputData->Lese_Vektor (counter, yindex),
			  &xpos, &ypos);
	  DrawSymbol (xpos, ypos,
		      drawable,
		      inputData->Lese_Vektor (counter).Lese_Nummer ());
	}
      }
      break;			// OUTPUT_POINTTYPCLASSES

    case OUTPUT_POINTTYPCOLOR:
      {
	short xpos, ypos;
	for (counter = 0; counter < number; counter++) {
	  WorldtoPixel2d (inputData->Lese_Vektor (counter, xindex),
			  inputData->Lese_Vektor (counter, yindex),
			  &xpos, &ypos);
	  if (eindeutig != NULL) {
	    if (eindeutig->Lese_i (counter))
	      DrawColorbox (xpos, ypos, drawable, inputData->Lese_Vektor (counter).Lese_Nummer ());
	    else
	      DrawBox (xpos, ypos, drawable);
	  } else
	    DrawColorbox (xpos, ypos, drawable, inputData->Lese_Vektor (counter).Lese_Nummer ());
	}
      }
      break;			// OUTPUT_POINTTYPCOLOR

    case OUTPUT_POINTTYPCROSS:
      XSegment * lines = NULL;
      int index = 0;
      short xpixel, ypixel;
      if ((lines = (XSegment *) malloc (sizeof (XSegment) * 2 * number)) == NULL) {
	printf ("Out of memory!\n");
	exit (1);
      }
      for (counter = 0; counter < 2 * number; counter += 2) {
	FillLines (&(lines[counter]),
		   inputData->Lese_Vektor (index, xindex),
		   inputData->Lese_Vektor (index, yindex),
		   &xpixel, &ypixel);

	index++;
      }
      XDrawSegments (display, drawable, gc, lines, number * 2);
      free (lines);
      break;			// case OUTPUT_POINTTYPCROSS

    }
  }
}

void 
Output2d_3d::DrawDVektorArray3d (int options, char pointtyp,
			      DVektorArray * inputData, BVektor * eindeutig)
{
  Drawable drawable;

  if (options & DRAWINWINDOW)
    drawable = window;
  else
    drawable = pixmap;

  if (options & DRAWGREY)
    XSetForeground (display, gc, shadowpix);
  else
    XSetForeground (display, gc, blackpix);

  if (Dataset != 0) {
    int counter, number = inputData->Lese_Groesse ();

    switch (pointtyp) {
    case OUTPUT_POINTTYPPOINT:
      XPoint * Points;
      if ((Points = (XPoint *) malloc (sizeof (XPoint) * number)) == NULL) {
	printf ("Out of memory!\n");
	exit (1);
      }
      for (counter = 0; counter < number; counter++)
	FillPoint (&(Points[counter]),
		   inputData->Lese_Vektor (counter, xindex),
		   inputData->Lese_Vektor (counter, yindex),
		   inputData->Lese_Vektor (counter, zindex));

      XDrawPoints (display, drawable, gc, Points, number, CoordModeOrigin);
      free (Points);
      break;			// case OUTPUT_POINTTYPPOINT

    case OUTPUT_POINTTYPCLASSES:
      {
	short xpos, ypos;
	for (counter = 0; counter < number; counter++) {
	  if (eindeutig != NULL) {
	    if (eindeutig->Lese_i (counter))
	      XSetForeground (display, gc, blackpix);
	    else
	      XSetForeground (display, gc, lightpix);
	  }
	  WorldtoPixel3d (inputData->Lese_Vektor (counter, xindex),
			  inputData->Lese_Vektor (counter, yindex),
			  inputData->Lese_Vektor (counter, zindex),
			  &xpos, &ypos);
	  DrawSymbol (xpos, ypos,
		      drawable,
		      inputData->Lese_Vektor (counter).Lese_Nummer ());
	}
      }
      break;			// OUTPUT_POINTTYPCLASSES

    case OUTPUT_POINTTYPCOLOR:
      {
	short xpos, ypos;
	int index;
//         double zpos;
	for (counter = number - 1; counter > 0; counter--) {
	  index = deepindex[counter];
	  WorldtoZvalue (inputData->Lese_Vektor (index, xindex),
			 inputData->Lese_Vektor (index, yindex),
			 inputData->Lese_Vektor (index, zindex),
			 &(deepvalues[counter]));
//                         &zpos);
	  //           deepvalues[counter]=zpos;
	}
	deepvalues.Quicksort (deepindex);

	for (counter = 0; counter < number; counter++) {
	  index = deepindex[counter];
	  WorldtoPixel3d (inputData->Lese_Vektor (index, xindex),
			  inputData->Lese_Vektor (index, yindex),
			  inputData->Lese_Vektor (index, zindex),
			  deepvalues[counter],
			  &xpos, &ypos);
	  if (eindeutig != NULL) {
	    if (eindeutig->Lese_i (index))
	      DrawColorbox (xpos, ypos, drawable, inputData->Lese_Vektor (index).Lese_Nummer ());
	    else
	      DrawBox (xpos, ypos, drawable);
	  } else
	    DrawColorbox (xpos, ypos, drawable, inputData->Lese_Vektor (index).Lese_Nummer ());
	}

/*         for(counter=0;counter<number;counter++)
   {
   WorldtoPixel3d(inputData->Lese_Vektor(counter,xindex),
   inputData->Lese_Vektor(counter,yindex),
   inputData->Lese_Vektor(counter,zindex),
   &xpos, &ypos);
   if(eindeutig!=NULL)
   {
   if(eindeutig->Lese_i(counter))
   DrawColorbox(xpos,ypos,drawable,inputData->Lese_Vektor(counter).Lese_Nummer());
   else
   DrawBox(xpos,ypos,drawable);
   }
   else
   DrawColorbox(xpos,ypos,drawable,inputData->Lese_Vektor(counter).Lese_Nummer());
   } */
      }
      break;			// OUTPUT_POINTTYPCOLOR

    case OUTPUT_POINTTYPCROSS:
      XSegment * lines;
      int index = 0;
      short xpixel, ypixel;
      if ((lines = (XSegment *) malloc (sizeof (XSegment) * 2 * number)) == NULL) {
	printf ("Out of memory!\n");
	exit (1);
      }
      for (counter = 0; counter < 2 * number; counter += 2) {
	FillLines (&(lines[counter]),
		   inputData->Lese_Vektor (index, xindex),
		   inputData->Lese_Vektor (index, yindex),
		   inputData->Lese_Vektor (index, zindex),
		   &xpixel, &ypixel);
	index++;
      }
      XDrawSegments (display, drawable, gc, lines, number * 2);
      free (lines);
      break;			// OUTPUT_POINTTYPCROSS

    }
  }
}

void 
Output2d_3d::FillPoint (XPoint * Point,
			double xpos,
			double ypos)
{
  WorldtoPixel2d (xpos, ypos,
		  &Point->x, &Point->y);
}

void 
Output2d_3d::FillLines (XSegment * lines,
			double xpos,
			double ypos,
			short *xpixel, short *ypixel)
{
  WorldtoPixel2d (xpos, ypos,
		  xpixel, ypixel);

  lines[0].y1 = lines[0].y2 = *ypixel;
  lines[0].x1 = (short) (*xpixel - CROSSWIDTH);
  lines[0].x2 = (short) (*xpixel + CROSSWIDTH);

  lines[1].x1 = lines[1].x2 = *xpixel;
  lines[1].y1 = (short) (*ypixel - CROSSWIDTH);
  lines[1].y2 = (short) (*ypixel + CROSSWIDTH);
}

void 
Output2d_3d::FillPoint (XPoint * Point,
			int number,
			double xpos)
{
  WorldtoPixel1d (number, xpos,
		  &Point->x, &Point->y);
}

void 
Output2d_3d::FillLines (XSegment * lines,
			int number,
			double xpos,
			short *xpixel, short *ypixel)
{
  WorldtoPixel1d (number, xpos,
		  xpixel, ypixel);

  lines[0].y1 = lines[0].y2 = *ypixel;
  lines[0].x1 = (short) (*xpixel - CROSSWIDTH);
  lines[0].x2 = (short) (*xpixel + CROSSWIDTH);

  lines[1].x1 = lines[1].x2 = *xpixel;
  lines[1].y1 = (short) (*ypixel - CROSSWIDTH);
  lines[1].y2 = (short) (*ypixel + CROSSWIDTH);
}

void 
Output2d_3d::FillPoint (XPoint * Point,
			double xpos,
			double ypos,
			double zpos)
{

  WorldtoPixel3d (xpos, ypos, zpos,
		  &Point->x, &Point->y);
}

void 
Output2d_3d::FillLines (XSegment * lines,
			double xpos,
			double ypos,
			double zpos,
			short *xpixel, short *ypixel)
{
  WorldtoPixel3d (xpos, ypos, zpos,
		  xpixel, ypixel);

  lines[0].y1 = lines[0].y2 = *ypixel;
  lines[0].x1 = (short) (*xpixel - CROSSWIDTH);
  lines[0].x2 = (short) (*xpixel + CROSSWIDTH);

  lines[1].x1 = lines[1].x2 = *xpixel;
  lines[1].y1 = (short) (*ypixel - CROSSWIDTH);
  lines[1].y2 = (short) (*ypixel + CROSSWIDTH);
}

Output2d::Output2d (Display * initdisplay,
	  GC initgc,
	  MoveableIcon * initOutputicon,
	  XFontStruct * fontstruct,
	  int initx, int inity, int initwidth, int initheight,
	  char *text):
Output (initdisplay,
	initgc,
	initOutputicon,
	fontstruct,
	initx, inity, initwidth, initheight,
	text)
{
}

void 
Output2d::DrawGraph (Parameter * parameter,
		     int xpos, int ypos, int drawwidth, int drawheight,
		     double range, double minxvalue)
{
  int counter;

  XSegment Lines[2];
  XPoint *points;
  XPoint arrow[4];
  if ((points = (XPoint *) malloc (sizeof (XPoint) * parameter->Lese_Parameter_Anzahl ())) == NULL) {
    printf ("Out of memory!\n");
    exit (1);
  }
//first draw axes
  Lines[0].x1 = xpos;
  Lines[0].y1 = ypos;
  Lines[0].x2 = xpos + drawwidth;
  Lines[0].y2 = ypos;
  Lines[1].x2 = Lines[1].x1 = xpos + (int) (drawwidth * (-minxvalue / range));
  Lines[1].y1 = ypos;
  Lines[1].y2 = ypos - drawheight;
  XDrawSegments (display, pixmap, gc, Lines, 2);

//and arrows on x-axis > 
  arrow[0].x = xpos + 1 + drawwidth - (int) (drawwidth * RULE_ARROWLENGTHMULT);
  arrow[0].y = ypos - (int) (drawheight * RULE_ARROWWIDTHMULT);
  arrow[1].x = xpos + 1 + drawwidth;
  arrow[1].y = ypos;
  arrow[2].x = xpos + 1 + drawwidth - (int) (drawwidth * RULE_ARROWLENGTHMULT);
  arrow[2].y = ypos + (int) (drawheight * RULE_ARROWWIDTHMULT);
// +1 looks nicer !
  XFillPolygon (display, pixmap, gc, arrow, 3, Convex, CoordModeOrigin);

//and arrows on y-axis /\

#if 0
/* guess - Guess - GUESS -- ARGHHHHHHHHHHH */
#endif
  arrow[0].x = Lines[1].x1 - (int) (drawwidth * RULE_ARROWWIDTHMULT);
  arrow[0].y = ypos - drawheight - 1 + (int) (drawheight * RULE_ARROWLENGTHMULT);
  arrow[1].x = Lines[1].x1;
  arrow[1].y = ypos - drawheight - 1;
  arrow[2].x = Lines[1].x1 + (int) (drawwidth * RULE_ARROWWIDTHMULT);
  arrow[2].y = ypos - drawheight - 1 + (int) (drawheight * RULE_ARROWLENGTHMULT);
  XFillPolygon (display, pixmap, gc, arrow, 3, Convex, CoordModeOrigin);


  for (counter = 0; counter < parameter->Lese_Parameter_Anzahl (); counter++) {
    points[counter].x = xpos + (int) (drawwidth * ((parameter->Lese_xParameter (counter) - minxvalue) / range));
    points[counter].y = ypos - (short) (drawheight * parameter->Lese_yParameter (counter));
  }

  XDrawLines (display, pixmap, gc, points, parameter->Lese_Parameter_Anzahl (), CoordModeOrigin);

// now draw values
  {
    char leftvalue[RULE_VALUELENGTH], rightvalue[RULE_VALUELENGTH];
    short lefttextwidth, righttextwidth, overlap;
    // first lower values
    sprintf (leftvalue, "%2.3f", parameter->Lese_xParameter (0));
    sprintf (rightvalue, "%2.3f", parameter->Lese_xParameter (parameter->Lese_Parameter_Anzahl () - 1));
    lefttextwidth = XTextWidth (fontstruct, leftvalue, strlen (leftvalue));
    righttextwidth = XTextWidth (fontstruct, rightvalue, strlen (rightvalue));
    points[parameter->Lese_Parameter_Anzahl () - 1].y = ypos;
    points[0].y = points[parameter->Lese_Parameter_Anzahl () - 1].y += fontstruct->max_bounds.ascent + fontstruct->max_bounds.descent,
      points[0].x -= lefttextwidth / 2;
    points[parameter->Lese_Parameter_Anzahl () - 1].x -= righttextwidth / 2;
    overlap = MAX (0, 5 + points[0].x + lefttextwidth - points[parameter->Lese_Parameter_Anzahl () - 1].x);
    points[0].x -= overlap / 2;
    points[parameter->Lese_Parameter_Anzahl () - 1].x += overlap / 2;
    XDrawString (display, pixmap, gc,
		 points[0].x, points[0].y,
		 leftvalue, strlen (leftvalue));
    XDrawString (display, pixmap, gc,
		 points[parameter->Lese_Parameter_Anzahl () - 1].x, points[parameter->Lese_Parameter_Anzahl () - 1].y,
		 rightvalue, strlen (rightvalue));
  }
  free (points);
}


InOutput::InOutput (Display * initdisplay,
	  GC initgc,
	  InputIcon * initOutputicon,
	  XFontStruct * fontstruct,
	  int initx, int inity, int initwidth, int initheight,
	  char *text):
Output2d_3d (initdisplay,
	     initgc,
	     initOutputicon,
	     fontstruct,
	     initx, inity, initwidth, initheight,
	     text)
{
  ruleresultinfo = NULL;
  selectedinput = NULL;
  viewicons = NULL;
  resultdisplaytyp = NULL;
  xindex = DEFAULTXINDEX;
  yindex = DEFAULTYINDEX;
  zindex = DEFAULTZINDEX;
  axismenu = NULL;
  float_to_pixel = 1.0;
  XDefineCursor (display, window, maincursor);

  viewicons = new SelectIconlist ();

  *viewicons += new SelectIcon (display, gc, this, fontstruct,
			 width - 2 * VIEWICON_SIZE - 3 * DEFAULT_BDWIDTH, 0,
				VIEWICON_SIZE, VIEWICON_SIZE,
		     "", d2_bits, d2_width, d2_height, OUTPUT_VIEW2D_SELECT,
				ICON_SELECTED);
  *viewicons += new SelectIcon (display, gc, this, fontstruct,
			     width - VIEWICON_SIZE - 2 * DEFAULT_BDWIDTH, 0,
				VIEWICON_SIZE, VIEWICON_SIZE,
		     "", d3_bits, d3_width, d3_height, OUTPUT_VIEW3D_SELECT,
				0);
  *viewicons += new SelectIcon (display, gc, this, fontstruct,
			    width - 2 * VIEWICON_SIZE - 3 * DEFAULT_BDWIDTH,
				VIEWICON_SIZE + 1 * DEFAULT_BDWIDTH,
				VIEWICON_SIZE, VIEWICON_SIZE,
				"", d2membertyp1_bits, d2membertyp1_width, d2membertyp1_height, OUTPUT_VIEW2DCLASSMEMBERSHIP_SELECT,
				0);
  *viewicons += new SelectIcon (display, gc, this, fontstruct,
				width - VIEWICON_SIZE - 2 * DEFAULT_BDWIDTH,
				VIEWICON_SIZE + 1 * DEFAULT_BDWIDTH,
				VIEWICON_SIZE, VIEWICON_SIZE,
				"", d3membertyp1_bits, d3membertyp1_width, d3membertyp1_height, OUTPUT_VIEW3DCLASSMEMBERSHIP_SELECT,
				0);

  pointtypicons = new SelectIconlist ();
  *pointtypicons += new SelectIcon (display, gc, this, fontstruct,
			    width - 2 * VIEWICON_SIZE - 3 * DEFAULT_BDWIDTH,
				    2 * VIEWICON_SIZE + 3 * DEFAULT_BDWIDTH,
				    VIEWICON_SIZE, VIEWICON_SIZE,
	  "", pointtyppoint_bits, pointtyppoint_width, pointtyppoint_height,
				    OUTPUT_POINTTYPPOINT,
				    ICON_SELECTABLE);
  *pointtypicons += new SelectIcon (display, gc, this, fontstruct,
				width - VIEWICON_SIZE - 2 * DEFAULT_BDWIDTH,
				    2 * VIEWICON_SIZE + 3 * DEFAULT_BDWIDTH,
				    VIEWICON_SIZE, VIEWICON_SIZE,
	  "", pointtypcross_bits, pointtypcross_width, pointtypcross_height,
				    OUTPUT_POINTTYPCROSS,
				    ICON_SELECTED | ICON_SELECTABLE);
  *pointtypicons += new SelectIcon (display, gc, this, fontstruct,
			    width - 2 * VIEWICON_SIZE - 3 * DEFAULT_BDWIDTH,
				    3 * VIEWICON_SIZE + 4 * DEFAULT_BDWIDTH,
				    VIEWICON_SIZE, VIEWICON_SIZE,
  "", clustertypsymbol_bits, clustertypsymbol_width, clustertypsymbol_height,
				    OUTPUT_POINTTYPCLASSES,
				    ICON_SELECTABLE);
  *pointtypicons += new SelectIcon (display, gc, this, fontstruct,
				width - VIEWICON_SIZE - 2 * DEFAULT_BDWIDTH,
				    3 * VIEWICON_SIZE + 4 * DEFAULT_BDWIDTH,
				    VIEWICON_SIZE, VIEWICON_SIZE,
    "", clustertypcolor_bits, clustertypcolor_width, clustertypcolor_height,
				    OUTPUT_POINTTYPCOLOR,
				    ICON_SELECTABLE);

  actionselecticons = new SelectIconlist ();
  *actionselecticons += new SelectIcon (display, gc, this, fontstruct,
				width - VIEWICON_SIZE - 2 * DEFAULT_BDWIDTH,
				    4 * VIEWICON_SIZE + 6 * DEFAULT_BDWIDTH,
					VIEWICON_SIZE, VIEWICON_SIZE,
				     "", move_bits, move_width, move_height,
					OUTPUTSTATUS_MOVE,
					ICON_SELECTABLE);
  *actionselecticons += new SelectIcon (display, gc, this, fontstruct,
			    width - 2 * VIEWICON_SIZE - 3 * DEFAULT_BDWIDTH,
				    4 * VIEWICON_SIZE + 6 * DEFAULT_BDWIDTH,
					VIEWICON_SIZE, VIEWICON_SIZE,
				     "", roll_bits, roll_width, roll_height,
					OUTPUTSTATUS_ROLL,
					ICON_SELECTED | ICON_SELECTABLE);
  *actionselecticons += new SelectIcon (display, gc, this, fontstruct,
				width - VIEWICON_SIZE - 2 * DEFAULT_BDWIDTH,
				    5 * VIEWICON_SIZE + 7 * DEFAULT_BDWIDTH,
					VIEWICON_SIZE, VIEWICON_SIZE,
		      "", movepoint_bits, movepoint_width, movepoint_height,
					OUTPUTSTATUS_MOVEPOINT,
					ICON_SELECTABLE);
  *actionselecticons += new SelectIcon (display, gc, this, fontstruct,
			    width - 2 * VIEWICON_SIZE - 3 * DEFAULT_BDWIDTH,
				    5 * VIEWICON_SIZE + 7 * DEFAULT_BDWIDTH,
					VIEWICON_SIZE, VIEWICON_SIZE,
		"", selectpoint_bits, selectpoint_width, selectpoint_height,
					OUTPUTSTATUS_SELECTPOINT,
					ICON_SELECTABLE);
  *actionselecticons += new SelectIcon (display, gc, this, fontstruct,
				width - VIEWICON_SIZE - 2 * DEFAULT_BDWIDTH,
				    6 * VIEWICON_SIZE + 8 * DEFAULT_BDWIDTH,
					VIEWICON_SIZE, VIEWICON_SIZE,
		"", createpoint_bits, createpoint_width, createpoint_height,
					OUTPUTSTATUS_CREATEPOINT,
					ICON_SELECTABLE);
  *actionselecticons += new SelectIcon (display, gc, this, fontstruct,
			    width - 2 * VIEWICON_SIZE - 3 * DEFAULT_BDWIDTH,
				    6 * VIEWICON_SIZE + 8 * DEFAULT_BDWIDTH,
					VIEWICON_SIZE, VIEWICON_SIZE,
		"", deletepoint_bits, deletepoint_width, deletepoint_height,
					OUTPUTSTATUS_DELETEPOINT,
					ICON_SELECTABLE);

  resultdisplaytyp = new SelectIconlist ();
  *resultdisplaytyp += new SelectIcon (display, gc, this, fontstruct,
				width - VIEWICON_SIZE - 2 * DEFAULT_BDWIDTH,
				   7 * VIEWICON_SIZE + 10 * DEFAULT_BDWIDTH,
				       VIEWICON_SIZE, VIEWICON_SIZE,
  "", drawonlyoriginal_bits, drawonlyoriginal_width, drawonlyoriginal_height,
				       OUTPUT_DRAWONLYORIGINAL_SELECT,
				       0);
  *resultdisplaytyp += new SelectIcon (display, gc, this, fontstruct,
			    width - 2 * VIEWICON_SIZE - 3 * DEFAULT_BDWIDTH,
				   7 * VIEWICON_SIZE + 10 * DEFAULT_BDWIDTH,
				       VIEWICON_SIZE, VIEWICON_SIZE,
       "", drawonlyresult_bits, drawonlyresult_width, drawonlyresult_height,
				       OUTPUT_DRAWONLYRESULT_SELECT,
				       0);
  *resultdisplaytyp += new SelectIcon (display, gc, this, fontstruct,
				width - VIEWICON_SIZE - 2 * DEFAULT_BDWIDTH,
				   8 * VIEWICON_SIZE + 11 * DEFAULT_BDWIDTH,
				       VIEWICON_SIZE, VIEWICON_SIZE,
			 "", drawboth_bits, drawboth_width, drawboth_height,
				       OUTPUT_DRAWBOTH_SELECT,
				       ICON_SELECTED);

  Scroll = new ScrollWindow (display, gc, this, fontstruct,
			     width - 26, 170,
			     24, 120,
			     0, 1,
			 SCROLL_BORDER_OPTION | SCROLL_COUNTERWINDOW_OPTION,
			     OUTPUT_CHANCECLASSINDEX_ACTION);

  fitscaleicon = new OnOffIcon (display,
				gc,
				this,
				fontstruct,
				0, 18, 16, 16,
				"", fitscale_bits,
				fitscale_width, fitscale_height,
				OUTPUT_FITSCALE, ICON_SELECTABLE);
  fitscaleicon->MapMe ();

  SetSelectedInput (ExposureMask
		    | StructureNotifyMask
		    | PointerMotionMask
		    | ButtonPressMask
		    | ButtonReleaseMask
		    | KeyPressMask);
}

InOutput::~InOutput ()
{
  if (axismenu != NULL)
    delete axismenu;
  delete viewicons;
  delete actionselecticons;
  delete resultdisplaytyp;
  delete pointtypicons;
  delete Scroll;
  if (ruleresultinfo != NULL)
    delete ruleresultinfo;
}

void 
InOutput::Unmapmyicons ()
{
  Scroll->UnmapMe ();
  axismenu->UnmapMe ();
  ghost->UnmapMe ();
  actionselecticons->UnmapMe ();
  viewicons->UnmapMe ();
  pointtypicons->UnmapMe ();
  resultdisplaytyp->UnmapMe ();
  fitscaleicon->UnmapMe ();
}

void 
InOutput::Mapmyicons ()
{
  if ((viewicons->GetSelectedvalue () == OUTPUT_VIEW2DCLASSMEMBERSHIP_SELECT) ||
    (viewicons->GetSelectedvalue () == OUTPUT_VIEW3DCLASSMEMBERSHIP_SELECT))
    Scroll->MapMe ();
  axismenu->MapMe ();
  ghost->MapMe ();
  actionselecticons->MapMe ();
  viewicons->MapMe ();
  pointtypicons->MapMe ();
  resultdisplaytyp->MapMe ();
  fitscaleicon->MapMe ();
}

void 
InOutput::SetDataset (Datensatz * newinputDataset)
{
  int counter;

  inputDataset = *newinputDataset;
  Dataset = &inputDataset;
  DVektorArray *inputData = &(Dataset->Lese_Daten ());
  if (Dataset->Lese_Daten_Typ () == Klassifiziert) {
    int number = inputData->Lese_Groesse (), maxclass = 0;

    for (counter = 0; counter < number; counter++)
      if (inputData->Lese_Vektor (counter).Lese_Nummer () < MAXINT)
	maxclass = MAX (maxclass, inputData->Lese_Vektor (counter).Lese_Nummer ());
    Scroll->SetMinMax (0, maxclass);
  }
  deepvalues.DVektor (inputData->Lese_Groesse (),
		      0,
		      NULL);
  deepindex.IVektor (inputData->Lese_Groesse (),
		     NULL);
//for(counter=0;counter<inputData->Lese_Groesse();counter++)
  for (counter = inputData->Lese_Groesse () - 1; counter > 0; counter--)
    deepindex[counter] = counter;

  selectedinput = &(Dataset->Lese_Daten ().Lese_Vektor (0));
  InitOutput (&(Dataset->Lese_Daten ()));
  Mapmyicons ();
}

void 
InOutput::DrawInputData (int options)
{
  Drawable drawable;
  if (options & DRAWINWINDOW)
    drawable = window;
  else
    drawable = pixmap;

// erase old drawing 
  if ((options & ERASEOLDDRAWING) || (Dataset == 0)) {
    XSetForeground (display, gc, whitepix);
    XFillRectangle (display, drawable, gc, 0, 0, width, height);
  }
  switch (viewicons->GetSelectedvalue ()) {
  case OUTPUT_VIEW2D_SELECT:
    if (Dataset->Lese_Daten ().Lese_Dim () > 1)
      DrawInputData2d (options);
    else
      DrawInputData1d (options);
    break;

  case OUTPUT_VIEW3D_SELECT:
    DrawInputData3d (options);
    break;

  case OUTPUT_VIEW2DCLASSMEMBERSHIP_SELECT:
    DrawClassMember2d (options);
    break;

  case OUTPUT_VIEW3DCLASSMEMBERSHIP_SELECT:
    DrawClassMember3d (options);
    break;
  }
};

void 
InOutput::DrawDifference2d (int options,
			    DVektorArray * originaldata,
			    DVektorArray * calculateddata)
{
  int counter, number = originaldata->Lese_Groesse ();
  Drawable drawable;

  if (options & DRAWINWINDOW)
    drawable = window;
  else
    drawable = pixmap;

//switch(typ)
  switch (pointtypicons->GetSelectedvalue ()) {
  case OUTPUT_POINTTYPPOINT:
    {
      XSegment *difflines = 0;
      XPoint *originalpoints = 0;
      XPoint *calculatedpoints = 0;

      if (((originalpoints = (XPoint *) malloc (sizeof (XPoint) * number)) == NULL) ||
	  ((calculatedpoints = (XPoint *) malloc (sizeof (XPoint) * number)) == NULL) ||
	  ((difflines = (XSegment *) malloc (sizeof (XSegment) * number)) == NULL)) {
	printf ("Out of memory!\n");
	exit (1);
      }
      for (counter = 0; counter < number; counter++) {
	FillPoint (&(originalpoints[counter]),
		   originaldata->Lese_Vektor (counter, xindex),
		   originaldata->Lese_Vektor (counter, yindex));

	FillPoint (&(calculatedpoints[counter]),
		   calculateddata->Lese_Vektor (counter, xindex),
		   calculateddata->Lese_Vektor (counter, yindex));

	difflines[counter].x1 = originalpoints[counter].x;
	difflines[counter].y1 = originalpoints[counter].y;
	difflines[counter].x2 = calculatedpoints[counter].x;
	difflines[counter].y2 = calculatedpoints[counter].y;
      }

      XSetForeground (display, gc, redpix);
      XDrawSegments (display, drawable, gc, difflines, number);
      XSetForeground (display, gc, shadowpix);
      XDrawPoints (display, drawable, gc, calculatedpoints, number, CoordModeOrigin);
      XSetForeground (display, gc, blackpix);
      XDrawPoints (display, drawable, gc, originalpoints, number, CoordModeOrigin);

      free (difflines);
      free (originalpoints);
      free (calculatedpoints);
    }
    break;			// OUTPUT_POINTTYPPOINT

  case OUTPUT_POINTTYPCROSS:
    {
      XSegment *difflines = 0;
      XSegment *originallines = 0;
      XSegment *calculatedlines = 0;
      int index = 0;
      if (((originallines = (XSegment *) malloc (sizeof (XSegment) * 2 * number)) == NULL) ||
	  ((calculatedlines = (XSegment *) malloc (sizeof (XSegment) * 2 * number)) == NULL) ||
	  ((difflines = (XSegment *) malloc (sizeof (XSegment) * number)) == NULL)) {
	printf ("Out of memory!\n");
	exit (1);
      }
      for (counter = 0; counter < 2 * number; counter += 2) {
	FillLines (&(originallines[counter]),
		   originaldata->Lese_Vektor (index, xindex),
		   originaldata->Lese_Vektor (index, yindex),
		   &difflines[index].x1, &difflines[index].y1);
	FillLines (&(calculatedlines[counter]),
		   calculateddata->Lese_Vektor (index, xindex),
		   calculateddata->Lese_Vektor (index, yindex),
		   &difflines[index].x2, &difflines[index].y2);
	index++;
      }

      XSetForeground (display, gc, redpix);
      XDrawSegments (display, drawable, gc, difflines, number);
      XSetForeground (display, gc, shadowpix);
      XDrawSegments (display, drawable, gc, calculatedlines, number * 2);
      XSetForeground (display, gc, blackpix);
      XDrawSegments (display, drawable, gc, originallines, number * 2);
      free (calculatedlines);
      free (originallines);
      free (difflines);
    }
    break;			// case OUTPUT_POINTTYPCROSS

  }
}

void 
InOutput::DrawDifference3d (int options,
			    DVektorArray * originaldata,
			    DVektorArray * calculateddata)
{
  int counter, number = originaldata->Lese_Groesse ();

  Drawable drawable;

  if (options & DRAWINWINDOW)
    drawable = window;
  else
    drawable = pixmap;

//switch(typ)
  switch (pointtypicons->GetSelectedvalue ()) {
  case OUTPUT_POINTTYPPOINT:
    {
      XSegment *difflines = 0;
      XPoint *originalpoints = 0;
      XPoint *calculatedpoints = 0;
      if (((originalpoints = (XPoint *) malloc (sizeof (XPoint) * number)) == NULL) ||
	  ((calculatedpoints = (XPoint *) malloc (sizeof (XPoint) * number)) == NULL) ||
	  ((difflines = (XSegment *) malloc (sizeof (XSegment) * number)) == NULL)) {
	printf ("Out of memory!\n");
	exit (1);
      }
      for (counter = 0; counter < number; counter++) {
	FillPoint (&(originalpoints[counter]),
		   originaldata->Lese_Vektor (counter, xindex),
		   originaldata->Lese_Vektor (counter, yindex),
		   originaldata->Lese_Vektor (counter, zindex));

	FillPoint (&(calculatedpoints[counter]),
		   calculateddata->Lese_Vektor (counter, xindex),
		   calculateddata->Lese_Vektor (counter, yindex),
		   calculateddata->Lese_Vektor (counter, zindex));
	difflines[counter].x1 = originalpoints[counter].x;
	difflines[counter].y1 = originalpoints[counter].y;
	difflines[counter].x2 = calculatedpoints[counter].x;
	difflines[counter].y2 = calculatedpoints[counter].y;
      }

      XSetForeground (display, gc, redpix);
      XDrawSegments (display, drawable, gc, difflines, number);
      XSetForeground (display, gc, shadowpix);
      XDrawPoints (display, drawable, gc, calculatedpoints, number, CoordModeOrigin);
      XSetForeground (display, gc, blackpix);
      XDrawPoints (display, drawable, gc, originalpoints, number, CoordModeOrigin);
      free (difflines);
      free (originalpoints);
      free (calculatedpoints);
    }
    break;			// case OUTPUT_POINTTYPPOINT

  case OUTPUT_POINTTYPCROSS:
    {
      XSegment *difflines = 0;
      XSegment *originallines = 0;
      XSegment *calculatedlines = 0;
      int index = 0;
      if (((originallines = (XSegment *) malloc (sizeof (XSegment) * 2 * number)) == NULL) ||
	  ((calculatedlines = (XSegment *) malloc (sizeof (XSegment) * 2 * number)) == NULL) ||
	  ((difflines = (XSegment *) malloc (sizeof (XSegment) * number)) == NULL)) {
	printf ("Out of memory!\n");
	exit (1);
      }
      for (counter = 0; counter < 2 * number; counter += 2) {
	FillLines (&(originallines[counter]),
		   originaldata->Lese_Vektor (index, xindex),
		   originaldata->Lese_Vektor (index, yindex),
		   originaldata->Lese_Vektor (index, zindex),
		   &difflines[index].x1, &difflines[index].y1);
	FillLines (&(calculatedlines[counter]),
		   calculateddata->Lese_Vektor (index, xindex),
		   calculateddata->Lese_Vektor (index, yindex),
		   calculateddata->Lese_Vektor (index, zindex),
		   &difflines[index].x2, &difflines[index].y2);

	index++;
      }
      XSetForeground (display, gc, redpix);
      XDrawSegments (display, drawable, gc, difflines, number);
      XSetForeground (display, gc, shadowpix);
      XDrawSegments (display, drawable, gc, calculatedlines, number * 2);
      XSetForeground (display, gc, blackpix);
      XDrawSegments (display, drawable, gc, originallines, number * 2);
      free (calculatedlines);
      free (originallines);
      free (difflines);
    }
    break;			// OUTPUT_POINTTYPCROSS

  }
}


void 
InOutput::UndrawPoint2d (short xpos, short ypos, int options, int number)
{
  Drawable drawable;
  if (options & DRAWINWINDOW)
    drawable = window;
  else
    drawable = pixmap;

  if (options & DRAWINWINDOW)
    XCopyArea (display, pixmap, window, gc,
	       xpos - 4 * CROSSWIDTH, ypos - 4 * CROSSWIDTH,
	       1 + 8 * CROSSWIDTH, 1 + 8 * CROSSWIDTH,
	       xpos - 4 * CROSSWIDTH, ypos - 4 * CROSSWIDTH);
  else {
    XSetForeground (display, gc, whitepix);
    switch (pointtypicons->GetSelectedvalue ()) {
    case OUTPUT_POINTTYPPOINT:
      if (options & DRAWMARKED) {
	XSegment lines[2];
	lines[0].y1 = lines[0].y2 = ypos;
	lines[0].x1 = xpos - CROSSWIDTH;
	lines[0].x2 = xpos + CROSSWIDTH;

	lines[1].x1 = lines[1].x2 = xpos;
	lines[1].y1 = ypos - CROSSWIDTH;
	lines[1].y2 = ypos + CROSSWIDTH;
	XDrawSegments (display, drawable, gc, lines, 2);
      } else
	XDrawPoint (display, drawable, gc, xpos, ypos);
      break;			// case OUTPUT_POINTTYPPOINT

    case OUTPUT_POINTTYPCROSS:
      XSegment lines[2];
      if (options & DRAWMARKED) {
	lines[0].y1 = lines[0].y2 = ypos;
	lines[0].x1 = xpos - 2 * CROSSWIDTH;
	lines[0].x2 = xpos + 2 * CROSSWIDTH;

	lines[1].x1 = lines[1].x2 = xpos;
	lines[1].y1 = ypos - 2 * CROSSWIDTH;
	lines[1].y2 = ypos + 2 * CROSSWIDTH;
      } else {
	lines[0].y1 = lines[0].y2 = ypos;
	lines[0].x1 = xpos - CROSSWIDTH;
	lines[0].x2 = xpos + CROSSWIDTH;

	lines[1].x1 = lines[1].x2 = xpos;
	lines[1].y1 = ypos - CROSSWIDTH;
	lines[1].y2 = ypos + CROSSWIDTH;
      }
      XDrawSegments (display, drawable, gc, lines, 2);
      break;			// case OUTPUT_POINTTYPCROSS

    case OUTPUT_POINTTYPCLASSES:
      XSetLineAttributes (display, gc, 3, LineSolid, CapProjecting, JoinMiter);
      DrawSymbol (xpos, ypos,
		  drawable, number);
      XSetLineAttributes (display, gc, 1, LineSolid, CapProjecting, JoinMiter);
      break;

    case OUTPUT_POINTTYPCOLOR:
//       XSetLineAttributes(display, gc, 3, LineSolid, CapProjecting, JoinMiter);
      UnDrawBox (xpos, ypos, drawable);
//       XSetLineAttributes(display, gc, 1, LineSolid, CapProjecting, JoinMiter);
      break;
    }
  }
}


void 
InOutput::DrawClassMember2d (int options)
{
  DVektorArray *inputData = &(Dataset->Lese_Daten ());
  DVektorArray *membership = &(Dataset->Lese_Zugehoerigkeiten ());
  int counter, Datanumber = inputData->Lese_Groesse ();

  Drawable drawable;

  if (options & DRAWINWINDOW)
    drawable = window;
  else
    drawable = pixmap;

  DrawAxes2d (options,
	      Dataset->Lese_Name_Pointer (xindex),
	      "U");
  XSegment *Lines;
  Lines = (XSegment *) malloc (sizeof (XSegment) * Datanumber);

  for (counter = 0; counter < Datanumber; counter++) {
    WorldtoPixel2d (inputData->Lese_Vektor (counter, xindex),
		    membership->Lese_Vektor (counter, Scroll->GetValue ()),
		    &(Lines[counter].x1), &(Lines[counter].y1));

    WorldtoPixel2d (inputData->Lese_Vektor (counter, xindex),
		    0,
		    &(Lines[counter].x2), &(Lines[counter].y2));
  }
  XDrawSegments (display, drawable, gc, Lines, Datanumber);
  free (Lines);
}

void 
InOutput::DrawClassMember3d (int options)
{
  DVektorArray *inputData = &(Dataset->Lese_Daten ());
  DVektorArray *membership = &(Dataset->Lese_Zugehoerigkeiten ());
  int counter, Datanumber = inputData->Lese_Groesse ();

  Drawable drawable;

  if (options & DRAWINWINDOW)
    drawable = window;
  else
    drawable = pixmap;

  DrawAxes3d (options,
	      Dataset->Lese_Name_Pointer (xindex),
	      Dataset->Lese_Name_Pointer (yindex),
	      "U");

  XSetForeground (display, gc, blackpix);
  XSegment *Lines;
  if ((Lines = (XSegment *) malloc (sizeof (XSegment) * Datanumber)) == NULL) {
    printf ("Out of memory!\n");
    exit (1);
  }
  for (counter = 0; counter < Datanumber; counter++) {
    WorldtoPixel3d (inputData->Lese_Vektor (counter, xindex),
		    inputData->Lese_Vektor (counter, yindex),
		    membership->Lese_Vektor (counter, Scroll->GetValue ()),
		    &(Lines[counter].x1), &(Lines[counter].y1));

    WorldtoPixel3d (inputData->Lese_Vektor (counter, xindex),
		    inputData->Lese_Vektor (counter, yindex),
		    0,
		    &(Lines[counter].x2), &(Lines[counter].y2));
  }
  XDrawSegments (display, drawable, gc, Lines, Datanumber);
  free (Lines);
}

void 
InOutput::DrawInputData1d (int options)
{
// draw axes
  DrawAxes2d (options,
	      "datano.",
	      Dataset->Lese_Name_Pointer (xindex));

  DrawDVektorArray1d (options,
		      pointtypicons->GetSelectedvalue (),
		      &(Dataset->Lese_Daten ()),
		      &(Dataset->Lese_eindeutig ()));
  if ((selectedinput != NULL) &&
      (viewicons->GetSelectedvalue () == OUTPUT_VIEW2D_SELECT))
    switch (pointtypicons->GetSelectedvalue ()) {
    case OUTPUT_POINTTYPCOLOR:
      {
	short xpos, ypos;
	Drawable drawable;

	if (options & DRAWINWINDOW)
	  drawable = window;
	else
	  drawable = pixmap;
	WorldtoPixel1d (Getvektorindex (selectedinput, &(Dataset->Lese_Daten ())),
			selectedinput->Lese_i (xindex),
			&xpos, &ypos);
//         XSetLineAttributes(display, gc, 3, LineSolid, CapProjecting, JoinMiter);
	//         XSetForeground(display,gc,redpix);
	DrawBox (xpos, ypos, drawable);
//         XSetLineAttributes(display, gc, 1, LineSolid, CapProjecting, JoinMiter);
      }
      break;

    case OUTPUT_POINTTYPCLASSES:
      {
	short xpos, ypos;
	Drawable drawable;

	if (options & DRAWINWINDOW)
	  drawable = window;
	else
	  drawable = pixmap;
	WorldtoPixel1d (Getvektorindex (selectedinput, &(Dataset->Lese_Daten ())),
			selectedinput->Lese_i (xindex),
			&xpos, &ypos);
	XSetLineAttributes (display, gc, 3, LineSolid, CapProjecting, JoinMiter);
	XSetForeground (display, gc, redpix);
	DrawSymbol (xpos, ypos,
		    drawable,
		    selectedinput->Lese_Nummer ());
	XSetLineAttributes (display, gc, 1, LineSolid, CapProjecting, JoinMiter);
      }

    default:
      DrawPoint2d (selectedinput, DRAWMARKED | options, pointtypicons->GetSelectedvalue ());
      break;
    }
}

void 
InOutput::DrawInputData2d (int options)
{
// draw axes
  DrawAxes2d (options,
	      Dataset->Lese_Name_Pointer (xindex),
	      Dataset->Lese_Name_Pointer (yindex));

  if (ruleresultinfo != NULL) {
    switch (resultdisplaytyp->GetSelectedvalue ()) {
    case OUTPUT_DRAWBOTH_SELECT:
      DrawDifference2d (options,
			&(Dataset->Lese_Daten ()),
			ruleresultinfo->result);
      break;

    case OUTPUT_DRAWONLYORIGINAL_SELECT:
      DrawDVektorArray2d (options,
			  pointtypicons->GetSelectedvalue (),
			  &(Dataset->Lese_Daten ()),
			  &(Dataset->Lese_eindeutig ()));
      break;

    case OUTPUT_DRAWONLYRESULT_SELECT:
      DrawDVektorArray2d (options | DRAWGREY,
			  pointtypicons->GetSelectedvalue (),
			  ruleresultinfo->result,
			  &(Dataset->Lese_eindeutig ()));
      break;
    }
  } else
    DrawDVektorArray2d (options,
			pointtypicons->GetSelectedvalue (),
			&(Dataset->Lese_Daten ()),
			&(Dataset->Lese_eindeutig ()));

  if ((selectedinput != NULL) &&
      (viewicons->GetSelectedvalue () == OUTPUT_VIEW2D_SELECT))
    switch (pointtypicons->GetSelectedvalue ()) {
    case OUTPUT_POINTTYPCOLOR:
      {
	short xpos, ypos;
	Drawable drawable;

	if (options & DRAWINWINDOW)
	  drawable = window;
	else
	  drawable = pixmap;
	WorldtoPixel2d (selectedinput->Lese_i (xindex),
			selectedinput->Lese_i (yindex),
			&xpos, &ypos);
//         XSetLineAttributes(display, gc, 3, LineSolid, CapProjecting, JoinMiter);
	//         XSetForeground(display,gc,redpix);
	DrawBox (xpos, ypos, drawable);
//         XSetLineAttributes(display, gc, 1, LineSolid, CapProjecting, JoinMiter);
      }
      break;

    case OUTPUT_POINTTYPCLASSES:
      {
	short xpos, ypos;
	Drawable drawable;

	if (options & DRAWINWINDOW)
	  drawable = window;
	else
	  drawable = pixmap;
	WorldtoPixel2d (selectedinput->Lese_i (xindex),
			selectedinput->Lese_i (yindex),
			&xpos, &ypos);
	XSetLineAttributes (display, gc, 3, LineSolid, CapProjecting, JoinMiter);
	XSetForeground (display, gc, redpix);
	DrawSymbol (xpos, ypos,
		    drawable,
		    selectedinput->Lese_Nummer ());
	XSetLineAttributes (display, gc, 1, LineSolid, CapProjecting, JoinMiter);
      }
    default:
      DrawPoint2d (selectedinput, DRAWMARKED | options, pointtypicons->GetSelectedvalue ());
      break;
    }
}

void 
InOutput::DrawInputData3d (int options)
{
// draw axes
  DrawAxes3d (options,
	      Dataset->Lese_Name_Pointer (xindex),
	      Dataset->Lese_Name_Pointer (yindex),
	      Dataset->Lese_Name_Pointer (zindex));

  if (ruleresultinfo != NULL) {
    switch (resultdisplaytyp->GetSelectedvalue ()) {
    case OUTPUT_DRAWBOTH_SELECT:
      DrawDifference3d (options,
			&(Dataset->Lese_Daten ()),
			ruleresultinfo->result);
      break;

    case OUTPUT_DRAWONLYORIGINAL_SELECT:
      DrawDVektorArray3d (options,
			  pointtypicons->GetSelectedvalue (),
			  &(Dataset->Lese_Daten ()),
			  &(Dataset->Lese_eindeutig ()));
      break;

    case OUTPUT_DRAWONLYRESULT_SELECT:
      DrawDVektorArray3d (options | DRAWGREY,
			  pointtypicons->GetSelectedvalue (),
			  ruleresultinfo->result,
			  &(Dataset->Lese_eindeutig ()));
      break;
    }
  } else
    DrawDVektorArray3d (options,
			pointtypicons->GetSelectedvalue (),
			&(Dataset->Lese_Daten ()),
			&(Dataset->Lese_eindeutig ()));

  if ((selectedinput != NULL) &&
      (viewicons->GetSelectedvalue () == OUTPUT_VIEW3D_SELECT))
    switch (pointtypicons->GetSelectedvalue ()) {
    case OUTPUT_POINTTYPCOLOR:
      {
	short xpos, ypos;
	Drawable drawable;

	if (options & DRAWINWINDOW)
	  drawable = window;
	else
	  drawable = pixmap;
	WorldtoPixel3d (selectedinput->Lese_i (xindex),
			selectedinput->Lese_i (yindex),
			selectedinput->Lese_i (zindex),
			&xpos, &ypos);
//         XSetLineAttributes(display, gc, 3, LineSolid, CapProjecting, JoinMiter);
	//         XSetForeground(display,gc,redpix);
	DrawBox (xpos, ypos, drawable);
//         XSetLineAttributes(display, gc, 1, LineSolid, CapProjecting, JoinMiter);
      }
      break;

    case OUTPUT_POINTTYPCLASSES:
      {
	short xpos, ypos;
	Drawable drawable;

	if (options & DRAWINWINDOW)
	  drawable = window;
	else
	  drawable = pixmap;
	WorldtoPixel3d (selectedinput->Lese_i (xindex),
			selectedinput->Lese_i (yindex),
			selectedinput->Lese_i (zindex),
			&xpos, &ypos);
	XSetForeground (display, gc, redpix);
	XSetLineAttributes (display, gc, 3, LineSolid, CapProjecting, JoinMiter);
	DrawSymbol (xpos, ypos,
		    drawable,
		    selectedinput->Lese_Nummer ());
	XSetLineAttributes (display, gc, 1, LineSolid, CapProjecting, JoinMiter);
      }

    default:
      DrawPoint3d (selectedinput, DRAWMARKED | options, pointtypicons->GetSelectedvalue ());
      break;
    }
}

void 
InOutput::InitOutput (DVektorArray * inputData)
{
  int dimnumber = Dataset->Lese_DatenDim ();
  int counter;

  if (axismenu != NULL)
    delete axismenu;
  axismenu = new Menu (display, gc, this, fontstruct, 18, 0);
  MenuIconInfo *Menuinfo;
  Menuinfo = (MenuIconInfo *) malloc (sizeof (MenuIconInfo) * (dimnumber + 1));

  Menuinfo[0].bartext = "axis1";
  Menuinfo[0].barstatus = STATUS_DEFAULT;
  Menuinfo[0].actionnumber = OUTPUT_CHANCEINDEX1_ACTION;
  Menuinfo[0].value = 0;

  for (counter = 1; counter <= dimnumber; counter++) {
    Menuinfo[counter].bartext = Dataset->Lese_Name_Pointer (counter - 1);
    Menuinfo[counter].barstatus = STATUS_DEFAULT | STATUS_MARKABLE;
    Menuinfo[counter].actionnumber = OUTPUT_CHANCEINDEX1_ACTION;
    Menuinfo[counter].value = counter - 1;
  }
  Menuinfo[1].barstatus |= STATUS_MARKED;

  if (dimnumber < 2)
    Menuinfo[0].barstatus |= STATUS_UNSELECTABLE;
  axismenu->AddPulldownMenu (Menuinfo, dimnumber + 1);

  Menuinfo[0].bartext = "axis2";
  Menuinfo[0].actionnumber = OUTPUT_CHANCEINDEX2_ACTION;
  if (dimnumber < 2)
    Menuinfo[0].barstatus |= STATUS_UNSELECTABLE;
  for (counter = 1; counter <= dimnumber; counter++)
    Menuinfo[counter].actionnumber = OUTPUT_CHANCEINDEX2_ACTION;
  Menuinfo[1].barstatus &= ~STATUS_MARKED;
  Menuinfo[2].barstatus |= STATUS_MARKED;
  axismenu->AddPulldownMenu (Menuinfo, dimnumber + 1);

  Menuinfo[0].bartext = "axis3";
  Menuinfo[0].actionnumber = OUTPUT_CHANCEINDEX3_ACTION;
  if (dimnumber < 3)
    Menuinfo[0].barstatus |= STATUS_UNSELECTABLE;
  Menuinfo[2].barstatus &= ~STATUS_MARKED;
  Menuinfo[dimnumber].barstatus |= STATUS_MARKED;
  for (counter = 1; counter <= dimnumber; counter++)
    Menuinfo[counter].actionnumber = OUTPUT_CHANCEINDEX3_ACTION;

  axismenu->AddPulldownMenu (Menuinfo, dimnumber + 1);
  viewicons->MakeIconSelectable (OUTPUT_VIEW2D_SELECT);

  if (dimnumber > 2) {
    viewicons->MakeIconSelectable (OUTPUT_VIEW3D_SELECT);
    viewicons->SelectanIcon (OUTPUT_VIEW3D_SELECT);
    SetIndex (0, 1, dimnumber - 1);
    actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_MOVEPOINT);
    actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_CREATEPOINT);
    XDefineCursor (display, window, maincursor);
  } else {
    SetIndex (0, 1);
    viewicons->SelectanIcon (OUTPUT_VIEW2D_SELECT);
    actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_ROLL);
    actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_MOVE);
    actionselecticons->SelectanIcon (OUTPUTSTATUS_MOVEPOINT);
    Action (OUTPUTSTATUS_MOVEPOINT, 0);
    XDefineCursor (display, window, movepointcursor);
  }
  free (Menuinfo);

  if (Dataset->Lese_Daten_Typ () == Klassifiziert) {
    pointtypicons->MakeIconSelectable (OUTPUT_POINTTYPCLASSES);
    pointtypicons->MakeIconSelectable (OUTPUT_POINTTYPCOLOR);
    viewicons->MakeIconSelectable (OUTPUT_VIEW2DCLASSMEMBERSHIP_SELECT);
    if (dimnumber > 1)
      viewicons->MakeIconSelectable (OUTPUT_VIEW3DCLASSMEMBERSHIP_SELECT);
  } else {
    pointtypicons->MakeIconUnselectable (OUTPUT_POINTTYPCLASSES);
    pointtypicons->MakeIconUnselectable (OUTPUT_POINTTYPCOLOR);
  }

//if(inputData->Lese_Dim()>2)
  Calculate3dparameter (RESET_AXES);

  Action (viewicons->GetSelectedvalue (), 0);
}

void 
InOutput::Action (int actionnumber, int value)
{
  switch (actionnumber) {
  case OUTPUT_FITSCALE:
    Action (viewicons->GetSelectedvalue (), 0);
    break;

  case OUTPUT_REDRAW:
    DrawInputData (ERASEOLDDRAWING);
    break;

  case OUTPUT_RESIZED:
    viewicons->MoveSelectIcons (value, 0);
    pointtypicons->MoveSelectIcons (value, 0);
    actionselecticons->MoveSelectIcons (value, 0);
    resultdisplaytyp->MoveSelectIcons (value, 0);
    Scroll->MoveMe (value, 0);
    Action (viewicons->GetSelectedvalue (), 0);
    break;

  case OUTPUT_REDRAWTOWINDOW:
    DrawInputData (ERASEOLDDRAWING | DRAWINWINDOW);
    break;

  case OUTPUT_CHANCEINDEX1_ACTION:
    SetIndex1 (value);
    Action (viewicons->GetSelectedvalue (), 0);
    break;

  case OUTPUT_CHANCEINDEX2_ACTION:
    SetIndex2 (value);
    Action (viewicons->GetSelectedvalue (), 0);
    break;

  case OUTPUT_CHANCEINDEX3_ACTION:
    SetIndex3 (value);
    Action (viewicons->GetSelectedvalue (), 0);
    break;

  case OUTPUT_VIEW2D_SELECT:
    Calculate_scale_2d (&(Dataset->Lese_Daten ()));
    Calculate2dparameter (&(Dataset->Lese_Daten ()));
    Scroll->UnmapMe ();
    axismenu->DelMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX1_ACTION);
    axismenu->DelMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX2_ACTION);
    axismenu->AddMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX3_ACTION);
    Action (OUTPUT_REDRAW, 0);
    UpdateOutput ();
    actionselecticons->MakeIconSelectable (OUTPUTSTATUS_MOVEPOINT);
    actionselecticons->MakeIconSelectable (OUTPUTSTATUS_CREATEPOINT);
    actionselecticons->MakeIconSelectable (OUTPUTSTATUS_SELECTPOINT);
    actionselecticons->MakeIconSelectable (OUTPUTSTATUS_DELETEPOINT);
    if ((actionselecticons->GetSelectedvalue () == OUTPUTSTATUS_ROLL) ||
	(actionselecticons->GetSelectedvalue () == OUTPUTSTATUS_MOVE)) {
      actionselecticons->SelectanIcon (OUTPUTSTATUS_MOVEPOINT);
      Action (OUTPUTSTATUS_MOVEPOINT, 0);
    }
    actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_ROLL);
    actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_MOVE);
    break;

  case OUTPUT_VIEW3D_SELECT:
    {
      DVektorArray *inputData = &(Dataset->Lese_Daten ());
      Calculate_scale_3d (MAX (inputData->Lese_Max (xindex), -inputData->Lese_Min (xindex)),
	  MAX (inputData->Lese_Max (yindex), -inputData->Lese_Min (yindex)),
	 MAX (inputData->Lese_Max (zindex), -inputData->Lese_Min (zindex)));
    }
    Scroll->UnmapMe ();
    axismenu->DelMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX1_ACTION);
    axismenu->DelMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX2_ACTION);
    axismenu->DelMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX3_ACTION);
    Action (OUTPUT_REDRAW, 0);
    UpdateOutput ();
    actionselecticons->MakeIconSelectable (OUTPUTSTATUS_ROLL);
    actionselecticons->MakeIconSelectable (OUTPUTSTATUS_MOVE);
    actionselecticons->MakeIconSelectable (OUTPUTSTATUS_SELECTPOINT);
    actionselecticons->MakeIconSelectable (OUTPUTSTATUS_DELETEPOINT);
    if ((actionselecticons->GetSelectedvalue () == OUTPUTSTATUS_MOVEPOINT) ||
     (actionselecticons->GetSelectedvalue () == OUTPUTSTATUS_CREATEPOINT)) {
      actionselecticons->SelectanIcon (OUTPUTSTATUS_ROLL);
      Action (OUTPUTSTATUS_ROLL, 0);
    }
    actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_MOVEPOINT);
    actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_CREATEPOINT);
    break;

  case OUTPUT_VIEW2DCLASSMEMBERSHIP_SELECT:
    Scroll->MapMe ();
    {
      DVektorArray *inputData = &(Dataset->Lese_Daten ());
      Calculate_scale_2d (MAX (inputData->Lese_Max (xindex), -inputData->Lese_Min (xindex)));
      Calculate2dparameter (inputData->Lese_Min (xindex),
			    0.0,
			    inputData->Lese_Max (xindex),
			    1.0);
    }
    axismenu->DelMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX1_ACTION);
    axismenu->AddMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX2_ACTION);
    axismenu->AddMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX3_ACTION);
    Action (OUTPUT_REDRAW, 0);
    UpdateOutput ();
    actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_MOVEPOINT);
    actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_CREATEPOINT);
    actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_MOVE);
    actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_ROLL);
    actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_SELECTPOINT);
    actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_DELETEPOINT);
    break;

  case OUTPUT_VIEW3DCLASSMEMBERSHIP_SELECT:
    {
      DVektorArray *inputData = &(Dataset->Lese_Daten ());
      Calculate_scale_3d (MAX (inputData->Lese_Max (xindex), -inputData->Lese_Min (xindex)),
	 MAX (inputData->Lese_Max (yindex), -inputData->Lese_Min (yindex)));
//       if(Dataset->Lese_Dim()<3)
      Calculate3dparameter (0);
    }
    Scroll->MapMe ();
    axismenu->DelMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX1_ACTION);
    axismenu->DelMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX2_ACTION);
    axismenu->AddMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX3_ACTION);
    Action (OUTPUT_REDRAW, 0);
    UpdateOutput ();
    actionselecticons->MakeIconSelectable (OUTPUTSTATUS_ROLL);
    actionselecticons->MakeIconSelectable (OUTPUTSTATUS_MOVE);
    actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_MOVEPOINT);
    actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_CREATEPOINT);
    actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_SELECTPOINT);
    actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_DELETEPOINT);
    if ((actionselecticons->GetSelectedvalue () != OUTPUTSTATUS_MOVE) &&
	(actionselecticons->GetSelectedvalue () != OUTPUTSTATUS_ROLL)) {
      actionselecticons->SelectanIcon (OUTPUTSTATUS_ROLL);
      Action (OUTPUTSTATUS_ROLL, 0);
    }
    break;

  case OUTPUT_CHANCECLASSINDEX_ACTION:
  case OUTPUT_POINTTYPPOINT:
  case OUTPUT_POINTTYPCROSS:
  case OUTPUT_POINTTYPCLASSES:
  case OUTPUT_POINTTYPCOLOR:
  case OUTPUT_DRAWONLYRESULT_SELECT:
  case OUTPUT_DRAWONLYORIGINAL_SELECT:
  case OUTPUT_DRAWBOTH_SELECT:
    Action (OUTPUT_REDRAW, 0);
    UpdateOutput ();
    break;

  case OUTPUTSTATUS_ROLL:
  case OUTPUTSTATUS_MOVE:
  case OUTPUTSTATUS_DELETEPOINT:
  case OUTPUTSTATUS_SELECTPOINT:
    XDefineCursor (display, window, maincursor);
    break;

  case OUTPUTSTATUS_MOVEPOINT:
    XDefineCursor (display, window, movepointcursor);
    break;

  case OUTPUTSTATUS_CREATEPOINT:
    XDefineCursor (display, window, createpointcursor);
    break;


  case OUTPUT_GHOST:
    Unmapmyicons ();
    break;

  case OUTPUT_ICONREQUESTFORRULERESULTINFO:
    if (ruleresultinfo == NULL)
      ruleresultinfo = new RuleResultInfo (Dataset);
    *((RuleResultInfo **) value) = ruleresultinfo;
    resultdisplaytyp->MakeallSelectable ();
    if ((pointtypicons->GetSelectedvalue () == OUTPUT_POINTTYPCLASSES) ||
	(pointtypicons->GetSelectedvalue () == OUTPUT_POINTTYPCOLOR)) {
      pointtypicons->SelectanIcon (OUTPUT_POINTTYPCROSS);
    }
    pointtypicons->MakeIconUnselectable (OUTPUT_POINTTYPCLASSES);
    pointtypicons->MakeIconUnselectable (OUTPUT_POINTTYPCOLOR);
    break;

  case OUTPUT_ICONLOSTINPUTCONNECTION:
    delete ruleresultinfo;
    ruleresultinfo = NULL;
    resultdisplaytyp->MakeallUnselectable ();
    Action (OUTPUT_REDRAW, 0);
    UpdateOutput ();
/*       if(Dataset->Lese_Daten_Typ()==Klassifiziert)
   {
   pointtypicons->MakeIconSelectable(OUTPUT_POINTTYPCLASSES);
   pointtypicons->MakeIconSelectable(OUTPUT_POINTTYPCOLOR);
   viewicons->MakeIconSelectable(OUTPUT_VIEW2DMEMBERSHIP_SELECT);
   if(Dataset->Lese_DatenDim()>1)
   viewicons->MakeIconSelectable(OUTPUT_VIEW3DMEMBERSHIP_SELECT);
   } */
    break;

  case OUTPUT_SAVEOBJECT:
    if ((char *) value != NULL) {
      if (ruleresultinfo != NULL) {
	switch (resultdisplaytyp->GetSelectedvalue ()) {
	case OUTPUT_DRAWBOTH_SELECT:
	case OUTPUT_DRAWONLYORIGINAL_SELECT:
	  inputDataset.File_speichern ((char *) value);
	  break;

	case OUTPUT_DRAWONLYRESULT_SELECT:
	  {
	    Datensatz tempdataset = inputDataset;
	    tempdataset.Lese_Daten () = *(ruleresultinfo->result);
	    tempdataset.File_speichern ((char *) value);
	  }
	  break;
	}
      } else
	inputDataset.File_speichern ((char *) value);
    }
    myicon->GetMainwindow ()->SetStatus (MAINNORMALSTATUS);
    break;

  case OUTPUT_CLASSIFIED:
    Scroll->SetMinMax (0, value - 1);
    pointtypicons->MakeIconSelectable (OUTPUT_POINTTYPCLASSES);
    pointtypicons->MakeIconSelectable (OUTPUT_POINTTYPCOLOR);
    viewicons->MakeIconSelectable (OUTPUT_VIEW2DCLASSMEMBERSHIP_SELECT);
    if (Dataset->Lese_DatenDim () > 1)
      viewicons->MakeIconSelectable (OUTPUT_VIEW3DCLASSMEMBERSHIP_SELECT);
    Action (OUTPUT_REDRAW, 0);
    UpdateOutput ();
    break;
  }
}

void 
InOutput::HandleEvent (XEvent * Event)
{
  static Time lastpresstime = 0;
  static InOutput *lastpressoutput = NULL;

  static short oldx, oldy;
  static DVektor *nearest = NULL;
  static char buttonpressedstatus = 0, acutebutton = 0;
  switch (Event->type) {
  case Expose:
    XCopyArea (display, pixmap, window, gc,
	       Event->xgraphicsexpose.x, Event->xgraphicsexpose.y,
	       Event->xgraphicsexpose.width, Event->xgraphicsexpose.height,
	       Event->xgraphicsexpose.x, Event->xgraphicsexpose.y);
    break;

  case UnmapNotify:
    XWithdrawWindow (display, window, DefaultScreen (display));
    break;

  case ConfigureNotify:
    if ((Event->xconfigure.width != width) ||
	(Event->xconfigure.height != height)) {
      short oldwidth = width;
      Resized (Event->xconfigure.width,
	       Event->xconfigure.height);
      Action (OUTPUT_RESIZED, width - oldwidth);
//         Action(OUTPUT_REDRAW,0);
    }
    break;			//case ConfigureNotify

  case ButtonPress:
    RaiseMe ();
    switch (actionselecticons->GetSelectedvalue ()) {
    case OUTPUTSTATUS_ROLL:
      oldx = Event->xbutton.x;
      oldy = Event->xbutton.y;
      switch (Event->xbutton.button) {
      case Button1:
	buttonpressedstatus |= BUTTON1PRESSED;
	acutebutton = Button1;
	break;

      case Button2:
	buttonpressedstatus |= BUTTON2PRESSED;
	acutebutton = Button2;
	break;

      case Button3:
	buttonpressedstatus |= BUTTON3PRESSED;
	acutebutton = Button3;
	break;
      }

      XDefineCursor (display, window, rotatecursor);
      break;			//case OUTPUTSTATUS_ROLL

    case OUTPUTSTATUS_MOVE:
      oldx = Event->xbutton.x;
      oldy = Event->xbutton.y;
      switch (Event->xbutton.button) {
      case Button1:
	buttonpressedstatus |= BUTTON1PRESSED;
	acutebutton = Button1;
	break;

      case Button2:
	buttonpressedstatus |= BUTTON2PRESSED;
	acutebutton = Button2;
	break;

      case Button3:
	buttonpressedstatus |= BUTTON3PRESSED;
	acutebutton = Button3;
	axisframe[0][3] = -0.5;
	axisframe[1][3] = -0.5;
	axisframe[2][3] = -1.0;
	Action (OUTPUT_REDRAW, 0);
	UpdateOutput ();
	break;
      }

      if ((Event->xbutton.button == Button1) ||
	  (Event->xbutton.button == Button2))
	XDefineCursor (display, window, moveaxescursor);
      break;			//case OUTPUTSTATUS_MOVE

    case OUTPUTSTATUS_MOVEPOINT:
      if (actionselecticons->IsMapped ()) {
	if (Dataset->Lese_Daten ().Lese_Dim () > 1) {
	  int dummy;
	  short options = 0;
	  if (Event->xbutton.button == Button1)
	    buttonpressedstatus |= BUTTON1PRESSED;

	  nearest = Getnearestpoint2d (Event->xbutton.x, Event->xbutton.y,
				       &dummy, &oldx, &oldy);
	  XWarpPointer (display, window, window,
			0, 0, width, height, oldx, oldy);
	  if (nearest == selectedinput)
	    options = DRAWMARKED;
	  UndrawPoint2d (oldx, oldy, options, nearest->Lese_Nummer ());
	  DrawAxes2d (0);	// perhaps one of the axes is destroyed

	  UndrawPoint2d (oldx, oldy, DRAWINWINDOW | options, nearest->Lese_Nummer ());
	  DrawPoint (oldx, oldy, DRAWINWINDOW | DRAWTEMP, nearest->Lese_Nummer (), pointtypicons->GetSelectedvalue ());
	} else {
	  int dummy;
	  short options = 0;
	  if (Event->xbutton.button == Button1)
	    buttonpressedstatus |= BUTTON1PRESSED;

	  nearest = Getnearestpoint1d (Event->xbutton.x, Event->xbutton.y,
				       &dummy, &oldx, &oldy);
	  XWarpPointer (display, window, window,
			0, 0, width, height, oldx, oldy);
	  if (nearest == selectedinput)
	    options = DRAWMARKED;
	  UndrawPoint2d (oldx, oldy, options, nearest->Lese_Nummer ());
	  DrawAxes2d (0);	// perhaps one of the axes is destroyed

	  UndrawPoint2d (oldx, oldy, DRAWINWINDOW | options, nearest->Lese_Nummer ());
	  DrawPoint (oldx, oldy, DRAWINWINDOW | DRAWTEMP, nearest->Lese_Nummer (), pointtypicons->GetSelectedvalue ());
	}
      }
      break;			// case OUTPUTSTATUS_MOVEPOINT

    case OUTPUTSTATUS_CREATEPOINT:
      if ((Event->xbutton.button == Button1) &&
	  (actionselecticons->IsMapped ())) {
	DVektorArray *inputData = &(Dataset->Lese_Daten ());
	int dim = Dataset->Lese_Daten ().Lese_Dim ();
	DVektor *newdate = new DVektor;
	newdate->Setze_Dim (dim);
	if (dim > 1) {
	  double newx, newy;
	  PixeltoWorld2d (Event->xmotion.x, Event->xmotion.y,
			  &newx, &newy);
	  newdate->Setze_i (xindex, newx);
	  newdate->Setze_i (yindex, newy);
//                  Dataset->Lese_Daten()+=(*newdate);
	} else {
	  double newx;
	  int dummy;
	  PixeltoWorld1d (Event->xmotion.x, Event->xmotion.y,
			  &dummy, &newx);
	  newdate->Setze_i (xindex, newx);
//                  Dataset->Lese_Daten()+=(*newdate);
	}
	*Dataset += (*newdate);

	// add membership info of data
	//                DVektorArray *membership=&(Dataset->Lese_Zugehoerigkeiten());
	//                DVektor *newmembership=new DVektor(membership->Lese_Dim(),0,NULL);
	//                (*membership)+=(*newmembership);

	// modify z-buffer infos
	{
	  short counter;
	  deepvalues.DVektor (inputData->Lese_Groesse (),
			      0,
			      NULL);
	  deepindex.IVektor (inputData->Lese_Groesse (),
			     NULL);
	  for (counter = inputData->Lese_Groesse () - 1; counter > 0; counter--)
//                for(counter=0;counter<inputData->Lese_Groesse();counter++)
	    deepindex[counter] = counter;
	}

	if (ruleresultinfo != NULL)
	  ruleresultinfo->result->Setze_Groesse (Dataset->Lese_Groesse ());

	Dataset->Lese_eindeutig ().Setze_Dim (Dataset->Lese_Groesse ());
	Dataset->Lese_eindeutig () = TRUE;
	myicon->Action (ICON_DATASETMODIFIED, (int) ruleresultinfo);
	Action (OUTPUT_REDRAW, 0);
	UpdateOutput ();
      }
      break;			// case OUTPUTSTATUS_DELETEPOINT

    case OUTPUTSTATUS_DELETEPOINT:
      if (actionselecticons->IsMapped ()) {
	DVektorArray *inputData = &(Dataset->Lese_Daten ());
	DVektor *nearest = 0;
	int index;
	if (Event->xbutton.button == Button1) {
	  switch (viewicons->GetSelectedvalue ()) {
	  case OUTPUT_VIEW2D_SELECT:
	    if (Dataset->Lese_Daten ().Lese_Dim () > 1)
	      nearest = Getnearestpoint2d (Event->xbutton.x, Event->xbutton.y,
					   &index, &oldx, &oldy);
	    else
	      nearest = Getnearestpoint1d (Event->xbutton.x, Event->xbutton.y,
					   &index, &oldx, &oldy);
	    break;

	  case OUTPUT_VIEW3D_SELECT:
	    nearest = Getnearestpoint3d (Event->xbutton.x, Event->xbutton.y,
					 &index, &oldx, &oldy);
	    break;
	  }
	  if (Dataset->Lese_Daten ().Lese_Groesse () > 1) {
//                  Dataset->Lese_Daten()-=(*nearest);
	    *Dataset -= (*nearest);
	    // delete membership info of data
	    //                  DVektorArray *membership=&(Dataset->Lese_Zugehoerigkeiten());
	    //                  DVektor *nearestmembership=&(membership->Lese_Vektor(index));
	    //                  (*membership)-=(*nearestmembership);

	    // modify z-buffer infos
	    {
	      short counter;
	      deepvalues.DVektor (inputData->Lese_Groesse (),
				  0,
				  NULL);
	      deepindex.IVektor (inputData->Lese_Groesse (),
				 NULL);
	      for (counter = inputData->Lese_Groesse () - 1; counter > 0; counter--)
//                  for(counter=0;counter<inputData->Lese_Groesse();counter++)
		deepindex[counter] = counter;
	    }

	    if (nearest == selectedinput) {
	      selectedinput = &(Dataset->Lese_Daten (0));
	      myicon->Action (ICON_SENDINGDVECTOR, (int) selectedinput);
	    }
	    myicon->Action (ICON_DATASETMODIFIED, (int) ruleresultinfo);
	    Action (OUTPUT_REDRAW, 0);
	    UpdateOutput ();
	  }
	  Dataset->Lese_eindeutig ().Setze_Dim (Dataset->Lese_Groesse ());
	  Dataset->Lese_eindeutig () = TRUE;

	}
      }
      break;			// case OUTPUTSTATUS_DELETEPOINT

    case OUTPUTSTATUS_SELECTPOINT:
      {
	int dummy;
	if (Event->xbutton.button == Button1) {
	  switch (viewicons->GetSelectedvalue ()) {
	  case OUTPUT_VIEW2D_SELECT:
	    if (Dataset->Lese_Daten ().Lese_Dim () > 1)
	      selectedinput = Getnearestpoint2d (Event->xbutton.x, Event->xbutton.y,
						 &dummy, &oldx, &oldy);
	    else
	      selectedinput = Getnearestpoint1d (Event->xbutton.x, Event->xbutton.y,
						 &dummy, &oldx, &oldy);
	    break;

	  case OUTPUT_VIEW3D_SELECT:
	    selectedinput = Getnearestpoint3d (Event->xbutton.x, Event->xbutton.y,
					       &dummy, &oldx, &oldy);
	    break;
	  }
	  Action (OUTPUT_REDRAW, 0);
	  UpdateOutput ();
	  myicon->Action (ICON_SENDINGDVECTOR, (int) selectedinput);
	}
      }
      break;
    }
    if (Event->xbutton.button == Button1) {
      if (((Event->xbutton.time - lastpresstime) < DOUBLECLICKTIME) &&
	  (this == lastpressoutput))
	Mapmyicons ();
      lastpresstime = Event->xbutton.time;
      lastpressoutput = this;
    }
    break;			//case ButtonPress

  case ButtonRelease:
    switch (Event->xbutton.button) {
    case Button1:
      buttonpressedstatus &= ~BUTTON1PRESSED;
      break;

    case Button2:
      buttonpressedstatus &= ~BUTTON2PRESSED;
      break;

    case Button3:
      buttonpressedstatus &= ~BUTTON3PRESSED;
      break;
    }
    if (buttonpressedstatus == 0) {
      switch (actionselecticons->GetSelectedvalue ()) {
      case OUTPUTSTATUS_ROLL:
      case OUTPUTSTATUS_MOVE:
	XDefineCursor (display, window, maincursor);
	break;
      }
    }
    if (buttonpressedstatus == 0)
      Action (OUTPUT_REDRAW, 0);
    else if (buttonpressedstatus & BUTTON1PRESSED)
      acutebutton = Button1;
    else if (buttonpressedstatus & BUTTON2PRESSED)
      acutebutton = Button2;
    else
      acutebutton = Button3;

    switch (actionselecticons->GetSelectedvalue ()) {
    case OUTPUTSTATUS_MOVEPOINT:
      myicon->Action (ICON_DATASETMODIFIED, (int) ruleresultinfo);
      Action (OUTPUT_REDRAW, 0);
      UpdateOutput ();
      DrawCoordinates (Event->xbutton.x, Event->xbutton.y);
      break;
    }
    nearest = NULL;
    break;			// case ButtonRelease

  case MotionNotify:
    switch (actionselecticons->GetSelectedvalue ()) {
    case OUTPUTSTATUS_ROLL:
      if (buttonpressedstatus != 0) {
	int xdiff, ydiff;
	int newx, newy;

	newx = Event->xmotion.x;
	newy = Event->xmotion.y;

	xdiff = newx - oldx;
	ydiff = newy - oldy;

	switch (acutebutton) {
	case Button1:
	  if (xdiff != 0)
	    AxisRoty (ANGLEPERPIXEL * xdiff);
	  if (ydiff != 0)
	    AxisRotx (ANGLEPERPIXEL * ydiff);
	  break;

	case Button2:
	  if (xdiff != 0)
	    AxisRoty (ANGLEPERPIXEL * xdiff);
	  if (ydiff != 0)
	    AxisRotz (ANGLEPERPIXEL * ydiff);
	  break;

	case Button3:
	  if (xdiff != 0)
	    AxisRotz (ANGLEPERPIXEL * xdiff);
	  if (ydiff != 0)
	    AxisRotx (ANGLEPERPIXEL * ydiff);
	  break;
	}

	Action (OUTPUT_REDRAWTOWINDOW, 0);
//                if(selectedinput!=NULL)
	//                  DrawPoint3d(selectedinput,DRAWMARKED|DRAWINWINDOW,pointtypicons->GetSelectedvalue());

	oldx = newx;
	oldy = newy;

      }
      break;			// case OUTPUTSTATUS_ROLL

    case OUTPUTSTATUS_MOVE:
      if (buttonpressedstatus != 0) {
	int xdiff, ydiff;
	int newx, newy;
	newx = Event->xmotion.x;
	newy = Event->xmotion.y;

	xdiff = newx - oldx;
	ydiff = newy - oldy;

	switch (acutebutton) {
	case Button1:
	  axisframe[0][3] += MOVEFACTOR * xdiff;
	  axisframe[1][3] -= MOVEFACTOR * ydiff;
	  break;

	case Button2:
	  axisframe[2][3] -= MOVEFACTOR * xdiff;
	  axisframe[2][3] = MIN (axisframe[2][3], 1.0 / 2);
	  break;
	}

	if ((acutebutton == Button1) ||
	    (acutebutton == Button2))
	  Action (OUTPUT_REDRAWTOWINDOW, 0);

	oldx = newx;
	oldy = newy;
      }
      break;

    case OUTPUTSTATUS_MOVEPOINT:
      if (actionselecticons->IsMapped ()) {
	if (Dataset->Lese_Daten ().Lese_Dim () > 1) {
	  double newx, newy;
	  PixeltoWorld2d (Event->xmotion.x, Event->xmotion.y,
			  &newx, &newy);
	  if (buttonpressedstatus & BUTTON1PRESSED) {
	    UndrawPoint2d (oldx, oldy, DRAWINWINDOW | DRAWTEMP, nearest->Lese_Nummer ());
	    nearest->Setze_i (xindex, newx);
	    nearest->Setze_i (yindex, newy);
	    DrawPoint (Event->xmotion.x, Event->xmotion.y,
		       DRAWINWINDOW | DRAWTEMP, nearest->Lese_Nummer (),
		       pointtypicons->GetSelectedvalue ());
	  }
	  oldx = Event->xmotion.x;
	  oldy = Event->xmotion.y;
	} else {
	  double newx;
	  int dummy;
	  PixeltoWorld1d (Event->xmotion.x, Event->xmotion.y,
			  &dummy, &newx);
	  if (buttonpressedstatus & BUTTON1PRESSED) {
	    short xpixel, ypixel;
	    UndrawPoint2d (oldx, oldy, DRAWINWINDOW | DRAWTEMP, nearest->Lese_Nummer ());
	    nearest->Setze_i (xindex, newx);
	    WorldtoPixel1d (Getvektorindex (nearest, &(Dataset->Lese_Daten ())),
			    newx, &xpixel, &ypixel);
	    DrawPoint (xpixel, ypixel,
		       DRAWINWINDOW | DRAWTEMP, nearest->Lese_Nummer (),
		       pointtypicons->GetSelectedvalue ());
	    oldx = xpixel;
	    oldy = ypixel;
	  } else {
	    oldx = Event->xmotion.x;
	    oldy = Event->xmotion.y;
	  }
	}
	if (nearest == selectedinput)
	  myicon->Action (ICON_SENDINGDVECTOR, (int) selectedinput);
      } else {
	oldx = Event->xmotion.x;
	oldy = Event->xmotion.y;
      }
      DrawCoordinates (oldx, oldy);
      break;

    case OUTPUTSTATUS_CREATEPOINT:
    case OUTPUTSTATUS_DELETEPOINT:
    case OUTPUTSTATUS_SELECTPOINT:
      if (viewicons->GetSelectedvalue () == OUTPUT_VIEW2D_SELECT)
	DrawCoordinates (Event->xmotion.x, Event->xmotion.y);
      break;

    default:
      if (viewicons->GetSelectedvalue () == OUTPUT_VIEW2DCLASSMEMBERSHIP_SELECT)
	DrawCoordinates (Event->xmotion.x, Event->xmotion.y);
      break;
    }
    break;			//case MotionNotify:

  case KeyPress:
    /* 
     * process key hit:
     */
    char c;
    int n = XLookupString((XKeyEvent *) Event, &c, 1, NULL, NULL);
    if (n == 1 && c == 'p') {
      DumpBitmap();
    }

    break;

  }
}


ClusterOutput::ClusterOutput (Display * initdisplay,
	       GC initgc,
	       ClusterIcon * initOutputicon,
	       XFontStruct * fontstruct,
	       int initx, int inity, int initwidth, int initheight,
	       char *text):
Output2d_3d (initdisplay,
	     initgc,
	     initOutputicon,
	     fontstruct,
	     initx, inity, initwidth, initheight,
	     text)
{
  selectedinput = NULL;
  Scroll = NULL;
  xindex = DEFAULTXINDEX;
  yindex = DEFAULTYINDEX;
  zindex = DEFAULTZINDEX;
  axismenu = NULL;
  exactness = 10;
  optionwindow = NULL;
  ellipsis_2d = NULL;
  ellipsis_3d = NULL;
  ellipsisnumber = 0;
  clustering = NULL;
  float_to_pixel = 1.0;
  selectedcluster = new BVektor;

  viewicons = new SelectIconlist ();
  *viewicons += new SelectIcon (display, gc, this, fontstruct,
			 width - 2 * VIEWICON_SIZE - 3 * DEFAULT_BDWIDTH, 0,
				VIEWICON_SIZE, VIEWICON_SIZE,
		      "", d2cluster_bits, d2cluster_width, d2cluster_height,
				OUTPUT_VIEW2D_SELECT,
				ICON_SELECTED);
  *viewicons += new SelectIcon (display, gc, this, fontstruct,
			     width - VIEWICON_SIZE - 2 * DEFAULT_BDWIDTH, 0,
				VIEWICON_SIZE, VIEWICON_SIZE,
		      "", d3cluster_bits, d3cluster_width, d3cluster_height,
				OUTPUT_VIEW3D_SELECT,
				0);
  *viewicons += new SelectIcon (display, gc, this, fontstruct,
				width - 2 * VIEWICON_SIZE - 3 * DEFAULT_BDWIDTH, VIEWICON_SIZE + DEFAULT_BDWIDTH,
				VIEWICON_SIZE, VIEWICON_SIZE,
	     "", d2membertyp1_bits, d2membertyp1_width, d2membertyp1_height,
				OUTPUT_VIEW2DMEMBERTYP1_SELECT,
				0);
  *viewicons += new SelectIcon (display, gc, this, fontstruct,
				width - VIEWICON_SIZE - 2 * DEFAULT_BDWIDTH, VIEWICON_SIZE + DEFAULT_BDWIDTH,
				VIEWICON_SIZE, VIEWICON_SIZE,
	     "", d3membertyp1_bits, d3membertyp1_width, d3membertyp1_height,
				OUTPUT_VIEW3DMEMBERTYP1_SELECT,
				0);
  *viewicons += new SelectIcon (display, gc, this, fontstruct,
				width - 2 * VIEWICON_SIZE - 3 * DEFAULT_BDWIDTH, 2 * VIEWICON_SIZE + 2 * DEFAULT_BDWIDTH,
				VIEWICON_SIZE, VIEWICON_SIZE,
	     "", d2membertyp2_bits, d2membertyp2_width, d2membertyp2_height,
				OUTPUT_VIEW2DMEMBERTYP2_SELECT,
				0);
  *viewicons += new SelectIcon (display, gc, this, fontstruct,
				width - VIEWICON_SIZE - 2 * DEFAULT_BDWIDTH, 2 * VIEWICON_SIZE + 2 * DEFAULT_BDWIDTH,
				VIEWICON_SIZE, VIEWICON_SIZE,
	     "", d3membertyp2_bits, d3membertyp2_width, d3membertyp2_height,
				OUTPUT_VIEW3DMEMBERTYP2_SELECT,
				0);
  *viewicons += new SelectIcon (display, gc, this, fontstruct,
				width - 2 * VIEWICON_SIZE - 3 * DEFAULT_BDWIDTH, 3 * VIEWICON_SIZE + 3 * DEFAULT_BDWIDTH,
				VIEWICON_SIZE, VIEWICON_SIZE,
			    "", d2guete_bits, d2guete_width, d2guete_height,
				OUTPUT_VIEW2DGUETE_SELECT,
				0);

  singlemembership = new SelectIconlist ();
  *singlemembership += new SelectIcon (display, gc, this, fontstruct,
				       width - VIEWICON_SIZE - 2 * DEFAULT_BDWIDTH, 3 * VIEWICON_SIZE + 3 * DEFAULT_BDWIDTH,
				       VIEWICON_SIZE, VIEWICON_SIZE,
  "", singlemembership_bits, singlemembership_width, singlemembership_height,
				       OUTPUT_SINGLEMEMBERSHIP_SELECT,
				       0);

  pointtypicons = new SelectIconlist ();
  *pointtypicons += new SelectIcon (display, gc, this, fontstruct,
				    width - 2 * VIEWICON_SIZE - 3 * DEFAULT_BDWIDTH, 4 * VIEWICON_SIZE + 5 * DEFAULT_BDWIDTH,
				    VIEWICON_SIZE, VIEWICON_SIZE,
	  "", pointtyppoint_bits, pointtyppoint_width, pointtyppoint_height,
				    OUTPUT_POINTTYPPOINT,
				    0);
  *pointtypicons += new SelectIcon (display, gc, this, fontstruct,
				    width - VIEWICON_SIZE - 2 * DEFAULT_BDWIDTH, 4 * VIEWICON_SIZE + 5 * DEFAULT_BDWIDTH,
				    VIEWICON_SIZE, VIEWICON_SIZE,
	  "", pointtypcross_bits, pointtypcross_width, pointtypcross_height,
				    OUTPUT_POINTTYPCROSS,
				    ICON_SELECTED);

  clustericons = new SelectIconlist ();
  *clustericons += new SelectIcon (display, gc, this, fontstruct,
				   width - VIEWICON_SIZE - 2 * DEFAULT_BDWIDTH, 5 * VIEWICON_SIZE + 7 * DEFAULT_BDWIDTH,
				   VIEWICON_SIZE, VIEWICON_SIZE,
				   "", clustertypellipsis_bits, clustertypellipsis_width, clustertypellipsis_height,
				   OUTPUT_CLUSTERTYPELLIPSIS_SELECT,
				   0);
  *clustericons += new SelectIcon (display, gc, this, fontstruct,
				   width - 2 * VIEWICON_SIZE - 3 * DEFAULT_BDWIDTH, 5 * VIEWICON_SIZE + 7 * DEFAULT_BDWIDTH,
				   VIEWICON_SIZE, VIEWICON_SIZE,
  "", clustertypsymbol_bits, clustertypsymbol_width, clustertypsymbol_height,
				   OUTPUT_CLUSTERTYPSYMBOL_SELECT,
				   0);
  *clustericons += new SelectIcon (display, gc, this, fontstruct,
				   width - VIEWICON_SIZE - 2 * DEFAULT_BDWIDTH, 6 * VIEWICON_SIZE + 8 * DEFAULT_BDWIDTH,
				   VIEWICON_SIZE, VIEWICON_SIZE,
       "", clustertypstar_bits, clustertypstar_width, clustertypstar_height,
				   OUTPUT_CLUSTERTYPSTAR_SELECT,
				   ICON_SELECTED);
  *clustericons += new SelectIcon (display, gc, this, fontstruct,
				   width - 2 * VIEWICON_SIZE - 3 * DEFAULT_BDWIDTH, 6 * VIEWICON_SIZE + 8 * DEFAULT_BDWIDTH,
				   VIEWICON_SIZE, VIEWICON_SIZE,
    "", clustertypcolor_bits, clustertypcolor_width, clustertypcolor_height,
				   OUTPUT_CLUSTERTYPCOLOR_SELECT,
				   0);

  actionselecticons = new SelectIconlist ();
  *actionselecticons += new SelectIcon (display, gc, this, fontstruct,
					width - 2 * VIEWICON_SIZE - 3 * DEFAULT_BDWIDTH, 7 * VIEWICON_SIZE + 10 * DEFAULT_BDWIDTH,
					VIEWICON_SIZE, VIEWICON_SIZE,
				     "", roll_bits, roll_width, roll_height,
					OUTPUTSTATUS_ROLL,
					0);
  *actionselecticons += new SelectIcon (display, gc, this, fontstruct,
					width - VIEWICON_SIZE - 2 * DEFAULT_BDWIDTH, 7 * VIEWICON_SIZE + 10 * DEFAULT_BDWIDTH,
					VIEWICON_SIZE, VIEWICON_SIZE,
				     "", move_bits, move_width, move_height,
					OUTPUTSTATUS_MOVE,
					0);
  *actionselecticons += new SelectIcon (display, gc, this, fontstruct,
					width - 2 * VIEWICON_SIZE - 3 * DEFAULT_BDWIDTH, 8 * VIEWICON_SIZE + 11 * DEFAULT_BDWIDTH,
					VIEWICON_SIZE, VIEWICON_SIZE,
		"", movecluster_bits, movecluster_width, movecluster_height,
					OUTPUTSTATUS_MOVECLUSTER_SELECT,
					0);
  *actionselecticons += new SelectIcon (display, gc, this, fontstruct,
					width - VIEWICON_SIZE - 2 * DEFAULT_BDWIDTH, 8 * VIEWICON_SIZE + 11 * DEFAULT_BDWIDTH,
					VIEWICON_SIZE, VIEWICON_SIZE,
	  "", createcluster_bits, createcluster_width, createcluster_height,
					OUTPUTSTATUS_CREATECLUSTER_SELECT,
					0);
  *actionselecticons += new SelectIcon (display, gc, this, fontstruct,
					width - 2 * VIEWICON_SIZE - 3 * DEFAULT_BDWIDTH, 9 * VIEWICON_SIZE + 12 * DEFAULT_BDWIDTH,
					VIEWICON_SIZE, VIEWICON_SIZE,
	  "", selectcluster_bits, selectcluster_width, selectcluster_height,
					OUTPUTSTATUS_SELECTCLUSTER_SELECT,
					0);
  *actionselecticons += new SelectIcon (display, gc, this, fontstruct,
					width - VIEWICON_SIZE - 2 * DEFAULT_BDWIDTH, 9 * VIEWICON_SIZE + 12 * DEFAULT_BDWIDTH,
					VIEWICON_SIZE, VIEWICON_SIZE,
	  "", deletecluster_bits, deletecluster_width, deletecluster_height,
					OUTPUTSTATUS_DESTROYCLUSTER_SELECT,
					0);

  go = new ActionIcon (display,
		       gc,
		       this,
		       fontstruct,
		       0, 17, 16, 16,
		       "", go_bits,
		       go_width, go_height,
		       OUTPUT_STARTCLUSTERING_ACTION,
		       0);
  reset = new ActionIcon (display,
			  gc,
			  this,
			  fontstruct,
			  0, 34, 16, 16,
			  "", reset_bits,
			  reset_width, reset_height,
			  OUTPUT_RESETCLUSTERING_ACTION,
			  0);

  makerules = new ActionIcon (display,
			      gc,
			      this,
			      fontstruct,
			      0, 51, 16, 16,
			      "", makerules_bits,
			      makerules_width, makerules_height,
			      OUTPUT_CREATERULES_ACTION,
			      0);

  classify = new ActionIcon (display,
			     gc,
			     this,
			     fontstruct,
			     0, 68, 16, 16,
			     "", classify_bits,
			     classify_width, classify_height,
			     OUTPUT_CLASSIFY_ACTION,
			     0);
  save = new ActionIcon (display,
			 gc,
			 this,
			 fontstruct,
			 0, 85, 16, 16,
			 "", save_bits,
			 save_width, save_height,
			 OUTPUT_SAVE, 0);
  load = new ActionIcon (display,
			 gc,
			 this,
			 fontstruct,
			 0, 102, 16, 16,
			 "", load_bits,
			 load_width, load_height,
			 OUTPUT_LOAD, ICON_SELECTABLE);
  fitscaleicon = new OnOffIcon (display,
				gc,
				this,
				fontstruct,
				0, 119, 16, 16,
				"", fitscale_bits,
				fitscale_width, fitscale_height,
				OUTPUT_FITSCALE, ICON_SELECTABLE);
  fitscaleicon->MapMe ();

  SetSelectedInput (ExposureMask
		    | StructureNotifyMask
		    | PointerMotionMask
		    | ButtonPressMask
		    | ButtonReleaseMask
		    | KeyPressMask);

  memberoutput = new MemberOutput (display,
				   gc,
				   fontstruct,
				   10, 10, text);
}

ClusterOutput::~ClusterOutput ()
{
  if (selectedcluster != NULL)
    delete selectedcluster;
  if (Scroll != NULL)
    delete Scroll;
  delete go;
  delete reset;
  delete makerules;
  delete classify;
  short counter;
  if (ellipsis_2d != NULL) {
    for (counter = 0; counter < ellipsisnumber; counter++)
      delete ellipsis_2d[counter];
    free (ellipsis_2d);
  }
  if (ellipsis_3d != NULL) {
    for (counter = 0; counter < ellipsisnumber; counter++)
      delete ellipsis_3d[counter];
    free (ellipsis_3d);
  }
  if (axismenu != NULL)
    delete axismenu;
  if (clustering != NULL)
    delete clustering;
  delete viewicons;
  delete clustericons;
  delete actionselecticons;
  delete pointtypicons;
  if (optionwindow != NULL)
    delete optionwindow;
}

void 
ClusterOutput::Unmapmyicons ()
{
  if (axismenu != NULL)
    axismenu->UnmapMe ();
  if (Scroll != NULL)
    Scroll->UnmapMe ();
  ghost->UnmapMe ();
  go->UnmapMe ();
  save->UnmapMe ();
  load->UnmapMe ();
  reset->UnmapMe ();
  makerules->UnmapMe ();
  classify->UnmapMe ();
  viewicons->UnmapMe ();
  clustericons->UnmapMe ();
  actionselecticons->UnmapMe ();
  singlemembership->UnmapMe ();
  pointtypicons->UnmapMe ();
  fitscaleicon->UnmapMe ();
}

void 
ClusterOutput::Mapmyicons ()
{
  if (axismenu != NULL)
    axismenu->MapMe ();
  if (Scroll != NULL) {
    if ((viewicons->GetSelectedvalue () == OUTPUT_VIEW2DMEMBERTYP1_SELECT) ||
	(viewicons->GetSelectedvalue () == OUTPUT_VIEW3DMEMBERTYP1_SELECT) ||
	(viewicons->GetSelectedvalue () == OUTPUT_VIEW2DMEMBERTYP2_SELECT))
      Scroll->MapMe ();
  }
  ghost->MapMe ();
  go->MapMe ();
  load->MapMe ();
  save->MapMe ();
  reset->MapMe ();
  makerules->MapMe ();
  classify->MapMe ();
  viewicons->MapMe ();
  clustericons->MapMe ();
  actionselecticons->MapMe ();
  singlemembership->MapMe ();
  pointtypicons->MapMe ();
  fitscaleicon->MapMe ();
}

void 
ClusterOutput::SetClustering (Clustering * newClustering)
{
  if (clustering != 0) {
    if (clustering->Lese_Vorgabe_Prototypen ().Lese_Groesse () > 0)
      newClustering->Setze_Cluster (clustering->Lese_Vorgabe_Prototypen ());
    delete clustering;
  }
  clustering = newClustering;


  switch (clustering->Lese_Art ()) {
  case OPTIMAL:
    clusternumber = MAXINT;
    break;
  case ANZAHL:
    clusternumber = clustering->Lese_Clusteranzahl ();
    break;
  case KLASSEN:
    break;
  default:
    break;
  }

  if (clustering->Lese_Clusteranzahl () != 0)
    save->MakeSelectable ();
  else
    save->MakeUnselectable ();
}

void 
ClusterOutput::SetClustertyp (ClusterInfo * values)
{
  clusternumber = values->clusternumber;
  char changed = 0;
  if (clustering != NULL) {
    if (clustering->Lese_Typ () != values->clustertyp)
      changed = 1;
  } else
    changed = 1;

  if (changed) {
    Clustering *newclustering;
    switch (values->clustertyp) {
    case FCM:
      newclustering = new FCM_Clustering ();
      break;

    case GK:
      newclustering = new GK_Clustering ();
      break;

    case GK_parallel:
      newclustering = new GK_parallel_Clustering ();
      break;

    case GG:
      newclustering = new GG_Clustering ();
      break;

/*    case GG_parallel: clustering=new GG_parallel_Clustering();
   break; */
    case GG_parallel:
      newclustering = new GG_parallel_Clustering ();
      break;
#if 0
    default:
      clustering = new FCM_Clustering ();
#else
    default:
      newclustering = new FCM_Clustering ();
#endif
      break;
    }
    SetClustering (newclustering);
    myicon->DrawMoveableIcon ();
    Action (OUTPUT_RESETCLUSTERING_ACTION, 0);
  }
  if (changed || (values->validitytyp != clustering->Lese_Guete ().Lese_Typ ()))
    if (values->validitytyp != CLASSIFY)
      clustering->Setze_Guete (values->validitytyp);

#if 1
// XXX: bullshit but necessary:
  validitytyp = values->validitytyp;
#endif
  clustering->Setze_Max_Iteration (values->repeatnumber);
  clusternumber = values->clusternumber;
#if 0
  clustering.Setze_Possib (values->ispossib);
#else
  clustering->Setze_Possib (values->ispossib);
#endif
  clustering->Setze_M (values->M);
  clustering->Setze_K (values->K);
  exactness = values->exactness;
  clustering->Setze_Delta_U (values->delta_u);
}

void 
ClusterOutput::SetDataset (Datensatz * inputDataset)
{
  Dataset = inputDataset;

  if (Dataset != 0) {
    myicon->Action (ICON_SENDINGDATASETNAME, (int) myicon->GetInputConnection ()->GetStartIcon ()->Gettext ());
    go->MakeSelectable ();
    classify->MakeUnselectable ();

    BVektor tempbvector;
    tempbvector.Setze_Dim (Dataset->Lese_DatenDim ());
    tempbvector = TRUE;
    clustering->Setze_relevant (tempbvector);
    if (Dataset->Lese_DatenDim () > 2)
      SetIndex (0, 1, Dataset->Lese_DatenDim () - 1);
    else
      SetIndex (0, 1);

    InitOutput (&(Dataset->Lese_Daten ()));

    Calculate3dparameter (RESET_AXES);

    Mapmyicons ();

    DVektorArray *inputData = &(Dataset->Lese_Daten ());
    short counter;
    deepvalues.DVektor (inputData->Lese_Groesse (),
			0,
			NULL);
    deepindex.IVektor (inputData->Lese_Groesse (),
		       NULL);
    for (counter = inputData->Lese_Groesse () - 1; counter > 0; counter--)
      deepindex[counter] = counter;

    if (clustering != NULL) {
      clustering->Berechne_fuer_neue_Daten (Dataset->Lese_Daten ());
      Calculate_ellipses ();
      if (!clustering->Ist_Fertig ())
	go->MakeSelectable ();
      else
	go->MakeUnselectable ();
    }
    SelectUnselectIcons ();
    SetSelectedinputvector (selectedinput);
  } else {
    ClearPixmap ();
    UpdateOutput ();
//  Action(OUTPUT_RESETCLUSTERING_ACTION,0);
    SetSelectedinputvector (NULL);
    memberoutput->UnmapMe ();
  }

}

void 
ClusterOutput::InitOutput (DVektorArray * inputData)
{
  int dimnumber = Dataset->Lese_DatenDim ();
  int counter;
  {
    if (axismenu != NULL)
      delete axismenu;

    axismenu = new Menu (display, gc, this, fontstruct, 18, 0);
    MenuIconInfo *Menuinfo;
    Menuinfo = (MenuIconInfo *) malloc (sizeof (MenuIconInfo) * (dimnumber + 1));

    Menuinfo[0].bartext = "axis1";
    Menuinfo[0].barstatus = STATUS_DEFAULT;
    Menuinfo[0].actionnumber = OUTPUT_CHANCEINDEX1_ACTION;
    Menuinfo[0].value = 0;

    for (counter = 1; counter <= dimnumber; counter++) {
      Menuinfo[counter].bartext = Dataset->Lese_Name_Pointer (counter - 1);
      Menuinfo[counter].barstatus = STATUS_DEFAULT | STATUS_MARKABLE;
      Menuinfo[counter].actionnumber = OUTPUT_CHANCEINDEX1_ACTION;
      Menuinfo[counter].value = counter - 1;
    }
    Menuinfo[1].barstatus |= STATUS_MARKED;
    if (dimnumber < 2)
      Menuinfo[0].barstatus |= STATUS_UNSELECTABLE;

    axismenu->AddPulldownMenu (Menuinfo, dimnumber + 1);

    Menuinfo[0].bartext = "axis2";
    Menuinfo[0].actionnumber = OUTPUT_CHANCEINDEX2_ACTION;
    if (dimnumber < 2)
      Menuinfo[0].barstatus |= STATUS_UNSELECTABLE;
    for (counter = 1; counter <= dimnumber; counter++)
      Menuinfo[counter].actionnumber = OUTPUT_CHANCEINDEX2_ACTION;
    Menuinfo[1].barstatus &= ~STATUS_MARKED;
    Menuinfo[2].barstatus |= STATUS_MARKED;
    axismenu->AddPulldownMenu (Menuinfo, dimnumber + 1);

    Menuinfo[0].bartext = "axis3";
    Menuinfo[0].actionnumber = OUTPUT_CHANCEINDEX3_ACTION;
    if (dimnumber < 3)
      Menuinfo[0].barstatus |= STATUS_UNSELECTABLE;

    Menuinfo[2].barstatus &= ~STATUS_MARKED;
    Menuinfo[dimnumber].barstatus |= STATUS_MARKED;
    for (counter = 1; counter <= dimnumber; counter++)
      Menuinfo[counter].actionnumber = OUTPUT_CHANCEINDEX3_ACTION;

    axismenu->AddPulldownMenu (Menuinfo, dimnumber + 1);
    free (Menuinfo);

    if (Scroll != NULL)
      delete Scroll;

    Scroll = new ScrollWindow (display, gc, this, fontstruct,
			       width - 26, 170,
			       24, 120,
			       0, Dataset->Lese_Daten ().Lese_Groesse () - 1,
			 SCROLL_BORDER_OPTION | SCROLL_COUNTERWINDOW_OPTION,
			       OUTPUT_CHANCEDATAINDEX_ACTION);
  }

  if (clustering->Lese_Clusteranzahl () == 0) {
//  if(inputData->Lese_Dim()>2)
    Calculate3dparameter (RESET_AXES);
  }
  SelectUnselectIcons ();
  Action (viewicons->GetSelectedvalue (), 0);
}

void 
ClusterOutput::SelectUnselectIcons ()
{
  if (Dataset != NULL) {
    if (clustering->Lese_Clusteranzahl () != 0) {
      singlemembership->MakeallSelectable ();
      viewicons->MakeIconSelectable (OUTPUT_VIEW2DMEMBERTYP1_SELECT);
      viewicons->MakeIconSelectable (OUTPUT_VIEW2DMEMBERTYP2_SELECT);
      if (Dataset->Lese_DatenDim () > 1) {
	viewicons->MakeIconSelectable (OUTPUT_VIEW3DMEMBERTYP1_SELECT);
	viewicons->MakeIconSelectable (OUTPUT_VIEW3DMEMBERTYP2_SELECT);
      }
      clustericons->MakeallSelectable ();
      actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_CREATECLUSTER_SELECT);
      actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_MOVECLUSTER_SELECT);
      actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_DESTROYCLUSTER_SELECT);
      actionselecticons->MakeIconSelectable (OUTPUTSTATUS_SELECTCLUSTER_SELECT);
      if (selectedcluster->Anzahl_gesetzt () > 0) {
	makerules->MakeSelectable ();
	classify->MakeSelectable ();
      } else {
	makerules->MakeUnselectable ();
	classify->MakeUnselectable ();
      }
      if (Dataset->Lese_DatenDim () < 2)
	clustericons->MakeIconUnselectable (OUTPUT_CLUSTERTYPELLIPSIS_SELECT);
    } else {
      singlemembership->MakeallUnselectable ();
      memberoutput->UnmapMe ();
      viewicons->MakeIconUnselectable (OUTPUT_VIEW2DMEMBERTYP1_SELECT);
      viewicons->MakeIconUnselectable (OUTPUT_VIEW3DMEMBERTYP1_SELECT);
      viewicons->MakeIconUnselectable (OUTPUT_VIEW2DMEMBERTYP2_SELECT);
      viewicons->MakeIconUnselectable (OUTPUT_VIEW3DMEMBERTYP2_SELECT);
      actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_SELECTCLUSTER_SELECT);
      makerules->MakeUnselectable ();
      classify->MakeUnselectable ();
      if (Dataset->Lese_DatenDim () > 2) {
	viewicons->MakeIconSelectable (OUTPUT_VIEW3D_SELECT);
	viewicons->MakeIconSelectable (OUTPUT_VIEW2D_SELECT);
	viewicons->SelectanIcon (OUTPUT_VIEW3D_SELECT);
	actionselecticons->MakeIconSelectable (OUTPUTSTATUS_ROLL);
	actionselecticons->MakeIconSelectable (OUTPUTSTATUS_MOVE);
	actionselecticons->SelectanIcon (OUTPUTSTATUS_ROLL);
	Action (OUTPUT_VIEW3D_SELECT, 0);
	Action (OUTPUTSTATUS_ROLL, 0);
      } else {
	viewicons->MakeIconSelectable (OUTPUT_VIEW2D_SELECT);
	viewicons->SelectanIcon (OUTPUT_VIEW2D_SELECT);
	viewicons->MakeIconUnselectable (OUTPUT_VIEW3D_SELECT);
	actionselecticons->MakeIconSelectable (OUTPUTSTATUS_CREATECLUSTER_SELECT);
	actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_MOVE);
	actionselecticons->SelectanIcon (OUTPUTSTATUS_CREATECLUSTER_SELECT);
	Action (OUTPUT_VIEW2D_SELECT, 0);
	if (clustering->Lese_Vorgabe_Prototypen ().Lese_Groesse () > 0) {
	  actionselecticons->MakeIconSelectable (OUTPUTSTATUS_MOVECLUSTER_SELECT);
	  actionselecticons->MakeIconSelectable (OUTPUTSTATUS_DESTROYCLUSTER_SELECT);
	}
      }
    }

    if (Dataset->Lese_DatenDim () > 2) {
      viewicons->MakeIconSelectable (OUTPUT_VIEW3D_SELECT);
      viewicons->MakeIconSelectable (OUTPUT_VIEW2D_SELECT);
    } else
      viewicons->MakeIconSelectable (OUTPUT_VIEW2D_SELECT);
    pointtypicons->MakeIconSelectable (OUTPUT_POINTTYPPOINT);
    pointtypicons->MakeIconSelectable (OUTPUT_POINTTYPCROSS);

    viewicons->MakeIconSelectable (OUTPUT_VIEW2D_SELECT);

    if (!clustering->Ist_Fertig ())
      go->MakeSelectable ();
    else
      go->MakeUnselectable ();
  } else {
    go->MakeUnselectable ();
    reset->MakeUnselectable ();
    makerules->MakeUnselectable ();
    classify->MakeUnselectable ();
    viewicons->MakeallUnselectable ();
    actionselecticons->MakeallUnselectable ();
    pointtypicons->MakeallUnselectable ();
    singlemembership->MakeallUnselectable ();
    singlemembership->UnselectanIcon (OUTPUT_SINGLEMEMBERSHIP_SELECT);
    if (axismenu != NULL)
      delete axismenu;
    axismenu = NULL;
  }
}

void 
ClusterOutput::DrawClusterPoint (short xpos, short ypos, int options)
{
  Drawable drawable;
  if (options & DRAWINWINDOW)
    drawable = window;
  else
    drawable = pixmap;

  XSetForeground (display, gc, redpix);
  if (options & DRAWTEMP)
    XSetFunction (display, gc, GXxor);

  XSegment lines[2];

  lines[0].y1 = lines[1].y2 = ypos - CLUSTERCROSSWIDTH;
  lines[0].y2 = lines[1].y1 = ypos + CLUSTERCROSSWIDTH;

  lines[0].x1 = lines[1].x1 = xpos - CLUSTERCROSSWIDTH;
  lines[0].x2 = lines[1].x2 = xpos + CLUSTERCROSSWIDTH;

  XDrawSegments (display, drawable, gc, lines, 2);

  XSetFunction (display, gc, GXcopy);
}

void 
ClusterOutput::UndrawClusterPoint (short xpos, short ypos, int options)
{
  Drawable drawable;
  if (options & DRAWINWINDOW)
    drawable = window;
  else
    drawable = pixmap;

  if (options & DRAWINWINDOW) {
    XCopyArea (display, pixmap, window, gc,
	       xpos - CLUSTERCROSSWIDTH, ypos - CLUSTERCROSSWIDTH,
	       1 + 2 * CLUSTERCROSSWIDTH, 1 + 2 * CLUSTERCROSSWIDTH,
	       xpos - CLUSTERCROSSWIDTH, ypos - CLUSTERCROSSWIDTH);
  } else {
    drawable = pixmap;
    XSetForeground (display, gc, whitepix);
    XSegment lines[2];

    lines[0].y1 = lines[1].y2 = ypos - CLUSTERCROSSWIDTH;
    lines[0].y2 = lines[1].y1 = ypos + CLUSTERCROSSWIDTH;

    lines[0].x1 = lines[1].x1 = xpos - CLUSTERCROSSWIDTH;
    lines[0].x2 = lines[1].x2 = xpos + CLUSTERCROSSWIDTH;

    XDrawSegments (display, drawable, gc, lines, 2);
  }
}



void 
ClusterOutput::DrawCluster (int options)
{
// erase old drawing
  Drawable drawable;
  if (options & DRAWINWINDOW)
    drawable = window;
  else
    drawable = pixmap;

// erase old drawing 
  if ((options & ERASEOLDDRAWING) || (Dataset == 0)) {
    XSetForeground (display, gc, whitepix);
    XFillRectangle (display, drawable, gc, 0, 0, width, height);
  }
  if (clustering->Lese_Clusteranzahl () == 0) {
    if (Dataset != NULL) {
      switch (viewicons->GetSelectedvalue ()) {
      case OUTPUT_VIEW2D_SELECT:
	if (Dataset->Lese_Daten ().Lese_Dim () > 1)
	  DrawData2d (options);
	else
	  DrawData1d (options);
	if (selectedinput != NULL)
	  DrawPoint2d (selectedinput,
		       options | DRAWMARKED,
		       pointtypicons->GetSelectedvalue ());
	break;

      case OUTPUT_VIEW3D_SELECT:
	DrawData3d (options);
	if (selectedinput != NULL)
	  DrawPoint3d (selectedinput,
		       options | DRAWMARKED,
		       pointtypicons->GetSelectedvalue ());
	break;
      }
    }
  } else {
    if (Dataset != NULL)
      switch (viewicons->GetSelectedvalue ()) {
      case OUTPUT_VIEW2D_SELECT:
	if (Dataset->Lese_Daten ().Lese_Dim () > 1)
	  DrawCluster2d (options);
	else
	  DrawCluster1d (options);
	break;

      case OUTPUT_VIEW3D_SELECT:
	DrawCluster3d (options);
	break;

      case OUTPUT_VIEW2DMEMBERTYP1_SELECT:
	Draw2dClusterMemberTyp1 (options);
	break;

      case OUTPUT_VIEW3DMEMBERTYP1_SELECT:
	Draw3dClusterMemberTyp1 (options);
	break;

      case OUTPUT_VIEW2DMEMBERTYP2_SELECT:
	Draw2dClusterMemberTyp2 (options);
	break;

      case OUTPUT_VIEW3DMEMBERTYP2_SELECT:
	Draw3dClusterMemberTyp2 (options);
	break;

      case OUTPUT_VIEW2DGUETE_SELECT:
	Draw2dGuete (options);
	break;
      }
  }
}

void 
ClusterOutput::SetSelectedinputvector (DVektor * input)
{
  selectedinput = input;
  if ((selectedinput != NULL) && (clustering->Lese_Clusteranzahl () != 0))
    if (selectedinput->Lese_Dim () == clustering->Lese_Vorgabe_Prototypen ().Lese_Dim ())
      memberoutput->SetResult (clustering->Berechne_Zugehoerigkeit (*selectedinput),
			       clustering);
  Action (OUTPUT_REDRAW, 0);
  UpdateOutput ();
}

void 
ClusterOutput::Drawredcross (short xpos, short ypos, Drawable drawable)
{
  XSegment lines[2];
  lines[0].x1 = lines[0].x2 = xpos;
  lines[0].y1 = ypos - SYMBOLHALFWIDTH;
  lines[0].y2 = ypos + SYMBOLHALFWIDTH;

  lines[1].x1 = xpos - SYMBOLHALFWIDTH;
  lines[1].x2 = xpos + SYMBOLHALFWIDTH;
  lines[1].y1 = lines[1].y2 = ypos;
  XSetForeground (display, gc, redpix);
  XDrawSegments (display, drawable, gc, lines, 2);
}

void 
ClusterOutput::DrawCluster1d (int options)
{
  DVektorArray *inputData = &(Dataset->Lese_Daten ());
// MIN, cause inputData->Lese_Groesse() may be !=
  // clustering->Lese_Zuordnung->Lese_Dim(), if dates are added or
  // deleteed
  int Datanumber = MIN (inputData->Lese_Groesse (),
			clustering->Lese_Zuordnung ().Lese_Dim ());
  DVektorArray *Cluster = &(clustering->Lese_Cluster ());
  int Clusternumber = clustering->Lese_Clusteranzahl ();

  int counter;

  Drawable drawable;

  if (options & DRAWINWINDOW)
    drawable = window;
  else
    drawable = pixmap;

// draw axes:
  DrawAxes2d (options,
	      "datano.",
	      Dataset->Lese_Name_Pointer (xindex));

// draw cluster:
  switch (clustericons->GetSelectedvalue ()) {
  case OUTPUT_CLUSTERTYPSYMBOL_SELECT:
    for (counter = 0; counter < Datanumber; counter++) {
      short xpos, ypos;
      int maxcluster = clustering->Lese_Zuordnung (counter);	// cluster with max membership 

      WorldtoPixel1d (counter,
		      inputData->Lese_Vektor (counter, xindex),
		      &xpos, &ypos);

      if (selectedcluster->Lese_i (maxcluster))
	XSetForeground (display, gc, blackpix);
      else
	XSetForeground (display, gc, shadowpix);
      DrawSymbol (xpos, ypos, drawable, maxcluster);

      if (selectedinput == &(inputData->Lese_Vektor (counter))) {
	XSetForeground (display, gc, redpix);
	XSetLineAttributes (display, gc, 3, LineSolid, CapProjecting, JoinMiter);
	DrawSymbol (xpos, ypos, drawable, maxcluster);
	XSetLineAttributes (display, gc, 1, LineSolid, CapProjecting, JoinMiter);
      }
    }
    break;

  case OUTPUT_CLUSTERTYPCOLOR_SELECT:
    for (counter = 0; counter < Datanumber; counter++) {
      short xpos, ypos;
      int maxcluster = clustering->Lese_Zuordnung (counter);	// cluster with max membership 

      WorldtoPixel1d (counter,
		      inputData->Lese_Vektor (counter, xindex),
		      &xpos, &ypos);

/*         if(selectedcluster->Lese_i(maxcluster))
   XSetForeground(display,gc,blackpix);
   else
   XSetForeground(display,gc,shadowpix); */
      DrawColorbox (xpos, ypos, drawable, maxcluster);
      if (selectedcluster->Lese_i (maxcluster))
	DrawBox (xpos, ypos, drawable);

/*         if(selectedinput==&(inputData->Lese_Vektor(counter)))
   {
   XSetForeground(display,gc,redpix);
   XSetLineAttributes(display, gc, 3, LineSolid, CapProjecting, JoinMiter);
   DrawSymbol(xpos,ypos,drawable,maxcluster);
   XSetLineAttributes(display, gc, 1, LineSolid, CapProjecting, JoinMiter);
   } */
    }
    break;

  case OUTPUT_CLUSTERTYPSTAR_SELECT:
    XPoint * Clusterpoints = (XPoint *) malloc (sizeof (XPoint) * Clusternumber);

    // first calculate Clusterpoint-coordinates
    for (counter = 0; counter < Clusternumber; counter++)
      WorldtoPixel1d (0,
		      Cluster->Lese_Vektor (counter, xindex),
		  &(Clusterpoints[counter].x), &(Clusterpoints[counter].y));

    short x1, y1;
    for (counter = 0; counter < Datanumber; counter++) {
      int maxcluster = clustering->Lese_Zuordnung (counter);

      WorldtoPixel1d (counter,
		      inputData->Lese_Vektor (counter, xindex),
		      &x1, &y1);

      if (selectedcluster->Lese_i (maxcluster))
	XSetForeground (display, gc, blackpix);
      else
	XSetForeground (display, gc, shadowpix);
      XDrawLine (display, drawable, gc,
		 x1, y1,
		 Clusterpoints[maxcluster].x, Clusterpoints[maxcluster].y);
      if (selectedinput == &(inputData->Lese_Vektor (counter))) {
	short xpos, ypos;
	WorldtoPixel1d (counter,
			selectedinput->Lese_i (xindex),
			&xpos, &ypos);
	Drawredcross (xpos, ypos, drawable);
      }
    }
    free (Clusterpoints);

    break;
  }
}

void 
ClusterOutput::DrawCluster2d (int options)
{
  DVektorArray *inputData = &(Dataset->Lese_Daten ());
// MIN, cause inputData->Lese_Groesse() may be !=
  // clustering->Lese_Zuordnung->Lese_Dim(), if dates are added or
  // deleteed
  int Datanumber = MIN (inputData->Lese_Groesse (),
			clustering->Lese_Zuordnung ().Lese_Dim ());
  DVektorArray *Cluster = &(clustering->Lese_Cluster ());
  int Clusternumber = clustering->Lese_Clusteranzahl ();

  int counter;

  Drawable drawable;

  if (options & DRAWINWINDOW)
    drawable = window;
  else
    drawable = pixmap;

// draw axes:
  DrawAxes2d (options,
	      Dataset->Lese_Name_Pointer (xindex),
	      Dataset->Lese_Name_Pointer (yindex));

#if 1
  XPoint *Clusterpoints = (XPoint *) malloc (sizeof (XPoint) * Clusternumber);
#endif

// draw cluster:
  switch (clustericons->GetSelectedvalue ()) {
  case OUTPUT_CLUSTERTYPSYMBOL_SELECT:
    for (counter = 0; counter < Datanumber; counter++) {
      short xpos, ypos;
      int maxcluster = clustering->Lese_Zuordnung (counter);	// cluster with max membership 

      WorldtoPixel2d (inputData->Lese_Vektor (counter, xindex),
		      inputData->Lese_Vektor (counter, yindex),
		      &xpos, &ypos);

      if (selectedcluster->Lese_i (maxcluster))
	XSetForeground (display, gc, blackpix);
      else
	XSetForeground (display, gc, shadowpix);
      DrawSymbol (xpos, ypos, drawable, maxcluster);

      if (selectedinput == &(inputData->Lese_Vektor (counter))) {
	XSetForeground (display, gc, redpix);
	XSetLineAttributes (display, gc, 3, LineSolid, CapProjecting, JoinMiter);
	DrawSymbol (xpos, ypos, drawable, maxcluster);
	XSetLineAttributes (display, gc, 1, LineSolid, CapProjecting, JoinMiter);
      }
    }
    break;

  case OUTPUT_CLUSTERTYPCOLOR_SELECT:
    for (counter = 0; counter < Datanumber; counter++) {
      short xpos, ypos;
      int maxcluster = clustering->Lese_Zuordnung (counter);	// cluster with max membership 

      WorldtoPixel2d (inputData->Lese_Vektor (counter, xindex),
		      inputData->Lese_Vektor (counter, yindex),
		      &xpos, &ypos);

/*         if(selectedcluster->Lese_i(maxcluster))
   XSetForeground(display,gc,blackpix);
   else
   XSetForeground(display,gc,shadowpix); */
      DrawColorbox (xpos, ypos, drawable, maxcluster);
      if (selectedcluster->Lese_i (maxcluster))
	DrawBox (xpos, ypos, drawable);

/*         if(selectedinput==&(inputData->Lese_Vektor(counter)))
   {
   XSetForeground(display,gc,redpix);
   XSetLineAttributes(display, gc, 3, LineSolid, CapProjecting, JoinMiter);
   DrawSymbol(xpos,ypos,drawable,maxcluster);
   XSetLineAttributes(display, gc, 1, LineSolid, CapProjecting, JoinMiter);
   } */
    }
    break;

  case OUTPUT_CLUSTERTYPSTAR_SELECT:
#if 0
    XPoint * Clusterpoints = (XPoint *) malloc (sizeof (XPoint) * Clusternumber);
#endif
    // first calculate Clusterpoint-coordinates
    for (counter = 0; counter < Clusternumber; counter++)
      WorldtoPixel2d (Cluster->Lese_Vektor (counter, xindex),
		      Cluster->Lese_Vektor (counter, yindex),
		  &(Clusterpoints[counter].x), &(Clusterpoints[counter].y));

    short x1, y1;
    for (counter = 0; counter < Datanumber; counter++) {
      int maxcluster = clustering->Lese_Zuordnung (counter);

      WorldtoPixel2d (inputData->Lese_Vektor (counter, xindex),
		      inputData->Lese_Vektor (counter, yindex),
		      &x1, &y1);

      if (selectedcluster->Lese_i (maxcluster))
	XSetForeground (display, gc, blackpix);
      else
	XSetForeground (display, gc, shadowpix);
      XDrawLine (display, drawable, gc,
		 x1, y1,
		 Clusterpoints[maxcluster].x, Clusterpoints[maxcluster].y);
    }
    free (Clusterpoints);
    if (selectedinput != NULL) {
      short xpos, ypos;
      WorldtoPixel2d (selectedinput->Lese_i (xindex),
		      selectedinput->Lese_i (yindex),
		      &xpos, &ypos);
      Drawredcross (xpos, ypos, drawable);
    }
    break;

  case OUTPUT_CLUSTERTYPELLIPSIS_SELECT:
    // first draw data
    DrawDVektorArray2d (options, pointtypicons->GetSelectedvalue (), &(Dataset->Lese_Daten ()), NULL);
    // now the ellipses
    for (counter = 0; counter < ellipsisnumber; counter++) {
      if (selectedcluster->Lese_i (counter))
	XSetForeground (display, gc, blackpix);
      else
	XSetForeground (display, gc, shadowpix);

      ellipsis_2d[counter]->DrawMe (this, drawable);
    }
    if (selectedinput != NULL) {
      short xpos, ypos;
      WorldtoPixel2d (selectedinput->Lese_i (xindex),
		      selectedinput->Lese_i (yindex),
		      &xpos, &ypos);
      Drawredcross (xpos, ypos, drawable);
    }
    break;
  }
}

void 
ClusterOutput::DrawCluster3d (int options)
{
  DVektorArray *inputData = &(Dataset->Lese_Daten ());
// MIN, cause inputData->Lese_Groesse() may be !=
  // clustering->Lese_Zuordnung->Lese_Dim(), if dates are added or
  // deleteed
  int Datanumber = MIN (inputData->Lese_Groesse (),
			clustering->Lese_Zuordnung ().Lese_Dim ());
  DVektorArray *Cluster = &(clustering->Lese_Cluster ());
  int Clusternumber = clustering->Lese_Clusteranzahl ();

  int counter;
  Drawable drawable;

  if (options & DRAWINWINDOW)
    drawable = window;
  else
    drawable = pixmap;

// draw axes:
  DrawAxes3d (options,
	      Dataset->Lese_Name_Pointer (xindex),
	      Dataset->Lese_Name_Pointer (yindex),
	      Dataset->Lese_Name_Pointer (zindex));

#if 1
  XPoint *Clusterpoints = (XPoint *) malloc (sizeof (XPoint) * Clusternumber);
#endif

// draw cluster:
  switch (clustericons->GetSelectedvalue ()) {
  case OUTPUT_CLUSTERTYPSYMBOL_SELECT:
    for (counter = 0; counter < Datanumber; counter++) {
      short xpos, ypos;
      int maxcluster = clustering->Lese_Zuordnung (counter);

      WorldtoPixel3d (inputData->Lese_Vektor (counter, xindex),
		      inputData->Lese_Vektor (counter, yindex),
		      inputData->Lese_Vektor (counter, zindex),
		      &xpos, &ypos);

      if (selectedcluster->Lese_i (maxcluster))
	XSetForeground (display, gc, blackpix);
      else
	XSetForeground (display, gc, shadowpix);
      DrawSymbol (xpos, ypos, drawable, maxcluster);

      if (selectedinput == &(inputData->Lese_Vektor (counter))) {
	XSetForeground (display, gc, redpix);
	XSetLineAttributes (display, gc, 3, LineSolid, CapProjecting, JoinMiter);
	DrawSymbol (xpos, ypos, drawable, maxcluster);
	XSetLineAttributes (display, gc, 1, LineSolid, CapProjecting, JoinMiter);
      }
    }
    break;

  case OUTPUT_CLUSTERTYPCOLOR_SELECT:
    for (counter = Datanumber - 1; counter > 0; counter--) {
      int index;
//         double zpos;
      index = deepindex[counter];
      WorldtoZvalue (inputData->Lese_Vektor (index, xindex),
		     inputData->Lese_Vektor (index, yindex),
		     inputData->Lese_Vektor (index, zindex),
		     &(deepvalues[counter]));
//                       &zpos);
      //         deepvalues[counter]=zpos;
    }
    deepvalues.Quicksort (deepindex);

    for (counter = 0; counter < Datanumber; counter++) {
      int index;
      short xpos, ypos;
      index = deepindex[counter];
      int maxcluster = clustering->Lese_Zuordnung (index);

      WorldtoPixel3d (inputData->Lese_Vektor (index, xindex),
		      inputData->Lese_Vektor (index, yindex),
		      inputData->Lese_Vektor (index, zindex),
		      deepvalues[counter],
		      &xpos, &ypos);

      DrawColorbox (xpos, ypos, drawable, maxcluster);
      if (selectedcluster->Lese_i (maxcluster))
	DrawBox (xpos, ypos, drawable);
    }
    break;

  case OUTPUT_CLUSTERTYPSTAR_SELECT:
#if 0
    XPoint * Clusterpoints = (XPoint *) malloc (sizeof (XPoint) * Clusternumber);
#endif
    // first calculate Clusterpoint-coordinates
    for (counter = 0; counter < Clusternumber; counter++) {
      WorldtoPixel3d (Cluster->Lese_Vektor (counter, xindex),
		      Cluster->Lese_Vektor (counter, yindex),
		      Cluster->Lese_Vektor (counter, zindex),
		  &(Clusterpoints[counter].x), &(Clusterpoints[counter].y));
    }

    short x1, y1;
    for (counter = 0; counter < Datanumber; counter++) {
      int maxcluster = clustering->Lese_Zuordnung (counter);

      WorldtoPixel3d (inputData->Lese_Vektor (counter, xindex),
		      inputData->Lese_Vektor (counter, yindex),
		      inputData->Lese_Vektor (counter, zindex),
		      &x1, &y1);

      if (selectedcluster->Lese_i (maxcluster))
	XSetForeground (display, gc, blackpix);
      else
	XSetForeground (display, gc, shadowpix);
      XDrawLine (display, drawable, gc,
		 x1, y1,
		 Clusterpoints[maxcluster].x, Clusterpoints[maxcluster].y);
    }
    free (Clusterpoints);
    if (selectedinput != NULL) {
      short xpos, ypos;
      WorldtoPixel3d (selectedinput->Lese_i (xindex),
		      selectedinput->Lese_i (yindex),
		      selectedinput->Lese_i (zindex),
		      &xpos, &ypos);
      Drawredcross (xpos, ypos, drawable);
    }
    break;

  case OUTPUT_CLUSTERTYPELLIPSIS_SELECT:
    DrawDVektorArray3d (options, pointtypicons->GetSelectedvalue (), &(Dataset->Lese_Daten ()), NULL);
    for (counter = 0; counter < ellipsisnumber; counter++) {
      if (selectedcluster->Lese_i (counter))
	XSetForeground (display, gc, blackpix);
      else
	XSetForeground (display, gc, shadowpix);
      ellipsis_3d[counter]->DrawMe (this,
				    drawable);
    }
    if (selectedinput != NULL) {
      short xpos, ypos;
      WorldtoPixel3d (selectedinput->Lese_i (xindex),
		      selectedinput->Lese_i (yindex),
		      selectedinput->Lese_i (zindex),
		      &xpos, &ypos);
      Drawredcross (xpos, ypos, drawable);
    }
    break;
  }
}

void 
ClusterOutput::Draw2dClusterMemberTyp1 (int options)
{
  DVektorArray *inputData = &(Dataset->Lese_Daten ());
// MIN, cause inputData->Lese_Groesse() may be !=
  // clustering->Lese_Zuordnung->Lese_Dim(), if dates are added or
  // deleteed
  int Datanumber = MIN (inputData->Lese_Groesse (),
			clustering->Lese_Zuordnung ().Lese_Dim ());
  DMatrix *membership = &(clustering->Lese_Zugehoerigkeit ());
  Drawable drawable;
  int counter;

  if (options & DRAWINWINDOW)
    drawable = window;
  else
    drawable = pixmap;

// draw axes:
  DrawAxes2d (options,
	      Dataset->Lese_Name_Pointer (xindex),
	      "U");
  XSegment *Lines;
  Lines = (XSegment *) malloc (sizeof (XSegment) * Datanumber);

  for (counter = 0; counter < Datanumber; counter++) {
    WorldtoPixel2d (inputData->Lese_Vektor (counter, xindex),
		    membership->Lese_i_j (Scroll->GetValue (), counter),
		    &(Lines[counter].x1), &(Lines[counter].y1));

    WorldtoPixel2d (inputData->Lese_Vektor (counter, xindex),
		    0,
		    &(Lines[counter].x2), &(Lines[counter].y2));
  }
  XDrawSegments (display, drawable, gc, Lines, Datanumber);
  free (Lines);
};

void 
ClusterOutput::Draw3dClusterMemberTyp1 (int options)
{
  DVektorArray *inputData = &(Dataset->Lese_Daten ());
// MIN, cause inputData->Lese_Groesse() may be !=
  // clustering->Lese_Zuordnung->Lese_Dim(), if dates are added or
  // deleteed
  int Datanumber = MIN (inputData->Lese_Groesse (),
			clustering->Lese_Zuordnung ().Lese_Dim ());
  DMatrix *membership = &(clustering->Lese_Zugehoerigkeit ());

  int counter;
  Drawable drawable;

  if (options & DRAWINWINDOW)
    drawable = window;
  else
    drawable = pixmap;

// draw axes:
  DrawAxes3d (options,
	      Dataset->Lese_Name_Pointer (xindex),
	      Dataset->Lese_Name_Pointer (yindex),
	      "U");

// draw clustermembership:
  XSetForeground (display, gc, blackpix);
  XSegment *Lines;
  if ((Lines = (XSegment *) malloc (sizeof (XSegment) * Datanumber)) == NULL) {
    printf ("Out of memory!\n");
    exit (1);
  }
  for (counter = 0; counter < Datanumber; counter++) {
    WorldtoPixel3d (inputData->Lese_Vektor (counter, xindex),
		    inputData->Lese_Vektor (counter, yindex),
		    membership->Lese_i_j (Scroll->GetValue (), counter),
		    &(Lines[counter].x1), &(Lines[counter].y1));

    WorldtoPixel3d (inputData->Lese_Vektor (counter, xindex),
		    inputData->Lese_Vektor (counter, yindex),
		    0,
		    &(Lines[counter].x2), &(Lines[counter].y2));
  }
  XDrawSegments (display, drawable, gc, Lines, Datanumber);
  free (Lines);
};


void 
ClusterOutput::Draw3dClusterMemberTyp2 (int options)
{
  DVektorArray *inputData = &(Dataset->Lese_Daten ());
// MIN, cause inputData->Lese_Groesse() may be !=
  // clustering->Lese_Zuordnung->Lese_Dim(), if dates are added or
  // deleteed
  int Datanumber = MIN (inputData->Lese_Groesse (),
			clustering->Lese_Zuordnung ().Lese_Dim ());
  int Clusternumber = clustering->Lese_Clusteranzahl ();
  DMatrix *membership = &(clustering->Lese_Zugehoerigkeit ());

  double datastep, clusterstep;

  int datacounter, clustercounter;
  Drawable drawable;

  if (Datanumber == 1)
    datastep = 1.0;
  else
    datastep = 1.0 / (Datanumber - 1);

  if (Clusternumber == 1)
    clusterstep = 1.0;
  else
    clusterstep = 1.0 / (Clusternumber - 1);

  if (options & DRAWINWINDOW)
    drawable = window;
  else
    drawable = pixmap;

// draw axes:
  DrawAxes3d (options,
	      "Clusterno.",
//           "U",
	      "Data-no.",
	      "U");

// draw clustermembership:
  XPoint *Points;
  if ((Points = (XPoint *) malloc (sizeof (XPoint) * Clusternumber)) == NULL) {
    printf ("Out of memory!\n");
    exit (1);
  }
  XSetForeground (display, gc, blackpix);

  for (datacounter = 0; datacounter < Datanumber; datacounter++) {
    for (clustercounter = 0; clustercounter < Clusternumber; clustercounter++)
      WorldtoPixel3d (clustercounter * clusterstep,
		      datacounter * datastep,
		      membership->Lese_i_j (clustercounter, datacounter),
		  &(Points[clustercounter].x), &(Points[clustercounter].y));

    XDrawLines (display, drawable, gc, Points, Clusternumber, CoordModeOrigin);
  }
  free (Points);
};


void 
ClusterOutput::Draw2dClusterMemberTyp2 (int options)
{
  int Clusternumber = clustering->Lese_Clusteranzahl ();
  DMatrix *membership = &(clustering->Lese_Zugehoerigkeit ());
  Drawable drawable;
  int clustercounter;

  double clusterstep;
  if (Clusternumber == 1)
    clusterstep = 1.0;
  else
    clusterstep = 1.0 / (Clusternumber - 1);

  if (options & DRAWINWINDOW)
    drawable = window;
  else
    drawable = pixmap;

// draw axes:
  DrawAxes2d (options,
	      "Clusterno.",
	      "U");

  XPoint *Points;
  if ((Points = (XPoint *) malloc (sizeof (XPoint) * Clusternumber)) == NULL) {
    printf ("Out of memory!\n");
    exit (1);
  }
//dataindex=Scroll->GetValue();
  for (clustercounter = 0; clustercounter < Clusternumber; clustercounter++)
    WorldtoPixel2d (clustercounter * clusterstep,
		 membership->Lese_i_j (clustercounter, Scroll->GetValue ()),
		    &(Points[clustercounter].x),
		    &(Points[clustercounter].y));

  XDrawLines (display, drawable, gc, Points, Clusternumber, CoordModeOrigin);
  free (Points);
};

void 
ClusterOutput::Draw2dGuete (int options)
{
  Guete *guete = &(clustering->Lese_Guete ());
  int Clusternumber = guete->Lese_getestet ();
  Drawable drawable;

  if (options & DRAWINWINDOW)
    drawable = window;
  else
    drawable = pixmap;

// draw axes:
  DrawAxes2d (options,
	      "no.",
	      "Validity");
  if (Clusternumber > 1) {
    static short textheight = fontstruct->max_bounds.ascent + fontstruct->max_bounds.descent;
    short textxpos, textypos;
    WorldtoPixel2d (0.0, 0.0, &textxpos, &textypos);
    textypos += textheight;

    char number[30];		// hope its long enought

    double ydiff = guete->Lese_Max_Wert () - guete->Lese_Min_Wert ();
    double guetestep = 1.0 / (ydiff * ARROWOVERLAP), clusterstep = (1.0 / (guete->Lese_Max_Anzahl () - guete->Lese_Min_Anzahl ())) / ARROWOVERLAP,
      min = guete->Lese_Min_Wert ();

    XPoint *Points;
    Points = (XPoint *) malloc (sizeof (XPoint) * Clusternumber);
    int clustercounter;
    for (clustercounter = 0; clustercounter < Clusternumber; clustercounter++) {
      int xoffset = guete->Lese_Anzahl (clustercounter) - guete->Lese_Min_Anzahl ();
      WorldtoPixel2d (xoffset * clusterstep, (guete->Lese_Werte (clustercounter) - min) * guetestep,
		  &(Points[clustercounter].x), &(Points[clustercounter].y));
      // print clusternumber on bottom
      sprintf (number, "%d", guete->Lese_Anzahl (clustercounter));
      XDrawString (display, drawable, gc,
		   Points[clustercounter].x,
		   textypos,
		   number, strlen (number));
    }
    XDrawLines (display, drawable, gc, Points, Clusternumber, CoordModeOrigin);
    // print values on y-axis
    sprintf (number, "%.6f", guete->Lese_Max_Wert ());

    WorldtoPixel2d (0.0, 1.0 / ARROWOVERLAP, &textxpos, &textypos);
    XDrawString (display, drawable, gc,
		 0,
		 textypos,
		 number, strlen (number));
    sprintf (number, "%.6f", guete->Lese_Min_Wert ());
    WorldtoPixel2d (0.0, 0.0, &textxpos, &textypos);
    XDrawString (display, drawable, gc,
		 0,
		 textypos,
		 number, strlen (number));
  } else {
    short textxpos, textypos;

    XPoint Points[2];
    WorldtoPixel2d (0.0, 1.0 / ARROWOVERLAP,
		    &(Points[0].x), &(Points[0].y));
    WorldtoPixel2d (1.0 / ARROWOVERLAP, 1.0 / ARROWOVERLAP,
		    &(Points[1].x), &(Points[1].y));
    XDrawLines (display, drawable, gc, Points, 2, CoordModeOrigin);

    // print values on y-axis
    WorldtoPixel2d (0.0, 1.0 / ARROWOVERLAP, &textxpos, &textypos);
    char number[8];
    sprintf (number, "%.6f", guete->Lese_Werte (0));
    XDrawString (display, drawable, gc,
		 0,
		 textypos,
		 number, strlen (number));
  }
};

void 
ClusterOutput::DrawData1d (int options)
{
// draw axes
  DrawAxes2d (options,
	      "datano.",
	      Dataset->Lese_Name_Pointer (xindex));
  DrawDVektorArray1d (options,
		      pointtypicons->GetSelectedvalue (),
		      &(Dataset->Lese_Daten ()),
		      NULL);

  short counter;
  for (counter = 0; counter < clustering->Lese_Vorgabe_Prototypen ().Lese_Groesse (); counter++) {
    short xpos, ypos;
    WorldtoPixel1d (0,
       clustering->Lese_Vorgabe_Prototypen ().Lese_Vektor (counter, xindex),
		    &xpos, &ypos);
    DrawClusterPoint (xpos, ypos, options);
  }
}

void 
ClusterOutput::DrawData2d (int options)
{
// draw axes
  DrawAxes2d (options,
	      Dataset->Lese_Name_Pointer (xindex),
	      Dataset->Lese_Name_Pointer (yindex));
  DrawDVektorArray2d (options, pointtypicons->GetSelectedvalue (), &(Dataset->Lese_Daten ()), NULL);

  short counter;
  for (counter = 0; counter < clustering->Lese_Vorgabe_Prototypen ().Lese_Groesse (); counter++) {
    short xpos, ypos;
    WorldtoPixel2d (clustering->Lese_Vorgabe_Prototypen ().Lese_Vektor (counter, xindex),
       clustering->Lese_Vorgabe_Prototypen ().Lese_Vektor (counter, yindex),
		    &xpos, &ypos);
    DrawClusterPoint (xpos, ypos, options);
  }
}

void 
ClusterOutput::DrawData3d (int options)
{
// draw axes

  DrawAxes3d (options,
	      Dataset->Lese_Name_Pointer (xindex),
	      Dataset->Lese_Name_Pointer (yindex),
	      Dataset->Lese_Name_Pointer (zindex));

  DrawDVektorArray3d (options, pointtypicons->GetSelectedvalue (), &(Dataset->Lese_Daten ()), NULL);

  short counter;
  for (counter = 0; counter < clustering->Lese_Vorgabe_Prototypen ().Lese_Groesse (); counter++) {
    short xpos, ypos;
    WorldtoPixel3d (clustering->Lese_Vorgabe_Prototypen ().Lese_Vektor (counter, xindex),
       clustering->Lese_Vorgabe_Prototypen ().Lese_Vektor (counter, yindex),
       clustering->Lese_Vorgabe_Prototypen ().Lese_Vektor (counter, zindex),
		    &xpos, &ypos);
    DrawClusterPoint (xpos, ypos, options);
  }
}

void 
ClusterOutput::Calculate_ellipses ()
{
  int Clusternumber = clustering->Lese_Clusteranzahl ();
  DVektorArray *Cluster = &(clustering->Lese_Cluster ());
  DMatrixArray clusterellipses2d;
  DMatrixArray clusterellipses3d;

  IVektor ivektor;
  if (Dataset->Lese_DatenDim () > 2) {
    ivektor.Setze_Dim (3);
    ivektor.Setze_i (0, xindex);
    ivektor.Setze_i (1, yindex);
    ivektor.Setze_i (2, zindex);
    clusterellipses3d = (clustering->Lese_Form (ivektor, Dataset->Lese_Daten ()));
  }
  ivektor.Setze_Dim (2);
  ivektor.Setze_i (0, xindex);
  ivektor.Setze_i (1, yindex);
  clusterellipses2d = (clustering->Lese_Form (ivektor, Dataset->Lese_Daten ()));

  short counter;

// first free the old ellipses
  if (ellipsis_2d != NULL) {
    for (counter = 0; counter < ellipsisnumber; counter++)
      delete ellipsis_2d[counter];

    free (ellipsis_2d);
  }
  if (ellipsis_3d != NULL) {
    for (counter = 0; counter < ellipsisnumber; counter++)
      delete ellipsis_3d[counter];

    free (ellipsis_3d);
  }
// now alloc new ellipses

  if ((clusterellipses2d.Lese_Groesse () > 0) && (Clusternumber > 0)) {
    // now alloc space for new ellipses
    ellipsisnumber = Clusternumber;
    ellipsis_2d = (Ellipsis_2d **) malloc (sizeof (Ellipsis_2d *) * ellipsisnumber);
    if (Dataset->Lese_DatenDim () > 2)
      ellipsis_3d = (Ellipsis_3d **) malloc (sizeof (Ellipsis_3d *) * ellipsisnumber);
    else
      ellipsis_3d = NULL;
    for (counter = 0; counter < ellipsisnumber; counter++) {
      ellipsis_2d[counter] = new Ellipsis_2d (clusterellipses2d.Lese_Matrix_Pointer (counter),
				     Cluster->Lese_Vektor (counter, xindex),
				    Cluster->Lese_Vektor (counter, yindex));
      if (Dataset->Lese_DatenDim () > 2)
	ellipsis_3d[counter] = new Ellipsis_3d (clusterellipses3d.Lese_Matrix_Pointer (counter),
				     Cluster->Lese_Vektor (counter, xindex),
				     Cluster->Lese_Vektor (counter, yindex),
				      Cluster->Lese_Vektor (counter, zindex)
	  );
    }
  } else {
    ellipsisnumber = 0;
    ellipsis_2d = NULL;
    ellipsis_3d = NULL;
  }
}

void 
ClusterOutput::Action (int actionnumber, int value)
{
  switch (actionnumber) {
  case OUTPUT_SAVE:
    myicon->GetMainwindow ()->Getsaveloadwindow ()
      ->ChangeFunction (WORK_AS_SAVE_SELECTOR,
			this,
			OUTPUT_SAVEOK,
			"SAVE CLUSTER");
    myicon->GetMainwindow ()->Getsaveloadwindow ()->MapMe ();
    myicon->GetMainwindow ()->SetStatus (MAINSAVECLUSTERSTATUS);
    break;

  case OUTPUT_SAVEOK:
    if ((char *) value != NULL)
      clustering->Speichern ((char *) value);
    myicon->GetMainwindow ()->SetStatus (MAINNORMALSTATUS);
    break;

  case OUTPUT_LOAD:
    myicon->GetMainwindow ()->Getsaveloadwindow ()
      ->ChangeFunction (WORK_AS_LOAD_SELECTOR,
			this,
			OUTPUT_RELOADOK,
			"LOAD CLUSTER");
    myicon->GetMainwindow ()->Getsaveloadwindow ()->MapMe ();
    myicon->GetMainwindow ()->SetStatus (MAINLOADCLUSTERSTATUS);
    break;

  case OUTPUT_RELOADOK:
    if ((char *) value != NULL) {
      Reset_Fehlerstatus ();
      Clustering *tempclustering = Clustering_Laden ((char *) value);
      if (Lese_Fehlerstatus () == KEINFEHLER) {
	Action (OUTPUT_LOADOK, (int) tempclustering);
	((ClusterIcon *) myicon)->SetIconname ((char *) value);
      } else {
	switch (Lese_Fehlerstatus ()) {
	case DATEIFORMATFEHLER:
	  myicon->GetMainwindow ()->printError ("Error in format of cluster-file!");
	  break;

	default:
	  break;
	}
	Reset_Fehlerstatus ();
      }
    }
    myicon->GetMainwindow ()->SetStatus (MAINNORMALSTATUS);
    break;


    break;

  case OUTPUT_LOADOK:
    if ((Clustering *) value != NULL) {
      if (clustering != NULL) {
	Action (OUTPUT_RESETCLUSTERING_ACTION, 0);
//           delete clustering;
      }
      SetClustering ((Clustering *) value);
      if (Dataset != NULL)
	clustering->Berechne_fuer_neue_Daten (Dataset->Lese_Daten ());

      short counter, readclusternumber = clustering->Lese_Clusteranzahl ();
      selectedcluster->Setze_Dim (readclusternumber);
      for (counter = 0; counter < readclusternumber; counter++)
	selectedcluster->Setze_i (counter, 1);

      if (Dataset != NULL)
	Calculate_ellipses ();
      SelectUnselectIcons ();
      reset->MakeSelectable ();
      go->MakeUnselectable ();
      if ((!clustering->Ist_Fertig ()) && (Dataset != NULL))
	go->MakeSelectable ();

      Action (OUTPUT_REDRAW, 0);
      UpdateOutput ();
    }
    myicon->GetMainwindow ()->SetStatus (MAINNORMALSTATUS);
    break;

  case OUTPUT_FITSCALE:
    Action (viewicons->GetSelectedvalue (), 0);
    break;

  case OUTPUT_SINGLEMEMBERSHIP_SELECT:
    singlemembership->UnselectanIcon (OUTPUT_SINGLEMEMBERSHIP_SELECT);
    memberoutput->SetResult (clustering->Berechne_Zugehoerigkeit (*selectedinput),
			     clustering);
    memberoutput->MapMe ();
    memberoutput->RaiseMe ();
    break;

  case OUTPUT_RESIZED:
    if (viewicons != NULL)
      viewicons->MoveSelectIcons (value, 0);
    if (clustericons != NULL)
      clustericons->MoveSelectIcons (value, 0);
    if (actionselecticons != NULL)
      actionselecticons->MoveSelectIcons (value, 0);
    if (pointtypicons != NULL)
      pointtypicons->MoveSelectIcons (value, 0);
    if (singlemembership != NULL)
      singlemembership->MoveSelectIcons (value, 0);
    if (Scroll != NULL)
      Scroll->MoveMe (value, 0);
    if (Dataset != NULL)
      Action (viewicons->GetSelectedvalue (), 0);
    break;

  case OUTPUT_REDRAW:
    DrawCluster (ERASEOLDDRAWING);
    break;

  case OUTPUT_REDRAWTOWINDOW:
    DrawCluster (ERASEOLDDRAWING | DRAWINWINDOW);
    break;

  case OUTPUT_CHANCEINDEX1_ACTION:
    SetIndex1 (value);
    Action (viewicons->GetSelectedvalue (), 0);
    break;

  case OUTPUT_CHANCEINDEX2_ACTION:
    SetIndex2 (value);
    Action (viewicons->GetSelectedvalue (), 0);
    break;

  case OUTPUT_CHANCEINDEX3_ACTION:
    SetIndex3 (value);
    Action (viewicons->GetSelectedvalue (), 0);
    break;

  case OUTPUT_RESETCLUSTERING_ACTION:
    if (clustering != NULL) {
      clustering->Reset ();
      go->MakeSelectable ();
      reset->MakeUnselectable ();
      if (viewicons->GetSelectedvalue () == OUTPUT_VIEW2DGUETE_SELECT) {
	clustering->Lese_Guete ().Reset ();
	viewicons->MakeIconUnselectable (OUTPUT_VIEW2DGUETE_SELECT);
      }
      SelectUnselectIcons ();
    }
    save->MakeUnselectable ();
    break;

  case OUTPUT_STARTCLUSTERING_ACTION:
    // if clustering not finished
    if (!clustering->Ist_Fertig ()) {
      XDefineCursor (display, window, waitcursor);
      XFlush (display);
/*         XGrabPointer(display,
   window,
   TRUE,
   StructureNotifyMask|LeaveWindowMask|EnterWindowMask|FocusChangeMask|ButtonMotionMask|ButtonPressMask|ButtonReleaseMask,
   ButtonMotionMask,
   GrabModeAsync,
   GrabModeAsync,
   None,
   waitcursor,
   CurrentTime); */

      // Use defined or by other clustering calculated cluster:
      if (clustering->Lese_Clusteranzahl () == 0) {
//           if(Usercluster->Lese_Groesse()>0)
	//             clustering->Setze_Cluster();
	//           else
	if (clustering->Lese_Vorgabe_Prototypen ().Lese_Groesse () == 0) {
	  Clustering *inputclustering = NULL;
	  if (myicon->GetInputConnection () != NULL)
	    myicon->GetInputConnection ()->SendActiontoStart (ICON_REQUESTFORCLUSTERING, (int) &inputclustering);
	  if (inputclustering != 0) {
	    if (inputclustering->Lese_Clusteranzahl () != 0)
	      clustering->Setze_Cluster (inputclustering->Lese_Cluster ());
	  }
	}
      }
      // no defined or.. cluster

      switch (validitytyp /* Lese_Guete().Lese_Typ() */ ) {

      case NONE:
	fprintf (stderr, "** internal Oops in ClusterOutput::Action\n");
	break;

      case CLASSIFY:
	// means (validitytyp==Classify-Validity)
	if (clusternumber < MAXINT)
	  clustering->Clustere (Dataset->Lese_Daten (), clusternumber);
	else
	  clustering->Clustere_Klassen (Dataset->Lese_Daten (), exactness);
	break;

      default:
	if (clusternumber < MAXINT)
	  clustering->Clustere (Dataset->Lese_Daten (), clusternumber);
	else
	  clustering->Clustere (Dataset->Lese_Daten ());
	break;
      }

      if (clustering->Ist_Fertig ())
	go->MakeUnselectable ();

      if (selectedcluster->Lese_Dim () != clustering->Lese_Clusteranzahl ()) {
	short copynumber, counter;
	BVektor *newbvector = new BVektor ();
	newbvector->Setze_Dim (clustering->Lese_Clusteranzahl ());
	copynumber = MIN (selectedcluster->Lese_Dim (), clustering->Lese_Clusteranzahl ());
	for (counter = 0; counter < copynumber; counter++)
	  newbvector->Setze_i (counter, selectedcluster->Lese_i (counter));
	for (counter = copynumber; counter < clustering->Lese_Clusteranzahl (); counter++)
	  newbvector->Setze_i (counter, 1);
	delete selectedcluster;
	selectedcluster = newbvector;
      }
      SelectUnselectIcons ();
      Calculate_ellipses ();
      Action (OUTPUT_REDRAW, 0);
      UpdateOutput ();
      XUngrabPointer (display, CurrentTime);
      switch (actionselecticons->GetSelectedvalue ()) {
      case OUTPUTSTATUS_MOVECLUSTER_SELECT:
	XDefineCursor (display, window, movepointcursor);
	break;

      default:
	XDefineCursor (display, window, maincursor);
      }
    }
    reset->MakeSelectable ();

    if (clustering->Lese_Guete ().Lese_getestet () > 0)
      viewicons->MakeIconSelectable (OUTPUT_VIEW2DGUETE_SELECT);
    save->MakeSelectable ();
    break;

  case OUTPUT_VIEW2D_SELECT:
    Calculate_scale_2d (&(Dataset->Lese_Daten ()));
    Calculate2dparameter (&(Dataset->Lese_Daten ()));
    if (Dataset->Lese_DatenDim () > 1) {
      axismenu->DelMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX1_ACTION);
      axismenu->DelMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX2_ACTION);
    }
    axismenu->AddMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX3_ACTION);
    actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_MOVE);
    Action (OUTPUT_REDRAW, 0);
    Scroll->UnmapMe ();
    UpdateOutput ();
    XDefineCursor (display, window, maincursor);
    actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_ROLL);
    if (clustering->Lese_Clusteranzahl () == 0) {
      actionselecticons->MakeIconSelectable (OUTPUTSTATUS_CREATECLUSTER_SELECT);
      actionselecticons->SelectanIcon (OUTPUTSTATUS_CREATECLUSTER_SELECT);
      if (clustering->Lese_Vorgabe_Prototypen ().Lese_Groesse () > 0) {
	actionselecticons->MakeIconSelectable (OUTPUTSTATUS_MOVECLUSTER_SELECT);
	actionselecticons->MakeIconSelectable (OUTPUTSTATUS_DESTROYCLUSTER_SELECT);
      }
    } else {
      actionselecticons->MakeIconSelectable (OUTPUTSTATUS_SELECTCLUSTER_SELECT);
      actionselecticons->SelectanIcon (OUTPUTSTATUS_SELECTCLUSTER_SELECT);
    }
    break;

  case OUTPUT_VIEW3D_SELECT:
    {
      DVektorArray *inputData = &(Dataset->Lese_Daten ());
      Calculate_scale_3d (MAX (inputData->Lese_Max (xindex), -inputData->Lese_Min (xindex)),
	  MAX (inputData->Lese_Max (yindex), -inputData->Lese_Min (yindex)),
	 MAX (inputData->Lese_Max (zindex), -inputData->Lese_Min (zindex)));
    }
    axismenu->DelMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX1_ACTION);
    axismenu->DelMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX2_ACTION);
    axismenu->DelMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX3_ACTION);
    Scroll->UnmapMe ();
    Action (OUTPUT_REDRAW, 0);
    UpdateOutput ();
    actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_CREATECLUSTER_SELECT);
    actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_MOVECLUSTER_SELECT);
    actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_DESTROYCLUSTER_SELECT);
    actionselecticons->MakeIconSelectable (OUTPUTSTATUS_ROLL);
    actionselecticons->MakeIconSelectable (OUTPUTSTATUS_MOVE);
    actionselecticons->SelectanIcon (OUTPUTSTATUS_ROLL);
    if (clustering->Lese_Clusteranzahl () == 0)
      actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_SELECTCLUSTER_SELECT);
    else
      actionselecticons->MakeIconSelectable (OUTPUTSTATUS_SELECTCLUSTER_SELECT);
    Action (OUTPUTSTATUS_ROLL, 0);
    break;

  case OUTPUT_VIEW2DMEMBERTYP1_SELECT:
    {
      DVektorArray *inputData = &(Dataset->Lese_Daten ());
      Calculate_scale_2d (MAX (inputData->Lese_Max (xindex), -inputData->Lese_Min (xindex)));
      Calculate2dparameter (inputData->Lese_Min (xindex),
			    0.0,
			    inputData->Lese_Max (xindex),
			    1.0);
    }
    axismenu->DelMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX1_ACTION);
    axismenu->AddMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX2_ACTION);
    axismenu->AddMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX3_ACTION);
    if (!(Scroll->SetMinMax (0, clustering->Lese_Clusteranzahl () - 1))) {	// only if value was not changed and not OUTPUT_CHANCEDATAINDEX_ACTION called

      Action (OUTPUT_REDRAW, 0);
      UpdateOutput ();
    }
    Scroll->MapMe ();
    actionselecticons->MakeallUnselectable ();
    break;

  case OUTPUT_VIEW3DMEMBERTYP1_SELECT:
    {
      DVektorArray *inputData = &(Dataset->Lese_Daten ());
      Calculate_scale_3d (MAX (inputData->Lese_Max (xindex), -inputData->Lese_Min (xindex)),
	 MAX (inputData->Lese_Max (yindex), -inputData->Lese_Min (yindex)));
    }
    axismenu->DelMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX1_ACTION);
    axismenu->DelMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX2_ACTION);
    axismenu->AddMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX3_ACTION);
    if (!(Scroll->SetMinMax (0, clustering->Lese_Clusteranzahl () - 1))) {	// only if value was not changed and not OUTPUT_CHANCEDATAINDEX_ACTION called

      Action (OUTPUT_REDRAW, 0);
      UpdateOutput ();
    }
    Scroll->MapMe ();
    actionselecticons->MakeallUnselectable ();
    actionselecticons->MakeIconSelectable (OUTPUTSTATUS_ROLL);
    actionselecticons->SelectanIcon (OUTPUTSTATUS_ROLL);
    Action (OUTPUTSTATUS_ROLL, 0);
    actionselecticons->MakeIconSelectable (OUTPUTSTATUS_MOVE);
    break;

  case OUTPUT_VIEW2DMEMBERTYP2_SELECT:
    {
      Calculate_scale_2d (1.0,
			  1.0);
      Calculate2dparameter (0.0,
			    0.0,
			    1.0,
			    1.0);
    }
    axismenu->AddMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX1_ACTION);
    axismenu->AddMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX2_ACTION);
    axismenu->AddMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX3_ACTION);
    if (!(Scroll->SetMinMax (0,
			     MIN (Dataset->Lese_Daten ().Lese_Groesse (),
			 clustering->Lese_Zuordnung ().Lese_Dim ()) - 1))) {	// only if value was not changed and not OUTPUT_CHANCEDATAINDEX_ACTION called

      Action (OUTPUT_REDRAW, 0);
      UpdateOutput ();
    }
    Scroll->MapMe ();
    actionselecticons->MakeallUnselectable ();
    break;

  case OUTPUT_VIEW3DMEMBERTYP2_SELECT:
    Calculate_scale_3d (1.0, 1.0, 1.0);
    axismenu->AddMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX1_ACTION);
    axismenu->AddMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX2_ACTION);
    axismenu->AddMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX3_ACTION);
    Scroll->UnmapMe ();
    Action (OUTPUT_REDRAW, 0);
    UpdateOutput ();
    actionselecticons->MakeallUnselectable ();
    actionselecticons->MakeIconSelectable (OUTPUTSTATUS_ROLL);
    actionselecticons->MakeIconSelectable (OUTPUTSTATUS_MOVE);
    actionselecticons->SelectanIcon (OUTPUTSTATUS_ROLL);
    Action (OUTPUTSTATUS_ROLL, 0);
    break;

  case OUTPUT_VIEW2DGUETE_SELECT:
    {
      Calculate_scale_2d (1.0,
			  1.0);
      Calculate2dparameter (0.0,
			    0.0,
			    1.0,
			    1.0);
    }
    axismenu->AddMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX1_ACTION);
    axismenu->AddMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX2_ACTION);
    axismenu->AddMenuStatus (STATUS_UNSELECTABLE, OUTPUT_CHANCEINDEX3_ACTION);
    Scroll->UnmapMe ();
    Action (OUTPUT_REDRAW, 0);
    UpdateOutput ();
    actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_ROLL);
    actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_CREATECLUSTER_SELECT);
    actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_MOVECLUSTER_SELECT);
    actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_DESTROYCLUSTER_SELECT);
    break;

  case OUTPUT_CHANCEDATAINDEX_ACTION:
  case OUTPUT_CLUSTERTYPSYMBOL_SELECT:
  case OUTPUT_CLUSTERTYPELLIPSIS_SELECT:
  case OUTPUT_CLUSTERTYPSTAR_SELECT:
  case OUTPUT_CLUSTERTYPCOLOR_SELECT:
  case OUTPUT_POINTTYPPOINT:
  case OUTPUT_POINTTYPCROSS:
    Action (OUTPUT_REDRAW, 0);
    UpdateOutput ();
    break;

  case OUTPUT_CREATERULES_ACTION:
    {
      RuleInfo *ruleinfo = new
      RuleInfo (clustering,
		Dataset->Lese_InputDim (),
		&clustering->Lese_relevant (),
		selectedcluster,
		Dataset);
      ruleinfo->datasetname = myicon->Gettext ();
      RuleOptionWindow *ruleoptionwindow = new RuleOptionWindow (display,
								 gc,
								 this,
								 fontstruct,
								 10, 10,
							      "Ruleoptions",
					     OUTPUT_RULESSELECTION_FINISHED,
								 ruleinfo,
								 RULEOPTION);
      classify->MakeUnselectable ();
      makerules->MakeUnselectable ();
      myicon->GetMainwindow ()->SetStatus (MAINRULEOPTIONDISPLAYINGSTATUS);
      ruleoptionwindow->MapMe ();
    }
    break;

  case OUTPUT_RULESSELECTION_FINISHED:
    if ((RuleInfo *) value != NULL) {
      myicon->GetMainwindow ()->Action (MAIN_CREATERULE_ACTION, value);
      delete ((RuleInfo *) value);
    }
    myicon->GetMainwindow ()->SetStatus (MAINNORMALSTATUS);
    makerules->MakeSelectable ();
    classify->MakeSelectable ();
    break;

  case OUTPUT_CLASSIFY_ACTION:
    {
      RuleInfo *ruleinfo = new RuleInfo (clustering,
					 Dataset->Lese_InputDim (),
					 &clustering->Lese_relevant (),
					 selectedcluster,
					 Dataset);
      ruleinfo->datasetname = myicon->Gettext ();
      RuleOptionWindow *ruleoptionwindow = new RuleOptionWindow (display,
								 gc,
								 this,
								 fontstruct,
								 10, 10,
							 "Classify-options",
					  OUTPUT_CLASSIFYSELECTION_FINISHED,
								 ruleinfo,
							    CLASSIFYOPTION);
      classify->MakeUnselectable ();
      makerules->MakeUnselectable ();
      myicon->GetMainwindow ()->SetStatus (MAINCLASSIFYOPTIONDISPLAYINGSTATUS);
      ruleoptionwindow->MapMe ();
    }
    break;

  case OUTPUT_CLASSIFYSELECTION_FINISHED:
    if ((RuleInfo *) value != NULL) {
      myicon->GetMainwindow ()->Action (MAIN_CREATECLASSIFY_ACTION, value);
      delete ((RuleInfo *) value);
    }
    classify->MakeSelectable ();
    makerules->MakeSelectable ();
    myicon->GetMainwindow ()->SetStatus (MAINNORMALSTATUS);
    break;

  case OUTPUT_DISPLAYCLUSTEROPTIONS_ACTION:
    if (optionwindow == NULL) {
      ClusterInfo cinfo (Dataset,
			 clustering->Lese_Possib (),
			 clustering->Lese_Typ (),
			 clusternumber,		//clustering->Lese_Clusteranzahl(),
			  clustering->Lese_Max_Iteration (),
			 clustering->Lese_M (),
			 clustering->Lese_K (),
			 exactness,
			 clustering->Lese_Delta_U (),
			 validitytyp
      /* clustering->Lese_Guete().Lese_Typ() */ ,
			 &clustering->Lese_relevant ());
      optionwindow = new ClusterOptionWindow (display,
					      gc,
					      this,
					      fontstruct,
					      20, 20,
					      "Change cluster-options",
					      OUTPUT_CLUSTEROPTION_OK,
					      &cinfo);
      myicon->GetMainwindow ()->SetStatus (MAINCLUSTEROPTIONDISPLAYINGSTATUS);
    }
    break;

  case OUTPUT_CLUSTEROPTION_OK:
    if ((ClusterInfo *) value != NULL) {
      SetClustertyp ((ClusterInfo *) value);
      if (Dataset != NULL) {
#if 0
	clustering.Setze_relevant (((ClusterInfo *) value)->resultbvector);
#else
	clustering->Setze_relevant (((ClusterInfo *) value)->resultbvector);
#endif
      }
    }
    optionwindow = NULL;
    myicon->GetMainwindow ()->SetStatus (MAINNORMALSTATUS);
    break;

  case OUTPUTSTATUS_MOVECLUSTER_SELECT:
    XDefineCursor (display, window, movepointcursor);
    break;

  case OUTPUTSTATUS_CREATECLUSTER_SELECT:
  case OUTPUTSTATUS_DESTROYCLUSTER_SELECT:
  case OUTPUTSTATUS_ROLL:
  case OUTPUTSTATUS_MOVE:
    XDefineCursor (display, window, maincursor);
    break;

  case OUTPUT_GHOST:
    Unmapmyicons ();
    break;

  case OUTPUT_DATASETMODIFIED:
    if (Dataset->Lese_Daten ().Lese_Groesse () != deepindex.Lese_Dim ()) {
      DVektorArray *inputData = &(Dataset->Lese_Daten ());
      short counter;
      deepvalues.DVektor (inputData->Lese_Groesse (),
			  0,
			  NULL);
      deepindex.IVektor (inputData->Lese_Groesse (),
			 NULL);
      for (counter = inputData->Lese_Groesse () - 1; counter > 0; counter--)
	deepindex[counter] = counter;
    }
    Action (OUTPUT_REDRAW, 0);
    UpdateOutput ();
    break;
  }
}

DVektor *
ClusterOutput::Getnearestusercluster1d (short xpos,
					short *xis, short *yis)
{
  short tempx, tempy;
  int counter, nearestcounter = 0, intdummy;
  double xoutput;
  double distance, mindistance = MAXDOUBLE;	// distance high 2

  PixeltoWorld1d (0, xpos, &intdummy, &xoutput);

  for (counter = 0; counter < clustering->Lese_Vorgabe_Prototypen ().Lese_Groesse (); counter++) {
    WorldtoPixel1d (0,
       clustering->Lese_Vorgabe_Prototypen ().Lese_Vektor (counter, xindex),
		    &tempx, &tempy);
    distance = (clustering->Lese_Vorgabe_Prototypen ().Lese_Vektor (counter, xindex) - xoutput) *
      (clustering->Lese_Vorgabe_Prototypen ().Lese_Vektor (counter, xindex) - xoutput);

    if (distance < mindistance) {
      mindistance = distance;
      nearestcounter = counter;
    }
  }
  WorldtoPixel1d (0,
		  clustering->Lese_Vorgabe_Prototypen ().Lese_Vektor (nearestcounter, xindex),
		  xis, yis);
  return &(clustering->Lese_Vorgabe_Prototypen ().Lese_Vektor (nearestcounter));
}

DVektor *
ClusterOutput::Getnearestusercluster2d (short xpos, short ypos,
					short *xis, short *yis)
{
  short tempx, tempy;
  int counter, nearestcounter = 0;
  double xoutput, youtput;
  double distance, mindistance = MAXDOUBLE;	// distance high 2

  PixeltoWorld2d (xpos, ypos, &xoutput, &youtput);

  for (counter = 0; counter < clustering->Lese_Vorgabe_Prototypen ().Lese_Groesse (); counter++) {
    WorldtoPixel2d (clustering->Lese_Vorgabe_Prototypen ().Lese_Vektor (counter, xindex),
       clustering->Lese_Vorgabe_Prototypen ().Lese_Vektor (counter, yindex),
		    &tempx, &tempy);
    distance = ((clustering->Lese_Vorgabe_Prototypen ().Lese_Vektor (counter, xindex) - xoutput) * (clustering->Lese_Vorgabe_Prototypen ().Lese_Vektor (counter, xindex) - xoutput)) +
      ((clustering->Lese_Vorgabe_Prototypen ().Lese_Vektor (counter, yindex) - youtput) * (clustering->Lese_Vorgabe_Prototypen ().Lese_Vektor (counter, yindex) - youtput));
    if (distance < mindistance) {
      mindistance = distance;
      nearestcounter = counter;
    }
  }
  WorldtoPixel2d (clustering->Lese_Vorgabe_Prototypen ().Lese_Vektor (nearestcounter, xindex),
		  clustering->Lese_Vorgabe_Prototypen ().Lese_Vektor (nearestcounter, yindex),
		  xis, yis);

  return &(clustering->Lese_Vorgabe_Prototypen ().Lese_Vektor (nearestcounter));
}


short 
ClusterOutput::GetnearestClusterindex (int from_me)
{
  double membershipvalue = 0;
  short maxcluster = 0, clustercounter = 0;
  DMatrix *membership = &(clustering->Lese_Zugehoerigkeit ());

  for (clustercounter = 0; clustercounter < clustering->Lese_Clusteranzahl (); clustercounter++)
    if (membership->Lese_i_j (clustercounter, from_me) > membershipvalue) {
      membershipvalue = membership->Lese_i_j (clustercounter, from_me);
      maxcluster = clustercounter;
    }
  return maxcluster;
}


void 
ClusterOutput::HandleEvent (XEvent * Event)
{
  static Time lastpresstime = 0;
  static ClusterOutput *lastpressoutput = NULL;

  static short oldx, oldy;
  static DVektor *nearest = NULL;
  static char buttonpressedstatus = 0, acutebutton = 0;
  switch (Event->type) {
  case Expose:
    XCopyArea (display, pixmap, window, gc,
	       Event->xgraphicsexpose.x, Event->xgraphicsexpose.y,
	       Event->xgraphicsexpose.width, Event->xgraphicsexpose.height,
	       Event->xgraphicsexpose.x, Event->xgraphicsexpose.y);
    break;

  case UnmapNotify:
    XWithdrawWindow (display, window, DefaultScreen (display));
    break;

  case ConfigureNotify:
    if ((Event->xconfigure.width != width) ||
	(Event->xconfigure.height != height)) {
      short oldwidth = width;
      Resized (Event->xconfigure.width,
	       Event->xconfigure.height);
      Action (OUTPUT_RESIZED, width - oldwidth);
//         Action(OUTPUT_REDRAW,0);
    }
    break;

  case ButtonPress:
    RaiseMe ();
    switch (actionselecticons->GetSelectedvalue ()) {
    case OUTPUTSTATUS_MOVE:
      oldx = Event->xbutton.x;
      oldy = Event->xbutton.y;
      switch (Event->xbutton.button) {
      case Button1:
	buttonpressedstatus |= BUTTON1PRESSED;
	acutebutton = Button1;
	break;

      case Button2:
	buttonpressedstatus |= BUTTON2PRESSED;
	acutebutton = Button2;
	break;

      case Button3:
	buttonpressedstatus |= BUTTON3PRESSED;
	acutebutton = Button3;
	axisframe[0][3] = -0.5;
	axisframe[1][3] = -0.5;
	axisframe[2][3] = -1.0;
	Action (OUTPUT_REDRAW, 0);
	UpdateOutput ();
	break;
      }

      if ((Event->xbutton.button == Button1) ||
	  (Event->xbutton.button == Button2))
	XDefineCursor (display, window, moveaxescursor);
      break;			//case OUTPUTSTATUS_MOVE

    case OUTPUTSTATUS_CREATECLUSTER_SELECT:
      if (actionselecticons->IsMapped ()) {
	short tempx, tempy;
	double newx, newy;
	int dummy;
	if (clustering->Lese_Vorgabe_Prototypen ().Lese_Dim () == 0)
	  clustering->Lese_Vorgabe_Prototypen ().Setze_Dim (clustering->Lese_relevant ().Lese_Dim ());
	int dim = clustering->Lese_Vorgabe_Prototypen ().Lese_Dim ();

	if (Dataset->Lese_DatenDim () > 1)
	  PixeltoWorld2d (Event->xbutton.x,
			  Event->xbutton.y,
			  &newx, &newy);
	else
	  PixeltoWorld1d (Event->xbutton.x,
			  Event->xbutton.y,
			  &dummy, &newx);
	DVektor *newcluster = new DVektor;
	newcluster->Setze_Dim (dim);
	newcluster->Setze_i (xindex, newx);
	if (Dataset->Lese_DatenDim () > 1) {
	  WorldtoPixel2d (newx, newy, &tempx, &tempy);
	  newcluster->Setze_i (yindex, newy);
	} else
	  WorldtoPixel1d (0, newx, &tempx, &tempy);
	clustering->Lese_Vorgabe_Prototypen () += (*newcluster);
	DrawClusterPoint (tempx, tempy, 0);
	DrawClusterPoint (tempx, tempy, DRAWINWINDOW);
	actionselecticons->MakeIconSelectable (OUTPUTSTATUS_MOVECLUSTER_SELECT);
	actionselecticons->MakeIconSelectable (OUTPUTSTATUS_DESTROYCLUSTER_SELECT);
      }
      break;			// case OUTPUTSTATUS_CREATECLUSTER_SELECT:

    case OUTPUTSTATUS_MOVECLUSTER_SELECT:
      if (Event->xbutton.button == Button1)
	buttonpressedstatus |= BUTTON1PRESSED;

      if (actionselecticons->IsMapped ()) {
	if (Dataset->Lese_DatenDim () > 1)
	  nearest = Getnearestusercluster2d (Event->xbutton.x, Event->xbutton.y, &oldx, &oldy);
	else
	  nearest = Getnearestusercluster1d (Event->xbutton.y, &oldx, &oldy);
	XWarpPointer (display, window, window,
		      0, 0, width, height, oldx, oldy);
	UndrawClusterPoint (oldx, oldy, 0);
	DrawAxes2d (0);		// perhaps one of the axes is destroyed

	UndrawClusterPoint (oldx, oldy, DRAWINWINDOW);
	DrawClusterPoint (oldx, oldy, DRAWINWINDOW | DRAWTEMP);
      }
      break;			// case OUTPUTSTATUS_MOVECLUSTER_SELECT:

    case OUTPUTSTATUS_DESTROYCLUSTER_SELECT:
      if (Event->xbutton.button == Button1)
	buttonpressedstatus |= BUTTON1PRESSED;

      if (actionselecticons->IsMapped ()) {
	if (Dataset->Lese_DatenDim () > 1)
	  nearest = Getnearestusercluster2d (Event->xbutton.x, Event->xbutton.y, &oldx, &oldy);
	else
	  nearest = Getnearestusercluster1d (Event->xbutton.x, &oldx, &oldy);
	XWarpPointer (display, window, window,
		      0, 0, width, height, oldx, oldy);
	UndrawClusterPoint (oldx, oldy, 0);
	DrawAxes2d (0);		// perhaps one of the axes is destroyed

	UpdateOutput ();
	clustering->Lese_Vorgabe_Prototypen () -= (*nearest);
	if (clustering->Lese_Vorgabe_Prototypen ().Lese_Groesse () == 0) {
	  actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_MOVECLUSTER_SELECT);
	  actionselecticons->MakeIconUnselectable (OUTPUTSTATUS_DESTROYCLUSTER_SELECT);
	}
      }
      break;			// case OUTPUTSTATUS_DESTROYCLUSTER_SELECT:

    case OUTPUTSTATUS_SELECTCLUSTER_SELECT:
      switch (Event->xbutton.button) {
      case Button1:
#if 0
	DVektor * nearest = NULL;
#endif
	int nearestindex;
	if (Event->xbutton.button == Button1)
	  buttonpressedstatus |= BUTTON1PRESSED;
	switch (viewicons->GetSelectedvalue ()) {
	case OUTPUT_VIEW2D_SELECT:
	  if (Dataset->Lese_DatenDim () > 1)
	    nearest = Getnearestpoint2d (Event->xbutton.x, Event->xbutton.y,
					 &nearestindex, &oldx, &oldy);
	  else
	    nearest = Getnearestpoint1d (Event->xbutton.x, Event->xbutton.y,
					 &nearestindex, &oldx, &oldy);
	  break;

	case OUTPUT_VIEW3D_SELECT:
	  nearest = Getnearestpoint3d (Event->xbutton.x, Event->xbutton.y,
				       &nearestindex, &oldx, &oldy);
	  break;
	}
	if (nearest != NULL) {
	  short index;
	  index = GetnearestClusterindex (nearestindex);
	  if (selectedcluster->Lese_i (index))
	    selectedcluster->Setze_i (index, 0);
	  else
	    selectedcluster->Setze_i (index, 1);
	  Action (OUTPUT_REDRAW, 0);
	  UpdateOutput ();
	  if (selectedcluster->Anzahl_gesetzt () > 0) {
	    makerules->MakeSelectable ();
	    classify->MakeSelectable ();
	  } else {
	    makerules->MakeUnselectable ();
	    classify->MakeUnselectable ();
	  }
	}
	break;			// case Button1:

      case Button2:		// select all clusters
	{
	  short counter;
	  for (counter = 0; counter < selectedcluster->Lese_Dim (); counter++)
	    selectedcluster->Setze_i (counter, 1);
	  Action (OUTPUT_REDRAW, 0);
	  UpdateOutput ();
	  if (selectedcluster->Anzahl_gesetzt () > 0)
	    makerules->MakeSelectable ();
	  else
	    makerules->MakeUnselectable ();
	}
	break;			// case Button2:

      case Button3:		// uselect all clusters
	{
	  short counter;
	  for (counter = 0; counter < selectedcluster->Lese_Dim (); counter++)
	    selectedcluster->Setze_i (counter, 0);
	  Action (OUTPUT_REDRAW, 0);
	  UpdateOutput ();
	  if (selectedcluster->Anzahl_gesetzt () > 0)
	    makerules->MakeSelectable ();
	  else
	    makerules->MakeUnselectable ();
	}
	break;			// case Button3:

      }
      break;			// case OUTPUTSTATUS_SELECTCLUSTER_SELECT:

    case OUTPUTSTATUS_ROLL:
      switch (Event->xbutton.button) {
      case Button1:
	oldx = Event->xbutton.x;
	oldy = Event->xbutton.y;
	buttonpressedstatus |= BUTTON1PRESSED;
	acutebutton = Button1;
	break;

      case Button2:
	oldx = Event->xbutton.x;
	oldy = Event->xbutton.y;
	buttonpressedstatus |= BUTTON2PRESSED;
	acutebutton = Button2;
	break;

      case Button3:
	oldx = Event->xbutton.x;
	oldy = Event->xbutton.y;
	buttonpressedstatus |= BUTTON3PRESSED;
	acutebutton = Button3;
	break;
      }
      XDefineCursor (display, window, rotatecursor);
      break;			// case OUTPUTSTATUS_ROLL:

    }
    if (Event->xbutton.button == Button1) {
      if (((Event->xbutton.time - lastpresstime) < DOUBLECLICKTIME) &&
	  (this == lastpressoutput))
	Mapmyicons ();
      lastpresstime = Event->xbutton.time;
      lastpressoutput = this;
    }
    break;			// case ButtonPress:

  case ButtonRelease:
    switch (Event->xbutton.button) {
    case Button1:
      buttonpressedstatus &= ~BUTTON1PRESSED;
      break;

    case Button2:
      buttonpressedstatus &= ~BUTTON2PRESSED;
      break;

    case Button3:
      buttonpressedstatus &= ~BUTTON3PRESSED;
      break;
    }
    if (buttonpressedstatus == 0) {
      switch (actionselecticons->GetSelectedvalue ()) {
      case OUTPUTSTATUS_ROLL:
      case OUTPUTSTATUS_MOVE:
	XDefineCursor (display, window, maincursor);
	break;
      }
    }
    if (buttonpressedstatus == 0)
      Action (OUTPUT_REDRAW, 0);
    else if (buttonpressedstatus & BUTTON1PRESSED)
      acutebutton = Button1;
    else if (buttonpressedstatus & BUTTON2PRESSED)
      acutebutton = Button2;
    else
      acutebutton = Button3;

    switch (actionselecticons->GetSelectedvalue ()) {
    case OUTPUTSTATUS_MOVECLUSTER_SELECT:
      Action (OUTPUT_REDRAW, 0);
      UpdateOutput ();
      break;			// case OUTPUTSTATUS_MOVECLUSTER_SELECT:

    }
    break;			// case ButtonRelease:

  case MotionNotify:
    switch (actionselecticons->GetSelectedvalue ()) {
    case OUTPUTSTATUS_ROLL:
      if (buttonpressedstatus != 0)	//)
       {
	int xdiff, ydiff;
	int newx, newy;

	newx = Event->xmotion.x;
	newy = Event->xmotion.y;

	xdiff = newx - oldx;
	ydiff = newy - oldy;

	switch (acutebutton) {
	case Button1:
	  if (xdiff != 0)
	    AxisRoty (ANGLEPERPIXEL * xdiff);
	  if (ydiff != 0)
	    AxisRotx (ANGLEPERPIXEL * ydiff);
	  break;

	case Button2:
	  if (xdiff != 0)
	    AxisRoty (ANGLEPERPIXEL * xdiff);
	  if (ydiff != 0)
	    AxisRotz (ANGLEPERPIXEL * ydiff);
	  break;

	case Button3:
	  if (xdiff != 0)
	    AxisRotz (ANGLEPERPIXEL * xdiff);
	  if (ydiff != 0)
	    AxisRotx (ANGLEPERPIXEL * ydiff);
	  break;
	}

	Action (OUTPUT_REDRAWTOWINDOW, 0);

	oldx = newx;
	oldy = newy;
      }
      break;			// case OUTPUTSTATUS_ROLL:

    case OUTPUTSTATUS_MOVE:
      if (buttonpressedstatus != 0) {
	int xdiff, ydiff;
	int newx, newy;
	newx = Event->xmotion.x;
	newy = Event->xmotion.y;

	xdiff = newx - oldx;
	ydiff = newy - oldy;

	switch (acutebutton) {
	case Button1:
	  axisframe[0][3] += MOVEFACTOR * xdiff;
	  axisframe[1][3] -= MOVEFACTOR * ydiff;
	  break;

	case Button2:
	  axisframe[2][3] -= MOVEFACTOR * xdiff;
	  axisframe[2][3] = MIN (axisframe[2][3], 1.0 / 2);
	  break;
	}

	if ((acutebutton == Button1) ||
	    (acutebutton == Button2))
	  Action (OUTPUT_REDRAWTOWINDOW, 0);

	oldx = newx;
	oldy = newy;
      }
      break;			// case OUTPUTSTATUS_MOVE

    case OUTPUTSTATUS_CREATECLUSTER_SELECT:
      DrawCoordinates (Event->xmotion.x, Event->xmotion.y);
      break;			// case OUTPUTSTATUS_CREATECLUSTER_SELECT:

    case OUTPUTSTATUS_MOVECLUSTER_SELECT:
      {
	double newx, newy;
	int intdummy;
	if (Dataset->Lese_DatenDim () > 1)
	  PixeltoWorld2d (Event->xmotion.x, Event->xmotion.y,
			  &newx, &newy);
	else
	  PixeltoWorld1d (Event->xbutton.x,
			  Event->xbutton.y,
			  &intdummy, &newx);

	if (buttonpressedstatus & BUTTON1PRESSED) {
	  UndrawClusterPoint (oldx, oldy, DRAWINWINDOW | DRAWTEMP);
	  nearest->Setze_i (xindex, newx);
	  if (Dataset->Lese_DatenDim () > 1) {
	    nearest->Setze_i (yindex, newy);
	    DrawClusterPoint (Event->xmotion.x, Event->xmotion.y, DRAWINWINDOW | DRAWTEMP);
	    oldx = Event->xmotion.x;
	  } else {
	    short tempx, tempy;
	    WorldtoPixel1d (0, newx, &tempx, &tempy);
	    oldx = tempx;
	    DrawClusterPoint (tempx, Event->xmotion.y, DRAWINWINDOW | DRAWTEMP);
	  }
	}
	oldy = Event->xmotion.y;
	DrawCoordinates (Event->xmotion.x, Event->xmotion.y);
      }
      break;			// case OUTPUTSTATUS_MOVECLUSTER_SELECT:

    case OUTPUTSTATUS_SELECTCLUSTER_SELECT:
      DrawCoordinates (Event->xmotion.x, Event->xmotion.y);
      break;			// case OUTPUTSTATUS_CREATECLUSTER_SELECT:

    default:
      if (viewicons->GetSelectedvalue () == OUTPUT_VIEW2DMEMBERTYP1_SELECT)
	DrawCoordinates (Event->xmotion.x, Event->xmotion.y);
      break;
    }
    break;			// case MotionNotify:

  case KeyPress:
    /* 
     * process key hit:
     */
    char c;
    int n = XLookupString((XKeyEvent *) Event, &c, 1, NULL, NULL);
    if (n == 1 && c == 'p') {
      DumpBitmap();
    }
    break;

  }
}


RuleOutput::RuleOutput (Display * initdisplay,
	    GC initgc,
	    RuleIcon * initOutputicon,
	    XFontStruct * fontstruct,
	    int initx, int inity,
	    int initwidth, int initheight,
	    char *text,
	    RuleInfo * Ruleinfo):
Output2d (initdisplay,
	  initgc,
	  initOutputicon,
	  fontstruct,
	  initx, inity,
	  initwidth, initheight,
	  text)
{
  input.Setze_Dim (0);
  menunumber = 0;
  indexmenues = NULL;
  acuteruleresultinfo = NULL;
  SetSelectedInput (ExposureMask
		    | StructureNotifyMask
		    | ButtonMotionMask
		    | ButtonPressMask
		    | ButtonReleaseMask
		    | KeyPressMask);

  rules = Ruleinfo->rules;
  selectedinputindex.Setze_Dim (rules.Lese_xDimension () + rules.Lese_yDimension ());

  short counter;
  selectedinputs.Setze_Dim (rules.Lese_xDimension ());
  for (counter = 0; counter < selectedinputs.Lese_Dim (); counter++)
    selectedinputs.Setze_i (counter, TRUE);

  selectedoutputs.Setze_Dim (rules.Lese_yDimension ());
  for (counter = 0; counter < selectedoutputs.Lese_Dim (); counter++)
    selectedoutputs.Setze_i (counter, TRUE);

  selectedrules.Setze_Dim (rules.Lese_Regelanzahl ());
  for (counter = 0; counter < selectedrules.Lese_Dim (); counter++)
    selectedrules.Setze_i (counter, TRUE);

  graphxpos.Setze_Dim (rules.Lese_xDimension () + rules.Lese_yDimension ());
  graphypos.Setze_Dim (rules.Lese_Regelanzahl ());

  Action (OUTPUT_REDRAW, 0);
}

RuleOutput::~RuleOutput ()
{
  short menucounter;
  if (indexmenues != NULL)
    for (menucounter = 0; menucounter < menunumber; menucounter++)
      indexmenues[menucounter]->UnmapMe ();

/*if(acuteruleresultinfo!=NULL)
   delete acuteruleresultinfo; */
}

void 
RuleOutput::Unmapmyicons ()
{
  short menucounter;
  ghost->UnmapMe ();
  if (indexmenues != NULL)
    for (menucounter = 0; menucounter < menunumber; menucounter++)
      indexmenues[menucounter]->UnmapMe ();
}

void 
RuleOutput::Mapmyicons ()
{
  short menucounter;
  ghost->MapMe ();
  if (indexmenues != NULL)
    for (menucounter = 0; menucounter < menunumber; menucounter++)
      indexmenues[menucounter]->MapMe ();
}

void 
RuleOutput::DrawInputinGraph (int xpos, int ypos,
			      int drawwidth, int drawheight,
			      double range, double minxvalue,
			      short index,
			      short rulenumber)
{
  XSegment Line;
  XSetForeground (display, gc, redpix);
  double xvalue = input.Lese_i (index);
  double yvalue = rulesoutput[rulenumber].Lese_InputHoehen ()[index];
  Line.x1 = Line.x2 = xpos + (int) (drawwidth * ((xvalue - minxvalue) / range));
  Line.y1 = ypos;
  Line.y2 = ypos - drawheight;
  if ((Line.x1 >= xpos) && (Line.x1 <= (xpos + drawwidth)))
    XDrawLine (display, pixmap, gc, Line.x1, Line.y1, Line.x2, Line.y2);
  Line.x1 = xpos;
  Line.x2 = xpos + drawwidth;
  Line.y1 = Line.y2 = ypos - (int) (yvalue * drawheight);
  XDrawLine (display, pixmap, gc, Line.x1, Line.y1, Line.x2, Line.y2);
}

void 
RuleOutput::DrawOutputinGraph (int xpos, int ypos,
			       int drawwidth, int drawheight,
			       double range, double minxvalue,
			       short index,
			       short ruleindex)
{
  XSetForeground (display, gc, redpix);
  XSetLineAttributes (display, gc, 3, LineSolid, CapButt, JoinMiter);
  DVektor X_Output = rulesoutput[ruleindex].Lese_X_Output (index);
  DVektor Y_Output = rulesoutput[ruleindex].Lese_Y_Output (index);
  short pointnumber = X_Output.Lese_Dim (), counter;
  XPoint *Points;
  if ((Points = (XPoint *) malloc (sizeof (XPoint) * pointnumber)) == NULL) {
    printf ("Out of memory!\n");
    exit (1);
  }
  for (counter = 0; counter < pointnumber; counter++) {
    Points[counter].x = xpos + (int) (drawwidth * ((X_Output.Lese_i (counter) - minxvalue) / range));
    Points[counter].y = ypos - (short) (drawheight * Y_Output.Lese_i (counter));
  }
  XDrawLines (display, pixmap, gc, Points, pointnumber, CoordModeOrigin);
  free (Points);

// draw result
  DVektor acuteresults = rulesoutput[ruleindex].Lese_Ergebnisse ();
  short xresult = xpos + (int) (drawwidth * (acuteresults.Lese_i (index) - minxvalue) / range);
  XFillArc (display,
	    pixmap,
	    gc,
	    xresult - SYMBOLHALFWIDTH, ypos - SYMBOLHALFWIDTH,
	    SYMBOLHALFWIDTH * 2, SYMBOLHALFWIDTH * 2,
	    0, 23040);

// final the valuetext
  char valuetext[7];
  sprintf (valuetext, "%2.3f", acuteresults.Lese_i (index));
  short textwidth = XTextWidth (fontstruct, valuetext, strlen (valuetext));
  XSetForeground (display, gc, blackpix);
  XDrawString (display, pixmap, gc,
	       xresult - textwidth / 2, ypos - 10,
	       valuetext, strlen (valuetext));

  XSetLineAttributes (display, gc, 1, LineSolid, CapProjecting, JoinMiter);
}


void 
RuleOutput::DrawRules ()
{
  int rulecounter, counter, textwidth = 0, drawingnumber, drawingwidth,
    drawingheight, xpos, ypos, textheight = fontstruct->max_bounds.ascent +
  fontstruct->max_bounds.descent;
  short resulttextypos;
  short ruletextlength = 0;

  short firsttextlength;	// used when drawing arrow

  Regel *rule;
  Parameter *parameter;
  double maxxvalue = -MAXDOUBLE;
  double minxvalue = MAXDOUBLE;
  double xrange;
  char *IS = " is ";
  char *AND = " and ";
  char *THEN = " then ";
  int rulenumber = rules.Lese_Regelanzahl ();
// first calculate drawingsize and maxxvalue

  {
    char temptext[15];		// should be enought

    sprintf (temptext, "R%d: if ", rulenumber);
    ruletextlength += strlen (temptext);
    textwidth += XTextWidth (fontstruct, temptext, strlen (temptext));
  }

  char first = 1;
  for (counter = 0; counter < rules.Lese_xDimension (); counter++) {
    if (!first) {
      ruletextlength += strlen (AND);
      textwidth += XTextWidth (fontstruct, AND, strlen (AND));
    }
    ruletextlength += strlen (rules.Lese_Namen (counter));
    textwidth += XTextWidth (fontstruct, rules.Lese_Namen (counter), strlen (rules.Lese_Namen (counter)));
    ruletextlength += strlen (IS);
    textwidth += XTextWidth (fontstruct, IS, strlen (IS));
    first = 0;
    for (rulecounter = 0; rulecounter < rulenumber; rulecounter++) {
      rule = &(rules.Lese_Regel (rulecounter));
      parameter = &(rule->Lese_Parameter (counter));
      maxxvalue = MAX (maxxvalue, parameter->Lese_xParameter (parameter->Lese_Parameter_Anzahl () - 1));
      minxvalue = MIN (minxvalue, parameter->Lese_xParameter (0));
    }
  }
  firsttextlength = textwidth;
  ruletextlength += strlen (THEN);
  textwidth += XTextWidth (fontstruct, THEN, strlen (THEN));
  first = 1;
  for (counter = rules.Lese_xDimension (); counter < rules.Lese_xDimension () + rules.Lese_yDimension (); counter++) {
    if (!first) {
      ruletextlength += strlen (AND);
      textwidth += XTextWidth (fontstruct, AND, strlen (AND));
    }
    ruletextlength += strlen (rules.Lese_Namen (counter));
    textwidth += XTextWidth (fontstruct, rules.Lese_Namen (counter), strlen (rules.Lese_Namen (counter)));
    ruletextlength += strlen (IS);
    textwidth += XTextWidth (fontstruct, IS, strlen (IS));
    first = 0;
    for (rulecounter = 0; rulecounter < rulenumber; rulecounter++) {
      rule = &(rules.Lese_Regel (rulecounter));
      parameter = &(rule->Lese_Parameter (counter));
      maxxvalue = MAX (maxxvalue, parameter->Lese_xParameter (parameter->Lese_Parameter_Anzahl () - 1));
      minxvalue = MIN (minxvalue, parameter->Lese_xParameter (0));
    }
  }

  minxvalue = MIN (0.0, minxvalue);
  xrange = maxxvalue - minxvalue;

  drawingnumber = rules.Lese_xDimension () +
    rules.Lese_yDimension ();
  drawingwidth = (int) (width - textwidth - SPACE_BETWEEN_IN_AND_OUT - 2 * RULE_BORDER) / drawingnumber;
  drawingheight = (int) (height - 2 * textheight - 2 * RULE_BORDER - ((rulenumber - 1) * RULE_VERT_SPACE)) / rulenumber;

// **************************************************************
  // resize window if drawing would be too small
  char resize = 0;
  int newheight, newwidth;
  if (drawingwidth < RULE_MINDRAWWIDTH) {
    drawingwidth = RULE_MINDRAWWIDTH;
    newwidth = textwidth + SPACE_BETWEEN_IN_AND_OUT + 2 * RULE_BORDER + drawingnumber * drawingwidth;
    resize = 1;
  } else
    newwidth = width;

  if (drawingheight < RULE_MINDRAWHEIGHT) {
    drawingheight = RULE_MINDRAWHEIGHT;
    newheight = 2 * RULE_BORDER +
      rulenumber * drawingheight +
      (rulenumber - 1) * RULE_VERT_SPACE
      + 2 * textheight;
    resize = 1;
  } else
    newheight = height;

  if (resize)
    Resized (newwidth, newheight);

  XSetForeground (display, gc, whitepix);
  XFillRectangle (display, pixmap, gc, 0, 0, width, height);

// **************************************************************
  // draw rules:
  // first draw arrow
  {
//     2        heighesty
    //     |\
    //     | \
    //  0---1 \
    //   |     \3
    //   |     /
    //  6---5 /
    //     | /
    //     |/
    //     4        lowesty
    short heighesty,		// used when drawing
      lowesty,			// the arrow between in and out
      arrowwidth = 3 * SPACE_BETWEEN_IN_AND_OUT / 2;	//

    xpos = firsttextlength +
      RULE_BORDER +
      rules.Lese_xDimension () * drawingwidth;
    heighesty = RULE_BORDER;
    lowesty = drawingheight * rulenumber +
      (rulenumber - 1) * RULE_VERT_SPACE + RULE_BORDER;
    short arrowheight = lowesty - heighesty;
    XSetForeground (display, gc, buttonpix);
    XPoint points[7];
    points[0].x = points[6].x = xpos;
    points[1].x = points[2].x = points[4].x = points[5].x = xpos + arrowwidth / 4;
    points[3].x = xpos + arrowwidth;

    points[0].y = points[1].y = heighesty + arrowheight / 4;
    points[5].y = points[6].y = heighesty + 3 * arrowheight / 4;
    points[2].y = heighesty;
    points[4].y = lowesty;
    points[3].y = heighesty + arrowheight / 2;
    XFillPolygon (display, pixmap, gc, points, 7, Convex, CoordModeOrigin);
  }

// **************************************************************
  // second R1: ... Rn:
  char *ruletext = (char *) malloc (sizeof (char) * (ruletextlength + 1));
  xpos = RULE_BORDER;
  for (rulecounter = 0; rulecounter < rulenumber; rulecounter++) {
    sprintf (ruletext, "R%d: if ", rulecounter + 1);
    if (selectedrules.Lese_i (rulecounter))
      XSetForeground (display, gc, blackpix);
    else
      XSetForeground (display, gc, shadowpix);
    XDrawString (display, pixmap, gc, xpos, (drawingheight) * (rulecounter + 1) + rulecounter * RULE_VERT_SPACE + RULE_BORDER, ruletext, strlen (ruletext));
  }

  xpos += XTextWidth (fontstruct, ruletext, strlen (ruletext));

// **************************************************************
  // then the inputs
  first = 1;
  for (counter = 0; counter < rules.Lese_xDimension (); counter++) {
    ruletext[0] = 0;
    if (!first)
      strcat (ruletext, AND);
    strcat (ruletext, rules.Lese_Namen (counter));
    strcat (ruletext, IS);
    first = 0;
    for (rulecounter = 0; rulecounter < rulenumber; rulecounter++) {
      ypos = (drawingheight) * (rulecounter + 1) +
	rulecounter * RULE_VERT_SPACE + RULE_BORDER;
      graphypos.Setze_i (rulecounter, ypos + RULE_VERT_SPACE / 2);
      if ((selectedrules.Lese_i (rulecounter)) &&
	  (selectedinputs.Lese_i (counter)))
	XSetForeground (display, gc, blackpix);
      else
	XSetForeground (display, gc, shadowpix);
      XDrawString (display, pixmap, gc, xpos, ypos, ruletext, strlen (ruletext));
      rule = &(rules.Lese_Regel (rulecounter));
      parameter = &(rule->Lese_Parameter (counter));
      DrawGraph (parameter, xpos + XTextWidth (fontstruct, ruletext, strlen (ruletext)), ypos,
		 drawingwidth, drawingheight, xrange, minxvalue);
      if ((status == OUTPUT_WORKASSINGLEPOINTCALCULATOR) &&
	  (selectedrules.Lese_i (rulecounter)) &&
	  (selectedinputs.Lese_i (counter)))
	DrawInputinGraph (xpos + XTextWidth (fontstruct, ruletext, strlen (ruletext)), ypos,
			  drawingwidth, drawingheight, xrange, minxvalue,
			  counter, rulecounter);
    }

    if (indexmenues != NULL)	// move menues over graphes
     {
      short menuxpos = xpos + XTextWidth (fontstruct, ruletext, strlen (ruletext));
      menuxpos += (drawingwidth - indexmenues[counter]->Getwidth ()) / 2;
      indexmenues[counter]->PlaceMe (menuxpos, 0);
    }
    xpos += XTextWidth (fontstruct, ruletext, strlen (ruletext)) + drawingwidth;
    graphxpos.Setze_i (counter, xpos);
  }

  xpos += SPACE_BETWEEN_IN_AND_OUT;
// **************************************************************
  // then outputs
  resulttextypos = 2 * RULE_BORDER +
    rulenumber * drawingheight +
    (rulenumber - 1) * RULE_VERT_SPACE
    + textheight;
  first = 1;
  for (counter = rules.Lese_xDimension (); counter < rules.Lese_xDimension () + rules.Lese_yDimension (); counter++) {
    ruletext[0] = 0;
    if (!first)
      strcat (ruletext, AND);
    else
      strcat (ruletext, THEN);
    strcat (ruletext, rules.Lese_Namen (counter));
    strcat (ruletext, IS);
    first = 0;
    for (rulecounter = 0; rulecounter < rulenumber; rulecounter++) {
      if ((selectedrules.Lese_i (rulecounter)) &&
	  (selectedoutputs.Lese_i (counter - rules.Lese_xDimension ())))
	XSetForeground (display, gc, blackpix);
      else
	XSetForeground (display, gc, shadowpix);
      ypos = (drawingheight) * (rulecounter + 1) + rulecounter * RULE_VERT_SPACE + RULE_BORDER,
	XDrawString (display, pixmap, gc, xpos, ypos, ruletext, strlen (ruletext));
      rule = &(rules.Lese_Regel (rulecounter));
      parameter = &(rule->Lese_Parameter (counter));
      DrawGraph (parameter,
	  xpos + XTextWidth (fontstruct, ruletext, strlen (ruletext)), ypos,
		 drawingwidth, drawingheight, xrange, minxvalue);
      if ((status == OUTPUT_WORKASSINGLEPOINTCALCULATOR) &&
	  (selectedrules.Lese_i (rulecounter)) &&
	  (selectedoutputs.Lese_i (counter - rules.Lese_xDimension ())))
	DrawOutputinGraph (xpos + XTextWidth (fontstruct, ruletext, strlen (ruletext)), ypos,
			   drawingwidth, drawingheight, xrange, minxvalue,
			   counter - rules.Lese_xDimension (), rulecounter);
    }

    if (status == OUTPUT_WORKASALLPOINTSCALCULATOR) {
      if (indexmenues != NULL)	// move menues over graphes
       {
	short menuxpos = xpos + XTextWidth (fontstruct, ruletext, strlen (ruletext));
	menuxpos += (drawingwidth - indexmenues[counter]->Getwidth ()) / 2;
	indexmenues[counter]->PlaceMe (menuxpos, 0);
      }
    }
    if ((rulesoutput.Lese_Ergebnisse ().Lese_Dim () > 0) &&
	(selectedoutputs.Lese_i (counter - rules.Lese_xDimension ())) &&
	(selectedrules.Anzahl_gesetzt () > 0)) {
      char resulttext[25];
      sprintf (resulttext, "result: %4.3f", rulesoutput.Lese_Ergebnisse ().Lese_i (counter - rules.Lese_xDimension ()));
      XSetForeground (display, gc, blackpix);
      XDrawString (display, pixmap, gc,
		xpos + XTextWidth (fontstruct, ruletext, strlen (ruletext)),
		   resulttextypos,
		   resulttext, strlen (resulttext));
    }
    xpos += XTextWidth (fontstruct, ruletext, strlen (ruletext)) + drawingwidth;
    graphxpos.Setze_i (counter, xpos);
  }
  free (ruletext);
};

void 
RuleOutput::CalculateResults (RuleResultInfo * ruleresultinfo)
{
  int datacounter, counter;
  short inputdim = ruleresultinfo->dataset->Lese_InputDim ();

  status = OUTPUT_WORKASALLPOINTSCALCULATOR;
  if (acuteruleresultinfo != ruleresultinfo) {
    InitOutput (ruleresultinfo->dataset);
    Action (OUTPUT_REDRAW, 0);
  }
// now calculate the results
  acuteruleresultinfo = ruleresultinfo;
  DVektor tempresult;
  DVektor tempinput (rules.Lese_xDimension (), 0, NULL);
  Regelbasis_Ausgabe temprulesoutput;

  for (datacounter = 0; datacounter < ruleresultinfo->dataset->Lese_Groesse (); datacounter++) {
    DVektor *acuteresult = &(ruleresultinfo->result->Lese_Vektor (datacounter));
    *(acuteresult) = ruleresultinfo->dataset->Lese_Daten (datacounter);

    for (counter = 0; counter < rules.Lese_xDimension (); counter++)
      tempinput.Setze_i (counter, acuteresult->Lese_i (selectedinputindex[counter]));

    for (counter = inputdim; counter < acuteresult->Lese_Dim (); counter++)
      acuteresult->Setze_i (counter, 0);	// erase all output-data

    rules.Berechne (tempinput,
		    selectedrules,
		    selectedinputs,
		    selectedoutputs,
		    temprulesoutput);
    tempresult = temprulesoutput.Lese_Ergebnisse ();

    for (counter = rules.Lese_xDimension (); counter < rules.Lese_xDimension () + rules.Lese_yDimension (); counter++)	// copy result in data
     {
      if (selectedoutputs.Lese_i (counter - rules.Lese_xDimension ()))
	acuteresult->Setze_i (inputdim + selectedinputindex[counter],
		    tempresult.Lese_i (counter - rules.Lese_xDimension ()));
    }
  }
}

void 
RuleOutput::SetSelectedinputvector (DVektor * selectedinput)
{
  short counter;
  input.Setze_Dim (rules.Lese_xDimension ());

  for (counter = 0; counter < rules.Lese_xDimension (); counter++)
    input.Setze_i (counter, selectedinput->Lese_i (selectedinputindex[counter]));
  CalculateResult ();
  Action (OUTPUT_REDRAW, 0);
}

void 
RuleOutput::SetDataset (Datensatz * inputDataset)
{
  if (inputDataset != NULL)	// means, (In!=NULL)
   {
    status = OUTPUT_WORKASSINGLEPOINTCALCULATOR;
    DVektor *selectedinput;
    InitOutput (inputDataset);
    myicon->Action (ICON_REQUESTFORDVECTOR, (int) &selectedinput);
    SetSelectedinputvector (selectedinput);
    Mapmyicons ();
  } else {
    status = OUTPUT_NOWORKSTATUS;
    acuteruleresultinfo = NULL;
    input.Setze_Dim (0);
    rulesoutput.Lese_Ergebnisse ().Setze_Dim (0);
    short menucounter;
    if (indexmenues != NULL) {
      for (menucounter = 0; menucounter < menunumber; menucounter++)
	delete indexmenues[menucounter];
      delete indexmenues;
    }
    menunumber = 0;
    indexmenues = NULL;
    Action (OUTPUT_REDRAW, 0);
  }
}

void 
RuleOutput::InitOutput (Datensatz * inputDataset)
{
  short menucounter, dimnumber, inputcounter, outputcounter;

  if (indexmenues != NULL) {
    for (menucounter = 0; menucounter < menunumber; menucounter++)
      delete indexmenues[menucounter];
    delete indexmenues;
  }
  if (status == OUTPUT_WORKASALLPOINTSCALCULATOR)
    menunumber = rules.Lese_xDimension () + rules.Lese_yDimension ();
  else
    menunumber = rules.Lese_xDimension ();

  indexmenues = (Menu **) malloc (sizeof (Menu *) * menunumber);

// first menues for inputs
  dimnumber = inputDataset->Lese_InputDim ();
  MenuIconInfo *Menuinfo;
  Menuinfo = (MenuIconInfo *) malloc (sizeof (MenuIconInfo) * (dimnumber + 1));

  Menuinfo[0].bartext = "inputindex";
  Menuinfo[0].barstatus = STATUS_DEFAULT;
  Menuinfo[0].actionnumber = OUTPUT_CHANCEINDEX_ACTION;
  Menuinfo[0].value = 0;

  for (inputcounter = 1; inputcounter <= dimnumber; inputcounter++) {
    Menuinfo[inputcounter].bartext = inputDataset->Lese_Name_Pointer (inputcounter - 1);
    Menuinfo[inputcounter].barstatus = STATUS_DEFAULT | STATUS_MARKABLE;
    Menuinfo[inputcounter].actionnumber = OUTPUT_CHANCEINDEX_ACTION;
  }

  for (inputcounter = 0; inputcounter < rules.Lese_xDimension (); inputcounter++) {
    if (inputcounter > dimnumber)
      selectedinputindex.Setze_i (inputcounter, dimnumber);
    else
      selectedinputindex.Setze_i (inputcounter, inputcounter);
    for (menucounter = 0; menucounter <= dimnumber; menucounter++) {
      Menuinfo[menucounter].value = inputcounter +
	(menucounter - 1) * (menunumber);
      if (inputcounter == menucounter - 1)
	Menuinfo[menucounter].barstatus |= STATUS_MARKED;
      else
	Menuinfo[menucounter].barstatus &= ~STATUS_MARKED;

      if ((dimnumber < rules.Lese_xDimension ()) && (menucounter == dimnumber) &&
	  (inputcounter >= dimnumber))
	Menuinfo[menucounter].barstatus |= STATUS_MARKED;
    }
    indexmenues[inputcounter] = new Menu (display,
					  gc,
					  this,
					  fontstruct,
					  0, 0);
    indexmenues[inputcounter]->AddPulldownMenu (Menuinfo, dimnumber + 1);
    indexmenues[inputcounter]->MapMe ();
  }
  free (Menuinfo);

// last menues for outputs
  if (status == OUTPUT_WORKASALLPOINTSCALCULATOR)	//&&(inputDataset->Lese_OutputDim()>0))
   {
    dimnumber = inputDataset->Lese_OutputDim ();
    MenuIconInfo *Menuinfo;
    Menuinfo = (MenuIconInfo *) malloc (sizeof (MenuIconInfo) * (dimnumber + 1));

    Menuinfo[0].bartext = "outputindex";
    Menuinfo[0].barstatus = STATUS_DEFAULT;
    Menuinfo[0].actionnumber = OUTPUT_CHANCEINDEX_ACTION;
    Menuinfo[0].value = 0;

    for (outputcounter = 1; outputcounter <= dimnumber; outputcounter++) {
      Menuinfo[outputcounter].bartext = inputDataset->Lese_Name_Pointer (inputDataset->Lese_InputDim () +
							 outputcounter - 1);
      Menuinfo[outputcounter].barstatus = STATUS_DEFAULT | STATUS_MARKABLE;
      Menuinfo[outputcounter].actionnumber = OUTPUT_CHANCEINDEX_ACTION;
    }

    for (outputcounter = rules.Lese_xDimension ();
	 outputcounter < rules.Lese_xDimension () + rules.Lese_yDimension ();
	 outputcounter++) {
      if (outputcounter - rules.Lese_xDimension () >= dimnumber)
	selectedinputindex.Setze_i (outputcounter, inputDataset->Lese_OutputDim () - 1);
      else
	selectedinputindex.Setze_i (outputcounter,
				  outputcounter - rules.Lese_xDimension ());
      for (menucounter = 0; menucounter <= dimnumber; menucounter++) {
	Menuinfo[menucounter].value = outputcounter +
	  (menucounter - 1) * (menunumber);
	if (outputcounter - rules.Lese_xDimension () == (menucounter - 1))
	  Menuinfo[menucounter].barstatus |= STATUS_MARKED;
	else
	  Menuinfo[menucounter].barstatus &= ~STATUS_MARKED;
	if ((dimnumber < rules.Lese_yDimension ()) && (menucounter == dimnumber) &&
	    (outputcounter - rules.Lese_xDimension () >= dimnumber))
	  Menuinfo[menucounter].barstatus |= STATUS_MARKED;
      }
      indexmenues[outputcounter] = new Menu (display,
					     gc,
					     this,
					     fontstruct,
					     0, 0);
      indexmenues[outputcounter]->AddPulldownMenu (Menuinfo, dimnumber + 1);
      indexmenues[outputcounter]->MapMe ();
    }
    free (Menuinfo);
  }
}

void 
RuleOutput::Action (int actionnumber, int value)
{
  switch (actionnumber) {
  case OUTPUT_REDRAW:
    DrawRules ();
    UpdateOutput ();
    break;

  case OUTPUT_CHANCEINDEX_ACTION:
    {
      DVektor *selectedinput;
      short menuindex = value % menunumber;
      short index = value / menunumber;
      selectedinputindex.Setze_i (menuindex, index);
      if (status == OUTPUT_WORKASSINGLEPOINTCALCULATOR) {
	myicon->Action (ICON_REQUESTFORDVECTOR, (int) &selectedinput);
	SetSelectedinputvector (selectedinput);
      } else			// status==OUTPUT_WORKASALLPOINTSCALCULATOR)
       {
	CalculateResults (acuteruleresultinfo);
	myicon->Action (ICON_RULERESULTSCHANGED, 0);
      }
    }
    break;

  case OUTPUT_SAVEOBJECT:
    if ((char *) value != NULL)
      rules.Speichern ((char *) value);
    myicon->GetMainwindow ()->SetStatus (MAINNORMALSTATUS);
    break;

  case OUTPUT_GHOST:
    Unmapmyicons ();
    break;
  }
}

void 
RuleOutput::CalculateResult ()
{
  rules.Berechne (input,
		  selectedrules,
		  selectedinputs,
		  selectedoutputs,
		  rulesoutput);
}

void 
RuleOutput::HandleEvent (XEvent * Event)
{
  static Time lastpresstime = 0;
  static RuleOutput *lastpressoutput = NULL;

  switch (Event->type) {
  case Expose:
    XCopyArea (display, pixmap, window, gc,
	       Event->xgraphicsexpose.x, Event->xgraphicsexpose.y,
	       Event->xgraphicsexpose.width, Event->xgraphicsexpose.height,
	       Event->xgraphicsexpose.x, Event->xgraphicsexpose.y);
    break;			// case Expose

  case ButtonPress:
    switch (Event->xbutton.button) {
    case Button2:
      {
	short counter;
	for (counter = 0; counter < graphxpos.Lese_Dim () - 1; counter++)
	  if (graphxpos.Lese_i (counter) > Event->xbutton.x)
	    break;
	if (counter >= selectedinputs.Lese_Dim ())	// in outputs
	{
	  short acuteindex = counter - selectedinputs.Lese_Dim ();
	  if (selectedoutputs.Lese_i (acuteindex))
	    selectedoutputs.Setze_i (acuteindex, FALSE);
	  else
	    selectedoutputs.Setze_i (acuteindex, TRUE);
	} else {
	  if (selectedinputs.Lese_i (counter))
	    selectedinputs.Setze_i (counter, FALSE);
	  else
	    selectedinputs.Setze_i (counter, TRUE);
	}
      }
      break;

    case Button3:
      {
	short counter;
	for (counter = 0; counter < graphypos.Lese_Dim () - 1; counter++)
	  if (graphypos.Lese_i (counter) > Event->xbutton.y)
	    break;
	if (selectedrules.Lese_i (counter))
	  selectedrules.Setze_i (counter, FALSE);
	else
	  selectedrules.Setze_i (counter, TRUE);
      }
      break;
    }
    // only if changed on/off selection
    if ((Event->xbutton.button == Button2) ||
	(Event->xbutton.button == Button3)) {
      if (status == OUTPUT_WORKASSINGLEPOINTCALCULATOR)
	CalculateResult ();
      if (status == OUTPUT_WORKASALLPOINTSCALCULATOR) {
	CalculateResults (acuteruleresultinfo);
	myicon->Action (ICON_RULERESULTSCHANGED, 0);
      }
      Action (OUTPUT_REDRAW, 0);
    }
    // now test for doubleclick
    if ((Event->xbutton.button == Button1) &&
	!(ghost->GetStatus () & ICON_MAPPED)) {
      if (((Event->xbutton.time - lastpresstime) < DOUBLECLICKTIME) &&
	  (this == lastpressoutput))
	Mapmyicons ();
      lastpresstime = Event->xbutton.time;
      lastpressoutput = this;
    }
    RaiseMe ();
    break;

  case UnmapNotify:
    XWithdrawWindow (display, window, DefaultScreen (display));
    break;

  case ConfigureNotify:
    if ((Event->xconfigure.width != width) ||
	(Event->xconfigure.height != height)) {
      Resized (Event->xconfigure.width, Event->xconfigure.height);
      Action (OUTPUT_REDRAW, 0);
    }
    break;			// case ConfigureNotify

  case KeyPress:
    /* 
     * process key hit:
     */
    char c;
    int n = XLookupString((XKeyEvent *) Event, &c, 1, NULL, NULL);
    if (n == 1 && c == 'p') {
      DumpBitmap();
    }

    break;

  }
}

ClassifyOutput::ClassifyOutput (Display * initdisplay,
		GC initgc,
		ClassifyIcon * initOutputicon,
		XFontStruct * fontstruct,
		int initx, int inity,
		int initwidth, int initheight,
		char *text,
		RuleInfo * Ruleinfo):
Output2d (initdisplay,
	  initgc,
	  initOutputicon,
	  fontstruct,
	  initx, inity,
	  initwidth, initheight,
	  text)
{
  input.Setze_Dim (0);
  menunumber = 0;
  indexmenues = NULL;
  SetSelectedInput (ExposureMask
		    | StructureNotifyMask
		    | ButtonMotionMask
		    | ButtonPressMask
		    | ButtonReleaseMask
		    | KeyPressMask);

  classify = Ruleinfo->classify;
  selectedindex.Setze_Dim (classify.Lese_Dimension ());

  short counter;
  selecteddims.Setze_Dim (classify.Lese_Dimension ());
  for (counter = 0; counter < selecteddims.Lese_Dim (); counter++)
    selecteddims.Setze_i (counter, TRUE);


  selectedrules.Setze_Dim (classify.Lese_Klassenanzahl ());
  for (counter = 0; counter < selectedrules.Lese_Dim (); counter++)
    selectedrules.Setze_i (counter, TRUE);

  graphxpos.Setze_Dim (classify.Lese_Dimension ());
  graphypos.Setze_Dim (classify.Lese_Klassenanzahl ());

  classify_yes_no = new OnOffIcon (display,
				   gc,
				   this,
				   fontstruct,
				   0, 18, 16, 16,
				   "", classify_bits,
				   classify_width, classify_height,
				   OUTPUT_CLASSIFYNEW, ICON_SELECTABLE);
  classify_yes_no->MapMe ();
  Action (OUTPUT_REDRAW, 0);
}

ClassifyOutput::~ClassifyOutput ()
{
  short menucounter;
  if (indexmenues != NULL)
    for (menucounter = 0; menucounter < menunumber; menucounter++)
      indexmenues[menucounter]->UnmapMe ();
}

void 
ClassifyOutput::Unmapmyicons ()
{
  short menucounter;
  ghost->UnmapMe ();
  classify_yes_no->UnmapMe ();
  if (indexmenues != NULL)
    for (menucounter = 0; menucounter < menunumber; menucounter++)
      indexmenues[menucounter]->UnmapMe ();
}

void 
ClassifyOutput::Mapmyicons ()
{
  short menucounter;
  ghost->MapMe ();
  classify_yes_no->MapMe ();
  if (indexmenues != NULL)
    for (menucounter = 0; menucounter < menunumber; menucounter++)
      indexmenues[menucounter]->MapMe ();
}

void 
ClassifyOutput::DrawClassresultGraph (int xpos, int ypos,
				      int drawwidth, int drawheight,
				      double range, double minxvalue,
				      short index,
				      short classnumber)
{
  XSegment Line;
  XSetForeground (display, gc, redpix);
  double xvalue = input.Lese_i (index);
  double yvalue = classifyoutput[classnumber].Lese_InputHoehen ()[index];
  Line.x1 = Line.x2 = xpos + (int) (drawwidth * ((xvalue - minxvalue) / range));
  Line.y1 = ypos;
  Line.y2 = ypos - drawheight;
  if ((Line.x1 >= xpos) && (Line.x1 <= (xpos + drawwidth)))
    XDrawLine (display, pixmap, gc, Line.x1, Line.y1, Line.x2, Line.y2);

  Line.x1 = xpos;
  Line.x2 = xpos + drawwidth;
  Line.y1 = Line.y2 = ypos - (int) (yvalue * drawheight);
  XDrawLine (display, pixmap, gc, Line.x1, Line.y1, Line.x2, Line.y2);
}

void 
ClassifyOutput::DrawRules ()
{
  int counter, classcounter, drawingnumber, drawingwidth, drawingheight,
    textwidth = 0, textheight = fontstruct->max_bounds.ascent +
  fontstruct->max_bounds.descent, xpos, ypos, classnumber = classify.Lese_Klassenanzahl ();
  double maxxvalue = -MAXDOUBLE, minxvalue = MAXDOUBLE, xrange;
  short classtextlength = 0;
  char *AND = " and ";

  Klasse *Class;
  Parameter *parameter;

// first calculate drawingsize and maxxvalue
  {
    char temptext[20];		// should be enought

    sprintf (temptext, "Class%d: ", classnumber);
    classtextlength += strlen (temptext);
    textwidth += XTextWidth (fontstruct, temptext, strlen (temptext));
  }
  char first = 1;

  for (counter = 0; counter < classify.Lese_Dimension (); counter++) {
    if (!first) {
      classtextlength += strlen (AND);
      textwidth += XTextWidth (fontstruct, AND, strlen (AND));
    }
    classtextlength += strlen (classify.Lese_Klasse (0).Lese_Name (counter));
    textwidth += XTextWidth (fontstruct,
			     classify.Lese_Klasse (0).Lese_Name (counter),
		     strlen (classify.Lese_Klasse (0).Lese_Name (counter)));
    first = 0;
    for (classcounter = 0; classcounter < classnumber; classcounter++) {
      Class = &(classify.Lese_Klasse (classcounter));
      parameter = &(Class->Lese_Parameter (counter));
      maxxvalue = MAX (maxxvalue, parameter->Lese_xParameter (parameter->Lese_Parameter_Anzahl () - 1));
      minxvalue = MIN (minxvalue, parameter->Lese_xParameter (0));
    }
  }
  minxvalue = MIN (0.0, minxvalue);
  xrange = maxxvalue - minxvalue;

  drawingnumber = classify.Lese_Dimension ();
  drawingwidth = (int) (width - CLASSIFY_RESULTSPACE - drawingnumber * CLASSIFY_HORI_SPACE - textwidth - 2 * CLASSIFY_BORDER) / drawingnumber;
  drawingheight = (int) (height - 2 * textheight - 2 * CLASSIFY_BORDER - ((classnumber - 1) * CLASSIFY_VERT_SPACE)) / classnumber;

// **************************************************************
  // resize window if drawing would be too small
  char resize = 0;
  int newheight, newwidth;
  if (drawingwidth < CLASSIFY_MINDRAWWIDTH) {
    drawingwidth = CLASSIFY_MINDRAWWIDTH;
    newwidth = textwidth + drawingnumber * CLASSIFY_HORI_SPACE + 2 * CLASSIFY_BORDER + CLASSIFY_RESULTSPACE + drawingnumber * drawingwidth;
    resize = 1;
  } else
    newwidth = width;

  if (drawingheight < CLASSIFY_MINDRAWHEIGHT) {
    drawingheight = CLASSIFY_MINDRAWHEIGHT;
    newheight = 2 * CLASSIFY_BORDER +
      classnumber * drawingheight +
      (classnumber - 1) * CLASSIFY_VERT_SPACE
      + 2 * textheight;
    resize = 1;
  } else
    newheight = height;

  if (resize)
    Resized (newwidth, newheight);

  XSetForeground (display, gc, whitepix);
  XFillRectangle (display, pixmap, gc, 0, 0, width, height);

//**************************************************************
  // second Class1: ... Classn:
  char *classtext = (char *) malloc (sizeof (char) * (classtextlength + 1));
  xpos = CLASSIFY_BORDER;
  for (classcounter = 0; classcounter < classnumber; classcounter++) {
//  sprintf(classtext,"Class%d: ",classcounter+1);
    sprintf (classtext, "Class%d: ", classify.Lese_Klasse (classcounter).Lese_Nummer ());
    if (selectedrules.Lese_i (classcounter))
      XSetForeground (display, gc, blackpix);
    else
      XSetForeground (display, gc, shadowpix);
    XDrawString (display, pixmap, gc, xpos, (drawingheight) * (classcounter + 1) + classcounter * CLASSIFY_VERT_SPACE + CLASSIFY_BORDER, classtext, strlen (classtext));
  }
  xpos += XTextWidth (fontstruct, classtext, strlen (classtext));

//*************************************************************
  // then the graphes
  first = 1;
  for (counter = 0; counter < classify.Lese_Dimension (); counter++) {
    classtext[0] = 0;
    if (!first)
      strcat (classtext, AND);
    strcat (classtext, classify.Lese_Klasse (0).Lese_Name (counter));
    first = 0;
    for (classcounter = 0; classcounter < classnumber; classcounter++) {
      ypos = (drawingheight) * (classcounter + 1) +
	classcounter * CLASSIFY_VERT_SPACE + CLASSIFY_BORDER;
      graphypos.Setze_i (classcounter, ypos + CLASSIFY_VERT_SPACE / 2);
      if ((selectedrules.Lese_i (classcounter)) &&
	  (selecteddims.Lese_i (counter)))
	XSetForeground (display, gc, blackpix);
      else
	XSetForeground (display, gc, shadowpix);
      XDrawString (display, pixmap, gc, xpos, ypos, classtext, strlen (classtext));
      Class = &(classify.Lese_Klasse (classcounter));
      parameter = &(Class->Lese_Parameter (counter));
      DrawGraph (parameter,
		 xpos + XTextWidth (fontstruct, classtext, strlen (classtext)) + CLASSIFY_HORI_SPACE,
		 ypos,
		 drawingwidth, drawingheight, xrange, minxvalue);
      if ((status == OUTPUT_WORKASSINGLEPOINTCALCULATOR) &&
	  (selectedrules.Lese_i (classcounter)) &&
	  (selecteddims.Lese_i (counter)))
	DrawClassresultGraph (xpos + XTextWidth (fontstruct, classtext, strlen (classtext)) + CLASSIFY_HORI_SPACE, ypos,
			      drawingwidth, drawingheight, xrange, minxvalue,
			      counter, classcounter);
    }

    if (indexmenues != NULL)	// move menues over graphes
     {
      short menuxpos = xpos + XTextWidth (fontstruct, classtext, strlen (classtext));
      menuxpos += (drawingwidth - indexmenues[counter]->Getwidth ()) / 2;
      indexmenues[counter]->PlaceMe (menuxpos, 0);
    }
    xpos += XTextWidth (fontstruct, classtext, strlen (classtext))
      + drawingwidth + CLASSIFY_HORI_SPACE;
    graphxpos.Setze_i (counter, xpos);
  }

  xpos += CLASSIFY_HORI_SPACE;
  if (status == OUTPUT_WORKASSINGLEPOINTCALCULATOR) {
    XSetForeground (display, gc, blackpix);
    for (classcounter = 0; classcounter < classnumber; classcounter++) {
      if (selectedrules.Lese_i (classcounter)) {
	ypos = drawingheight * classcounter + drawingheight / 2 +
	  classcounter * CLASSIFY_VERT_SPACE + CLASSIFY_BORDER;
	char resulttext[10];
	sprintf (resulttext, "%1.5f", classifyoutput[classcounter].Lese_Uebereinstimmung ());
	XDrawString (display, pixmap, gc,
		     xpos,
		     ypos,
		     resulttext, strlen (resulttext));
      }
    }

    if ((selectedrules.Anzahl_gesetzt () > 0) &&
	(selecteddims.Anzahl_gesetzt ())) {
      ypos = drawingheight * classcounter +
	classcounter * CLASSIFY_VERT_SPACE + CLASSIFY_BORDER;
      char resulttext[25];

      int tempxpos = xpos;
      if (classifyoutput.Lese_geklappt ()) {
	sprintf (resulttext, "result %d", classifyoutput.Lese_Ergebnis ());
	XSetForeground (display, gc, blackpix);
	tempxpos = width - (2 * SYMBOLHALFWIDTH + CLASSIFY_HORI_SPACE + XTextWidth (fontstruct, resulttext, strlen (resulttext)));
      } else {
	sprintf (resulttext, "no match found");
	tempxpos = width - (CLASSIFY_HORI_SPACE + XTextWidth (fontstruct, resulttext, strlen (resulttext)));
	XSetForeground (display, gc, redpix);
      }

      XDrawString (display, pixmap, gc,
		   tempxpos,
		   ypos,
		   resulttext, strlen (resulttext));
      if (classifyoutput.Lese_geklappt ()) {
	int symbolxpos = tempxpos + XTextWidth (fontstruct, resulttext, strlen (resulttext));
	DrawSymbol (symbolxpos + 2 * SYMBOLHALFWIDTH, ypos - textheight / 2, pixmap, classifyoutput.Lese_Ergebnis ());
      }
    }
  }
  if ((classifyok.Lese_Dim () == classnumber) &&
      (status == OUTPUT_WORKASALLPOINTSCALCULATOR)) {
    XSetForeground (display, gc, blackpix);
    for (classcounter = 0; classcounter < classnumber; classcounter++) {
      ypos = drawingheight * classcounter + drawingheight / 2 +
	classcounter * CLASSIFY_VERT_SPACE + CLASSIFY_BORDER -
	textheight / 2;
      XSetForeground (display, gc, blackpix);
      char resulttext[15];
      sprintf (resulttext, "right %d", classifyok.Lese_i (classcounter));
      XDrawString (display, pixmap, gc,
		   xpos,
		   ypos,
		   resulttext, strlen (resulttext));
      XSetForeground (display, gc, redpix);
      ypos += textheight;
      sprintf (resulttext, "wrong %d", classifyerrors.Lese_i (classcounter));
      XDrawString (display, pixmap, gc,
		   xpos,
		   ypos,
		   resulttext, strlen (resulttext));
    }
    {
      ypos = drawingheight * classcounter + drawingheight / 2 +
	classcounter * CLASSIFY_VERT_SPACE + CLASSIFY_BORDER;
      char resulttext[30];
      sprintf (resulttext, "not classified  %d", noclassfound);
      int tempxpos = width - (CLASSIFY_HORI_SPACE + XTextWidth (fontstruct, resulttext, strlen (resulttext)));
      XDrawString (display, pixmap, gc,
		   tempxpos,
		   height,
		   resulttext, strlen (resulttext));
    }
  }
  free (classtext);
};

void 
ClassifyOutput::CalculateResult ()
{
  classify.Berechne (input,
		     selectedrules,
		     selecteddims,
		     classifyoutput);
}

void 
ClassifyOutput::CalculateResults (Datensatz * dataset)
{
  status = OUTPUT_WORKASALLPOINTSCALCULATOR;
  int datacounter, counter;
  noclassfound = 0;

  DVektor tempinput;
  if (Dataset != dataset) {
    Dataset = dataset;
    InitOutput (Dataset);
    Action (OUTPUT_REDRAW, 0);
  }
  dataset->Lese_eindeutig ().Setze_Dim (dataset->Lese_Daten ().Lese_Groesse ());

  if (Dataset->Lese_Daten_Typ () == Klassifiziert) {
    classifyerrors.Setze_Dim (classify.Lese_Klassenanzahl ());
    classifyok.Setze_Dim (classify.Lese_Klassenanzahl ());
    for (counter = 0; counter < classify.Lese_Klassenanzahl (); counter++) {
      classifyerrors.Setze_i (counter, 0);
      classifyok.Setze_i (counter, 0);
    }
  } else {
    classifyerrors.Setze_Dim (0);
    classifyok.Setze_Dim (0);
  }

  if (classify_yes_no->IsSelected ()) {
    DVektorArray *membership = &(Dataset->Lese_Zugehoerigkeiten ());
    DVektorArray *inputData = &(Dataset->Lese_Daten ());
#if 0
    membership->Setze_Groesse (inputData.Lese_Groesse ());
#else
    membership->Setze_Groesse (inputData->Lese_Groesse ());
#endif
//  membership->Setze_Dim(classify.Lese_Klassenanzahl());
    membership->Setze_Dim (classify.Lese_Klassen_Nummern ().Lese_Dim ());
  }
  for (datacounter = 0; datacounter < Dataset->Lese_Groesse (); datacounter++) {
    tempinput.Setze_Dim (classify.Lese_Dimension ());

    for (counter = 0; counter < classify.Lese_Dimension (); counter++)
      tempinput.Setze_i (counter, dataset->Lese_Daten (datacounter).Lese_i (selectedindex[counter]));

    Klassifizierungs_Ausgabe tempclassifyoutput;
    classify.Berechne (tempinput,
		       selectedrules,
		       selecteddims,
		       tempclassifyoutput);
    dataset->Lese_eindeutig ().Setze_i (datacounter, tempclassifyoutput.Lese_eindeutig ());

    // Add 1 to classifyerrors or classifyok
    if (Dataset->Lese_Daten_Typ () == Klassifiziert) {
      if (tempclassifyoutput.Lese_geklappt ()) {
	if (dataset->Lese_Daten (datacounter).Lese_Nummer () == tempclassifyoutput.Lese_Ergebnis ())
	  classifyok.Setze_i (tempclassifyoutput.Lese_Aus_dieser_Regel (), classifyok.Lese_i (tempclassifyoutput.Lese_Aus_dieser_Regel ()) + 1);
	else
	  classifyerrors.Setze_i (tempclassifyoutput.Lese_Aus_dieser_Regel (), classifyerrors.Lese_i (tempclassifyoutput.Lese_Aus_dieser_Regel ()) + 1);
      } else
	noclassfound++;
    }
    if (classify_yes_no->IsSelected ()) {
      short classcounter;
      DVektorArray *membership = &(Dataset->Lese_Zugehoerigkeiten ());

      if (tempclassifyoutput.Lese_geklappt ()) {
//      for(classcounter=0;classcounter<classify.Lese_Klassenanzahl();classcounter++)
	//        membership->Setze_Vektor(datacounter,
	//                                 classcounter,
	//                                 tempclassifyoutput[classcounter].Lese_Uebereinstimmung());
	for (classcounter = 0; classcounter < classify.Lese_Klassen_Nummern ().Lese_Dim (); classcounter++)
	  membership->Setze_Vektor (datacounter,
				    classcounter,
	  tempclassifyoutput.Lese_Zugehoerigkeiten ().Lese_i (classcounter));
	dataset->Lese_Daten (datacounter).Setze_Nummer (tempclassifyoutput.Lese_Ergebnis ());
      } else {
	dataset->Lese_Daten (datacounter).Setze_Nummer (MAXINT);
	for (classcounter = 0; classcounter < classify.Lese_Klassen_Nummern ().Lese_Dim (); classcounter++)
//      for(classcounter=0;classcounter<classify.Lese_Klassenanzahl();classcounter++)
	  membership->Setze_Vektor (datacounter,
				    classcounter,
				    0.0);
      }
    }
  }

  if (classify_yes_no->IsSelected ()) {
    dataset->Setze_Daten_Typ (Klassifiziert);
    myicon->GetOutputConnectlist ()
      ->SendActiontoallEnds (ICON_DATASETCLASSIFIED,
			     classify.Lese_Klassen_Nummern ().Lese_Dim ());
  }
  Action (OUTPUT_REDRAW, 0);
}

void 
ClassifyOutput::SetSelectedinputvector (DVektor * selectedinput)
{
  short counter;
  input.Setze_Dim (classify.Lese_Dimension ());
  for (counter = 0; counter < classify.Lese_Dimension (); counter++)
#if 0
    input.Setze_i (counter, selectedinput.Lese_i (selectedindex[counter]));
#else
    input.Setze_i (counter, selectedinput->Lese_i (selectedindex[counter]));
#endif

  CalculateResult ();
  Action (OUTPUT_REDRAW, 0);
}

void 
ClassifyOutput::SetDataset (Datensatz * inputDataset)
{
  if (inputDataset != NULL)	// means, (In!=NULL)
   {
    status = OUTPUT_WORKASSINGLEPOINTCALCULATOR;
    DVektor *selectedinput;
    InitOutput (inputDataset);
    myicon->Action (ICON_REQUESTFORDVECTOR, (int) &selectedinput);
    SetSelectedinputvector (selectedinput);
    Mapmyicons ();
  } else {
    status = OUTPUT_NOWORKSTATUS;
    input.Setze_Dim (0);
    short menucounter;
    if (indexmenues != NULL) {
      for (menucounter = 0; menucounter < menunumber; menucounter++)
	delete indexmenues[menucounter];
      delete indexmenues;
    }
    menunumber = 0;
    indexmenues = NULL;
    Action (OUTPUT_REDRAW, 0);
  }
}

void 
ClassifyOutput::InitOutput (Datensatz * inputDataset)
{
  short menucounter, dimnumber, counter;

  if (indexmenues != NULL) {
    for (menucounter = 0; menucounter < menunumber; menucounter++)
      delete indexmenues[menucounter];
    delete indexmenues;
  }
  menunumber = classify.Lese_Dimension ();
  indexmenues = (Menu **) malloc (sizeof (Menu *) * menunumber);

  dimnumber = inputDataset->Lese_DatenDim ();
  MenuIconInfo *Menuinfo;
  Menuinfo = (MenuIconInfo *) malloc (sizeof (MenuIconInfo) * (dimnumber + 1));

  Menuinfo[0].bartext = "index";
  Menuinfo[0].barstatus = STATUS_DEFAULT;
  Menuinfo[0].actionnumber = OUTPUT_CHANCEINDEX_ACTION;
  Menuinfo[0].value = 0;

  for (counter = 1; counter <= dimnumber; counter++) {
    Menuinfo[counter].bartext = inputDataset->Lese_Name_Pointer (counter - 1);
    Menuinfo[counter].barstatus = STATUS_DEFAULT | STATUS_MARKABLE;
    Menuinfo[counter].actionnumber = OUTPUT_CHANCEINDEX_ACTION;
  }

  for (counter = 0; counter < classify.Lese_Dimension (); counter++) {
    if (counter > dimnumber)
      selectedindex.Setze_i (counter, dimnumber);
    else
      selectedindex.Setze_i (counter, counter);
    for (menucounter = 0; menucounter <= dimnumber; menucounter++) {
      Menuinfo[menucounter].value = counter +
	(menucounter - 1) * (menunumber);
      if (counter == menucounter - 1)
	Menuinfo[menucounter].barstatus |= STATUS_MARKED;
      else
	Menuinfo[menucounter].barstatus &= ~STATUS_MARKED;

      if ((dimnumber < classify.Lese_Dimension ()) && (menucounter == dimnumber) &&
	  (counter >= dimnumber))
	Menuinfo[menucounter].barstatus |= STATUS_MARKED;
    }
    indexmenues[counter] = new Menu (display,
				     gc,
				     this,
				     fontstruct,
				     0, 0);
    indexmenues[counter]->AddPulldownMenu (Menuinfo, dimnumber + 1);
    indexmenues[counter]->MapMe ();
  }
  free (Menuinfo);
}

void 
ClassifyOutput::Action (int actionnumber, int value)
{
  switch (actionnumber) {
  case OUTPUT_CLASSIFYNEW:
    if ((value == SWITCHON) &&
	(status == OUTPUT_WORKASALLPOINTSCALCULATOR))
      CalculateResults (Dataset);
    break;

  case OUTPUT_REDRAW:
    DrawRules ();
    UpdateOutput ();
    break;

  case OUTPUT_CHANCEINDEX_ACTION:
    {
      DVektor *selectedinput;
      short menuindex = value % menunumber;
      short index = value / menunumber;
      selectedindex.Setze_i (menuindex, index);
      if (status == OUTPUT_WORKASSINGLEPOINTCALCULATOR) {
	myicon->Action (ICON_REQUESTFORDVECTOR, (int) &selectedinput);
	SetSelectedinputvector (selectedinput);
      } else			// status==OUTPUT_WORKASALLPOINTSCALCULATOR)
       {
	CalculateResults (Dataset);	//acuteruleresultinfo);

	myicon->Action (ICON_CLASSIFYRESULTSCHANGED, 0);
      }
    }
    break;

  case OUTPUT_SAVEOBJECT:
    if ((char *) value != NULL)
      classify.Speichern ((char *) value);
    myicon->GetMainwindow ()->SetStatus (MAINNORMALSTATUS);
    break;

  case OUTPUT_GHOST:
    Unmapmyicons ();
    break;
  }
}

void 
ClassifyOutput::HandleEvent (XEvent * Event)
{
  static Time lastpresstime = 0;
  static ClassifyOutput *lastpressoutput = NULL;

  switch (Event->type) {
  case Expose:
    XCopyArea (display, pixmap, window, gc,
	       Event->xgraphicsexpose.x, Event->xgraphicsexpose.y,
	       Event->xgraphicsexpose.width, Event->xgraphicsexpose.height,
	       Event->xgraphicsexpose.x, Event->xgraphicsexpose.y);
    break;			// case Expose

  case ButtonPress:
    switch (Event->xbutton.button) {
    case Button2:
      {
	short counter;
	for (counter = 0; counter < graphxpos.Lese_Dim () - 1; counter++)
	  if (graphxpos.Lese_i (counter) > Event->xbutton.x)
	    break;

	if (selecteddims.Lese_i (counter))
	  selecteddims.Setze_i (counter, FALSE);
	else
	  selecteddims.Setze_i (counter, TRUE);
      }
      break;

    case Button3:
      {
	short counter;
	for (counter = 0; counter < graphypos.Lese_Dim () - 1; counter++)
	  if (graphypos.Lese_i (counter) > Event->xbutton.y)
	    break;
	if (selectedrules.Lese_i (counter))
	  selectedrules.Setze_i (counter, FALSE);
	else
	  selectedrules.Setze_i (counter, TRUE);
      }
      break;
    }
    // only if changed on/off selection
    if ((Event->xbutton.button == Button2) ||
	(Event->xbutton.button == Button3)) {
      if (status == OUTPUT_WORKASSINGLEPOINTCALCULATOR)
	CalculateResult ();
      if (status == OUTPUT_WORKASALLPOINTSCALCULATOR) {
	CalculateResults (Dataset);
	myicon->Action (ICON_CLASSIFYRESULTSCHANGED, 0);
      }
      Action (OUTPUT_REDRAW, 0);
    }
    // now test for doubleclick
    if ((Event->xbutton.button == Button1) &&
	!(ghost->GetStatus () & ICON_MAPPED)) {
      if (((Event->xbutton.time - lastpresstime) < DOUBLECLICKTIME) &&
	  (this == lastpressoutput))
	Mapmyicons ();
      lastpresstime = Event->xbutton.time;
      lastpressoutput = this;
    }
    RaiseMe ();
    break;

  case UnmapNotify:
    XWithdrawWindow (display, window, DefaultScreen (display));
    break;

  case ConfigureNotify:
    if ((Event->xconfigure.width != width) ||
	(Event->xconfigure.height != height)) {
      Resized (Event->xconfigure.width, Event->xconfigure.height);
      Action (OUTPUT_REDRAW, 0);
    }
    break;			// case ConfigureNotify

  case KeyPress:
    /* 
     * process key hit:
     */
    char c;
    int n = XLookupString((XKeyEvent *) Event, &c, 1, NULL, NULL);
    if (n == 1 && c == 'p') {
      DumpBitmap();
    }

    break;

  }
}
