/************vv********************************************/
/*   filename: cluster.cc                               */
/*                                                      */
/********************************************************/
/* programmed by: Oliver Wagner                         */
/* last change: 26-04-95                                */
/********************************************************/

#include "cluster.h"
#include <limits.h>
/****************************************************************/
/*              Alles zur Klasse Clustering                     */
/****************************************************************/
Clustering::Clustering (Guete_Typ Guete, int Soll_Iterationen,
	    char Possib)
{
  Possibilistisch = Possib;
  Guete_Verfahren = NULL;
  Setze_Guete (Guete);
  Max_Iterationen = Soll_Iterationen;
  Iterationen = 0;
  Possib_Zaehler = 0;
  M = DEFAULT_M;
  K = DEFAULT_K;
  Delta_U = DEFAULT_DELTA_U;
  ClusterFertig = FALSE;
  Fertig = FALSE;
  Art = KEINE;
  zu_berechnen = 0;
};

Clustering::~Clustering ()
{
  delete Guete_Verfahren;
  Guete_Verfahren = NULL;
};

void 
Clustering::Setze_Guete (Guete_Typ Damit)
{
  if (Guete_Verfahren != NULL)
    delete Guete_Verfahren;
  switch (Damit) {
  case S:
    Guete_Verfahren = new S_Guete;
    break;
  case CS:
    Guete_Verfahren = new CS_Guete;
    break;
  case PK:
    Guete_Verfahren = new PK_Guete;
    break;
  case D1:
    Guete_Verfahren = new D1_Guete;
    break;
  case FHV:
    Guete_Verfahren = new FHV_Guete;
    break;
  case DPA:
    Guete_Verfahren = new DPA_Guete;
    break;
  case PD:
    Guete_Verfahren = new PD_Guete;
    break;
#if 0
  default:
    break;
#else
  default:
    Guete_Verfahren = 0;
    break;
#endif
  }
};

void 
Clustering::Reset ()
{
  Fertig = FALSE;
  ClusterFertig = FALSE;
  Cluster. ~ DVektorArray ();
  Zugehoerigkeit. ~ DMatrix ();
  Iterationen = 0;
  Possib_Zaehler = 0;
  Art = KEINE;
  eta. ~ DVektor ();
  Zuordnung. ~ IVektor ();
  Klassen. ~ IVektor ();
  neuer_Proto. ~ DVektor ();
  Loesche_Buffer ();
  Reset_Fehlerstatus ();
};

void 
Clustering::Clustere (DVektorArray & Das, int Clusteranzahl)
{
  int i;
  DVektorArray Uebergabe;

  if (((Art != ANZAHL) || (Cluster.Lese_Groesse () != Clusteranzahl)) &&
      (ClusterFertig != TRUE)) {
    Die_waren_relevant = Die_sind_relevant;
    Reset ();
    Art = ANZAHL;
    Cluster. ~ DVektorArray ();
    for (i = 0; (i < Vorgabe_Cluster.Lese_Groesse ()) &&
	 (i < Clusteranzahl); i++) {
      Vorgabe_Cluster[i].Setze_Nummer (i);
      Cluster += Vorgabe_Cluster[i];
    }
    for (i = Cluster.Lese_Groesse (); i < Clusteranzahl; i++) {
      Cluster += Das.Zufallsvektor ();
      Cluster[i].Setze_Nummer (i);
    }
  }
/* Zunaechst das normale Clustern */
  if (ClusterFertig != TRUE) {	/* Zunaechst einmal fertig Clustern */
    Uebergabe = Das.Projiziere (Die_sind_relevant);
    ClusterBuffer = Cluster.Projiziere (Die_sind_relevant);
/*********************************/
    /* Die Cluster runterprojizieren */
/*********************************/
    ClusterBuffer = Mein_Clustern (Uebergabe, ClusterBuffer);

    if (ClusterBuffer.Lese_Groesse () != 0) {
      Cluster = ClusterBuffer.Deprojiziere (Die_sind_relevant);		/* fuer Ausgabe */
      if (ClusterFertig) {
	if (Cluster.Lese_Groesse () != 0) {
	  Guete_Verfahren->Berechne (Uebergabe, Zugehoerigkeit,
				     ClusterBuffer, M);
	  if (Possibilistisch)
	    Possib_Zaehler = 1;
	  else
	    Fertig = TRUE;
	} else
	  Fertig = TRUE;
	ClusterBuffer. ~ DVektorArray ();
      }
      Cluster.Aktualisiere_Min_etc ();
    } else {
      Reset ();
      ClusterFertig = Fertig = TRUE;
    }
  }
/* Jetzt vielleicht noch possibilistisches Clustern */
  if ((ClusterFertig) && (Possibilistisch) && (Fertig == FALSE) &&
      (Cluster.Lese_Groesse () != 0) && (Iterationen < Max_Iterationen)) {
    Uebergabe = Das.Projiziere (Die_sind_relevant);
    ClusterBuffer = Cluster.Projiziere (Die_sind_relevant);

    Mache_Possibilistisch (Uebergabe, ClusterBuffer);
    Guete_Verfahren->Berechne (Uebergabe, Zugehoerigkeit, ClusterBuffer, M);

    Cluster = ClusterBuffer.Deprojiziere (Die_sind_relevant);
    if (Fertig) {
      Guete_Verfahren->Berechne (Uebergabe, Zugehoerigkeit,
				 ClusterBuffer, M);
      ClusterBuffer. ~ DVektorArray ();
    }
    Cluster.Aktualisiere_Min_etc ();
  }
  Berechne_Zuordnung ();	/* Auf jeden Fall vor Formberechnung !!!!! */

  if (Cluster.Lese_Groesse () != 0) {
    Berechne_Form (Das, Cluster);
  }
  Iterationen = 0;		/* bei erneutem Aufruf wieder alle */
};

void 
Clustering::Clustere (DVektorArray & Das)
{
  int i;
  DVektorArray Uebergabe;

  if ((Art != OPTIMAL) && (ClusterFertig != TRUE)) {	/* Reseten und Anfangscluster auffuellen */
    Die_waren_relevant = Die_sind_relevant;
    Reset ();
    Guete_Verfahren->Reset ();
    Art = OPTIMAL;
    for (i = 0; (i < Vorgabe_Cluster.Lese_Groesse ()) /* && (i < 2) */ ; i++) {
      Vorgabe_Cluster[i].Setze_Nummer (0);
      Cluster += Vorgabe_Cluster[i];
    }
    for (i = Cluster.Lese_Groesse (); i < 2; i++) {	/* mindestens zwei */
      Cluster += Das.Zufallsvektor ();
      Cluster[i].Setze_Nummer (i);
    }
    zu_berechnen = 2;
  }
  if (ClusterFertig != TRUE) {
    DVektorArray NiederCluster;

    Uebergabe = Das.Projiziere (Die_sind_relevant);
/********************/
    while ((Cluster.Lese_Groesse () < Das.Lese_Groesse ()) &&
	   (Iterationen < Max_Iterationen)) {	/* Hoechsten n-1 Cluster :-) */
      if (zu_berechnen > Cluster.Lese_Groesse ()) {	/* neues Cluster notwendig */
	Cluster += Das.Zufallsvektor ();
	Cluster[Cluster.Lese_Groesse () - 1].Setze_Nummer (Cluster.Lese_Groesse () - 1);
      }
      Cluster = Cluster.Projiziere (Die_sind_relevant);

      NiederCluster = Mein_Clustern (Uebergabe, Cluster);
      if (ClusterFertig) {	/* jetzt fertig ? */
	if (NiederCluster.Lese_Groesse () != 0) {
	  Guete_Verfahren->Berechne (Uebergabe, Zugehoerigkeit,
				     NiederCluster, M);

	  if (Cluster.Lese_Groesse () == Guete_Verfahren->Lese_Bestes ()) {	/* besser geworden ? */
	    Cluster = NiederCluster.Deprojiziere (Die_sind_relevant);
	    Speichere_Buffer ();
	    zu_berechnen++;
	    ClusterFertig = FALSE;
	  } else {		/* wieder schlechter */
	    Uebernehme_Buffer ();
	    Loesche_Buffer ();
	    if (Possibilistisch)
	      Possib_Zaehler = 1;
	    else
	      Fertig = TRUE;
	    break;
	  }
	} else {
	  if (zu_berechnen > 2) {	/* Dann gibt es schon was */
	    Uebernehme_Buffer ();
	    Loesche_Buffer ();
	    if (Possibilistisch)
	      Possib_Zaehler = 1;
	    else
	      Fertig = TRUE;
	  } else {
	    Reset ();
	    Fertig = TRUE;
	  }
	  break;
	}
      } else
	/* Cluster zumindest fuer die Ausgabe erzeugen */
      {
	Cluster = NiederCluster.Deprojiziere (Die_sind_relevant);
	break;
      }
    }
    Cluster.Aktualisiere_Min_etc ();
  }
  /* Jetzt vielleicht noch possibilistisches Clustern */
  if ((ClusterFertig) && (Possibilistisch) && (Fertig == FALSE) &&
      (Cluster.Lese_Groesse () != 0) && (Iterationen < Max_Iterationen)) {
    Uebergabe = Das.Projiziere (Die_sind_relevant);
    ClusterBuffer = Cluster.Projiziere (Die_sind_relevant);
    Mache_Possibilistisch (Uebergabe, ClusterBuffer);
    Guete_Verfahren->Berechne (Uebergabe, Zugehoerigkeit,
			       ClusterBuffer, M);
    Cluster = ClusterBuffer.Deprojiziere (Die_sind_relevant);
    if (Fertig) {
      ClusterBuffer. ~ DVektorArray ();
    }
    Cluster.Aktualisiere_Min_etc ();
  }
  Berechne_Zuordnung ();	/* Auf jeden Fall vor Formberechnung !!!!! */
  if (Cluster.Lese_Groesse () != 0) {
    Berechne_Form (Das, Cluster);
  }
  Iterationen = 0;		/* bei erneutem Aufruf wieder alle */
};

void 
Clustering::Clustere_Klassen (DVektorArray & Das, int Fehlerprozent)
{
  int i, j, k, Klasse, Zaehler;
  double Abstand, DBuffer;
  DVektorArray Uebergabe;
  // Zur Sicherheit: 
  Klasse=0;

  if ((Art != KLASSEN) && (ClusterFertig != TRUE)) {
    Die_waren_relevant = Die_sind_relevant;
/*****************************************/
    /* Reseten und Anfangscluster auffuellen */
/*****************************************/
    Reset ();
    Art = KLASSEN;
    /* Zunaechst Anzahl der Klassen feststellen */
    for (i = 0; i < Das.Lese_Groesse (); i++) {
      if (!Klassen.Ist_drinnen (Das[i].Lese_Nummer ())) {	/* neu aufnehmen */
	Klassen += (Das[i].Lese_Nummer ());
      }
    }
    BVektor benutzt (Klassen.Lese_Dim ());

    /* Zunaechst die Vorgabe-Cluster uebernehmen */
    for (i = 0; i < Vorgabe_Cluster.Lese_Groesse (); i++) {
      Abstand = MAXDOUBLE;
      for (j = 0; j < Das.Lese_Groesse (); j++) {
	DBuffer = Vorgabe_Cluster[i].Abstand (Das[j], Die_sind_relevant);
	if (DBuffer < Abstand) {
	  Abstand = DBuffer;
	  Klasse = Das[j].Lese_Nummer ();
	}
      }
      benutzt.Setze_i (Klassen.Lese_Index (Klasse), TRUE);
      Cluster += Vorgabe_Cluster[i];
    }

    /* jetzt noch fuer die Klassen, die nicht bedient sind */
    /* einen Prototyp in die Mitte setzen */
    i = 0;
    neuer_Proto.Setze_Dim (Die_sind_relevant.Lese_Dim ());
    while (benutzt.Anzahl_gesetzt () != benutzt.Lese_Dim ()) {
      if (!(benutzt[i])) {
	neuer_Proto.NullVektor ();
	Zaehler = 0;
	Klasse = Klassen[i];
	for (j = 0; j < Das.Lese_Groesse (); j++) {
	  if (Das[j].Lese_Nummer () == Klasse) {
	    neuer_Proto += Das[j];
	    Zaehler++;
	  }
	}
	neuer_Proto *= 1.0 / ((double) Zaehler);
	Cluster += neuer_Proto;
	benutzt.Setze_i (i, TRUE);
      }
      i++;
    }
    neuer_Proto. ~ DVektor ();
    zu_berechnen = Cluster.Lese_Groesse ();
  }
/********************* Jetzt geht es los ************************/
  if (ClusterFertig != TRUE) {
    DVektorArray NiederCluster;

    /* Zunaechst einmal Clustern */
    Uebergabe = Das.Projiziere (Die_sind_relevant);
    while ((Cluster.Lese_Groesse () < Das.Lese_Groesse ()) &&
	   (Iterationen < Max_Iterationen)) {	/* Hoechsten n-1 Cluster :-) */
      Cluster = Cluster.Projiziere (Die_sind_relevant);
      if (zu_berechnen > Cluster.Lese_Groesse ()) {	/* neues Cluster notwendig */
	Cluster += neuer_Proto;
	neuer_Proto. ~ DVektor ();	/* wird nicht mehr gebraucht */
      }
      NiederCluster = Mein_Clustern (Uebergabe, Cluster);

      /* Jetzt testen, ob das funktioniert hat */
      if (ClusterFertig) {	/* jetzt fertig ? */
	if (NiederCluster.Lese_Dim () == 0) {	/* So gings nicht *//* aber vielleicht gab es schon vorher was */
	  Zugehoerigkeit = UBuffer;
	  UBuffer. ~ DMatrix ();
	  Cluster = ClusterBuffer;
	  ClusterBuffer. ~ DVektorArray ();
	  Berechne_Zuordnung ();
	  break;
	  if (Possibilistisch == FALSE) {
	    Fertig = TRUE;
	  }
	} else {
	  /* Testen, ob noch weitere Prototypen notwendig sind: */
	  int Falsche = 0;

	  /* zunaechst jedem Cluster die Klasse zuordnen */
	  for (i = 0; i < NiederCluster.Lese_Groesse (); i++) {
	    DBuffer = 0;
	    for (j = 0; j < Das.Lese_Groesse (); j++) {
	      if (Zugehoerigkeit[i][j] > DBuffer) {
		DBuffer = Zugehoerigkeit[i][j];
		NiederCluster[i].Setze_Nummer (Das[j].Lese_Nummer ());
	      }
	    }
	  }
	  Berechne_Zuordnung ();	/* Jedes Datum einem Cluster zuordnen */

	  /* testen, wieviele falsch sind */
	  for (i = 0; i < Das.Lese_Groesse (); i++) {
	    if (Das[i].Lese_Nummer () != NiederCluster[Zuordnung[i]].Lese_Nummer ())
	      Falsche++;
	  }
	  if ((100.0 * Falsche) / ((double) Das.Lese_Groesse ()) >= Fehlerprozent) {	/* neuer Proto notwendig ? */
	    int diese_Klasse, dieses_Cluster, Falsche_akt, Falsche = 0;
	    // Zur Sicherheit:
	    diese_Klasse=0;
	    dieses_Cluster=0;

	    /* jetzt testen, welches Cluster die meiszten falschen enthaelt */
	    for (i = 0; i < NiederCluster.Lese_Groesse (); i++) {
	      for (j = 0; j < Klassen.Lese_Dim (); j++) {
		Falsche_akt = 0;
		for (k = 0; k < Das.Lese_Groesse (); k++) {
		  if ((Das[k].Lese_Nummer () == Klassen[j]) &&
		      (Das[k].Lese_Nummer () != NiederCluster[i].Lese_Nummer ()) &&
		      (Zuordnung[k] == i))
		    Falsche_akt++;
		}
		if (Falsche_akt > Falsche) {
		  diese_Klasse = Klassen[j];
		  dieses_Cluster = i;
		  Falsche = Falsche_akt;
		}
	      }
	    }
	    neuer_Proto.Setze_Dim (Die_sind_relevant.Anzahl_gesetzt ());
	    neuer_Proto.NullVektor ();
	    j = 0;
	    for (i = 0; i < Das.Lese_Groesse (); i++) {
	      if ((Das[i].Lese_Nummer () == diese_Klasse) &&
		  (Zuordnung[i] == dieses_Cluster)) {
		j++;
		neuer_Proto += Uebergabe[i];
	      }
	    }
	    neuer_Proto *= 1.0 / (double) j;
	    zu_berechnen++;
	    ClusterFertig = FALSE;
	  }
	  Cluster = NiederCluster.Deprojiziere (Die_sind_relevant);
	  if (ClusterFertig) {	/* wirklich immer noch fertig */
	    ClusterBuffer. ~ DVektorArray ();
#if 0
	    UBuffer. ~ Zugehoerigkeit ();
#endif
	    if (Possibilistisch == FALSE)
	      Fertig = TRUE;
	    else if (Iterationen < Max_Iterationen)
	      Possib_Zaehler = 1;
	    break;
	  } else {		/* weiter suchen */
	    ClusterBuffer = Cluster;
	    UBuffer = Zugehoerigkeit;
	  }
	}
      } else
	/* Clustern nocht nicht fertig,aber zumindest fuer die Ausgabe erzeugen */
      {
	Cluster = NiederCluster.Deprojiziere (Die_sind_relevant);
	Berechne_Zuordnung ();

	break;
      }
    }
    Cluster.Aktualisiere_Min_etc ();
  }
  /* Jetzt vielleicht noch possibilistisches Clustern */
  if ((ClusterFertig) && (Possibilistisch) && (Fertig == FALSE) &&
      (Cluster.Lese_Groesse () != 0) && (Iterationen < Max_Iterationen)) {
    Uebergabe = Das.Projiziere (Die_sind_relevant);
    ClusterBuffer = Cluster.Projiziere (Die_sind_relevant);
    Mache_Possibilistisch (Uebergabe, ClusterBuffer);
    Guete_Verfahren->Berechne (Uebergabe, Zugehoerigkeit,
			       ClusterBuffer, M);
    Cluster = ClusterBuffer.Deprojiziere (Die_sind_relevant);
    if (Fertig) {
      ClusterBuffer. ~ DVektorArray ();
    }
    Berechne_Zuordnung ();
    Cluster.Aktualisiere_Min_etc ();
  }
  if (Cluster.Lese_Groesse () != 0) {
    Berechne_Form (Das, Cluster);
  }
  Iterationen = 0;		/* bei erneutem Aufruf wieder alle */
};

DVektor 
Clustering::Berechne_Zugehoerigkeit (DVektor & Datum)
return Result (Cluster.Lese_Groesse (), 0, NULL);
{
  double Summe, Max = 0;
  int i, j, Dim;
  DVektor Dist (Cluster.Lese_Groesse (), 0, NULL);

  if ((Possibilistisch) && (ClusterFertig) && (eta.Lese_Dim () != 0)) {
    Berechne_Dist (Dist, Datum, Cluster);
    for (i = 0; i < Cluster.Lese_Groesse (); i++)
      Result[i] = 1.0 / (1.0 + pow (Dist[i] / eta[i],
				1.0 / (M - 1.0)));
  } else {
    Dim = Berechne_Dist (Dist, Datum, Cluster);
    if (Dim > 0) {
      /* Falls I_k != \emptyset */
      for (i = 0; i < Cluster.Lese_Groesse (); i++)
	if (Dist[i] == 0)
	  Result[i] = 1.0 / (double) Dim;
	else
	  Result[i] = 0.0;
    }
    /* if(Nulltest) */ 
    else {
      /* Falls I_k == \emptyset */
      for (i = 0; i < Cluster.Lese_Groesse (); i++) {
	if (Dist[i] == MAXDOUBLE)
	  Result[i] = 0,0;
	else {
	  Summe = 0.0;
	  for (j = 0; j < Cluster.Lese_Groesse (); j++)
	    Summe += pow (Dist[i] / Dist[j], 1.0 / (M - 1.0));
	  Result[i] = 1.0 / Summe;
	  if (Result[i] > Max) {
	    Max = Result[i];
	    Result.Setze_Nummer (Cluster[i].Lese_Nummer ());
	  }
	}
      }
    }				/* else */
  }
};

void 
Clustering::Berechne_fuer_neue_Daten (DVektorArray & Diese_Daten)
{
  DVektorArray Proj_Daten, Proj_Cluster;

  if ((Diese_Daten.Lese_Dim () != Die_waren_relevant.Lese_Dim ()) &&
      (Cluster.Lese_Groesse () != 0)) {
    Fehlermeldung ("Clustering::Berechne_fuer_neue_Daten(...)", DIMFEHLER);
  } else {
    if (Cluster.Lese_Groesse () != 0) {		/* gibt es was zu tun ? */
      Zugehoerigkeit.Setze_Dim (Cluster.Lese_Groesse (), Diese_Daten.Lese_Groesse ());
      Proj_Daten = Diese_Daten.Projiziere (Die_waren_relevant);
      Proj_Cluster = Cluster.Projiziere (Die_waren_relevant);

      if (Possibilistisch && ClusterFertig && Possib_Zaehler)
	Berechne_possib_U (Proj_Daten, Proj_Cluster, Zugehoerigkeit);
      else
	Berechne_U (Proj_Daten, Proj_Cluster, Zugehoerigkeit);

      if ((Art == OPTIMAL) && (Cluster.Lese_Groesse () > 2) && (!ClusterFertig)) {	/* Das vorherige Ergebnis */
	Tausche_Buffer ();
	Zugehoerigkeit.Setze_Dim (Cluster.Lese_Groesse (), Diese_Daten.Lese_Groesse ());
	Proj_Cluster = Cluster.Projiziere (Die_waren_relevant);

	Berechne_U (Proj_Daten, Proj_Cluster, Zugehoerigkeit);
	Tausche_Buffer ();
      }
//      Cluster=Cluster.Deprojiziere(Die_waren_relevant);/* gibt es schon */
      Berechne_Zuordnung ();
      Berechne_Form (Diese_Daten, Cluster);
    }
  }
};

void 
Clustering::Speichern (char *Name)
{
  FILE *File;

  if ((File = fopen (Name, "w")) == NULL) {
    Fehlermeldung ("Clustering::_File_speichern", DATEIOEFFNENFEHLER);
  } else {
    switch (Lese_Typ ()) {
/** Typ speichern **/
    case FCM:
      fprintf (File, "FCM\n");
      break;
    case GK:
      fprintf (File, "GK\n");
      break;
    case GK_parallel:
      fprintf (File, "GK_parallel\n");
      break;
    case GG:
      fprintf (File, "GG\n");
      break;
    case GG_parallel:
      fprintf (File, "GG_parallel\n");
      break;
    }

    switch (Art) {
/** Art speichern **/
    case KEINE:
      fprintf (File, "KEINE\n");
      break;
    case OPTIMAL:
      fprintf (File, "OPTIMAL\n");
      break;
    case ANZAHL:
      fprintf (File, "ANZAHL\n");
      break;
    case KLASSEN:
      fprintf (File, "KLASSEN\n");
      break;
    }

    switch (Guete_Verfahren->Lese_Typ ()) {
/** Typ speichern **/
    case S:
      fprintf (File, "S\n");
      break;
    case PK:
      fprintf (File, "PK\n");
      break;
    case CS:
      fprintf (File, "CS\n");
      break;
    case D1:
      fprintf (File, "D1\n");
      break;
    case FHV:
      fprintf (File, "FHV\n");
      break;
    case DPA:
      fprintf (File, "DPA\n");
      break;
    case PD:
      fprintf (File, "PD\n");
      break;
    default:
      break;
    }

    fprintf (File, "%d %d %d %d %lf\n", Iterationen,
	     Max_Iterationen, M, K, Delta_U);
    if (Fertig)
      fprintf (File, "1 ");
    else
      fprintf (File, "0 ");
    if (ClusterFertig)
      fprintf (File, "1 ");
    else
      fprintf (File, "0 ");
    if (Possibilistisch) {
      fprintf (File, "1 ");
      fprintf (File, "%d ", Possib_Zaehler);
      if (ClusterFertig && Possibilistisch && Possib_Zaehler) {
	eta.Speichern (File);
      }
    } else
      fprintf (File, "0 ");

    Guete_Verfahren->Speichern (File);
    Vorgabe_Cluster.Speichern (File);
    Cluster.Speichern (File);
    ClusterBuffer.Speichern (File);
    Die_waren_relevant.Speichern (File);
    Die_sind_relevant.Speichern (File);
    Mein_Speichern (File);
    fclose (File);
  }
};

Clustering *
Clustering_Laden (char *Name)
{
  Clustering *Result = 0;
  FILE *File;
  char Typ_String[20];

  if ((File = fopen (Name, "r")) == NULL) {
    Fehlermeldung ("Clustering_Laden(char* Name)", DATEIOEFFNENFEHLER);
    return (NULL);
  }
  do {
    if (!(fscanf (File, "%20s", Typ_String)))
      break;

    if (!strcmp (Typ_String, "FCM"))
      Result = new FCM_Clustering ();
    else if (!strcmp (Typ_String, "GK"))
      Result = new GK_Clustering ();
    else if (!strcmp (Typ_String, "GK_parallel"))
      Result = new GK_parallel_Clustering ();
    else if (!strcmp (Typ_String, "GG"))
      Result = new GG_Clustering ();
    else if (!strcmp (Typ_String, "GG_parallel"))
      Result = new GG_parallel_Clustering ();
    else {
      Fehlermeldung ("Clustering_Laden(char* Name)", DATEIFORMATFEHLER);
      return (NULL);		/* Fehler */
    }

    if (!(fscanf (File, "%20s", Typ_String)))
      break;
    if (!strcmp (Typ_String, "KEINE"))
      Result->Art = KEINE;
    else if (!strcmp (Typ_String, "OPTIMAL"))
      Result->Art = OPTIMAL;
    else if (!strcmp (Typ_String, "ANZAHL"))
      Result->Art = ANZAHL;
    else if (!strcmp (Typ_String, "KLASSEN"))
      Result->Art = KLASSEN;
    else
      break;			/* Fehler */

    if (!(fscanf (File, "%20s", Typ_String)))
      break;			/* Fehler */
    delete Result->Guete_Verfahren;
    if (!strcmp (Typ_String, "S"))
      Result->Guete_Verfahren = new S_Guete;
    else if (!strcmp (Typ_String, "PK"))
      Result->Guete_Verfahren = new PK_Guete;
    else if (!strcmp (Typ_String, "CS"))
      Result->Guete_Verfahren = new CS_Guete;
    else if (!strcmp (Typ_String, "D1"))
      Result->Guete_Verfahren = new D1_Guete;
    else if (!strcmp (Typ_String, "FHV"))
      Result->Guete_Verfahren = new FHV_Guete;
    else if (!strcmp (Typ_String, "DPA"))
      Result->Guete_Verfahren = new DPA_Guete;
    else if (!strcmp (Typ_String, "PD"))
      Result->Guete_Verfahren = new PD_Guete;
    else
      break;			/* Fehler */

    if (!(fscanf (File, "%d %d %d %d %lf\n", &(Result->Iterationen),
		  &(Result->Max_Iterationen), &(Result->M),
		  &(Result->K), &(Result->Delta_U))))
      break;
    if (!(fscanf (File, "%c %c %c", &(Typ_String[0]), &(Typ_String[1]),
		  &(Typ_String[2]))))
      break;

    if (Typ_String[0] == '1')
      Result->Fertig = TRUE;
    else
      Result->Fertig = FALSE;
    if (Typ_String[1] == '1')
      Result->ClusterFertig = TRUE;
    else
      Result->ClusterFertig = FALSE;
    if (Typ_String[2] == '1') {
      Result->Possibilistisch = TRUE;
      if (!(fscanf (File, "%d ", &(Result->Possib_Zaehler))))
	break;
      if (Result->ClusterFertig && Result->Possibilistisch && Result->Possib_Zaehler) {
	Result->eta.Laden (File);
      }
    } else
      Result->Possibilistisch = FALSE;

    if ((!Result->Guete_Verfahren->Laden (File)) ||
	(!Result->Vorgabe_Cluster.Laden (File)) ||
	(!Result->Cluster.Laden (File)) ||
	(!Result->ClusterBuffer.Laden (File)))
      break;
    if (Result->Art == OPTIMAL) {
      if (Result->Cluster.Lese_Groesse () == Result->ClusterBuffer.Lese_Groesse ())
	Result->zu_berechnen = Result->Cluster.Lese_Groesse () + 1;
      else
	Result->zu_berechnen = Result->Cluster.Lese_Groesse ();
    }
    if ((!Result->Die_waren_relevant.Laden (File)) ||
	(!Result->Die_sind_relevant.Laden (File)) ||
	(!Result->Mein_Laden (File)))
      break;
    fclose (File);
    return (Result);
  } while (TRUE);

  Fehlermeldung ("Clustering_Laden(char* Name)", DATEIFORMATFEHLER);
  delete Result;
  fclose (File);
  return (NULL);
};

void 
Clustering::Berechne_Zuordnung ()
{
  int DatenNr, ClusterNr;

  Zuordnung.Setze_Dim (Zugehoerigkeit.Lese_Dim_n ());
  if (Zugehoerigkeit.Lese_Dim_n () != 0) {
    for (DatenNr = 0; DatenNr < Zugehoerigkeit.Lese_Dim_n (); DatenNr++) {	/* alle Daten zuordnen */
      Zuordnung[DatenNr] = 0;	/* default */
      for (ClusterNr = 1; ClusterNr < Zugehoerigkeit.Lese_Dim_m (); ClusterNr++) {
	if (Zugehoerigkeit[ClusterNr][DatenNr] >
	    Zugehoerigkeit[Zuordnung[DatenNr]][DatenNr])
	  Zuordnung[DatenNr] = ClusterNr;
      }
    }
  }
};

DVektorArray 
Clustering::Berechne_v (DMatrix & MatrixU,
			DVektorArray & DieDaten)
return Result (DieDaten.Lese_Dim (), MatrixU.Lese_Dim_m (), NULL);
{
  int i, j, k;
  double SummeA, SummeB;

  for (i = 0; i < MatrixU.Lese_Dim_m (); i++) {		/* alle Cluster */
    SummeB = 0;
    for (k = 0; k < DieDaten.Lese_Groesse (); k++) {
      /*    E U_ik berechnen   */
      SummeB += pow (MatrixU.Lese_i_j (i, k), M);
    }
    for (j = 0; j < DieDaten.Lese_Dim (); j++) {	/* alle Dimensionen */
      SummeA = 0;
      for (k = 0; k < DieDaten.Lese_Groesse (); k++)
	SummeA += pow (MatrixU.Lese_i_j (i, k), M) 
	          * (DieDaten.Lese_Vektor (k, j));
      Result.Setze_Vektor (i, j, SummeA / SummeB);
    }	/* for ( j = ... */
  }   /* for (i = ... */
}

void 
Clustering::Berechne_U (DVektorArray & Daten,
			DVektorArray & Cluster, DMatrix & U)
{
  double Summe;
  int i, j, k, Dim;
  DVektor Dist (Cluster.Lese_Groesse (), 0, NULL);

/**********************************************/
  /* U fuer vorgegebene Prototypen berechnen    */
/**********************************************/
  for (k = 0; (Lese_Fehlerstatus () == KEINFEHLER) &&
       (k < Daten.Lese_Groesse ()); k++) {
    /* d_ik's  fuer ein k bestimmen */
    Dim = Berechne_Dist (Dist, Daten[k], Cluster);
    if (Dim > 0) {
      /* Falls I_k != \emptyset */
      for (i = 0; i < Cluster.Lese_Groesse (); i++)
	if (Dist[i] == 0)
	  U.Setze_i_j (i, k, 1.0 / ((double) Dim));
	else
	  U.Setze_i_j (i, k, 0);
    }
    /* if(Nulltest) */ 
    else {
      /* Falls I_k == \emptyset */
      for (i = 0; i < Cluster.Lese_Groesse (); i++) {
	if (Dist[i] == MAXDOUBLE)
	  U.Setze_i_j (i, k, 0);
	else {
	  Summe = 0;
	  for (j = 0; j < Cluster.Lese_Groesse (); j++)
	    Summe += pow (Dist[i] / Dist[j], 1.0 / (M - 1.0));
	  U.Setze_i_j (i, k, 1.0 / Summe);
	}
      }
    }				/* else */
  }				/* for(k=0 ; k < Daten.Lese_Groesse() ; k++) */
};

DVektor 
Clustering::Berechne_eta (DVektorArray & Daten,
			  DVektorArray & Cluster, DMatrix & U)
return Result (Cluster.Lese_Groesse (), 0, NULL);
{
  int i, k;
  double Potenz, Ni, Buffer;

  for (i = 0; i < Cluster.Lese_Groesse (); i++) {
    Buffer = 0;
    Ni = 0;
    for (k = 0; k < Daten.Lese_Groesse (); k++) {
      Potenz = pow (U[i][k], M);
      Buffer += Potenz * Die_Norm.d_quadr (Daten[k], Cluster[i]);
      Ni += Potenz;
    }
    Result[i] = (K / Ni) * Buffer;
  }
}

void 
Clustering::Speichere_Buffer ()
{
  ClusterBuffer = Cluster;
  UBuffer = Zugehoerigkeit;
};

void 
Clustering::Uebernehme_Buffer ()
{
  Zugehoerigkeit = UBuffer;
  Cluster = ClusterBuffer;
};

void 
Clustering::Tausche_Buffer ()
{
  Zugehoerigkeit.Tausche (UBuffer);
  Cluster.Tausche (ClusterBuffer);
};

void 
Clustering::Loesche_Buffer ()
{
  ClusterBuffer. ~ DVektorArray ();
  UBuffer. ~ DMatrix ();
};

void 
Errechne_Covarianz (DVektorArray & Daten, DMatrix & U,
		    DMatrixArray & CMatrix,
		    DVektorArray & Cluster, double M)
{
  DVektor DVektorBuffer;
  int i, k;
  double Summe, Buffer;

  for (i = 0; i < U.Lese_Dim_m (); i++) {	/* fuer alle Cluster */
    CMatrix[i].NullMatrix ();
    Summe = 0.0;
    for (k = 0; k < U.Lese_Dim_n (); k++) {
      DVektorBuffer = Daten[k] - Cluster[i];
      Buffer = pow (U.Lese_i_j (i, k), M);
      DVektorBuffer *= sqrt (Buffer);
      CMatrix[i] += DVektorBuffer.Produkt ();
      Summe += Buffer;
    }
    // verhindere entartete Cluster
    double minwert=MINFLOAT;
    while (fabs(CMatrix[i].Determinante())<=0.001) {
       int lauf;
       double spur,wert;
       spur=0.0;
       for(lauf=0; lauf<CMatrix[i].Lese_Dim_n (); lauf++)
	  spur+=CMatrix[i].Lese_i_j(lauf,lauf);
       spur=spur/CMatrix[i].Lese_Dim_n ()*0.01;
       if (fabs(spur)<(100.0*minwert))
	  spur=100.0*minwert;
       for(lauf=0; lauf<CMatrix[i].Lese_Dim_n (); lauf++) {
	  wert=CMatrix[i].Lese_i_j(lauf,lauf)+spur;
	  CMatrix[i].Setze_i_j(lauf,lauf,wert);
       }
    }
    CMatrix[i] /= Summe;
  }				/* for(i=0 ; i < U.Lese_Dim_m() ; i++) */
};

/**********************************************************/
/**********************************************************/
FCM_Clustering::FCM_Clustering (Guete_Typ Guete, int Soll_Iterationen,
		char Possib)
{
  Possibilistisch = Possib;
  sprintf (Name, "%s", "FCM");
};

Cluster_Typ 
FCM_Clustering::Lese_Typ ()
{
  return (FCM);
};

DMatrixArray 
FCM_Clustering::Lese_Form (IVektor & Dimensionen, DVektorArray & Daten)
return Result (Cluster.Lese_Groesse (), Dimensionen.Lese_Dim (),
	       Dimensionen.Lese_Dim (), NULL);
{
  int i, j;
  char Was_tun = TRUE;

  if (Dimensionen.Lese_Dim () == 2) {
    if (Dimensionen[0] == Dimensionen[1])
      Was_tun = FALSE;
  } else {
    if ((Dimensionen[0] == Dimensionen[1]) ||
	(Dimensionen[1] == Dimensionen[2]) ||
	(Dimensionen[0] == Dimensionen[2]))
      Was_tun = FALSE;
  }

  if (Was_tun) {
    for (i = 0; i < Cluster.Lese_Groesse (); i++) {
      for (j = 0; j < Dimensionen.Lese_Dim (); j++) {
	if (Die_sind_relevant[Dimensionen[j]])
	  Result[i][j][j] = Radien[i];
	else
	  Result[i][j][j] = 0;
      }
    }
  }
};

DVektorArray 
FCM_Clustering::Mein_Clustern (DVektorArray & Daten,
			       DVektorArray & Vorgabe_Protos)
return Result (Daten.Lese_Dim (), Vorgabe_Protos.Lese_Groesse (), NULL);
{
  int l;
  DMatrix U[2];

  if (Zugehoerigkeit.Lese_Dim_m () != Vorgabe_Protos.Lese_Groesse ()) {
    U[0] = DMatrix (Vorgabe_Protos.Lese_Groesse (), Daten.Lese_Groesse (), NULL);
    Berechne_U (Daten, Vorgabe_Protos, U[0]);
  } else
    U[0] = Zugehoerigkeit;

  U[1] = DMatrix (Vorgabe_Protos.Lese_Groesse (), Daten.Lese_Groesse (), NULL);

  /* Schleifen-Anfang */
  l = 0;
  while ((ClusterFertig != TRUE) && (Iterationen < Max_Iterationen)) {
    Iterationen++;
/*************************/
    /* II. v_i's berechnen   */
/*************************/
    Result = Berechne_v (U[l], Daten);
/*************************/
    /* III. U_neu bestimmen  */
/*************************/
    l = (l + 1) % 2;

    Berechne_U (Daten, Result, U[l]);
/*************************/
    /* IV. ||U[l-1]-U[l]||   */
/*************************/
    if (Die_Norm.d (U[0], U[1]) <= Delta_U) {
      ClusterFertig = TRUE;
    }
  };
  Zugehoerigkeit = U[l];
};

int 
FCM_Clustering::Berechne_Dist (DVektor & Dist, DVektor & Datum,
			       DVektorArray & Cluster)
{
  int Result = 0, i;

  for (i = 0; i < Cluster.Lese_Groesse (); i++) {
    Dist[i] = Die_Norm.d_quadr (Datum, Cluster[i]);
    if (Dist[i] == 0)
      Result++;			/* Wieviele sind 0 ? */
  }

  return (Result);
};

void 
FCM_Clustering::Berechne_possib_U (DVektorArray & Daten,
				   DVektorArray & Cluster, DMatrix & U)
{
  int i, j;

  for (i = 0; i < Cluster.Lese_Groesse (); i++) {	/* alle Cluster */
    for (j = 0; j < Daten.Lese_Groesse (); j++) {
      U[i][j] = 1.0 / (1.0 + pow (Die_Norm.d_quadr (Daten[j], Cluster[i]) / eta[i],
			      1.0 / (M - 1.0)));
    }
  }
};

void 
FCM_Clustering::Mache_Possibilistisch (DVektorArray & Daten,
				       DVektorArray & Cluster)
{
  int l;
  DMatrix U[2];
  char Test = FALSE;

  U[0] = Zugehoerigkeit;
  U[1].DMatrix (U[0].Lese_Dim_m (), U[0].Lese_Dim_n (), NULL);

  l = 0;

  if (eta.Lese_Dim () == 0)	/* Am Anfang */
    eta = Berechne_eta (Daten, Cluster, U[l]);

  if (Possib_Zaehler <= 2) {
    while ((Test == FALSE) && (Iterationen < Max_Iterationen)) {
      Iterationen++;
      /*    v_i's berechnen    */
      Cluster = Berechne_v (U[l], Daten);
      l = (l + 1) % 2;
/*************************/
      /*    U_neu berechnen    */
/*************************/
      Berechne_possib_U (Daten, Cluster, U[l]);

      if (Die_Norm.d (U[0], U[1]) < Delta_U) {
	Test = TRUE;
	Possib_Zaehler++;
	if (Possib_Zaehler <= 2)
	  eta = Berechne_eta (Daten, Cluster, U[l]);
	else {
	  Fertig = TRUE;
	}
      }
    };
  }				/* if(Possib_Zaehler <= 2) */
  Zugehoerigkeit = U[l];
};

void 
FCM_Clustering::Berechne_Form (DVektorArray & Daten,
			       DVektorArray & Cluster)
{
  int DatenNr;
  double Max;

  Radien.Setze_Dim (Cluster.Lese_Groesse ());
  Radien.NullVektor ();

  for (DatenNr = 0; DatenNr < Daten.Lese_Groesse (); DatenNr++) {	/* alle Daten */
    Max = Cluster[Zuordnung[DatenNr]].Abstand (Daten[DatenNr],
					       Die_sind_relevant);
    if (Max > Radien[Zuordnung[DatenNr]])
      Radien[Zuordnung[DatenNr]] = Max;
  }
};

/**********************************************************/
/**********************************************************/
GK_Clustering::GK_Clustering (Guete_Typ Guete, int Soll_Iterationen,
	       char Possib)
{
  sprintf (Name, "%s", "GK");
};

GK_Clustering::~GK_Clustering ()
{
};

Cluster_Typ 
GK_Clustering::Lese_Typ ()
{
  return (GK);
};

DMatrixArray 
GK_Clustering::Lese_parallel_Form (IVektor & Dimensionen,
				   DVektorArray & Daten)
return Result (Cluster.Lese_Groesse (), Dimensionen.Lese_Dim (),
	       Dimensionen.Lese_Dim (), NULL);
{
  int ClusterNr;

  if ((Dimensionen.Lese_Dim () == 2) &&
      (Dimensionen[0] != Dimensionen[1])) {
    for (ClusterNr = 0; ClusterNr < Cluster.Lese_Groesse (); ClusterNr++) {
      Result[ClusterNr][0][0] = Laengen[ClusterNr][Dimensionen[0]];
      Result[ClusterNr][1][1] = Laengen[ClusterNr][Dimensionen[1]];
    }
  }
/************** Jetzt 3D *******************/
  else {			/* (Dimensionen.Lese_Dim == 3) */
    if ((Dimensionen[0] != Dimensionen[1]) &&
	(Dimensionen[1] != Dimensionen[2]) &&
	(Dimensionen[0] != Dimensionen[2])) {
      for (ClusterNr = 0; ClusterNr < Cluster.Lese_Groesse (); ClusterNr++) {
	Result[ClusterNr][0][0] = Laengen[ClusterNr][Dimensionen[0]];
	Result[ClusterNr][1][1] = Laengen[ClusterNr][Dimensionen[1]];
	Result[ClusterNr][2][2] = Laengen[ClusterNr][Dimensionen[2]];
      }
    }
  }
};

DMatrixArray 
GK_Clustering::Lese_Form (IVektor & Dim,
			  DVektorArray & Daten)
return Result (Cluster.Lese_Groesse (), Dim.Lese_Dim (),
	       Dim.Lese_Dim (), NULL);
{
  int ClusterNr, DatenNr, DimNr, i;
  double Winkel, a, b, Nullstellen[2], Faktor[2], Verhaeltnis, Max;

/************* 2D-Fall *********************/
  if ((Dim.Lese_Dim () == 2) && (Dim[0] != Dim[1]) &&
      Die_sind_relevant[Dim[0]] &&
      Die_sind_relevant[Dim[1]]) {
    DMatrix KS (2, 2, NULL), KS2 (2, 2, NULL);
    DVektor Rotiert (2, 0, NULL);
    char Welche_drinnen;

    for (ClusterNr = 0; ClusterNr < Cluster.Lese_Groesse (); ClusterNr++) {
      Welche_drinnen = FALSE;
      for (i = 0; i < Zuordnung.Lese_Dim (); i++) {
	if (Zuordnung[i] == ClusterNr) {
	  Welche_drinnen = TRUE;
	  break;
	}
      }
      if (Welche_drinnen) {
	if (Covar[ClusterNr][Dim[0]][Dim[0]] == Covar[ClusterNr][Dim[1]][Dim[1]]) {	/*Sonderfall */
	  Winkel = 0;
	  Result[ClusterNr][0][0] = 1.0;	/* x -   */
	  Result[ClusterNr][1][0] = 0.0;	/* Achse */
	  Result[ClusterNr][0][1] = 0.0;	/* y -   */
	  Result[ClusterNr][1][1] = 1.0;	/* Achse */
	} else {
	  Winkel = (atan (2 * Covar[ClusterNr][Dim[0]][Dim[1]] /
			  (Covar[ClusterNr][Dim[0]][Dim[0]] - Covar[ClusterNr][Dim[1]][Dim[1]])) / 2);
	  if (Winkel < 0)
	    Winkel += M_PI_2;
	  Result[ClusterNr][0][0] = cos (Winkel);
	  Result[ClusterNr][1][0] = sin (Winkel);
	  Result[ClusterNr][0][1] = -Result[ClusterNr][1][0];
	  Result[ClusterNr][1][1] = Result[ClusterNr][0][0];
	}
	/* Die Achsenverhaeltnisse der Ellipsen feststellen */
	a = -(Covar[ClusterNr][Dim[0]][Dim[0]] + Covar[ClusterNr][Dim[1]][Dim[1]]);
	b = (Covar[ClusterNr][Dim[0]][Dim[0]] * Covar[ClusterNr][Dim[1]][Dim[1]] -
	Covar[ClusterNr][Dim[0]][Dim[1]] * Covar[ClusterNr][Dim[0]][Dim[1]]);
	Nullstellen[0] = -sqrt (a * a / 4 - b) - a / 2.0;
	Nullstellen[1] = sqrt (a * a / 4 - b) - a / 2.0;

	/* lange zu kurzer Seite */
	Verhaeltnis = sqrt (fabs (Nullstellen[1] / Nullstellen[0]));	/* >= 1 */

	/* Jetzt den Faktor herausfinden */
	KS = Result[ClusterNr].Gauss_Inverse ();
	KS2 = KS;
	KS.Zeile_Multiplizieren (0, 1.0 / Verhaeltnis);
	KS2.Zeile_Multiplizieren (1, 1.0 / Verhaeltnis);

	Faktor[0] = Faktor[1] = 0.0;
	for (DatenNr = 0; DatenNr < Daten.Lese_Groesse (); DatenNr++) {
	  if (ClusterNr == Zuordnung[DatenNr]) {
	    Rotiert[0] = KS[0][0] * (Daten[DatenNr][Dim[0]] - Cluster[ClusterNr][Dim[0]]) +
	      KS[0][1] * (Daten[DatenNr][Dim[1]] - Cluster[ClusterNr][Dim[1]]);
	    Rotiert[1] = KS[1][0] * (Daten[DatenNr][Dim[0]] - Cluster[ClusterNr][Dim[0]]) +
	      KS[1][1] * (Daten[DatenNr][Dim[1]] - Cluster[ClusterNr][Dim[1]]);
	    Max = Rotiert * Rotiert;
	    if (Max > Faktor[0])
	      Faktor[0] = Max;
	    Rotiert[0] = KS2[0][0] * (Daten[DatenNr][Dim[0]] - Cluster[ClusterNr][Dim[0]]) +
	      KS2[0][1] * (Daten[DatenNr][Dim[1]] - Cluster[ClusterNr][Dim[1]]);
	    Rotiert[1] = KS2[1][0] * (Daten[DatenNr][Dim[0]] - Cluster[ClusterNr][Dim[0]]) +
	      KS2[1][1] * (Daten[DatenNr][Dim[1]] - Cluster[ClusterNr][Dim[1]]);
	    Max = Rotiert * Rotiert;
	    if (Max > Faktor[1])
	      Faktor[1] = Max;
	  }
	}
	/* Jetzt noch in die Form eintragen */
	if (Faktor[0] < Faktor[1]) {	/* Dann nicht tauschen */
	  Result[ClusterNr].Spalte_Multiplizieren (
					 0, Verhaeltnis * sqrt (Faktor[0]));
	  Result[ClusterNr].Spalte_Multiplizieren (1, sqrt (Faktor[0]));
	} else {
	  Result[ClusterNr].Spalte_Multiplizieren (
					 1, Verhaeltnis * sqrt (Faktor[1]));
	  Result[ClusterNr].Spalte_Multiplizieren (0, sqrt (Faktor[1]));
	}
      }
    }
  }
/************** Jetzt 3D *******************/
  else if (Dim.Lese_Dim () == 3) {
    if ((Dim[0] != Dim[1]) && (Dim[1] != Dim[2]) && (Dim[0] != Dim[2])) {
      DMatrix KS (3, 3, NULL);
      DVektorArray Rotiert (3, Daten.Lese_Groesse (), NULL);
      double Achse[3];
      char Welche_drinnen;

      /* Zunaechst die Verhaletnisse der Achsen herausfinden */
      for (ClusterNr = 0; ClusterNr < Cluster.Lese_Groesse (); ClusterNr++) {
	Welche_drinnen = FALSE;
	for (i = 0; i < Zuordnung.Lese_Dim (); i++) {
	  if (Zuordnung[i] == ClusterNr) {
	    Welche_drinnen = TRUE;
	    break;
	  }
	}
	if (Welche_drinnen) {
	  Achse[0] = 0.0;
	  Achse[1] = 0.0;
	  Achse[2] = 0.0;
	  Result[ClusterNr] = (Covar[ClusterNr].Projiziere (Dim)).Eigenmatrix ();
	  KS = Result[ClusterNr].Gauss_Inverse ();
	  for (DatenNr = 0; DatenNr < Daten.Lese_Groesse (); DatenNr++) {
	    if (Zuordnung[DatenNr] == ClusterNr) {
	      for (DimNr = 0; DimNr < 3; DimNr++) {
		Rotiert[DatenNr][DimNr] =
		  KS[DimNr][0] * (Daten[DatenNr][Dim[0]] - Cluster[ClusterNr][Dim[0]]) +
		  KS[DimNr][1] * (Daten[DatenNr][Dim[1]] - Cluster[ClusterNr][Dim[1]]) +
		  KS[DimNr][2] * (Daten[DatenNr][Dim[2]] - Cluster[ClusterNr][Dim[2]]);
		Achse[DimNr] += Zugehoerigkeit[ClusterNr][DatenNr] *
		  Rotiert[DatenNr][DimNr] * Rotiert[DatenNr][DimNr];
	      }
	    }
	  }			/* for(DatenNr=0 ; DatenNr < Daten.Lese_Groesse() ; DatenNr++) */
	  /* Jetzt stehen die Achsenverhaeltnisse der Ellipsen fest */
	  /* Jetzt den Faktor herausfinden, indem jeder Punkt getestet wird */
	  Faktor[0] = 0.0;
	  for (DatenNr = 0; DatenNr < Daten.Lese_Groesse (); DatenNr++) {
	    if (Zuordnung[DatenNr] == ClusterNr) {
	      Max = 0.0;
	      for (DimNr = 0; DimNr < 3; DimNr++)
		Max += Rotiert[DatenNr][DimNr] * Rotiert[DatenNr][DimNr] /
		  Achse[DimNr];
	      if (Max > Faktor[0])
		Faktor[0] = Max;
	    }
	  }
	  Result[ClusterNr].Spalte_Multiplizieren (0, sqrt (Achse[0] * Faktor[0]));
	  Result[ClusterNr].Spalte_Multiplizieren (1, sqrt (Achse[1] * Faktor[0]));
	  Result[ClusterNr].Spalte_Multiplizieren (2, sqrt (Achse[2] * Faktor[0]));
	}
      }
    }
  }
};

void 
GK_Clustering::Mein_Speichern (FILE * File)
{
  AMatrizen.Speichern (File);
  if (Art == OPTIMAL)
    ABuffer.Speichern (File);
};

int 
GK_Clustering::Mein_Laden (FILE * File)
{
  if (!AMatrizen.Laden (File))
    return (0);
  if (Art == OPTIMAL) {
    if (!ABuffer.Laden (File))
      return (0);
  }
  return (1);
};

DVektorArray 
GK_Clustering::Mein_Clustern (DVektorArray & Daten,
			      DVektorArray & Vorgabe_Protos)
return Result (Daten.Lese_Dim (), Vorgabe_Protos.Lese_Groesse (), NULL);
{
  int i, l, Clusteranzahl = Vorgabe_Protos.Lese_Groesse ();
  DMatrix U[2];

  /* Initialisierungen */
  if (Zugehoerigkeit.Lese_Dim_m () != Vorgabe_Protos.Lese_Groesse ()) {
    AMatrizen.Setze_Dim (Daten.Lese_Dim (), Daten.Lese_Dim ());
    AMatrizen.Setze_Groesse (Clusteranzahl);
    if (rho.Lese_Dim () != Clusteranzahl) {	/* Dann eben alle auf 1 setzen */
      rho.Setze_Dim (Clusteranzahl);
      for (i = 0; i < Clusteranzahl; i++) {
	rho[i] = 1;
	AMatrizen[i].Einheitsmatrix ();
      }
    }
    U[0] = DMatrix (Clusteranzahl, Daten.Lese_Groesse (), NULL);
    Berechne_U (Daten, Vorgabe_Protos, U[0]);
  } else
    U[0] = Zugehoerigkeit;
  U[1] = DMatrix (Clusteranzahl, Daten.Lese_Groesse (), NULL);

/********************/
  /* Schleifen-Anfang */
/********************/
  l = 0;
  while ((ClusterFertig != TRUE) && (Iterationen < Max_Iterationen)) {
    Iterationen++;
/***********************/
    /* II. v_i's berechnen */
/***********************/
    Result = Berechne_v (U[l], Daten);
/************************/
    /* III. A[] bestimmen  */
/************************/
    Berechne_Ai (Daten, U[l], Result);
    if (Lese_Fehlerstatus () != KEINFEHLER) {
      ClusterFertig = TRUE;
      Reset_Fehlerstatus ();
      Result. ~ DVektorArray ();
      return (Result);
    }
/***********************/
    /* IV. U_neu bestimmen */
/***********************/
    l = (l + 1) % 2;
    Berechne_U (Daten, Result, U[l]);
/*************************/
    /* IV. ||U[l-1]-U[l]||   */
/*************************/
    if (Die_Norm.d (U[0], U[1]) <= Delta_U) {
      ClusterFertig = TRUE;
    }
  };

  Zugehoerigkeit = U[l];
};

int 
GK_Clustering::Berechne_Dist (DVektor & Dist, DVektor & Datum,
			      DVektorArray & Cluster)
{
  int Result = 0, i;

  for (i = 0; i < Cluster.Lese_Groesse (); i++) {
    Dist[i] = Die_A_Norm.d_quadr (AMatrizen[i], Datum, Cluster[i]);
    if (Dist[i] == 0)
      Result++;			/* Wieviele sind 0 ? */
  }

  return (Result);
};

void 
GK_Clustering::Berechne_possib_U (DVektorArray & Daten,
				  DVektorArray & Cluster, DMatrix & U)
{
  int i, j;

  for (i = 0; i < Cluster.Lese_Groesse (); i++) {	/* alle Cluster */
    for (j = 0; j < Daten.Lese_Groesse (); j++) {
      U.Setze_i_j (i, j, 1.0 / (1.0 + pow (Die_A_Norm.d_quadr (AMatrizen[i],
					     Daten[j], Cluster[i]) / eta[i],
				       1.0 / (M - 1.0))));
    }
  }
};

void 
GK_Clustering::Mache_Possibilistisch (DVektorArray & Daten,
				      DVektorArray & Cluster)
{
  int l;
  DMatrix U[2];
  char Test = FALSE;

  U[0] = Zugehoerigkeit;
  U[1].DMatrix (U[0].Lese_Dim_m (), U[0].Lese_Dim_n (), NULL);

  l = 0;
  if (eta.Lese_Dim () == 0.0) {	/* Am Anfang */
    Berechne_Ai (Daten, U[l], Cluster);
    eta = Berechne_eta (Daten, Cluster, U[l]);
  }
  if (Possib_Zaehler <= 2) {
    while ((Test == FALSE) && (Iterationen < Max_Iterationen)) {
      Iterationen++;
      /*    v_i's berechnen    */
      Cluster = Berechne_v (U[l], Daten);
      Berechne_Ai (Daten, U[l], Cluster);
      l = (l + 1) % 2;
/*************************/
      /*    U_neu berechnen    */
/*************************/
      Berechne_possib_U (Daten, Cluster, U[l]);
      if (Die_Norm.d (U[0], U[1]) < Delta_U) {
	Test = TRUE;
	Possib_Zaehler++;
	if (Possib_Zaehler <= 2) {
	  Berechne_Ai (Daten, U[l], Cluster);
	  eta = Berechne_eta (Daten, Cluster, U[l]);
	} else {
	  Fertig = TRUE;
	}
      }
    };
  }				/* if(Possib_Zaehler <= 2) */
  Zugehoerigkeit = U[l];
};

void 
GK_Clustering::Berechne_Ai (DVektorArray & Daten, DMatrix & U,
			    DVektorArray & Cluster)
{
  int i;
  double Buffer;

  Berechne_Covarianz (Daten, U, AMatrizen, Cluster);
  Reset_Fehlerstatus ();
  for (i = 0; i < Cluster.Lese_Groesse (); i++) {
    Buffer = pow (rho[i] * AMatrizen[i].Determinante (), 1.0 / Daten.Lese_Dim ());
    AMatrizen[i] = AMatrizen[i].Gauss_Inverse ();
    if (Lese_Fehlerstatus () != KEINFEHLER)
      break;
    AMatrizen[i] *= Buffer;
  }				/* for(i=0 ; i < Cluster.Lese_Groesse() ; i++) */
}

DVektor 
GK_Clustering::Berechne_eta (DVektorArray & Daten,
			     DVektorArray & Cluster, DMatrix & U)
return Result (Cluster.Lese_Groesse (), 0, NULL);
{
  int i, k;
  double Potenz, Ni, Buffer;

  for (i = 0; i < Cluster.Lese_Groesse (); i++) {
    Buffer = 0;
    Ni = 0;
    for (k = 0; k < Daten.Lese_Groesse (); k++) {
      Potenz = pow (U[i][k], M);
      Buffer += Potenz * Die_A_Norm.d_quadr (AMatrizen[i], Daten[k], Cluster[i]);
      Ni += Potenz;
    }
    Result[i] = (K / Ni) * Buffer;
  }
}

void 
GK_Clustering::parallel_Form_Berechnen (DVektorArray & Daten,
					DVektorArray & Cluster)
{
  int ClusterNr, Dim, DatenNr;
  double Max;
  DVektorArray Achsen (Daten.Lese_Dim (), Cluster.Lese_Groesse (), NULL);
  DVektor Faktor (Cluster.Lese_Groesse (), 0, NULL);

  Laengen.Setze_Dim (Daten.Lese_Dim ());
  Laengen.Setze_Groesse (Cluster.Lese_Groesse ());

  for (DatenNr = 0; DatenNr < Daten.Lese_Groesse (); DatenNr++) {
    for (Dim = 0; Dim < Daten.Lese_Dim (); Dim++)
      Achsen[Zuordnung[DatenNr]][Dim] +=
	(Daten[DatenNr][Dim] - Cluster[Zuordnung[DatenNr]][Dim]) *
	(Daten[DatenNr][Dim] - Cluster[Zuordnung[DatenNr]][Dim]);
  }				/* for(DatenNr=0 ; DatenNr < Daten.Lese_Groesse() ; DatenNr++) */
  /* Jetzt stehen die Achsenverhaeltnisse der Ellipsen fest */

  /* Jetzt den Faktor herausfinden, indem jeder Punkt getestet wird */
  for (DatenNr = 0; DatenNr < Daten.Lese_Groesse (); DatenNr++) {
    Max = 0;
    for (Dim = 0; Dim < Daten.Lese_Dim (); Dim++) {
      if (Achsen[Zuordnung[DatenNr]][Dim] != 0)
	Max += (Daten[DatenNr][Dim] - Cluster[Zuordnung[DatenNr]][Dim]) *
	  (Daten[DatenNr][Dim] - Cluster[Zuordnung[DatenNr]][Dim]) /
	  Achsen[Zuordnung[DatenNr]][Dim];
    }
    if (Max > Faktor[Zuordnung[DatenNr]])
      Faktor[Zuordnung[DatenNr]] = Max;
  }

  /* Jetzt noch in die Form eintragen */
  for (ClusterNr = 0; ClusterNr < Cluster.Lese_Groesse (); ClusterNr++) {
    for (Dim = 0; Dim < Daten.Lese_Dim (); Dim++) {
      Laengen[ClusterNr][Dim] = sqrt (Achsen[ClusterNr][Dim] * Faktor[ClusterNr]);
    }
  }
};

void 
GK_Clustering::Berechne_Form (DVektorArray & Daten,
			      DVektorArray & Cluster)
{
  DVektor VBuffer (Daten.Lese_Dim (), 0, NULL);
  int ClusterNr, DatenNr;


  Covar.Setze_Dim (Daten.Lese_Dim (), Daten.Lese_Dim ());
  Covar.Setze_Groesse (Cluster.Lese_Groesse ());

/************ jetzt geht es aber wirklich los *************/
  for (ClusterNr = 0; ClusterNr < Cluster.Lese_Groesse (); ClusterNr++) {
    /* Das werden die Kovarianzmatrizen fuer den Cluster */
    Covar[ClusterNr].NullMatrix ();
    for (DatenNr = 0; DatenNr < Daten.Lese_Groesse (); DatenNr++) {
      if (Zuordnung[DatenNr] == ClusterNr) {
	VBuffer = Daten[DatenNr] - Cluster[ClusterNr];
	Covar[ClusterNr] += VBuffer.Produkt (VBuffer);
      }
    }
  }
};

void 
GK_Clustering::Speichere_Buffer ()
{
  ClusterBuffer = Cluster;
  UBuffer = Zugehoerigkeit;
  ABuffer = AMatrizen;
  rho_Buffer = rho;
};

void 
GK_Clustering::Uebernehme_Buffer ()
{
  Zugehoerigkeit = UBuffer;
  Cluster = ClusterBuffer;
  AMatrizen = ABuffer;
  rho = rho_Buffer;
};

void 
GK_Clustering::Tausche_Buffer ()
{
  Zugehoerigkeit.Tausche (UBuffer);
  Cluster.Tausche (ClusterBuffer);
  AMatrizen.Tausche (ABuffer);
  rho.Tausche (rho_Buffer);
};

void 
GK_Clustering::Loesche_Buffer ()
{
  ClusterBuffer. ~ DVektorArray ();
  UBuffer. ~ DMatrix ();
  ABuffer. ~ DMatrixArray ();
  rho_Buffer. ~ DVektor ();
};

/**********************************************************/
/**********************************************************/
GK_parallel_Clustering::GK_parallel_Clustering (Guete_Typ Guete, int Soll_Iterationen,
			char Possib)
{
  Possibilistisch = Possib;
  sprintf (Name, "%s", "GK_parallel");
};

Cluster_Typ 
GK_parallel_Clustering::Lese_Typ ()
{
  return (GK_parallel);
};

void 
GK_parallel_Clustering::Berechne_Ai (DVektorArray & Daten, DMatrix & U,
				     DVektorArray & Cluster)
{
  int i, k, v;
  double Summe1, Summe2;

  for (i = 0; i < Cluster.Lese_Groesse (); i++) {	/* fuer alle Cluster */
    /* Zunaechst die Summen berechnen */
    Summe2 = 1.0;
    for (v = 0; v < Daten.Lese_Dim (); v++) {
      Summe1 = 0.0;
      for (k = 0; k < Daten.Lese_Groesse (); k++) {
	Summe1 += pow (U.Lese_i_j (i, k), M) *
	  (Daten[k][v] - Cluster[i][v]) *
	  (Daten[k][v] - Cluster[i][v]);
      }
      if (Summe1 == 0) {
	Fehlermeldung ("GK_parallel_Clustering::Berechne_Ai(...)", INVERTIERFEHLER);
	i = Cluster.Lese_Groesse ();
	break;
      }
      AMatrizen[i].Setze_i_j (v, v, 1.0 / (double) Summe1);
      Summe2 *= Summe1;
    }				/* for(v=0 ; v < Daten.Lese_Dim() ; v++) */
    if (Lese_Fehlerstatus () == KEINFEHLER) {
      for (v = 0; v < Daten.Lese_Dim (); v++)
	AMatrizen[i].Setze_i_j (v, v,
		   pow (rho[i] * Summe2, 1.0 / ((double) Daten.Lese_Dim ())) *
				AMatrizen[i].Lese_i_j (v, v));
    }
  }				/* for(i=0 ; i < U.Lese_Dim_m() ; i++) */
};

void 
GK_parallel_Clustering::Berechne_Form (DVektorArray & Daten,
				       DVektorArray & Cluster)
{
  parallel_Form_Berechnen (Daten, Cluster);
};

/**********************************************************/
/**********************************************************/
GG_Clustering::GG_Clustering (Guete_Typ Guete, int Soll_Iterationen,
	       char Possib)
{
  sprintf (Name, "%s", "GG");
};

Cluster_Typ 
GG_Clustering::Lese_Typ ()
{
  return (GG);
};

void 
GG_Clustering::Mein_Speichern (FILE * File)
{
  rho.Speichern (File);
  AMatrizen.Speichern (File);
  Faktor.Speichern (File);
  p.Speichern (File);
  if (Art == OPTIMAL) {
    ABuffer.Speichern (File);
    Faktor_Buffer.Speichern (File);
    p_Buffer.Speichern (File);
  }
};

int 
GG_Clustering::Mein_Laden (FILE * File)
{
  if ((!rho.Laden (File)) ||
      (!AMatrizen.Laden (File)) ||
      (!Faktor.Laden (File)) ||
      (!p.Laden (File)))
    return (0);
  if (Art == OPTIMAL) {
    if ((!ABuffer.Laden (File)) ||
	(!Faktor_Buffer.Laden (File)) ||
	(!p_Buffer.Laden (File)))
      return (0);
  }
  return (1);
};

DVektorArray 
GG_Clustering::Mein_Clustern (DVektorArray & Daten,
			      DVektorArray & Vorgabe_Protos)
return Result (Daten.Lese_Dim (), Vorgabe_Protos.Lese_Groesse (), NULL);
{
  int i, l;
  DMatrix U[2];

  if (Zugehoerigkeit.Lese_Dim_m () != Vorgabe_Protos.Lese_Groesse ()) {
    p.Setze_Dim (Vorgabe_Protos.Lese_Groesse ());
    p = 1.0 / (double) p.Lese_Dim ();
    Faktor.Setze_Dim (Cluster.Lese_Groesse ());
    Faktor = 1.0 / p[0];
    AMatrizen.Setze_Dim (Daten.Lese_Dim (), Daten.Lese_Dim ());
    AMatrizen.Setze_Groesse (Vorgabe_Protos.Lese_Groesse ());
    for (i = 0; i < Vorgabe_Protos.Lese_Groesse (); i++)
      AMatrizen[i].Einheitsmatrix ();

    U[0].Setze_Dim (Vorgabe_Protos.Lese_Groesse (),
		    Daten.Lese_Groesse ());
    Berechne_U (Daten, Vorgabe_Protos, U[0]);
  } else
    U[0] = Zugehoerigkeit;
  U[1].Setze_Dim (Vorgabe_Protos.Lese_Groesse (),
		  Daten.Lese_Groesse ());

/********************/
  /* Schleifen-Anfang */
/********************/
  l = 0;

  while ((ClusterFertig != TRUE) && (Iterationen < Max_Iterationen)) {
    Iterationen++;
/***********************/
    /* I. v_i's berechnen */
/***********************/
    Result = Berechne_v (U[l], Daten);
/***************************/
    /* II. A[] etc. bestimmen */
/***************************/
    Berechne_Ai (Daten, U[l], Result);
    if (Lese_Fehlerstatus () != KEINFEHLER) {
      ClusterFertig = TRUE;
      Reset_Fehlerstatus ();
      Result. ~ DVektorArray ();
      return (Result);
    }
/********************/
    /* III. p bestimmen */
/********************/
    Berechne_p (U[l]);
/***********************/
    /* IV. U_neu bestimmen */
/***********************/
    Berechne_Faktor ();
    l = (l + 1) % 2;
    Berechne_U (Daten, Result, U[l]);
    if (Lese_Fehlerstatus () != KEINFEHLER) {
      ClusterFertig = TRUE;
      Reset_Fehlerstatus ();
      Result. ~ DVektorArray ();
      return (Result);
    }
/*************************/
    /*  V. ||U[l-1]-U[l]||   */
/*************************/
    if (Die_Norm.d (U[0], U[1]) <= Delta_U)
      ClusterFertig = TRUE;
  };
  Zugehoerigkeit = U[l];
};

void 
GG_Clustering::Berechne_Ai (DVektorArray & Daten, DMatrix & U,
			    DVektorArray & Cluster)
{
  int i;

  Berechne_Covarianz (Daten, U, AMatrizen, Cluster);
  for (i = 0; i < Cluster.Lese_Groesse (); i++) {
    AMatrizen[i] = AMatrizen[i].Gauss_Inverse ();
    if (Lese_Fehlerstatus () != KEINFEHLER)
      break;
  };
};

void 
GG_Clustering::Berechne_Faktor ()
{
  int i;

  for (i = 0; i < AMatrizen.Lese_Groesse (); i++)
    Faktor[i] = p[i] / sqrt (AMatrizen[i].Determinante ());
};

int 
GG_Clustering::Berechne_Dist (DVektor & Dist, DVektor & Datum,
			      DVektorArray & Cluster)
{
  int Result = 0, i;
  double Summe;

  for (i = 0; i < Cluster.Lese_Groesse (); i++) {
    Summe = Die_A_Norm.d_quadr (AMatrizen[i], Datum, Cluster[i]);
    if (Summe < 1300) {
      Dist[i] = exp (.5 * Summe) * Faktor[i];
      if (Dist[i] == 0)
	Result++;
    } else
      Dist[i] = MAXDOUBLE;
  }
  return (Result);
};

void 
GG_Clustering::Berechne_possib_U (DVektorArray & Daten,
				  DVektorArray & Cluster, DMatrix & U)
{
  int i, j;
  double Buffer;

  for (i = 0; i < Cluster.Lese_Groesse (); i++) {	/* alle Cluster */
    for (j = 0; j < Daten.Lese_Groesse (); j++) {
      Buffer = Die_A_Norm.d_quadr (AMatrizen[i], Daten[j], Cluster[i]);
      if (Buffer < 1400) {
	Buffer = exp (.5 * Buffer) * Faktor[i];
	U[i][j] = 1.0 / (1.0 + pow (Buffer / eta[i],
					1.0 / (M - 1.0)));
      } else
	U[i][j] = 0;
    }
  }
};

void 
GG_Clustering::Mache_Possibilistisch (DVektorArray & Daten,
				      DVektorArray & Cluster)
{
  int l;
  DMatrix U[2];
  char Test = FALSE;

  U[0] = Zugehoerigkeit;
  U[1].DMatrix (U[0].Lese_Dim_m (), U[0].Lese_Dim_n (), NULL);

  l = 0;
  Berechne_p (U[l]);

  if (eta.Lese_Dim () == 0) {	/* Am Anfang */
    p.Setze_Dim (Cluster.Lese_Groesse ());
    Berechne_Ai (Daten, U[l], Cluster);
    Berechne_Faktor ();
    eta = Berechne_eta (Daten, Cluster, U[l]);
  }
  if (Possib_Zaehler <= 2) {
    while ((Test == FALSE) && (Iterationen < Max_Iterationen)) {
      Iterationen++;
      /*    v_i's berechnen    */
      Cluster = Berechne_v (U[l], Daten);

      l = (l + 1) % 2;
/*************************/
      /*    U_neu berechnen    */
/*************************/
      Berechne_possib_U (Daten, Cluster, U[l]);
      if (Die_Norm.d (U[0], U[1]) < Delta_U) {
	Test = TRUE;
	Possib_Zaehler++;
	if (Possib_Zaehler <= 2) {
	  Berechne_p (U[l]);
	  Berechne_Ai (Daten, U[l], Cluster);
	  Berechne_Faktor ();
	  eta = Berechne_eta (Daten, Cluster, U[l]);
	} else {
	  Fertig = TRUE;
	}
      }
    };
  }				/* if(Possib_Zaehler <= 2) */
  Zugehoerigkeit = U[l];
};

DVektor 
GG_Clustering::Berechne_eta (DVektorArray & Daten,
			     DVektorArray & Cluster, DMatrix & U)
return Result (Cluster.Lese_Groesse (), 0, NULL);
{
  int i, k;
  double Potenz, Ni, Buffer, Test;

  for (i = 0; i < Cluster.Lese_Groesse (); i++) {
    Buffer = 0;
    Ni = 0;
    for (k = 0; k < Daten.Lese_Groesse (); k++) {
      Test = Die_A_Norm.d_quadr (AMatrizen[i], Daten[k], Cluster[i]);
      if (Test < 1400) {
	Test = exp (.5 * Test) * Faktor[i];
      } else
	Test = MAXDOUBLE;
      Potenz = pow (U[i][k], M);
      Buffer += Potenz * Test;
      Ni += Potenz;
    }
    Result[i] = (K / Ni) * Buffer;
  }
}
void 
GG_Clustering::Berechne_p (DMatrix & U)
{
  double Teiler = 0.0, Buffer;
  int i, k;

  for (i = 0; i < U.Lese_Dim_m (); i++) {
    p[i] = 0;
    for (k = 0; k < U.Lese_Dim_n (); k++) {
      Buffer = pow (U[i][k], M);
      p[i] += Buffer;
      Teiler += Buffer;
    }
  }
  for (i = 0; i < U.Lese_Dim_m (); i++)
    p[i] /= Teiler;
};
/**********************************************************/
GG_parallel_Clustering::GG_parallel_Clustering (Guete_Typ Guete, int Soll_Iterationen,
			char Possib)
{
  sprintf (Name, "%s", "GG_parallel");
};

Cluster_Typ 
GG_parallel_Clustering::Lese_Typ ()
{
  return (GG_parallel);
};

void 
GG_parallel_Clustering::Berechne_Ai (DVektorArray & Daten, DMatrix & U,
				     DVektorArray & Cluster)
{
/******* Maximum-Likelihood-Schaetzer *********/
  int i, k, v;
  double Summe;

  for (i = 0; i < Cluster.Lese_Groesse (); i++) {	/* fuer alle Cluster */
    /* Zunaechst die Summen berechnen */
    for (v = 0; v < Daten.Lese_Dim (); v++) {
      Summe = 0;
      for (k = 0; k < Daten.Lese_Groesse (); k++) {
	Summe += pow (U[i][k], M) *
	  (Daten[k][v] - Cluster[i][v]) *
	  (Daten[k][v] - Cluster[i][v]);
      }
      AMatrizen[i][v][v] = Summe;
    }				/* for(v=0 ; v < Daten.Lese_Dim() ; v++) */
    Summe = 0;
    for (k = 0; k < Daten.Lese_Groesse (); k++) {
      Summe += pow (U[i][k], M);
    }
    for (v = 0; v < Daten.Lese_Dim (); v++) {
      if (AMatrizen[i][v][v] == 0) {
	Fehlermeldung ("GG_parallel_Clustering::Berechne_Ai(...)", INVERTIERFEHLER);
	i = Cluster.Lese_Groesse ();
	break;
      } else {			/* Zuviele Cluster gesucht */
	AMatrizen[i][v][v] = Summe / AMatrizen[i][v][v];
      }
    }
  }				/* for(i=0 ; i < U.Lese_Dim_m() ; i++) */
}

void 
GG_parallel_Clustering::Berechne_Faktor ()
{
  int i, j;

  for (i = 0; i < AMatrizen.Lese_Groesse (); i++) {
    Faktor[i] = 1;
    for (j = 0; j < AMatrizen.Lese_Dim_m (); j++) {
      Faktor[i] = Faktor[i] * AMatrizen[i][j][j];
    }
    Faktor[i] = 1.0 / (p[i] * Faktor[i]);
  }
};

void 
GG_parallel_Clustering::Berechne_Form (DVektorArray & Daten,
				       DVektorArray & Cluster)
{
  parallel_Form_Berechnen (Daten, Cluster);
};

void 
GG_Clustering::Speichere_Buffer ()
{
  ClusterBuffer = Cluster;
  UBuffer = Zugehoerigkeit;
  ABuffer = AMatrizen;
  p_Buffer = p;
  Faktor_Buffer = Faktor;
};

void 
GG_Clustering::Uebernehme_Buffer ()
{
  Zugehoerigkeit = UBuffer;
  Cluster = ClusterBuffer;
  AMatrizen = ABuffer;
  p = p_Buffer;
  Faktor = Faktor_Buffer;
};


void 
GG_Clustering::Tausche_Buffer ()
{
  Zugehoerigkeit.Tausche (UBuffer);
  Cluster.Tausche (ClusterBuffer);
  AMatrizen.Tausche (ABuffer);
  p.Tausche (p_Buffer);
  Faktor.Tausche (Faktor_Buffer);
};

void 
GG_Clustering::Loesche_Buffer ()
{
  ClusterBuffer. ~ DVektorArray ();
  UBuffer. ~ DMatrix ();
  ABuffer. ~ DMatrixArray ();
  p_Buffer. ~ DVektor ();
  Faktor_Buffer. ~ DVektor ();
};
