//**************************************************************
//*   filename: mainwindow.cc                                  *
//*                                                            *
//**************************************************************
//* programmed by: Thomas Wagner                               *
//* last change:  (XXX: not updated)                           *
//**************************************************************
#include <stdio.h>
#include <stdlib.h>

#include "mainwindow.h"
#include "saveloadwindow.h"

extern unsigned long buttonpix, whitepix, blackpix;
extern unsigned long shadowpix, lightpix, selectpix;
extern XContext Windowcontext;
extern Cursor movestartcursor, moveendcursor, delconnectcursor;
extern Cursor maincursor, waitcursor, createstartcursor;
extern Cursor deleteiconcursor, diskcursor;

MainWindow::MainWindow (Display * initdisplay,
	    GC initgc,
	    XFontStruct * initfontstruct,
	    int initx, int inity,
	    int initwidth, int initheight,
	    char *text):
BigWindow (initdisplay,
	   initgc,
	   DefaultRootWindow (initdisplay),
	   initfontstruct,
	   initx, inity, initwidth,
	   initheight,
	   text,
	   DEFAULT_BDWIDTH)
{
  XWMHints *p_xwmh;		/* Hints for the window manager  */
  XSizeHints *p_xsh;		/* Size hints for window manager */
  XClassHint *p_ch;		/* Class hint for window manager */
  XSetWindowAttributes xswa;
  XTextProperty blabla;
  moveableiconlist = new MoveableIconlist ();
  status = MAINNORMALSTATUS;
  Done = FALSE;

  if ((p_xwmh = XAllocWMHints ()) == NULL) {
    printf ("Error allocating Window Manager hints!\n");
    exit (1);
  }
  p_xwmh->flags = (InputHint | StateHint);
  p_xwmh->input = False;	//True;

  p_xwmh->initial_state = NormalState;

  if ((p_xsh = XAllocSizeHints ()) == NULL) {
    printf ("Error allocating size hints!\n");
    exit (1);
  }
  p_xsh->flags = (PPosition | PSize | PMinSize);
  p_xsh->width = p_xsh->min_width = width;
  p_xsh->height = p_xsh->min_height = height;
  p_xsh->x = x;
  p_xsh->y = y;

  if ((p_ch = XAllocClassHint ()) == NULL) {
    printf ("Error allocating class hint!\n");
    exit (1);
  }
  p_ch->res_name = text;
  p_ch->res_class = "XHello";

/* work around a bug in gcc v2.5.8 optimizer (copy text var): */
  {
    XTextProperty tp_dummy;
    char *text_ptr = text;
    XStringListToTextProperty (&text_ptr, 1, &tp_dummy);
    XSetWMProperties (display, window, &tp_dummy, &tp_dummy, &text_ptr, 0,
		      p_xsh, p_xwmh, p_ch);
  }

  xswa.colormap = DefaultColormap (display, DefaultScreen (display));
  xswa.bit_gravity = CenterGravity;

  XChangeWindowAttributes (display, window, (CWColormap | CWBitGravity),
			   &xswa);

  SetSelectedInput (ExposureMask
		    | StructureNotifyMask
		    | ButtonPressMask
		    | ButtonReleaseMask
		    | PointerMotionMask
		    | ButtonMotionMask | FocusChangeMask
    );

  XDefineCursor (display, window, maincursor);
  XSaveContext (display, window, Windowcontext, (caddr_t) this);

  allconnections = new Connectlist ();
  menu = new Menu (display, gc, this, fontstruct, 0, 0);

  MenuIconInfo Actions[4] =
  {
    {"Connection & Co.", STATUS_DEFAULT, 1, 0},
    {"new Connection", STATUS_DEFAULT, MAIN_CREATECONNECTION_ACTION, 0},
    {"delete Connection", STATUS_DEFAULT, MAIN_DESTROYCONNECTION_ACTION, 0},
    {"Quit", STATUS_DEFAULT, MAIN_QUIT_ACTION, 0}
  };
  menu->AddPulldownMenu (Actions, 4);

  MenuIconInfo Input[4] =
  {
    {"Inputobject", STATUS_DEFAULT, 1, 0},
    {"save Input", STATUS_DEFAULT, MAIN_SAVEINPUT_ACTION, 0},
    {"load Input", STATUS_DEFAULT, MAIN_NEWINPUT_ACTION, 0},
    {"delete Input", STATUS_DEFAULT, MAIN_DELETEINPUT_ACTION, 0}
  };
  menu->AddPulldownMenu (Input, 4);

  MenuIconInfo Cluster[4] =
  {
    {"Clusterobject", STATUS_DEFAULT, 1, 0},
    {"new Clusterobject", STATUS_DEFAULT, MAIN_CREATECLUSTERSELECTION_ACTION, 0},
    {"delete Clusterobject", STATUS_DEFAULT, MAIN_DELETECLUSTER_ACTION, 0},
    {"load Clusterobject", STATUS_DEFAULT, MAIN_LOADCLUSTER_OKACTION, 0}
  };
  menu->AddPulldownMenu (Cluster, 4);

  MenuIconInfo Clustersub[5] =
  {
    {"FCM Clustering", STATUS_DEFAULT, MAIN_CREATECLUSTERSELECTION_ACTION, FCM},
  {"GK Clustering", STATUS_DEFAULT, MAIN_CREATECLUSTERSELECTION_ACTION, GK},
    {"GK parallel Clustering", STATUS_DEFAULT, MAIN_CREATECLUSTERSELECTION_ACTION, GK_parallel},
  {"GG Clustering", STATUS_DEFAULT, MAIN_CREATECLUSTERSELECTION_ACTION, GG},
    {"GG parallel Clustering", STATUS_DEFAULT, MAIN_CREATECLUSTERSELECTION_ACTION, GG_parallel}
  };

  menu->GetPulldownMenu (2)->GetMenuIcon (0)->
    SetSonPulldownMenu (new PulldownMenu (display,
					  gc,
					  this,
					  menu,
					  fontstruct,
					  50, 30,
					  30,
					  Clustersub, 5));

  MenuIconInfo Rule[4] =
  {
    {"Ruleobject", STATUS_DEFAULT, 1, 0},
    {"save Ruleobject", STATUS_DEFAULT, MAIN_SAVERULE_ACTION, 0},
    {"load Ruleobject", STATUS_DEFAULT, MAIN_LOADRULE_ACTION, 0},
    {"delete Ruleobject", STATUS_DEFAULT, MAIN_DELETERULE_ACTION, 0}
  };
  menu->AddPulldownMenu (Rule, 4);

  MenuIconInfo Classify[4] =
  {
    {"Classifyobject", STATUS_DEFAULT, 1, 0},
    {"save Classifyobject", STATUS_DEFAULT, MAIN_SAVECLASSIFY_ACTION, 0},
    {"load Classifyobject", STATUS_DEFAULT, MAIN_LOADCLASSIFY_ACTION, 0},
    {"delete Classifyobject", STATUS_DEFAULT, MAIN_DELETECLASSIFY_ACTION, 0}
  };
  menu->AddPulldownMenu (Classify, 4);

  saveloadwindow = new SaveLoadWindow (display,
				       gc,
				       this,
				       fontstruct,
				       100, 100,
				       MAIN_CREATEINPUT_OK,
				       "LOAD INPUT",
				       WORK_AS_LOAD_SELECTOR);
  errorbox = new ErrorBox (display,
			   gc,
			   this,
			   fontstruct,
			   0, height - 17,
			   350, 15);
}

MainWindow::~MainWindow ()
{
  UnmapMe ();
  delete menu;
  delete saveloadwindow;
  delete errorbox;
  delete moveableiconlist;
  delete allconnections;
}

void 
MainWindow::printError (char *newtext)
{
  errorbox->ChangeText (newtext);
  errorbox->MapMe ();
  XBell (display, 100);
}

void 
MainWindow::HandleEventqueue ()
{
  while (!Done) {
    XEvent theEvent;		// Structure for current event

    Window_Info *window_info;

    XNextEvent (display, &theEvent);

    if (! XFindContext (display, theEvent.xany.window, Windowcontext,
		       (caddr_t *) & window_info)) {
      window_info->HandleEvent (&theEvent);
    }
  }				// while (!Done)

}

void 
MainWindow::SetStatus (MAINSTATUS newstatus)
{
  status = newstatus;
  if (status == MAINNORMALSTATUS) {
    menu->MakeSelectable ();
    XDefineCursor (display, window, maincursor);
  } else {
    menu->MakeUnselectable ();
  }

  errorbox->UnmapMe ();
}

void 
MainWindow::Action (int actionnumber, int value)
{
  switch (actionnumber) {

  case MAIN_NEWINPUT_ACTION:
    saveloadwindow->ChangeFunction (WORK_AS_LOAD_SELECTOR,
				    this,
				    MAIN_CREATEINPUT_OK,
				    "LOAD INPUT");
    saveloadwindow->MapMe ();
    SetStatus (MAINLOADINPUTSTATUS);
    break;

  case MAIN_CREATEINPUT_OK:
    {
      if ((char *) value != NULL) {
	Reset_Fehlerstatus ();
	Datensatz tempdataset = File_einlesen ((char *) value);
	SetStatus (MAINNORMALSTATUS);
	if (Lese_Fehlerstatus () == KEINFEHLER) {
	  InputIcon *temp = new InputIcon (display,
					   gc,
					   fontstruct,
					   this,
					   20, 70,
				    MOVEABLEICON_WIDTH, MOVEABLEICON_HEIGHT,
					   (char *) value);
	  temp->GetOutput ()->SetDataset (&tempdataset);
	  AddnewMoveableIcon (temp);
	} else {
	  switch (Lese_Fehlerstatus ()) {
	  case DATEIFORMATFEHLER:
	    printError ("Error in format of input-file!");
	    break;

	  default:
	    break;
	  }
	  Reset_Fehlerstatus ();
	}
      } else {
	SetStatus (MAINNORMALSTATUS);
      }
    }
    break;

  case MAIN_SAVEINPUT_ACTION:
    SetStatus (MAINSAVEINPUTSTATUS);
    XDefineCursor (display, window, diskcursor);
    break;

  case MAIN_SAVERULE_ACTION:
    SetStatus (MAINSAVERULESTATUS);
    XDefineCursor (display, window, diskcursor);
    break;

  case MAIN_LOADRULE_ACTION:
    saveloadwindow->ChangeFunction (WORK_AS_LOAD_SELECTOR,
				    this,
				    MAIN_LOADRULE_OK,
				    "LOAD RULE-OBJECT");
    saveloadwindow->MapMe ();
    SetStatus (MAINLOADRULESTATUS);
    break;

  case MAIN_LOADRULE_OK:
    if ((char *) value != NULL) {
      RuleInfo tempRuleinfo;
      tempRuleinfo.datasetname = (char *) value;
      Reset_Fehlerstatus ();
      tempRuleinfo.rules = Regelbasis_einlesen ((char *) value);
      SetStatus (MAINNORMALSTATUS);
      if (Lese_Fehlerstatus () == KEINFEHLER) {
	RuleIcon *temp = new RuleIcon (display,
				       gc,
				       fontstruct,
				       this,
				       20, 70,
				    MOVEABLEICON_WIDTH, MOVEABLEICON_HEIGHT,
				       (char *) value,
				       &tempRuleinfo);
	AddnewMoveableIcon (temp);
      } else {
	switch (Lese_Fehlerstatus ()) {
	case DATEIFORMATFEHLER:
	  printError ("Error in format of rule-file!");
	  break;

	default:
	  break;
	}
	Reset_Fehlerstatus ();
      }
    } else {
      SetStatus (MAINNORMALSTATUS);
    }
    break;

  case MAIN_SAVECLASSIFY_ACTION:
    SetStatus (MAINSAVECLASSIFYSTATUS);
    XDefineCursor (display, window, diskcursor);
    break;

  case MAIN_LOADCLASSIFY_ACTION:
    saveloadwindow->ChangeFunction (WORK_AS_LOAD_SELECTOR,
				    this,
				    MAIN_LOADCLASSIFY_OK,
				    "LOAD CLASSIFY-OBJECT");
    saveloadwindow->MapMe ();
    SetStatus (MAINLOADCLASSIFYSTATUS);
    break;

  case MAIN_LOADCLASSIFY_OK:
    if ((char *) value != NULL) {
      RuleInfo tempRuleinfo;
      tempRuleinfo.datasetname = (char *) value;
      Reset_Fehlerstatus ();
      tempRuleinfo.classify = Klassifizierung_einlesen ((char *) value);
      SetStatus (MAINNORMALSTATUS);
      if (Lese_Fehlerstatus () == KEINFEHLER) {
	ClassifyIcon *temp = new ClassifyIcon (display,
					       gc,
					       fontstruct,
					       this,
					       20, 70,
				    MOVEABLEICON_WIDTH, MOVEABLEICON_HEIGHT,
					       (char *) value,
					       &tempRuleinfo);
	AddnewMoveableIcon (temp);
//           temp->DrawMoveableIcon();
      } else {
	switch (Lese_Fehlerstatus ()) {
	case DATEIFORMATFEHLER:
	  printError ("Error in format of classify-file!");
	  break;

	default:
	  break;
	}
	Reset_Fehlerstatus ();
      }
    } else {
      SetStatus (MAINNORMALSTATUS);
    }
    break;

  case MAIN_LOADCLUSTER_OKACTION:
    saveloadwindow->ChangeFunction (WORK_AS_LOAD_SELECTOR,
				    this,
				    MAIN_LOADCLUSTER_OK,
				    "LOAD CLUSTER-OBJECT");
    saveloadwindow->MapMe ();
    SetStatus (MAINLOADCLUSTERSTATUS);
    break;

  case MAIN_LOADCLUSTER_OK:
    if ((char *) value != NULL) {
      Reset_Fehlerstatus ();
      Clustering *tempclustering = Clustering_Laden ((char *) value);
      if (Lese_Fehlerstatus () == KEINFEHLER) {
	ClusterIcon *temp = new ClusterIcon (display,
					     gc,
					     fontstruct,
					     this,
					     20, 70,
				    MOVEABLEICON_WIDTH, MOVEABLEICON_HEIGHT,
					     (char *) value,
					     NULL);
//                                       tempclustering);
	AddnewMoveableIcon (temp);
	temp->GetOutput ()->Action (OUTPUT_LOADOK, (int) tempclustering);
	temp->DrawMoveableIcon ();
      } else {
	switch (Lese_Fehlerstatus ()) {
	case DATEIFORMATFEHLER:
	  printError ("Error in format of cluster-file!");
	  break;

	default:
	  break;
	}
	Reset_Fehlerstatus ();
      }
    }
    SetStatus (MAINNORMALSTATUS);
    break;

  case MAIN_MAPME_ACTION:
    MapMe ();
    break;

  case MAIN_CREATERULE_ACTION:
    {
      RuleIcon *temp = new RuleIcon (display,
				     gc,
				     fontstruct,
				     this,
				     20, 70,
				     MOVEABLEICON_WIDTH, MOVEABLEICON_HEIGHT,
				     ((RuleInfo *) value)->datasetname,
				     (RuleInfo *) value);
      AddnewMoveableIcon (temp);
    }
    break;

  case MAIN_CREATECLASSIFY_ACTION:
    {
      ClassifyIcon *temp = new ClassifyIcon (display,
					     gc,
					     fontstruct,
					     this,
					     20, 70,
				    MOVEABLEICON_WIDTH, MOVEABLEICON_HEIGHT,
					  ((RuleInfo *) value)->datasetname,
					     (RuleInfo *) value);
      AddnewMoveableIcon (temp);
    }
    break;

  case MAIN_CREATECLUSTERSELECTION_ACTION:
    {
      ClusterInfo cinfo (NULL,
			 0,
			 value,
			 MAXINT,
			 MAXINT,
			 2,
			 1,
			 0,
			 DEFAULT_DELTA_U,
			 S,
			 NULL);
      ClusterOptionWindow *clusteroptionwindow =
      new ClusterOptionWindow (display,
			       gc,
			       this,
			       fontstruct,
			       20, 20,
			       "Create Cluster-Object",
			       MAIN_CREATECLUSTERING_OK,
			       &cinfo);
      clusteroptionwindow->MapMe ();
      SetStatus (MAINNEWCLUSTERINGSTATUS);
    }
    break;

  case MAIN_DELETECLUSTER_ACTION:
    SetStatus (MAINDELETECLUSTERINGSTATUS);
    XDefineCursor (display, window, deleteiconcursor);
    break;

  case MAIN_DELETECLASSIFY_ACTION:
    SetStatus (MAINDELETECLASSIFYSTATUS);
    XDefineCursor (display, window, deleteiconcursor);
    break;

  case MAIN_DELETEINPUT_ACTION:
    SetStatus (MAINDELETEINPUTSTATUS);
    XDefineCursor (display, window, deleteiconcursor);
    break;

  case MAIN_DELETERULE_ACTION:
    SetStatus (MAINDELETERULESTATUS);
    XDefineCursor (display, window, deleteiconcursor);
    break;

  case MAIN_CREATECLUSTERING_OK:
    if ((ClusterInfo *) value != NULL) {
      ClusterIcon *temp =
      new ClusterIcon (display,
		       gc,
		       fontstruct,
		       this,
		       20, 70,
		       MOVEABLEICON_WIDTH, MOVEABLEICON_HEIGHT,
		       "unplugged",
		       (ClusterInfo *) value);
//                                     NULL);
      AddnewMoveableIcon (temp);
    }
    SetStatus (MAINNORMALSTATUS);
    break;

  case MAIN_CREATECONNECTION_ACTION:
    XDefineCursor (display, window, createstartcursor);
    SetStatus (MAINSTARTCONNECTSTATUS);
    break;

  case MAIN_QUIT_ACTION:
    {
      Dialog *dialog = new Dialog (display,
				   gc,
				   this,
				   fontstruct,
				   10, 10, 200, 80,
				   "Do you really want to quit!",
				   DISPLAY_YES | DISPLAY_NO,
				   MAIN_QUIT_DIALOG_FINISHED_ACTION);
      SetStatus (MAINQUITREQUESTSTATUS);
      dialog->MapMe ();
    }
    break;

  case MAIN_QUIT_DIALOG_FINISHED_ACTION:
    if (value == DIALOG_YES_ACTION)
      Done = TRUE;
    SetStatus (MAINNORMALSTATUS);
    break;

  case MAIN_DESTROYCONNECTION_ACTION:
    if (allconnections->GetConnection () != NULL) {
      SetStatus (MAINDESTROYCONNECTSTATUS);
      XDefineCursor (display, window, delconnectcursor);
    }
    break;

  default:
    printf ("MainWindow::Action:  don't know action: %d\n", actionnumber);
  }
}

void 
MainWindow::DelMoveableIcon (MoveableIcon * Icon)
{
  *moveableiconlist -= Icon;
  XClearWindow (display, window);
  allconnections->DrawAllArrow ();
}

void 
MainWindow::AddnewMoveableIcon (MoveableIcon * NewIcon)
{
  short minarea = MAXSHORT;
  short xcounter = 0, ycounter = 0, xpos = 0, ypos = 0;
  short bestxpos = 0, bestypos = MAIN_TOP_SPACE;
  short xrepeats = width / (NewIcon->Getwidth () + 2 * DEFAULT_BDWIDTH);
  short yrepeats = (height - MAIN_TOP_SPACE) / (NewIcon->Getheight () 
                   + 2 * DEFAULT_BDWIDTH);
  MoveableIconlist *temp;

  // first find a good position for new icon
  while (ycounter < yrepeats) {
    xcounter = 0;
    while (xcounter < xrepeats) {
      short area = 0;
      temp = moveableiconlist;
      while (temp != NULL) {
	xpos = xcounter * (NewIcon->Getwidth () + 2 * DEFAULT_BDWIDTH);
	ypos = ycounter * (NewIcon->Getheight () + 2 * DEFAULT_BDWIDTH) +
	  MAIN_TOP_SPACE;
	NewIcon->PlaceMe (xpos, ypos);
	MoveableIcon *icon = temp->GetIcon ();
	short cutwide, cutheight;

	if (icon != NULL) {
	  cutwide = 0;
	  cutheight = 0;
	  // calculate wide of cut
	  {
	    MoveableIcon *left, *right;
	    if (NewIcon->Getx () < icon->Getx ()) {
	      left = NewIcon;
	      right = icon;
	    } else {
	      right = NewIcon;
	      left = icon;
	    }
	    if ((left->Getx () + left->Getwidth ()) > right->Getx ()) {
	      // cut != 0

	      if ((right->Getx () + right->Getwidth ()) 
		     < (left->Getx () + left->Getwidth ())) {
		// right(-icon) in left(-icon)
		cutwide = right->Getwidth ();
	      } else {
		cutwide = left->Getx () + left->Getwidth () - right->Getx ();
	      }
	    }
	  }
	  // calculate height of cut
	  {
	    MoveableIcon *higher, *lower;
	    if (NewIcon->Gety () < icon->Gety ()) {
	      higher = NewIcon;
	      lower = icon;
	    } else {
	      lower = NewIcon;
	      higher = icon;
	    }

	    if ((higher->Gety () + higher->Getheight ()) > lower->Gety ()) {
	      // cut != 0

	      if ((lower->Gety () + lower->Getheight ()) 
		  < (higher->Gety () + higher->Getheight ())) {
		// lower(-icon) in higher(-icon)
		cutheight = lower->Getheight ();
	      } else {
		cutheight = higher->Gety () + higher->Getheight ()
			    - lower->Gety ();
	      }
	    }
	  }
	  area = MAX (area, cutwide * cutheight);
	}			// if(icon!=NULL)

	temp = temp->Getnext ();
      }
      if (area == 0)		// found one of the best positions
       {
	bestxpos = xpos;
	bestypos = ypos;
	ycounter = yrepeats;
	xcounter = xrepeats;
      } else {			// found a better position, but cutarea != 0
	if (area < minarea) {
	  bestxpos = xpos;
	  bestypos = ypos;
	  minarea = area;
	}
      }				// if(area==0)

      xcounter++;
    }
    ycounter++;
  }

  // now find the index, display in ruleicon
  if (NewIcon->GetTyp () == MOVEABLEICONTYP_RULE) {
    temp = moveableiconlist;
    short maxindex = 0;
    while (temp != NULL) {
      MoveableIcon *icon = temp->GetIcon ();
      if (icon != NULL) {
	if (icon->GetTyp () == MOVEABLEICONTYP_RULE) {
	  if (((RuleIcon *) icon)->GetIndex () != 0)
	    maxindex = MAX (((RuleIcon *) icon)->GetIndex (), maxindex);
	}
      }
      temp = temp->Getnext ();
    }
    ((RuleIcon *) NewIcon)->SetIndex (maxindex + 1);
  }
  // now find the index, display in classifyicon
  if (NewIcon->GetTyp () == MOVEABLEICONTYP_CLASSIFY) {
    temp = moveableiconlist;
    short maxindex = 0;
    while (temp != NULL) {
      MoveableIcon *icon = temp->GetIcon ();
      if (icon != NULL) {
	if (icon->GetTyp () == MOVEABLEICONTYP_CLASSIFY) {
	  if (((ClassifyIcon *) icon)->GetIndex () != 0)
	    maxindex = MAX (((ClassifyIcon *) icon)->GetIndex (), maxindex);
	}
      }
      temp = temp->Getnext ();
    }
    ((ClassifyIcon *) NewIcon)->SetIndex (maxindex + 1);
  }
  // final add new icon to list
  *moveableiconlist += NewIcon;
  NewIcon->PlaceMe (bestxpos, bestypos);
  NewIcon->MapMe ();
}

void 
MainWindow::HandleEvent (XEvent * Event)
{
  switch (Event->type) {
  case Expose:
    XSetForeground (display, gc, buttonpix);
    XFillRectangle (display, window, gc,
		    Event->xgraphicsexpose.x, Event->xgraphicsexpose.y,
	       Event->xgraphicsexpose.width, Event->xgraphicsexpose.height);
    if (Event->xexpose.count == 0)
      allconnections->DrawAllArrow ();
    break;

  case ConfigureNotify:
    if ((Event->xconfigure.width != width) ||
	(Event->xconfigure.height != height)) {
      errorbox->MoveMe (0, Event->xconfigure.height - height);
      width = Event->xconfigure.width;
      height = Event->xconfigure.height;
      XResizeWindow (display, window, width, height);
      XSetForeground (display, gc, buttonpix);
      XFillRectangle (display, window, gc, 0, 0, width, height);
    }
    if (Event->xconfigure.x != x)
      x = Event->xconfigure.x;
    if (Event->xconfigure.y != y)
      y = Event->xconfigure.y;
    break;

  case ButtonPress:
    switch (Event->xbutton.button) {
    case Button1:
      int xpos, ypos;
      switch (status) {
      case MAINDESTROYCONNECTSTATUS:
	// destroy connector
	Connection * nearest;
	xpos = Event->xbutton.x;
	ypos = Event->xbutton.y;

	nearest = allconnections->FindnearestConnection (xpos, ypos);
	nearest->GetStartIcon ()->Action (ICON_REMOVECONNECTOUT, (int) nearest);
	nearest->GetEndIcon ()->Action (ICON_REMOVECONNECTIN, 0);
	delete nearest;
	XDefineCursor (display, window, maincursor);
	SetStatus (MAINNORMALSTATUS);
	XClearWindow (display, window);
	allconnections->DrawAllArrow ();
	break;			// case MAINDESTROYCONNECTSTATUS:

      case MAINNORMALSTATUS:
	char pointtyp;
	if (allconnections->GetConnection () != NULL) {
	  xpos = Event->xbutton.x;
	  ypos = Event->xbutton.y;

	  acuteconnection = allconnections->FindnearestConnection (xpos, ypos, &pointtyp);
	  if (pointtyp == STARTPOINT) {
	    XDefineCursor (display, window, movestartcursor);
	    acuteconnection->SetStart (xpos, ypos);
	    SetStatus (MAINMOVECONNECTSTARTSTATUS);
	  } else {
	    XDefineCursor (display, window, moveendcursor);
	    acuteconnection->SetEnd (xpos, ypos);
	    SetStatus (MAINMOVECONNECTENDSTATUS);
	  }
	  acuteconnection->SetStatus (CONNECTSTATUS_INVISIBLE);
	  XClearWindow (display, window);
	  allconnections->DrawAllArrow ();
	  acuteconnection->DrawtempArrow ();
	}
	break;			// case MAINNORMALSTATUS:

      default:
	break;
      }
      break;			// case Button1

    case Button3:
      switch (status) {
      case MAINSAVEINPUTSTATUS:
      case MAINSAVECLASSIFYSTATUS:
      case MAINSAVERULESTATUS:
      case MAINDELETECLUSTERINGSTATUS:
      case MAINDELETECLASSIFYSTATUS:
      case MAINDELETEINPUTSTATUS:
      case MAINDELETERULESTATUS:
	SetStatus (MAINNORMALSTATUS);
	break;

      case MAINDESTROYCONNECTSTATUS:
	SetStatus (MAINNORMALSTATUS);
	break;

      case MAINSTARTCONNECTSTATUS:
	// stop creating connection
	SetStatus (MAINNORMALSTATUS);
	break;

      case MAINENDCONNECTSTATUS:
	// or delete start-point of connection
	acuteconnection->ErasetempArrow ();
	*(acuteconnection->GetStartIcon ()->GetOutputConnectlist ()) -= acuteconnection;
	delete acuteconnection;
	SetStatus (MAINNORMALSTATUS);
	break;

      case MAINMOVECONNECTSTARTSTATUS:
	acuteconnection->ErasetempArrow ();
	acuteconnection->SetStart (acuteconnection->GetStartIcon ()->x + (int) (acuteconnection->GetStartIcon ()->width / 2),
				   acuteconnection->GetStartIcon ()->y + (int) (acuteconnection->GetStartIcon ()->height / 2));
	acuteconnection->SetStatus (CONNECTSTATUS_VISIBLE);
	acuteconnection->DrawArrow ();
	SetStatus (MAINNORMALSTATUS);
	break;

      case MAINMOVECONNECTENDSTATUS:
	acuteconnection->ErasetempArrow ();
	acuteconnection->SetEnd (acuteconnection->GetEndIcon ()->x + (int) (acuteconnection->GetEndIcon ()->width / 2),
				 acuteconnection->GetEndIcon ()->y + (int) (acuteconnection->GetEndIcon ()->height / 2));
	acuteconnection->SetStatus (CONNECTSTATUS_VISIBLE);
	acuteconnection->DrawArrow ();
	SetStatus (MAINNORMALSTATUS);
	break;

      default:
	break;
      }
      XDefineCursor (display, window, maincursor);
      break;			// case Button3 

    }
    break;			// case ButtonPress:

  case ButtonRelease:
    switch (Event->xbutton.button) {
    case Button1:
      break;			// case Button1

    }
    break;			// case ButtonRelease

  case MotionNotify:
    switch (status) {
      int xpos, ypos;
    case MAINENDCONNECTSTATUS:
      xpos = Event->xmotion.x;
      ypos = Event->xmotion.y;
      acuteconnection->ErasetempArrow ();
      acuteconnection->DrawtempArrow (xpos, ypos);
      break;			// case MAINENDCONNECTSTATUS

    case MAINMOVECONNECTSTARTSTATUS:
      xpos = Event->xmotion.x;
      ypos = Event->xmotion.y;
      acuteconnection->ErasetempArrow ();
      acuteconnection->SetStart (xpos, ypos);
      acuteconnection->DrawtempArrow ();
      break;			// case MAINMOVECONNECTSTARTSTATUS

    case MAINMOVECONNECTENDSTATUS:
      xpos = Event->xmotion.x;
      ypos = Event->xmotion.y;
      acuteconnection->ErasetempArrow ();
      acuteconnection->SetEnd (xpos, ypos);
      acuteconnection->DrawtempArrow ();
      break;			// case MAINMOVECONNECTENDSTATUS

    default:
      break;
    }
    break;			// switch(status)

  }
}
