/********************************************************/
/*   filename: matrix.cc                                */
/*                                                      */
/********************************************************/
/* programmed by: Oliver Wagner                         */
/* last change: 26-04-95                                */
/********************************************************/
#include "matrix.h"

/********************************************************/
/*              Alles zur Klasse DMatrix                */
/********************************************************/
DMatrix::DMatrix (int mGroesse, int nGroesse, double *Inhalt)
{
  int i;

  n = nGroesse;
  m = mGroesse;
  if ((n != 0) && (m != 0)) {
    if ((Daten = (double *) malloc (sizeof (double) * m * n)) == NULL)
        Fehlermeldung ("DMatrix", SPEICHERFEHLER);

    if (Inhalt != NULL)
      for (i = 0; i < n * m; i++)
	Daten[i] = Inhalt[i];
    else
      for (i = 0; i < n * m; i++)
	Daten[i] = 0;
  } else
    Daten = NULL;
};

DMatrix::DMatrix (const DMatrix & Daraus)
{
  int i;

  n = Daraus.n;
  m = Daraus.m;
  if ((n != 0) && (m != 0)) {
    if ((Daten = (double *) malloc (sizeof (double) * m * n)) == NULL)
        Fehlermeldung ("DMatrix::DMatrix(const DMatrix& Daraus)", SPEICHERFEHLER);

    for (i = 0; i < n * m; i++)
      Daten[i] = Daraus.Daten[i];
  } else
    Daten = NULL;
};

DMatrix::~DMatrix ()
{
  if ((n != 0) && (m != 0)) {
    n = m = 0;
    free (Daten);
    Daten = NULL;
  }
};

void 
DMatrix::print ()
{
  int i, j;

  printf ("\nDMatrix :");
  for (i = 0; i < m; i++) {
    if (i != 0)
      printf ("         ");	/* Versatz */
    for (j = 0; j < n; j++)
      printf ("%lf ", Daten[i * n + j]);
    printf ("\n");
  }
};

void 
DMatrix::Setze_Dim (int Neues_m, int Neues_n)
{

  if ((m != Neues_m) || (n != Neues_n)) {
    if ((m != 0) && (n != 0))
      free (Daten);
    m = Neues_m;
    n = Neues_n;
    if ((Daten = (double *) malloc (sizeof (double) * m * n)) == NULL)
        Fehlermeldung ("DMatrix", SPEICHERFEHLER);
  }
};

void 
DMatrix::Einheitsmatrix (int Groesse)
{
  int i;

  Setze_Dim (Groesse, Groesse);

  for (i = 0; i < Groesse * Groesse; i++)
    Daten[i] = 0;
  for (i = 0; i < Groesse * Groesse; i += Groesse + 1)
    Daten[i] = 1;
};

void 
DMatrix::Einheitsmatrix ()
{
  int i;

  for (i = 0; i < m * n; i++)
    Daten[i] = 0;
  for (i = 0; i < m * n; i += n + 1)
    Daten[i] = 1;
};

void 
DMatrix::NullMatrix ()
{
  int i;

  for (i = 0; i < n * m; i++)
    Daten[i] = 0;
};

DMatrix 
DMatrix::Projiziere (BVektor & Diese)
return Result (Diese.Anzahl_gesetzt (), Diese.Anzahl_gesetzt (), NULL);
{
  int i, j, a, b;

  a = 0;
  for (i = 0; i < m; i++) {
    b = 0;
    if (Diese[i]) {
      for (j = 0; j < n; j++) {
	if (Diese[j])
	  Result[a][b++] = (*this)[i][j];
      }
      a++;
    }
  }
}

DMatrix 
DMatrix::Projiziere (IVektor & Diese)
return Result (Diese.Lese_Dim (), Diese.Lese_Dim (), NULL);
{
  int i, j;

  for (i = 0; i < Result.m; i++) {
    for (j = 0; j < Result.n; j++) {
      Result[i][j] = (*this)[Diese[i]][Diese[j]];
    }
  }
}

DMatrix 
DMatrix::Deprojiziere (BVektor & Diese)
return Result (Diese.Lese_Dim (), Diese.Lese_Dim (), NULL);
{
  int i, j, a, b;

  a = 0;
  for (i = 0; i < Result.m; i++) {
    if (Diese[i]) {
      b = 0;
      for (j = 0; j < Result.n; j++) {
	if (Diese[j])
	  Result[i][j] = (*this)[a][b++];
      }
      a++;
    }
  }
}

DMatrix 
DMatrix::Gauss_Inverse ()
return Result (m, m, NULL);
{
  DMatrix *Buffer;
  int Spalte, Zeile;
  double Faktor;
  #define Inverse_eps 1e-12
  #define Inverse2_eps 1e-50


  Result.Einheitsmatrix (m);
  Buffer = new DMatrix (*this);

  for (Spalte = 0; Spalte < m; Spalte++) {
    if ((Buffer->Lese_i_j (Spalte, Spalte) <= Inverse_eps) &&
	(Buffer->Lese_i_j (Spalte, Spalte) >= -Inverse_eps)) {	/* moeliche Zeile zum Tauschen suchen */
      for (Zeile = Spalte + 1; Zeile < m; Zeile++) {
	if ((Buffer->Lese_i_j (Zeile, Spalte) >= Inverse_eps) ||
	    (Buffer->Lese_i_j (Zeile, Spalte) <= -Inverse_eps)) {
	  Buffer->Tausche_Zeilen (Spalte, Zeile);
	  Result.Tausche_Zeilen (Spalte, Zeile);
	  break;
	}
      }

      /* nullwert in letzter zeile/spalte */
      if (Zeile == m && 
	  (Buffer->Lese_i_j (Spalte, Spalte) <= Inverse2_eps) &&
	  (Buffer->Lese_i_j (Spalte, Spalte) >= -Inverse2_eps)) {
	printf("%g\n",Buffer->Lese_i_j (Spalte, Spalte));
	Fehlermeldung ("DMatrix::Gauss_Inverse()", INVERTIERFEHLER);
	return (Result);
      } else if (Zeile==m) {
	printf("%g\n",Buffer->Lese_i_j (Spalte, Spalte));
	printf ("Warning: matrix inversion might be unreliable\n");
      }
    }
    for (Zeile = 0; Zeile < m; Zeile++) {
      if ((Zeile != Spalte) &&
	  ((Buffer->Lese_i_j (Zeile, Spalte) >= Inverse_eps) ||
	   (Buffer->Lese_i_j (Zeile, Spalte) <= -Inverse_eps))) {
	/* Aufraeumen */
	Faktor = -Buffer->Lese_i_j (Zeile, Spalte) /
	  Buffer->Lese_i_j (Spalte, Spalte);

	Buffer->Zeile_plus_Zeile (Zeile, Spalte, Faktor);
	Result.Zeile_plus_Zeile (Zeile, Spalte, Faktor);
      }
    }
  }
  /* Diagonale auf 1 bringen */
  for (Spalte = 0; Spalte < m; Spalte++)
    Result.Zeile_Multiplizieren (Spalte, 1 / Buffer->Lese_i_j (Spalte, Spalte));

  delete Buffer;
};

double 
DMatrix::Determinante ()
{
  double Result = 1;
  DMatrix *Buffer;
  int Spalte, Zeile;
  double Faktor;

  Buffer = new DMatrix (*this);

  for (Spalte = 0; Spalte < m; Spalte++) {
    if (Buffer->Lese_i_j (Spalte, Spalte) == 0) {
      Result = 0;
      break;
    } else {
      for (Zeile = Spalte + 1; Zeile < m; Zeile++) {
	/* Aufraeumen */
	Faktor = -Buffer->Lese_i_j (Zeile, Spalte) /
	  Buffer->Lese_i_j (Spalte, Spalte);
	Buffer->Zeile_plus_Zeile (Zeile, Spalte, Faktor);
      }
    }
    /* Diagonale aufmultiplizieren */
    for (Spalte = 0; Spalte < m; Spalte++)
      Result *= Buffer->Lese_i_j (Spalte, Spalte);
  }

  delete Buffer;
  return (Result);
};

double 
DMatrix::Zeilensumme (int i)
{
  int x, bis;
  double Result = 0;

  bis = x + n;
  for (x = i * n; x < (i + 1) * n; x++)
    Result += Daten[x];
  return (Result);
};

void 
DMatrix::Tausche_Zeilen (int x, int y)
{
  double Buffer;
  int i;

  for (i = 0; i < n; i++) {
    Buffer = Lese_i_j (x, i);
    Setze_i_j (x, i, Lese_i_j (y, i));
    Setze_i_j (y, i, Buffer);
  }
};

void 
DMatrix::Tausche_Spalten (int x, int y)
{
  double Buffer;
  int i;

  for (i = 0; i < m; i++) {
    Buffer = Lese_i_j (i, x);
    Setze_i_j (i, x, Lese_i_j (i, y));
    Setze_i_j (i, y, Buffer);
  }
};

void 
DMatrix::Zeile_Multiplizieren (int Zeile, double Damit)
{
  int i;

  for (i = Zeile * n; i < (Zeile + 1) * n; i++)
    Daten[i] *= Damit;
};

void 
DMatrix::Spalte_Multiplizieren (int Spalte, double Damit)
{
  int i;

  for (i = Spalte; i < n * m; i += n)
    Daten[i] *= Damit;
};

void 
DMatrix::Zeile_plus_Zeile (int Diese, int plusDiese, double Faktor)
{
  int i;

  for (i = 0; i < n; i++)
    Setze_i_j (Diese, i, Lese_i_j (Diese, i) + Lese_i_j (plusDiese, i) * Faktor);
};

DVektor 
DMatrix::Spalte_aus_Matrix (int i)
  return Result (m, i, NULL);
{
  int j;

  for (j = 0; j < m; j++)
    Result[j] = Lese_i_j (j, i);
};

DVektor 
DMatrix::Zeile_aus_Matrix (int i)
  return Result (n, i, NULL);
{
  int j;
  for (j = 0; j < n; j++)
    Result[j] = Lese_i_j (i, j);
};

void 
DMatrix::Setze_Spalte (int Die, DVektor & Damit)
{
  int i;

  for (i = 0; i < m; i++)
    Daten[Die + i * n] = Damit[i];
};

DVektorArray 
DMatrix::Eigenvektoren ()
{
  IVektor Iterationen (m, NULL);

  Iterationen = ITERATIONEN;
  return (Eigenvektoren (Iterationen));
}

DVektorArray 
DMatrix::Eigenvektoren (IVektor & cnt)
return Result (n, n, NULL);
{
  DVektor Skal, valreal (n, 0, NULL), valim (n, 0, NULL);
  DMatrix Uebergabe_Matrix (*this), Eivec (n, n, NULL);
  int low, high, i;

  if (n < 2)
    Fehlermeldung ("DMatrix::Eigenvektoren(IVektor& cnt)", DIMFEHLER);

  Uebergabe_Matrix.Balance (Skal, low, high);
  Uebergabe_Matrix.Elmhes (low, high, cnt);
  Uebergabe_Matrix.Elmtrans (low, high, cnt, Eivec);
  Uebergabe_Matrix.Hqr2 (low, high, valreal, valim,
			 Eivec, cnt);
  Eivec.Balback (low, high, Skal);
  Eivec.Norm_1 (valim);

  for (i = 0; i < n; i++)
    Result[i] = Eivec.Spalte_aus_Matrix (i);
};

DMatrix 
DMatrix::Eigenmatrix ()
{
  IVektor Iterationen (m, NULL);

  Iterationen = ITERATIONEN;
  return (Eigenmatrix (Iterationen));
}

DMatrix 
DMatrix::Eigenmatrix (IVektor & cnt)
return Result (n, n, NULL);
{
  DVektor Skal, valreal (n, 0, NULL), valim (n, 0, NULL);
  DMatrix Uebergabe_Matrix (*this);
  int low, high;

  if (n < 2)
    Fehlermeldung ("DMatrix::Eigenvektoren(IVektor& cnt)", DIMFEHLER);

  Uebergabe_Matrix.Balance (Skal, low, high);
  Uebergabe_Matrix.Elmhes (low, high, cnt);
  Uebergabe_Matrix.Elmtrans (low, high, cnt, Result);
  Uebergabe_Matrix.Hqr2 (low, high, valreal, valim,
			 Result, cnt);
  Result.Balback (low, high, Skal);
  Result.Norm_1 (valim);
};

void 
DMatrix::Norm_1 (DVektor & wi)
{
  int i, j;
  double maxi;

  j = 0;
  while (j < n) {
    if (wi[j] == 0.0) {
      maxi = Lese_i_j (0, j);
      for (i = 1; i < n; i++) {
	if (fabs (Lese_i_j (i, j)) > fabs (maxi))
	  maxi = Lese_i_j (i, j);
      }
      if (maxi != 0.0) {
	maxi = 1.0 / maxi;
	for (i = 0; i < n; i++)
	  Lese_i_j (i, j) *= maxi;
      }
      j++;
    } else {
      Fehlermeldung ("DMatrix::Norm_1(DMatrix wi)", KOMPLEXFEHLER);
    }
  }
}

void 
DMatrix::Balback (int low, int high, DVektor & skal)
{
  int i, j, k;
  double s;

  for (i = low; i <= high; i++) {
    s = skal[i];
    for (j = 0; j < n; j++)
      Lese_i_j (i, j) *= s;
  }

  for (i = low - 1; i >= 0; i--) {
    k = (int) floor (skal[i] + 0.4999);
    if (k != i)
      Tausche_Zeilen (i, k);
  }

  for (i = high + 1; i < n; i++) {
    k = (int) floor (skal[i] + 0.4999);
    if (k != i)
      Tausche_Zeilen (i, k);
  }
}

char 
DMatrix::Hqrvec (int low, int high, DVektor & wr, DVektor & wi,
		 DMatrix & eivec)
{
  int na, en, i, j, k, l, m;
  double norm, p, q, r, s = 0, t, w, x, y, z = 1;

  norm = 0.0;
  k = 0;
  for (i = 0; i < n; i++) {
    for (j = k; j < n; j++)
      norm += fabs (Lese_i_j (i, j));
    k = i;
  }
  if (norm == 0.0)
    return 1;			/* Nullmatrix */

  for (en = n - 1; en >= 0; en--) {
    p = wr[en];
    q = wi[en];
    na = en - 1;
    if (q == 0.0) {
      m = en;
      Setze_i_j (en, en, 1.0);
      for (i = na; i >= 0; i--) {
	w = Lese_i_j (i, i) - p;
	r = Lese_i_j (i, en);
	for (j = m; j <= na; j++)
	  r += Lese_i_j (i, j) * Lese_i_j (j, en);
	if (wi[i] < 0.0) {
	  z = w;
	  s = r;
	} else {
	  m = i;
	  if (wi[i] == 0.0) {
	    if (w != 0.0)
	      Setze_i_j (i, en, -r / w);
	    else
	      Setze_i_j (i, en, -r / (MACH_EPS * norm));
	  } else {
	    x = Lese_i_j (i, i + 1);
	    y = Lese_i_j (i + 1, i);
	    q = wr[i] - p;
	    q = q * q + wi[i] * wi[i];
	    t = (x * s - z * r) / q;
	    Setze_i_j (i, en, t);
	    if (fabs (x) > fabs (z))
	      Setze_i_j (i + 1, en, (-r - w * t) / x);
	    else
	      Setze_i_j (i + 1, en, (-s - y * t) / z);
	  }
	}
      }
    } else {
      /* komplexe Loesung */
      Fehlermeldung ("DMatrix::hqrvec(int low, int high, DVektor& wr, DVektor& wi, DMatrix& eivec))", KOMPLEXFEHLER);
    }
  }

  for (i = 0; i < n; i++) {
    if ((i < low) || (i > high))
      for (k = i + 1; k < n; k++)
	eivec.Setze_i_j (i, k, Lese_i_j (i, k));
  }
  for (j = n - 1; j >= low; j--) {
    if (j < high)
      m = j;
    else
      m = high;
    if (wi[j] < 0.0) {
      l = j - 1;
      for (i = low; i <= high; i++) {
	y = 0.0;
	z = 0.0;
	for (k = low; k <= m; k++) {
	  y += eivec.Lese_i_j (i, k) * Lese_i_j (k, l);
	  z += eivec.Lese_i_j (i, k) * Lese_i_j (k, j);
	}
	eivec.Setze_i_j (i, l, y);
	eivec.Setze_i_j (i, j, z);
      }
    } else {
      if (wi[j] == 0.0) {
	for (i = low; i <= high; i++) {
	  z = 0.0;
	  for (k = low; k <= m; k++)
	    z += eivec[i][k] * Lese_i_j (k, j);
	  eivec.Setze_i_j (i, j, z);
	}
      }
    }
  }
  return 0;
}

char 
DMatrix::Hqr2 (int low, int high, DVektor & wr,
	       DVektor & wi, DMatrix & eivec, IVektor & cnt)
{
  int na, en, iter, i, j, k, l, m;
  int enditer, stop;
  double p = 0, q = 0, r = 0, s, t, w, x, y, z;

  for (i = 0; i < n; i++) {
    if ((i < low) || (i > high)) {
      wr[i] = Lese_i_j (i, i);
      wi[i] = 0.0;
      cnt[i] = 0;
    }
  }
  en = high;
  t = 0.0;

  while (en >= low) {
    iter = 0;
    na = en - 1;
    enditer = FALSE;
    do {
      l = en;
      while (l > low) {
	if (fabs (Lese_i_j (l, l - 1)) <=
	MACH_EPS * (fabs (Lese_i_j (l - 1, l - 1)) + fabs (Lese_i_j (l, l))))
	  goto cont;
	l--;
      }

    cont:x = Lese_i_j (en, en);
      if (l == en) {
	Setze_i_j (en, en, x + t);
	wr[en] = Lese_i_j (en, en);
	wi[en] = 0.0;
	cnt[en] = iter;
	en--;
	enditer = TRUE;
      } else {
	y = Lese_i_j (na, na);
	w = Lese_i_j (en, na) * Lese_i_j (na, en);
	if (l == na) {
	  p = (y - x) * 0.5;
	  q = p * p + w;
	  z = sqrt (fabs (q));
	  x = x + t;
	  Setze_i_j (en, en, x);
	  Setze_i_j (na, na, y + t);
	  cnt[en] = -iter;
	  cnt[na] = iter;
	  if ((q < 0) && (q + 1E-5 > 0))
	    q = 0;
	  if (q >= 0.0) {
	    if (p < 0.0)
	      z = p - z;
	    else
	      z = p + z;
	    wr[na] = x + z;
	    s = x - w / z;
	    wr[en] = s;
	    wi[na] = 0.0;
	    wi[en] = 0.0;
	    x = Lese_i_j (en, na);
	    r = sqrt (x * x + z * z);
	    p = x / r;
	    q = z / r;
	    for (j = na; j < n; j++) {
	      z = Lese_i_j (na, j);
	      Setze_i_j (na, j, q * z + p * Lese_i_j (en, j));
	      Setze_i_j (en, j, q * Lese_i_j (en, j) - p * z);
	    }
	    for (i = 0; i <= en; i++) {
	      z = Lese_i_j (i, na);
	      Setze_i_j (i, na, q * z + p * Lese_i_j (i, en));
	      Setze_i_j (i, en, q * Lese_i_j (i, en) - p * z);
	    }
	    for (i = low; i <= high; i++) {
	      z = eivec[i][na];
	      eivec[i][na] = q * z + p * eivec[i][en];
	      eivec[i][en] = q * eivec[i][en] - p * z;
	    }
	  }
	  /* q>=0.0 */ 
	  else {
	    Fehlermeldung ("DMatrix::Hqr2(int low, int high,DVektor& wr, DVektor& wi, DMatrix& eivec, IVektor& cnt)", KOMPLEXFEHLER);
	    wr[na] = x + p;
	    wr[en] = x + p;
	    wi[na] = z;
	    wi[en] = -z;
	  }
	  en = en - 2;
	  enditer = TRUE;
	} else {		/* l==na */
	  enditer = FALSE;
	  if (iter >= MAXIT) {
	    cnt[en] = 1 + MAXIT;
	    return en;
	  }
	  if ((iter != 0) && ((iter % KORITER) == 0)) {
	    t += x;
	    for (i = low; i <= en; i++)
	      Setze_i_j (i, i, Lese_i_j (i, i) - x);
	    s = fabs (Lese_i_j (en, na)) + fabs (Lese_i_j (na, en - 2));
	    y = 0.75 * s;
	    x = y;
	    w = -0.4375 * s * s;
	  }
	  iter++;
	  m = en - 1;
	  stop = (en - 2 < l);
	  while ((m >= l) && (!stop)) {
	    m--;
	    z = Lese_i_j (m, m);
	    r = x - z;
	    s = y - z;
	    p = (r * s - w) / Lese_i_j (m + 1, m) + Lese_i_j (m, m + 1);
	    q = Lese_i_j (m + 1, m + 1) - z - r - s;
	    r = Lese_i_j (m + 2, m + 1);
	    s = fabs (p) + fabs (q) + fabs (r);
	    p = p / s;
	    q = q / s;
	    r = r / s;
	    stop = (m <= l);
	    if (m > l)
	      stop = (fabs (Lese_i_j (m, m - 1)) * (fabs (q) + fabs (r)) <=
		      MACH_EPS * fabs (p) * (fabs (Lese_i_j (m - 1, m - 1)) + fabs (z) + fabs (Lese_i_j (m + 1, m + 1))));
	  }
	  for (i = m + 2; i <= en; i++)
	    Setze_i_j (i, i - 2, 0.0);
	  for (i = m + 3; i <= en; i++)
	    Setze_i_j (i, i - 3, 0.0);
	  for (k = m; k <= na; k++) {
	    if (k != m) {
	      p = Lese_i_j (k, k - 1);
	      q = Lese_i_j (k + 1, k - 1);
	      if (k != na)
		r = Lese_i_j (k + 2, k - 1);
	      else
		r = 0.0;
	      x = fabs (p) + fabs (q) + fabs (r);
	      if (x == 0.0)
		goto weiter;
	      p = p / x;
	      q = q / x;
	      r = r / x;
	    }
	    s = sqrt (p * p + q * q + r * r);
	    if (p < 0.0)
	      s = -s;
	    if (k != m)
	      Setze_i_j (k, k - 1, -s * x);
	    else if (l != m)
	      Setze_i_j (k, k - 1, -Lese_i_j (k, k - 1));
	    p = p + s;
	    x = p / s;
	    y = q / s;
	    z = r / s;
	    q = q / p;
	    r = r / p;
	    for (j = k; j < n; j++) {
	      p = Lese_i_j (k, j) + q * Lese_i_j (k + 1, j);
	      if (k != na) {
		p = p + r * Lese_i_j (k + 2, j);
		Setze_i_j (k + 2, j, Lese_i_j (k + 2, j) - p * z);
	      }
	      Setze_i_j (k + 1, j, Lese_i_j (k + 1, j) - p * y);
	      Setze_i_j (k, j, Lese_i_j (k, j) - p * x);
	    }
	    if (k + 3 < en)
	      j = k + 3;
	    else
	      j = en;
	    for (i = 0; i <= j; i++) {
	      p = x * Lese_i_j (i, k) + y * Lese_i_j (i, k + 1);
	      if (k != na) {
		p += z * Lese_i_j (i, k + 2);
		Setze_i_j (i, k + 2, Lese_i_j (i, k + 2) - p * r);
	      }
	      Setze_i_j (i, k + 1, Lese_i_j (i, k + 1) - p * q);
	      Setze_i_j (i, k, Lese_i_j (i, k) - p);
	    }
	    for (i = low; i <= high; i++) {
	      p = x * eivec.Lese_i_j (i, k) + y * eivec.Lese_i_j (i, k + 1);
	      if (k != na) {
		p = p + z * eivec.Lese_i_j (i, k + 2);
		eivec.Setze_i_j (i, k + 2, eivec.Lese_i_j (i, k + 2) - p * r);
	      }
	      eivec.Setze_i_j (i, k + 1, eivec.Lese_i_j (i, k + 1) - p * q);
	      eivec.Setze_i_j (i, k, eivec.Lese_i_j (i, k) - p);
	    }
	  weiter:;
	  }			/* k */
	}			/* l!=na */
      }				/* l!=en */
    } while (!enditer);
  }
  if (Hqrvec (low, high, wr, wi, eivec) != 0)
    return 99;
  return 0;
};

void 
DMatrix::Elmtrans (int low, int high, IVektor & perm, DMatrix & Result)
{
  int i, j, k;

  Result.Einheitsmatrix ();
  for (i = high - 1; i >= low + 1; i--) {
    j = perm[i];
    for (k = i + 1; k <= high; k++)
      Result.Setze_i_j (k, i, Lese_i_j (k, i - 1));
    if (i != j) {
      for (k = i; k <= high; k++) {
	Result.Setze_i_j (i, k, Result.Lese_i_j (j, k));
	Result.Setze_i_j (j, k, 0.0);
      }
      Result.Setze_i_j (j, i, 1.0);
    }
  }
}

char 
DMatrix::Elmhes (int low, int high, IVektor & perm)
{
  int i, j, m;
  double x, y;

  for (m = low + 1; m < high; m++) {
    i = m;
    x = 0.0;
    for (j = m; j <= high; j++) {
      if (fabs (Lese_i_j (j, m - 1)) > fabs (x)) {
	x = Lese_i_j (j, m - 1);
	i = j;
      }
    }

    perm[m] = i;
    if (i != m) {
      for (j = m - 1; j < n; j++)
	swap (Lese_i_j (i, j), Lese_i_j (m, j));
      for (j = 0; j <= high; j++)
	swap (Lese_i_j (j, i), Lese_i_j (j, m));
    }
    if (x != 0.0) {
      for (i = m + 1; i <= high; i++) {
	y = Lese_i_j (i, m - 1);
	if (y != 0.0) {
	  Setze_i_j (i, m - 1, y / x);
	  y = Lese_i_j (i, m - 1);
	  for (j = m; j < n; j++)
	    Lese_i_j (i, j) -= y * Lese_i_j (m, j);
	  for (j = 0; j <= high; j++)
	    Lese_i_j (j, m) += y * Lese_i_j (j, i);
	}
      }
    }
  }
};

char 
DMatrix::Balance (DVektor & Skal, int &low, int &high)
{
  int i, j, k, m;
  char contin;
  double b2, r, c, f, g, s;

  if (this->n != this->m) {
    Fehlermeldung ("DMatrix::Balance(DVektor& Skal, int& low, int& high)", DIMFEHLER);
    return (1);
  }
  Skal.Setze_Dim (this->n);
  b2 = BASIS * BASIS;
  m = 0;
  k = n - 1;
  do {
    contin = FALSE;
    for (j = k; j >= 0; j--) {
      r = 0.0;
      for (i = 0; i <= k; i++) {
	if (i != j)
	  r += fabs (Lese_i_j (j, i));
      }
      if (r == 0.0) {
	Skal[k] = j;
	if (j != k) {
	  for (i = 0; i <= k; i++)
	    swap (Lese_i_j (i, j), Lese_i_j (i, k));
	  for (i = m; i < n; i++)
	    swap (Lese_i_j (j, i), Lese_i_j (k, i));
	}
	k--;
	contin = TRUE;
      }
    }
  } while (contin);

  do {
    contin = FALSE;
    for (j = m; j <= k; j++) {
      c = 0.0;
      for (i = m; i <= k; i++) {
	if (i != j)
	  c += fabs (Lese_i_j (i, j));
      }
      if (c == 0.0) {
	Skal[m] = j;
	if (j != m) {
	  for (i = 0; i <= k; i++)
	    swap (Lese_i_j (i, j), Lese_i_j (i, m));
	  for (i = m; i < n; i++)
	    swap (Lese_i_j (j, i), Lese_i_j (m, i));
	}
	m++;
	contin = TRUE;
      }
    }
  } while (contin);

  low = m;
  high = k;
  for (i = m; i <= k; i++)
    Skal[i] = 1.0;
  do {
    contin = FALSE;
    for (i = m; i <= k; i++) {
      c = 0.0;
      r = 0.0;
      for (j = m; j <= k; j++)
	if (j != i) {
	  c += fabs (Lese_i_j (j, i));
	  r += fabs (Lese_i_j (i, j));
	}
      g = r / BASIS;
      f = 1.0;
      s = c + r;
      while (c < g) {
	f *= BASIS;
	c *= b2;
      }
      g = r * BASIS;
      while (c >= g) {
	f /= BASIS;
	c /= b2;
      }
      if (((c + r) / f) < (0.95 * s)) {
	g = 1.0 / f;
	Skal[i] = Skal[i] * f;
	contin = TRUE;
	for (j = m; j < n; j++)
	  Lese_i_j (i, j) *= g;
	for (j = 0; j <= k; j++)
	  Lese_i_j (j, i) *= f;
      }
    }
  } while (contin);

  return (0);
};

void 
DMatrix::Speichern (FILE * File)
{
  int i, j, k = 0;

  fprintf (File, "%d %d\n", m, n);
  for (i = 0; i < m; i++) {
    for (j = 0; j < n; j++)
      fprintf (File, "%lf ", Daten[k++]);
    fprintf (File, "\n");
  }
};

int 
DMatrix::Laden (FILE * File)
{
  int i, j, k = 0;

  if (!(fscanf (File, "%d %d", &i, &j))) {
#if 0
    this. ~ DMatrix ();
#endif
    return (0);
  }
  Setze_Dim (i, j);
  for (i = 0; i < m; i++) {
    for (j = 0; j < n; j++)
      if (!(fscanf (File, "%lf", &Daten[k++]))) {
#if 0
	this. ~ DMatrix ();
#endif
	return (0);
      }
  }
  return (1);
};

void 
DMatrix::Quicksort (int a, int b, int Index)
{
  int r, i, j;
  double x, DBuffer;

  if (b > a) {
    i = a;
    j = b;
    /* Waehle r einmal in der Mitte */
    r = (a + b) / 2;
    x = Lese_i_j (r, Index);
    do {
      while (Lese_i_j (i, Index) < x)
	i++;
      while (Lese_i_j (j, Index) > x)
	j--;
      if (i <= j) {
	/* tauschen */
	for (r = 0; r < n; r++) {
	  DBuffer = Lese_i_j (i, r);
	  Setze_i_j (i, r, Lese_i_j (j, r));
	  Setze_i_j (j, r, DBuffer);
	}
	i++;
	j--;
      }
    } while (j >= i);
    Quicksort (a, j, Index);
    Quicksort (i, b, Index);
  }
};

void 
DMatrix::Tausche (DMatrix & Damit)
{
  double *dBuffer;
  int i;

  dBuffer = Damit.Daten;
  Damit.Daten = Daten;
  Daten = dBuffer;
  i = Damit.m;
  Damit.m = m;
  m = i;
  i = Damit.n;
  Damit.n = n;
  n = i;
};

DMatrix & DMatrix::operator = (const DMatrix & Das) {
  int i;

  if ((Das.n != n) || (Das.m != m)) {
    if ((n != 0) && (m != 0))
      free (Daten);
    m = Das.m;
    n = Das.n;
    if ((Daten = (double *) malloc (sizeof (double) * m * n)) == NULL)
        Fehlermeldung ("DMatrix::operator=(DMatrix& Das)", SPEICHERFEHLER);
  }
  for (i = 0; i < m * n; i++)
    Daten[i] = Das.Daten[i];
};

void 
DMatrix::operator+= (const DMatrix & Das)
{
  int i;

  for (i = 0; i < (m * n); i++)
    Daten[i] = Daten[i] + Das.Daten[i];
};

void 
DMatrix::operator*= (double Das)
{
  int i;

  for (i = 0; i < m * n; i++)
    Daten[i] *= Das;
};

void 
DMatrix::operator/= (double Das)
{
  int i;

  for (i = 0; i < m * n; i++)
    Daten[i] /= Das;
};

char 
DMatrix::operator== (double Das)
{
  int i;

  for (i = 0; i < m * n; i++)
    if (Daten[i] != Das)
      return (FALSE);
  return (TRUE);
};

char 
DMatrix::operator== (DMatrix & Das)
{
  int i;

  if ((n == Das.n) && (m == Das.m)) {
    for (i = 0; i < m * n; i++)
      if (Daten[i] != Das.Daten[i])
	return (FALSE);
    return (TRUE);
  }
  return (FALSE);
};


DMatrix 
DMatrix::operator+ (DMatrix & Das)
return Result (m, n, NULL);
{
  int i;

  for (i = 0; i < (m * n); i++)
    Result.Daten[i] = Daten[i] + Das.Daten[i];
};

DMatrix 
DMatrix::operator- (DMatrix & Das)
return Result (m, n, NULL);
{
  int i;

  for (i = 0; i < (m * n); i++)
    Result.Daten[i] = Daten[i] - Das.Daten[i];
};

DMatrix 
DMatrix::operator* (DMatrix & Das)
return Result (m, Das.n, NULL);
{
  int Spalte, Zeile, i, dort = 0;
  double Buf;

  for (Zeile = 0; Zeile < m; Zeile++)
    for (Spalte = 0; Spalte < Das.n; Spalte++) {
      Buf = 0;
      for (i = 0; i < n; i++)
	Buf += Lese_i_j (Zeile, i) * Das.Lese_i_j (i, Spalte);
      Result.Daten[dort++] = Buf;
    }
};

DVektor 
DMatrix::operator* (DVektor & Das)
return Result (m, 0, NULL);
{
  int i, j;
  double Buffer;

  for (i = 0; i < m; i++) {
    Buffer = 0;
    for (j = 0; j < n; j++)
      Buffer += Lese_i_j (i, j) * Das.Lese_i (j);
    Result.Setze_i (i, Buffer);
  }
};

DMatrix 
d2_Rot (double alpha)
  return Result (2, 2, NULL);
{
  Result[0][0] = cos (alpha);
  Result[0][1] = -sin (alpha);
  Result[1][1] = sin (alpha);
  Result[1][1] = cos (alpha);
};

DMatrix 
d3_xRot (double alpha)
return Result (3, 3, NULL);
{
  Result[0][0] = 1;
  Result[0][1] = 0;
  Result[0][2] = 0;
  Result[1][0] = 0;
  Result[1][1] = cos (alpha);
  Result[1][2] = -sin (alpha);
  Result[2][0] = 0;
  Result[2][1] = sin (alpha);
  Result[2][2] = cos (alpha);
};

DMatrix 
d3_yRot (double alpha)
return Result (3, 3, NULL);
{
  Result[0][0] = cos (alpha);
  Result[0][1] = 0;
  Result[0][2] = sin (alpha);
  Result[1][0] = 0;
  Result[1][1] = 1;
  Result[1][2] = 0;
  Result[2][0] = -sin (alpha);
  Result[2][1] = 0;
  Result[2][2] = cos (alpha);
};

DMatrix 
d3_zRot (double alpha)
return Result (3, 3, NULL);
{
  Result[0][0] = cos (alpha);
  Result[0][1] = -sin (alpha);
  Result[0][2] = 0;
  Result[1][0] = sin (alpha);
  Result[1][1] = cos (alpha);
  Result[1][2] = 0;
  Result[2][0] = 0;
  Result[2][1] = 0;
  Result[2][2] = 1;
};

/****************************************************************/
/*              Alles zur Klasse Matrix Array                   */
/****************************************************************/
DMatrixArray::DMatrixArray (int ArrayGroesse, int mDim, int nDim,
	      DMatrix * Inhalt)
{
  int i;

  Groesse = ArrayGroesse;
  m = mDim;
  n = nDim;
  if (Groesse != 0) {
    if ((Daten = (DMatrix **) malloc (sizeof (DMatrix *) * Groesse)) == NULL)
      Fehlermeldung ("DMatrixArray:DMatrixArray()", SPEICHERFEHLER);
    if (Inhalt != NULL) {
      for (i = 0; i < Groesse; i++)
	Daten[i] = new DMatrix (Inhalt[i]);
    } else {
      for (i = 0; i < Groesse; i++)
	Daten[i] = new DMatrix (m, n, NULL);
    }
  } else
    Daten = NULL;
};

DMatrixArray::DMatrixArray (const DMatrixArray & Daraus)
{
  int i;

  Groesse = Daraus.Groesse;
  m = Daraus.m;
  n = Daraus.n;
  if (Groesse != 0) {
    if ((Daten = (DMatrix **) malloc (sizeof (DMatrix *) * Groesse)) == NULL)
      Fehlermeldung ("DMatrixArray:DMatrixArray(const DMatrixArray& Daraus)", SPEICHERFEHLER);

    for (i = 0; i < Groesse; i++)
      Daten[i] = new DMatrix (*(Daraus.Daten[i]));
  } else
    Daten = NULL;
};

DMatrixArray::~DMatrixArray ()
{
  int i;

  for (i = 0; i < Groesse; i++)
    delete Daten[i];
  m = n = Groesse = 0;
  free (Daten);
  Daten = NULL;
};

void 
DMatrixArray::print ()
{
  int i;

  for (i = 0; i < Groesse; i++)
    Daten[i]->print ();
};

void 
DMatrixArray::Setze_Groesse (int Neue_Groesse)
{
  int i;
  DMatrix **Buffer;

  if (Neue_Groesse != Groesse) {
    for (i = Neue_Groesse; i < Groesse; i++)
      delete Daten[i];		/* Ueberfl"ussige L"oschen */
    Buffer = Daten;
    if (Neue_Groesse != 0) {
      if ((Daten = (DMatrix **) malloc (sizeof (DMatrix *) * Neue_Groesse)) == NULL)
	Fehlermeldung ("DMatrixArray:Setze_Groesse()", SPEICHERFEHLER);
      for (i = 0; (i < Groesse) && (i < Neue_Groesse); i++) {	/* Alte "ubernehmen */
	Daten[i] = Buffer[i];
#if 0
	Daten[i].NullMatrix ();
#else
	Daten[i]->NullMatrix ();
#endif
      }
      for (; i < Neue_Groesse; i++)
	Daten[i] = new DMatrix (m, n, NULL);
    } else
      Daten = NULL;
    Groesse = Neue_Groesse;
    free (Buffer);
  }
};

void 
DMatrixArray::Setze_Dim (int Neues_m, int Neues_n)
{
  int i;

  if ((m != Neues_m) || (n != Neues_n)) {
    m = Neues_m;
    n = Neues_n;
    for (i = 0; i < Groesse; i++)
      Daten[i]->Setze_Dim (Neues_m, Neues_n);
  }
};

void 
DMatrixArray::Setze_Groesse_Dim (int Neue_Groesse,
				 int Neues_m, int Neues_n)
{
  int i;
  DMatrix **Buffer;

  if (Neue_Groesse != Groesse) {
    for (i = Neue_Groesse; i < Groesse; i++)
      delete Daten[i];		/* Ueberfl"ussige L"oschen */
    Buffer = Daten;
    if (Neue_Groesse != 0) {
      if ((Daten = (DMatrix **) malloc (sizeof (DMatrix *) * Neue_Groesse)) == NULL)
	Fehlermeldung ("DMatrixArray:Setze_Groesse_Dim()", SPEICHERFEHLER);
      for (i = 0; (i < Groesse) && (i < Neue_Groesse); i++) {	/* Alte "ubernehmen */
	Daten[i] = Buffer[i];
	Daten[i]->Setze_Dim (Neues_m, Neues_n);
      }
      for (; i < Neue_Groesse; i++)
	Daten[i] = new DMatrix (Neues_m, Neues_n, NULL);
    } else
      Daten = NULL;
    Groesse = Neue_Groesse;
    m = Neues_m;
    n = Neues_n;
    free (Buffer);
  }
};

void 
DMatrixArray::NullMatrix ()
{
  int i;

  for (i = 0; i < Groesse; i++)
    Daten[i]->NullMatrix ();
};

DMatrixArray 
DMatrixArray::Projiziere (BVektor & Diese)
return Result (Groesse, Diese.Anzahl_gesetzt (), Diese.Anzahl_gesetzt (), NULL);
{
  int i;
  for (i = 0; i < Groesse; i++)
    Result[i] = Daten[i]->Projiziere (Diese);
};

DMatrixArray 
DMatrixArray::Projiziere (IVektor & Diese)
return Result (Groesse, Diese.Lese_Dim (), Diese.Lese_Dim (), NULL);
{
  int i;
  for (i = 0; i < Groesse; i++)
    Result[i] = Daten[i]->Projiziere (Diese);
};

DMatrixArray 
DMatrixArray::Deprojiziere (BVektor & Diese)
return Result (Groesse, Diese.Anzahl_gesetzt (), Diese.Anzahl_gesetzt (), NULL);
{
  int i;
  for (i = 0; i < Groesse; i++)
    Result[i] = Daten[i]->Deprojiziere (Diese);
};

void 
DMatrixArray::Setze_Matrix (int Nummer, DMatrix & Damit)
{
  if ((Nummer >= Groesse) || (Damit.Lese_Dim_m () != m) || (Damit.Lese_Dim_n () != n))
    Fehlermeldung ("DMatrixArray::Setze_Matrix(int Nummer, DMatrix& Damit)", DIMFEHLER);
  else
    (*(Daten[Nummer])) = Damit;
};

void 
DMatrixArray::Speichern (FILE * File)
{
  int i;

  fprintf (File, "%d %d %d\n", Groesse, m, n);
  for (i = 0; i < Groesse; i++)
    Daten[i]->Speichern (File);
};

int 
DMatrixArray::Laden (FILE * File)
{
  int i, j, k;

  if (!(fscanf (File, "%d %d %d", &i, &j, &k))) {
#if 0
    this. ~ DMatrixArray ();
#endif
    return (0);
  }
  Setze_Groesse_Dim (i, j, k);
  for (i = 0; i < Groesse; i++) {
    if (!(Daten[i]->Laden (File))) {
#if 0
      this. ~ DMatrixArray ();
#endif
      return (0);
    }
  }
  return (1);
};

void 
DMatrixArray::Tausche (DMatrixArray & Damit)
{
  DMatrix **dBuffer;
  int i;

  dBuffer = Damit.Daten;
  Damit.Daten = Daten;
  Daten = dBuffer;
  i = Damit.m;
  Damit.m = m;
  m = i;
  i = Damit.n;
  Damit.n = n;
  n = i;
  i = Damit.Groesse;
  Damit.Groesse = Groesse;
  Groesse = i;
};

void 
DMatrixArray::operator= (const DMatrixArray & Das)
{
  int i;

  if (Das.Groesse != Groesse) {
    for (i = 0; i < Groesse; i++)
      delete Daten[i];
    free (Daten);
    if ((Daten = (DMatrix **) malloc (sizeof (DMatrix *) * Das.Groesse)) == NULL)
      Fehlermeldung ("DMatrixArray::operator=()", SPEICHERFEHLER);
    for (i = 0; i < Das.Groesse; i++)
      Daten[i] = new DMatrix (*(Das.Daten[i]));
  } else {
    for (i = 0; i < Das.Groesse; i++)
      (*(Daten[i])) = (*(Das.Daten[i]));
  }
  Groesse = Das.Groesse;
  m = Das.m;
  n = Das.n;
};

void 
DMatrixArray::operator+= (DMatrix & Das)
{
  DMatrix **Buffer;
  int i;

  Buffer = Daten;
  if ((Daten = (DMatrix **) malloc (sizeof (DMatrix *) * Groesse + 1)) == NULL)
    Fehlermeldung ("DMatrixArray::operator+=()", SPEICHERFEHLER);

  for (i = 0; i < Groesse; i++)
    Daten[i] = Buffer[i];
  (*(Daten[Groesse])) = Das;
  Groesse = Groesse + 1;

  free (Buffer);
};

void 
swap (double &a, double &b)
{
  double Buffer;

  Buffer = a;
  a = b;
  b = Buffer;
}
