//**************************************************************
//*   filename: menu.cc                                        *
//*                                                            *
//**************************************************************
//* programmed by: Thomas Wagner                               *
//* last change: 23-05-95                                      *
//**************************************************************
#include <stdio.h>
#include <stdlib.h>

#include "menu.h"
#include "window.h"
#include "include_bmps/mark.bmp"

extern unsigned long buttonpix, blackpix, shadowpix, lightpix;
extern Cursor maincursor;

MenuIcon::MenuIcon (Display * initdisplay,
	  GC initgc,
	  BigWindow * initMenuParentWindow,
	  Window_Info * Parentwindow,
	  XFontStruct * fontstruct,
	  int initx, int inity,
	  int initwidth, int initheight,
	  MenuIconInfo * Infos):
Icon (initdisplay,
      initgc,
      Parentwindow->GetWindow (),
      fontstruct,
      initx, inity, initwidth, initheight,
      Infos->bartext,
      NULL, 0, 0,
      0, 0)
{
  parent = Parentwindow;
  actionnumber = Infos->actionnumber;
  value = Infos->value;
  status = Infos->barstatus;
  menuparentwindow = initMenuParentWindow;
  sonpulldownmenu = NULL;
  if (status & STATUS_CENTER)
    textx = (width - XTextWidth (fontstruct, text, strlen (text))) / 2;
  else
    textx = MENUHORIZONTALSPACE;

  texty = (int) ((height + fontstruct->max_bounds.ascent +
      fontstruct->max_bounds.descent) / 2 - fontstruct->max_bounds.descent);

  DrawMenuIcon ();

  SetSelectedInput (ExposureMask
		    | ButtonPressMask
		    | ButtonReleaseMask
		    | LeaveWindowMask
		    | EnterWindowMask
		    | PointerMotionMask
		    | StructureNotifyMask
		    | OwnerGrabButtonMask);
  XMapWindow (display, window);
};

MenuIcon::~MenuIcon ()
{
  UnmapMe ();
}

void 
MenuIcon::AddMenuStatus (int addstatus)
{
  SetStatus (status | addstatus);
  DrawMenuIcon ();
  DrawIcon (~ICON_SELECTED);
}

void 
MenuIcon::DelMenuStatus (int delstatus)
{
  SetStatus (status & ~delstatus);
  DrawMenuIcon ();
  DrawIcon (~ICON_SELECTED);
}

void 
MenuIcon::DrawMenuIcon ()
{
  XSetForeground (display, gc, buttonpix);
  XFillRectangle (display, pixmap_up, gc, 0, 0, width, height);

  XSetForeground (display, gc, shadowpix);
  XFillRectangle (display, pixmap_down, gc, 0, 0, width, height);

  DrawBorder (pixmap_down, HIGH);
  XSetForeground (display, gc, buttonpix);
  XFillRectangle (display, pixmap_down, gc, 1, 1, width - 2, height - 2);
  XSetForeground (display, gc, shadowpix);

  if (!(status & STATUS_UNSELECTABLE))
    XSetForeground (display, gc, blackpix);
  XDrawString (display, pixmap_up, gc, textx, texty, text, strlen (text));

  XSetForeground (display, gc, lightpix);
  XDrawString (display, pixmap_down, gc, textx, texty, text, strlen (text));

  if ((status & STATUS_MARKED) &&
      (GetPulldownMenu () != NULL)) {
    if (GetPulldownMenu ()->GetParentmenu ()->Getmarkpixmap () != 0) {
      XSetForeground (display, gc, blackpix);
      XSetBackground (display, gc, buttonpix);
      XCopyPlane (display, GetPulldownMenu ()->GetParentmenu ()->Getmarkpixmap (),
		  pixmap_up, gc,
		  0, 0, MARKSIZE, MARKSIZE,
	     (int) (width - MARKSIZE), (int) ((height - MARKSIZE) * .5), 1);
      XSetForeground (display, gc, lightpix);
      XSetBackground (display, gc, buttonpix);
      XCopyPlane (display, GetPulldownMenu ()->GetParentmenu ()->Getmarkpixmap (),
		  pixmap_down, gc,
		  0, 0, MARKSIZE, MARKSIZE,
	     (int) (width - MARKSIZE), (int) ((height - MARKSIZE) * .5), 1);
    }
  }
  if (!(status & STATUS_TOP) &&
      (sonpulldownmenu != NULL)) {
    XSegment lines[4];
    lines[0].x1 = width - MARKSIZE;
    lines[0].y1 = height - 1 - MARKSIZE;
    lines[0].x2 = width - 2;
    lines[0].y2 = height - 1 - MARKSIZE / 2;
    lines[1].x1 = width - MARKSIZE;
    lines[1].y1 = height - 1 - MARKSIZE;
    lines[1].x2 = width - MARKSIZE;
    lines[1].y2 = height - 2;
    lines[2].x1 = width - MARKSIZE - 1;
    lines[2].y1 = height - 1 - MARKSIZE;
    lines[2].x2 = width - MARKSIZE - 1;
    lines[2].y2 = height - 2;
    lines[3].x1 = width - MARKSIZE;
    lines[3].y1 = height - 2;
    lines[3].x2 = width - 2;
    lines[3].y2 = height - 1 - MARKSIZE / 2;

    XSetForeground (display, gc, lightpix);
    XDrawSegments (display, pixmap_up, gc, lines, 3);
    XSetForeground (display, gc, shadowpix);
    XDrawSegments (display, pixmap_up, gc, &(lines[3]), 1);

    XSetForeground (display, gc, blackpix);
    XSetForeground (display, gc, shadowpix);
    XDrawSegments (display, pixmap_down, gc, lines, 3);
    XSetForeground (display, gc, buttonpix);
    XSetForeground (display, gc, lightpix);
    XDrawSegments (display, pixmap_down, gc, &(lines[3]), 1);
  }
}

void 
MenuIcon::SetSonPulldownMenu (PulldownMenu * setpulldownmenu)
{
  sonpulldownmenu = setpulldownmenu;
  if (!(status & STATUS_TOP)) {
    setpulldownmenu->PlaceMe (GetPulldownMenu ()->Getx () + width - MARKSIZE,
			      GetPulldownMenu ()->Gety () + y);
    DrawMenuIcon ();
  }
}

PulldownMenu *
MenuIcon::GetPulldownMenu ()
{
  if (status & STATUS_TOP)
    return NULL;
  else
    return (PulldownMenu *) parent;
}

void 
MenuIcon::Changetomarked ()
{
  unsigned short counter;
  for (counter = 0; counter < GetPulldownMenu ()->Geticonnumber (); counter++) {
    if (GetPulldownMenu ()->GetMenuIcon (counter) == this) {
      if (!(status & STATUS_MARKED)) {
	status |= STATUS_MARKED;
	DrawMenuIcon ();
      }
    } else {
      if (GetPulldownMenu ()->GetMenuIcon (counter)->status & STATUS_MARKED) {
	GetPulldownMenu ()->GetMenuIcon (counter)->status &= ~STATUS_MARKED;
	GetPulldownMenu ()->GetMenuIcon (counter)->DrawMenuIcon ();
      }
    }
  }
}

void 
MenuIcon::HandleEvent (XEvent * Event)
{
  switch (Event->type) {
  case Expose:
    if (Event->xexpose.count == 0)
      if (status & STATUS_SELECTED)
	DrawIcon (ICON_SELECTED);
      else
	DrawIcon (~ICON_SELECTED);
    break;

  case UnmapNotify:
    MapMe ();
    menuparentwindow->Action (actionnumber, value);
    break;

  case ButtonPress:
    if (Event->xbutton.button == Button1) {
      XRaiseWindow (display, menuparentwindow->GetWindow ());
      DrawIcon (ICON_SELECTED);
      ((Menu *) parent)->SelectMe ();
      if ((status & STATUS_TOP) && !(status & STATUS_UNSELECTABLE)) {
	sonpulldownmenu->MapMe ();
      }
    }
    break;

  case ButtonRelease:
    if (Event->xbutton.button == Button1) {
      if (Event->xbutton.x <= width &&
	  Event->xbutton.y <= height &&
	  Event->xbutton.y >= -DEFAULT_BDWIDTH &&
	  Event->xbutton.x >= -DEFAULT_BDWIDTH) {
	if (status & STATUS_TOP)
	  sonpulldownmenu->UnmapMe ();
	else
	  GetPulldownMenu ()->UnmapMe ();
	DrawIcon (~ICON_SELECTED);
	if (!(status & STATUS_TOP) &&
	    !(status & STATUS_UNSELECTABLE) &&
	    !(sonpulldownmenu != NULL)) {
	  if (status & STATUS_MARKABLE)
	    Changetomarked ();
//             menuparentwindow->Action(actionnumber, value);
	  UnmapMe ();
	}
      }
    }
    break;

  case EnterNotify:
    if (Event->xcrossing.state & Button1MotionMask) {
      if (status & STATUS_TOP) {
	if (((Menu *) parent)->IsSelected ()) {
	  if (!(status & STATUS_UNSELECTABLE))
	    ((Menu *) parent)->MapmyPulldownMenu (this);
	  else
	    ((Menu *) parent)->MapmyPulldownMenu (NULL);
	  DrawIcon (ICON_SELECTED);
	}
      } else {
	status |= STATUS_SELECTED;
	DrawIcon (ICON_SELECTED);
	if (Event->xcrossing.x > (width - MARKSIZE))
	  GetPulldownMenu ()->EnteredMe (this);
	else
	  GetPulldownMenu ()->EnteredMe (NULL);
      }
    }
    break;

  case MotionNotify:
    if ((Event->xmotion.x > (width - MARKSIZE)) &&
	!(status & STATUS_TOP))
      GetPulldownMenu ()->EnteredMe (this);
    break;

  case LeaveNotify:
    if (!(Event->xcrossing.state & Button1Mask) &&
	(status & STATUS_TOP)) {
      ((Menu *) parent)->UnmapallPulldownMenus ();
      ((Menu *) parent)->UnselectMe ();
    }
    status &= ~STATUS_SELECTED;
    DrawIcon (~ICON_SELECTED);
    break;
  }
}

Menu::Menu (Display * initdisplay,
      GC initgc,
      BigWindow * ParentWindow,
      XFontStruct * fontstruct,
      int initx, int inity):
Window_Info (initdisplay,
	     initgc,
	     ParentWindow->GetWindow (),
	     fontstruct,
	     initx, inity, 100, 20,
	     "", MENUEBORDERWIDTH)
{
  pixmap = 0;
  markpixmap = 0;
  parentwindow = ParentWindow;
  pulldownnumber = 0;
  status = MENUSTATUS_DEFAULT;
  height = fontstruct->max_bounds.descent +
    fontstruct->max_bounds.ascent +
    MENUVERTDISTANCE;
  width = 2 * MENUBORDER;

  if ((markpixmap = XCreateBitmapFromData (display, window, mark_bits,
					mark_width, mark_height)) == None) {
    printf ("Failed to create markbmp-bitmap in menu.cc!\n");
    exit (0);
  }
  SetSelectedInput (ExposureMask
		    | EnterWindowMask
		    | LeaveWindowMask
		    | ButtonPressMask
		    | ButtonReleaseMask
		    | OwnerGrabButtonMask);

}

Menu::~Menu ()
{
  UnmapMe ();
  unsigned short counter;
  for (counter = 0; counter < pulldownnumber; counter++) {
    delete menuicons[counter]->GetSonPulldownMenu ();
    delete menuicons[counter];
  }
  free (menuicons);
  XFreePixmap (display, pixmap);
  XFreePixmap (display, markpixmap);
}

void 
Menu::MakeUnselectable ()
{
  unsigned short counter;
  for (counter = 0; counter < pulldownnumber; counter++)
    menuicons[counter]->AddMenuStatus (STATUS_UNSELECTABLE);
}

void 
Menu::MakeSelectable ()
{
  unsigned short counter;
  for (counter = 0; counter < pulldownnumber; counter++)
    menuicons[counter]->DelMenuStatus (STATUS_UNSELECTABLE);
}

void 
Menu::AddPulldownMenu (MenuIconInfo * Infos,
		       unsigned short number)
{
  unsigned short counter, iconheight, iconwidth, depth, topiconwidth;
  pulldownnumber++;
  iconwidth = MINMENUWIDTH;
  for (counter = 0; counter < number; counter++)
    iconwidth = MAX (iconwidth,
		     XTextWidth (fontstruct, Infos[counter].bartext, strlen (Infos[counter].bartext)));

  iconwidth += MENUHORIZONTALSPACE;
  iconheight = height - MENUVERTDISTANCE;

  topiconwidth = XTextWidth (fontstruct, Infos[0].bartext, strlen (Infos[0].bartext)) +
    2 * MENUHORIZONTALSPACE;
  width += topiconwidth;

  XResizeWindow (display, window, width, height);

  depth = DefaultDepth (display, DefaultScreen (display));
  if (pixmap)
    XFreePixmap (display, pixmap);
  pixmap = XCreatePixmap (display, window, width, height, depth);
  DrawMenu ();

  MenuIcon **tempicons = (MenuIcon **) malloc (sizeof (MenuIcon *) * (pulldownnumber));
  for (counter = 0; counter < pulldownnumber - 1; counter++)
    tempicons[counter] = menuicons[counter];
  Infos[0].barstatus |= STATUS_CENTER | STATUS_TOP;
  tempicons[pulldownnumber - 1] =
    new MenuIcon (display, gc, parentwindow, this, fontstruct,
		  width - topiconwidth - MENUBORDER, MENUBORDER,
		  topiconwidth, iconheight,
		  Infos);

  menuicons = tempicons;
  menuicons[pulldownnumber - 1]->SetSonPulldownMenu (
	      new PulldownMenu (display, gc, parentwindow, this, fontstruct,
			 x + width - iconwidth - MENUBORDER, y + height + 2,
				iconwidth,
				Infos + 1, number - 1));

// place the pulldown-menu
  if (pulldownnumber == 1)
    menuicons[0]->GetSonPulldownMenu ()->PlaceMe (x, y + height + 2 * MENUBORDER);
  if (pulldownnumber > 1) {
    menuicons[pulldownnumber - 1]->GetSonPulldownMenu ()->
      PlaceMe ((x + width - topiconwidth) +
	       (topiconwidth - menuicons[pulldownnumber - 1]->GetSonPulldownMenu ()->Getwidth ()),
	       y + height + 2 * MENUBORDER);
    if (pulldownnumber > 2)
      menuicons[pulldownnumber - 2]->GetSonPulldownMenu ()->
	MoveMe ((menuicons[pulldownnumber - 2]->GetSonPulldownMenu ()->Getwidth () -
		 menuicons[pulldownnumber - 2]->Getwidth ()) / 2,
		0);
  }
  XMapWindow (display, window);
}

void 
Menu::PlaceMe (unsigned short newx, unsigned short newy)
{
  short xdiff = newx - x, ydiff = newy - y;
  unsigned short counter;
  x = newx;
  y = newy;
  XMoveWindow (display, window, x, y);
  for (counter = 0; counter < pulldownnumber; counter++)
    menuicons[counter]->GetSonPulldownMenu ()->MoveMe (xdiff, ydiff);
}

void 
Menu::DrawMenu ()
{
  XSegment linesrightlow[2] =
  {
    {width - 1, 1, width - 1, height - 1},
    {1, height - 1, width - 1, height - 1}
  }, lineslefthigh[2] =
  {
    {0, 0, width - 2, 0},
    {0, 0, 0, height - 2}
  };

  XSetForeground (display, gc, buttonpix);
  XFillRectangle (display, pixmap, gc, 0, 0, width, height);

  XSetForeground (display, gc, shadowpix);
  XDrawSegments (display, pixmap, gc, linesrightlow, 2);

  XSetForeground (display, gc, lightpix);
  XDrawSegments (display, pixmap, gc, lineslefthigh, 2);
}

PulldownMenu *
Menu::GetPulldownMenu (short index)
{
  return (menuicons[index]->GetSonPulldownMenu ());
}

void 
Menu::AddMenuStatus (int addstatus, int menuidentifier)
{
  int counter;
  for (counter = 0; (counter < pulldownnumber) && (menuicons[counter]->GetActionnumber () != menuidentifier); counter++);
  if (menuicons[counter]->GetActionnumber () == menuidentifier)
    menuicons[counter]->AddMenuStatus (addstatus);
}

void 
Menu::DelMenuStatus (int delstatus, int menuidentifier)
{
  int counter;
  for (counter = 0; (counter < pulldownnumber) && (menuicons[counter]->GetActionnumber () != menuidentifier); counter++);
  if (menuicons[counter]->GetActionnumber () == menuidentifier)
    menuicons[counter]->DelMenuStatus (delstatus);
}

void 
Menu::MapmyPulldownMenu (MenuIcon * my)
{
  unsigned short counter;
  for (counter = 0; counter < pulldownnumber; counter++) {
    if (menuicons[counter] == my)
      my->GetSonPulldownMenu ()->MapMe ();
    else
      menuicons[counter]->GetSonPulldownMenu ()->UnmapMe ();
  }
}

void 
Menu::UnmapallPulldownMenus ()
{
  unsigned short counter;
  for (counter = 0; counter < pulldownnumber; counter++)
    menuicons[counter]->GetSonPulldownMenu ()->UnmapMe ();
}

void 
Menu::HandleEvent (XEvent * Event)
{
  switch (Event->type) {
  case Expose:
    if (Event->xexpose.count == 0)
      XCopyArea (display, pixmap, window, gc,
		 0, 0, width, height, 0, 0);
    break;
  }
}


PulldownMenu::PulldownMenu (Display * initdisplay,
	      GC initgc,
	      BigWindow * ParentWindow,
	      Menu * ParentMenu,
	      XFontStruct * fontstruct,
	      int initx, int inity, int initwidth,
	      MenuIconInfo * Infos,
	      unsigned short number):
Window_Info (initdisplay,
	     initgc,
	     ParentWindow->GetWindow (),
	     fontstruct,
	     initx, inity, initwidth, 20,
	     "", 1)
{
  unsigned short counter, iconheight;
  iconnumber = number;
  parentmenu = ParentMenu;
  parentwindow = ParentWindow;

  iconheight = fontstruct->max_bounds.descent +
    fontstruct->max_bounds.ascent +
    MENUVERTDISTANCE;
  height = number * iconheight;

  width = initwidth;
  for (counter = 0; counter < number; counter++)
    width = MAX (width,
		 XTextWidth (fontstruct, Infos[counter].bartext, strlen (Infos[counter].bartext)));

  width += MENUHORIZONTALSPACE * 2;
  width += 2 * PULLDOWNBORDER;
  height += 2 * PULLDOWNBORDER;

  SizeMe (width, height);
  short depth = DefaultDepth (display, DefaultScreen (display));
  pixmap = XCreatePixmap (display, window, width, height, depth);
  DrawBorder (pixmap, HIGH);

  menuicons = (MenuIcon **) malloc (sizeof (MenuIcon *) * (number));
  for (counter = 0; counter < number; counter++) {
    menuicons[counter] = new MenuIcon (display,
				       gc,
				       parentwindow,
				       this,
				       fontstruct,
		      PULLDOWNBORDER, PULLDOWNBORDER + iconheight * counter,
				     width - 2 * PULLDOWNBORDER, iconheight,
				       &Infos[counter]);
  }

  SetSelectedInput (ExposureMask);
}

void 
PulldownMenu::MapMe ()
{
/*unsigned short counter;
   for(counter=0; counter<iconnumber ;counter++)
   {
   if(menuicons[counter]->GetSonPulldownMenu()!=NULL)
   menuicons[counter]->GetSonPulldownMenu()->UnmapMe();
   } */
  XMapWindow (display, window);
  XRaiseWindow (display, window);
}

void 
PulldownMenu::MoveMe (short xdiff, short ydiff)
{
  x += xdiff;
  y += ydiff;
  XMoveWindow (display, window, x, y);

  unsigned short counter;
  for (counter = 0; counter < iconnumber; counter++) {
    if (menuicons[counter]->GetSonPulldownMenu () != NULL)
      menuicons[counter]->GetSonPulldownMenu ()->MoveMe (xdiff, ydiff);
  }
}

void 
PulldownMenu::UnmapMe ()
{
  XUnmapWindow (display, window);
  unsigned short counter;
  for (counter = 0; counter < iconnumber; counter++) {
    if (menuicons[counter]->GetSonPulldownMenu () != NULL)
      menuicons[counter]->GetSonPulldownMenu ()->UnmapMe ();
  }
}

PulldownMenu::~PulldownMenu ()
{
  unsigned short counter;
  for (counter = 0; counter < iconnumber; counter++)
    delete menuicons[counter];
  free (menuicons);
  XFreePixmap (display, pixmap);
}

void 
PulldownMenu::EnteredMe (MenuIcon * me)
{
  unsigned short counter;
  for (counter = 0; counter < iconnumber; counter++) {
    if (menuicons[counter]->GetSonPulldownMenu () != NULL) {
      if (menuicons[counter] != me)
	menuicons[counter]->GetSonPulldownMenu ()->UnmapMe ();
      else if (!(menuicons[counter]->GetStatus () & STATUS_UNSELECTABLE))
	menuicons[counter]->GetSonPulldownMenu ()->MapMe ();
    }
  }
}

void 
PulldownMenu::HandleEvent (XEvent * Event)
{
  switch (Event->type) {
  case Expose:
    XCopyArea (display, pixmap, window, gc,
	       Event->xgraphicsexpose.x, Event->xgraphicsexpose.y,
	       Event->xgraphicsexpose.width, Event->xgraphicsexpose.height,
	       Event->xgraphicsexpose.x, Event->xgraphicsexpose.y);
    break;
  }
}
