//**************************************************************
//*   filename: moveicon.cc                                    *
//*                                                            *
//**************************************************************
//* programmed by: Thomas Wagner                               *
//* last change: 25-05-95                                      *
//**************************************************************
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "moveicon.h"

extern unsigned long blackpix, buttonpix, shadowpix, lightpix, yellowbuttonpix,
  yellowshadowpix, yellowlightpix, bluebuttonpix, blueshadowpix, bluelightpix,
  redbuttonpix, redshadowpix, redlightpix, greenbuttonpix, greenshadowpix,
  greenlightpix;

extern Cursor movecursor, delconnectcursor, createendcursor, maincursor,
  waitcursor;

MoveableIconlist::MoveableIconlist ()
{
  moveableicon = NULL;
  next = NULL;
}

MoveableIconlist::~MoveableIconlist ()
{
}

void 
MoveableIconlist::operator+= (MoveableIcon * NewIcon)
{
  if (moveableicon == NULL) {
    moveableicon = NewIcon;
    first = this;
  } else {
    MoveableIconlist *temp = this;
    while (temp->next != NULL)
      temp = temp->next;
    temp->next = new MoveableIconlist ();
    temp = temp->next;
    temp->first = this;
    temp->moveableicon = NewIcon;
  }
}

void 
MoveableIconlist::operator-= (MoveableIcon * DeleteIcon)
{
  if (moveableicon == NULL)
    printf ("OhOh - icon not in this list\n");
  else {
    MoveableIconlist *temp = this, *last;
    if (temp->moveableicon == DeleteIcon)	// at first position
     {
      if (temp->next != NULL) {
	temp->moveableicon = temp->next->moveableicon;
	temp->next = temp->next->next;
      } else
	temp->moveableicon = NULL;
    } else {
      last = temp;
      temp = temp->next;
      while ((temp->next != NULL) &&
	     (temp->moveableicon != DeleteIcon)) {
	last = temp;
	temp = temp->next;
      }

      if (temp->moveableicon == NULL)
	printf ("OhOh - icon not in this list\n");
      else {
	last->next = temp->next;
      }
    }				// if(temp->moveableicon==DeleteIcon)

  }
}

MoveableIcon::MoveableIcon (Display * initdisplay,
	      GC initgc,
	      XFontStruct * fontstruct,
	      MainWindow * initMainwindow,
	      int initx, int inity,
	      int initwidth, int initheight,
	      char *inittext):
Icon (initdisplay,
      initgc,
      initMainwindow->GetWindow (),
      fontstruct,
      initx, inity, initwidth, initheight,
      inittext,
      NULL, 0, 0, DEFAULT_BDWIDTH, 0)
{
  short length = strlen (text), counter, textstart = 0;

  for (counter = length - 1; counter >= 0; counter--)
    if (text[counter] == '/')	// found last "/"
     {
      textstart = counter + 1;
      break;
    }
  text += textstart;

  Mainwindow = initMainwindow;
  In = NULL;
  Outputs = new Connectlist ();
  if (width < XTextWidth (fontstruct, text, strlen (text)))
    width = XTextWidth (fontstruct, text, strlen (text)) + 4;

  SetSelectedInput (ExposureMask |
		    ButtonPressMask |
		    ButtonReleaseMask |
		    LeaveWindowMask |
		    EnterWindowMask |
		    ButtonMotionMask);
};

MoveableIcon::~MoveableIcon ()
{
  Mainwindow->DelMoveableIcon (this);
}

void 
MoveableIcon::HandleEvent (XEvent * Event)
{
  static int oldx, oldy;
  static Time lastpresstime = 0;
  static MoveableIcon *lastpressicon = NULL;
  switch (Event->type) {
  case Expose:
    if (status & ICON_SELECTED)
      XCopyArea (display, pixmap_down, window, gc,
		 Event->xgraphicsexpose.x, Event->xgraphicsexpose.y,
		 Event->xgraphicsexpose.width, Event->xgraphicsexpose.height,
		 Event->xgraphicsexpose.x, Event->xgraphicsexpose.y);
    else
      XCopyArea (display, pixmap_up, window, gc,
		 Event->xgraphicsexpose.x, Event->xgraphicsexpose.y,
		 Event->xgraphicsexpose.width, Event->xgraphicsexpose.height,
		 Event->xgraphicsexpose.x, Event->xgraphicsexpose.y);
    break;

  case ButtonPress:
    if (Event->xbutton.button == Button1) {
      RaiseMe ();
      Mainwindow->RaiseMe ();
      int iconoldx, iconoldy;
      char DeleteIcon = FALSE;
      switch (Mainwindow->GetStatus ()) {
      case MAINNORMALSTATUS:
	XGrabPointer (display,
		      window,
		      TRUE,
		      0,
		      GrabModeAsync,
		      GrabModeAsync,
		      window,
		      movecursor,
		      CurrentTime);

	oldx = Event->xbutton.x_root;
	oldy = Event->xbutton.y_root;

	iconoldx = x;
	iconoldy = y;
	x -= (int) (width / 2) - Event->xmotion.x;
	y -= (int) (height / 2) - Event->xmotion.y;
	XMoveWindow (display, window, x, y);

	if ((x != iconoldx) || (y != iconoldy)) {
	  if (In != NULL) {
	    In->EraseArrow ();
	    In->MoveEnd (x - iconoldx, y - iconoldy);
	  }
	  Connectlist *temp = Outputs;
	  Connection *connection = temp->GetConnection ();
	  while (connection != NULL) {
	    connection->EraseArrow ();
	    connection->MoveStart (x - iconoldx, y - iconoldy);
	    temp = temp->GetNext ();
	    if (temp != NULL)
	      connection = temp->GetConnection ();
	    else
	      connection = NULL;
	  }
	}
	DrawIcon (ICON_SELECTED);
	status |= ICON_SELECTED;
	break;			// // case MAINNORMALSTATUS

      case MAINDELETEINPUTSTATUS:
	if (GetTyp () == MOVEABLEICONTYP_INPUT)
	  DeleteIcon = TRUE;
	else
	  Mainwindow->printError ("Not an Input-Object!");
	break;

      case MAINDELETECLASSIFYSTATUS:
	if (GetTyp () == MOVEABLEICONTYP_CLASSIFY)
	  DeleteIcon = TRUE;
	else
	  Mainwindow->printError ("Not a Classify-Object!");
	break;

      case MAINDELETECLUSTERINGSTATUS:
	if (GetTyp () == MOVEABLEICONTYP_CLUSTER)
	  DeleteIcon = TRUE;
	else
	  Mainwindow->printError ("Not a Cluster-Object!");
	break;

      case MAINDELETERULESTATUS:
	if (GetTyp () == MOVEABLEICONTYP_RULE)
	  DeleteIcon = TRUE;
	else
	  Mainwindow->printError ("Not a Rule-Object!");
	break;

      case MAINSAVEINPUTSTATUS:
	if (GetTyp () == MOVEABLEICONTYP_INPUT) {
	  Mainwindow->Getsaveloadwindow ()
	    ->ChangeFunction (WORK_AS_SAVE_SELECTOR,
			      ((InputIcon *) this)->GetOutput (),
			      OUTPUT_SAVEOBJECT,
			      "SAVE INPUT-OBJECT");
	  Mainwindow->Getsaveloadwindow ()->MapMe ();
	  Mainwindow->SetStatus (MAIN_SAVELOADDISPLAYING_STATUS);
	} else
	  Mainwindow->printError ("Not a Rule-Object!");
	break;

      case MAINSAVERULESTATUS:
	if (GetTyp () == MOVEABLEICONTYP_RULE) {
	  Mainwindow->Getsaveloadwindow ()
	    ->ChangeFunction (WORK_AS_SAVE_SELECTOR,
			      ((RuleIcon *) this)->GetOutput (),
			      OUTPUT_SAVEOBJECT,
			      "SAVE RULE-OBJECT");
	  Mainwindow->Getsaveloadwindow ()->MapMe ();
	  Mainwindow->SetStatus (MAIN_SAVELOADDISPLAYING_STATUS);
	} else
	  Mainwindow->printError ("Not a Rule-Object!");
	break;


      case MAINSAVECLASSIFYSTATUS:
	if (GetTyp () == MOVEABLEICONTYP_CLASSIFY) {
	  Mainwindow->Getsaveloadwindow ()
	    ->ChangeFunction (WORK_AS_SAVE_SELECTOR,
			      ((ClassifyIcon *) this)->GetOutput (),
			      OUTPUT_SAVEOBJECT,
			      "SAVE CLASSIFY-OBJECT");
	  Mainwindow->Getsaveloadwindow ()->MapMe ();
	  Mainwindow->SetStatus (MAIN_SAVELOADDISPLAYING_STATUS);
	} else
	  Mainwindow->printError ("Not a Classify-Object!");
	break;

      default:
	break;
      }

      if (DeleteIcon == TRUE) {
	// first delete connections
	if (In != NULL) {
	  In->SendActiontoStart (ICON_REMOVECONNECTOUT, (int) In);
	  delete In;
	}
	Outputs->SendActiontoallEnds (ICON_REMOVECONNECTIN, 0);
	delete Outputs;
	UnmapMe ();
#if 0
	/* XXX: would you believe this ??? */
	delete this;
#endif
	Mainwindow->SetStatus (MAINNORMALSTATUS);
      }
    }
    break;			// case ButtonPress

  case ButtonRelease:
    switch (Event->xbutton.button) {
    case Button1:
      {
	XUngrabPointer (display, CurrentTime);
	status &= ~ICON_SELECTED;
	DrawIcon (~ICON_SELECTED);

	if (((Event->xbutton.time - lastpresstime) < DOUBLECLICKTIME) &&
	    (this == lastpressicon) &&
	    (Mainwindow->GetStatus () == MAINNORMALSTATUS))
	  Doubleclicked ();
	lastpresstime = Event->xbutton.time;
	lastpressicon = this;
      }

      if (Event->xbutton.button == Button1 &&
	  Event->xbutton.x <= width &&
	  Event->xbutton.y <= height &&
	  Event->xbutton.y >= -DEFAULT_BDWIDTH &&
	  Event->xbutton.x >= -DEFAULT_BDWIDTH) {
	Connection *acuteconnection = Mainwindow->GetAcuteconnection ();
	switch (Mainwindow->GetStatus ()) {
	case MAINSTARTCONNECTSTATUS:
	  Mainwindow->SetAcuteconnection (
					   new Connection (display,
							   gc,
							   Mainwindow,
			    (int) (x + width * .5), (int) (y + height * .5),
			  (int) (x + width * .5), (int) (y + height * .5)));
	  Connectout (Mainwindow->GetAcuteconnection ());
	  break;

	case MAINENDCONNECTSTATUS:
	  Connectin (Mainwindow->GetAcuteconnection ());
	  break;

	case MAINMOVECONNECTSTARTSTATUS:
	  if (this == Mainwindow->GetAcuteconnection ()->GetStartIcon ()) {
	    acuteconnection->ErasetempArrow ();
	    acuteconnection->SetStart (acuteconnection->GetStartIcon ()->x + (int) (acuteconnection->GetStartIcon ()->width / 2),
				       acuteconnection->GetStartIcon ()->y + (int) (acuteconnection->GetStartIcon ()->height / 2));
	    acuteconnection->SetStatus (CONNECTSTATUS_VISIBLE);
	    acuteconnection->DrawArrow ();
	    Mainwindow->SetStatus (MAINNORMALSTATUS);
	    XDefineCursor (display, Mainwindow->GetWindow (), maincursor);
	  } else
	    ChangeConnectout (Mainwindow->GetAcuteconnection ());
	  break;

	case MAINMOVECONNECTENDSTATUS:
	  if (this == Mainwindow->GetAcuteconnection ()->GetEndIcon ()) {
	    acuteconnection->ErasetempArrow ();
	    acuteconnection->SetEnd (acuteconnection->GetEndIcon ()->x + (int) (acuteconnection->GetEndIcon ()->width / 2),
				     acuteconnection->GetEndIcon ()->y + (int) (acuteconnection->GetEndIcon ()->height / 2));
	    acuteconnection->SetStatus (CONNECTSTATUS_VISIBLE);
	    acuteconnection->DrawArrow ();
	    Mainwindow->SetStatus (MAINNORMALSTATUS);
	    XDefineCursor (display, Mainwindow->GetWindow (), maincursor);
	    In = Mainwindow->GetAcuteconnection ();
	  } else
	    Connectin (Mainwindow->GetAcuteconnection ());
	  break;

	default:
	  break;
	}
      } {
	int oldx = x, oldy = y;
	if (x < 0)
	  x = 0;

	if (x > Mainwindow->width - width)
	  x = Mainwindow->width - width - 2 * DEFAULT_BDWIDTH;

	if (y < MAIN_TOP_SPACE)
	  y = MAIN_TOP_SPACE;

	if (y > Mainwindow->height - height)
	  y = Mainwindow->height - height - 2 * DEFAULT_BDWIDTH;

	if ((oldx != x) || (oldy != y)) {
	  XMoveWindow (display, window, x, y);
	  CenterConnections ();
	}
      }
      break;			//case Button1:

    case Button3:
      if (Mainwindow->GetStatus () == MAINNORMALSTATUS)
	Pressedbutton3 ();
      break;
    }
    break;			// case ButtonRelease:

  case EnterNotify:
    if (!(Event->xmotion.state & Button1MotionMask) &&
	(status & ICON_SELECTED)) {
      status &= ~ICON_SELECTED;
      DrawIcon (~ICON_SELECTED);
      XUngrabPointer (display, CurrentTime);
    }
    break;

  case MotionNotify:
    if ((status & ICON_SELECTED) &&
	(Mainwindow->GetStatus () == MAINNORMALSTATUS)) {
      /* Next calculate difference and move */
      int newx, newy;

      newx = Event->xmotion.x_root;
      newy = Event->xmotion.y_root;

      x += newx - oldx;
      y += newy - oldy;

      if (In != 0) {
	In->EraseArrow ();
	In->MoveEnd (newx - oldx, newy - oldy);
      }
      Connectlist *temp = Outputs;
      Connection *acuteconnection = temp->GetConnection ();
      while (acuteconnection != NULL) {
	acuteconnection->EraseArrow ();
	acuteconnection->MoveStart (newx - oldx, newy - oldy);
	temp = temp->GetNext ();
	if (temp != NULL)
	  acuteconnection = temp->GetConnection ();
	else
	  acuteconnection = NULL;
      }

      XMoveWindow (display, window, x, y);
      oldx = newx;
      oldy = newy;
    }
    break;
  }
}

void 
MoveableIcon::CenterConnections ()
{
  Connection *connection;
  if (In != 0) {
    connection = In;
    In->EraseArrow ();
    In->SetEnd (x + width / 2, y + height / 2);
    In->DrawArrow ();
  }
  Connectlist *temp = Outputs;
  connection = temp->GetConnection ();
  while (connection != NULL) {
    connection->EraseArrow ();
    connection->SetStart (x + width / 2, y + height / 2);
    temp = temp->GetNext ();
    if (temp != NULL)
      connection = temp->GetConnection ();
    else
      connection = NULL;
  }
  Outputs->DrawAllArrow ();
}

InputIcon::InputIcon (Display * initdisplay,
	   GC initgc,
	   XFontStruct * fontstruct,
	   MainWindow * initMainwindow,
	   int initx, int inity, int initwidth, int initheight,
	   char *inittext):
MoveableIcon (initdisplay,
	      initgc,
	      fontstruct,
	      initMainwindow,
	      initx, inity, initwidth, initheight,
	      inittext)
{
  InOutputwindow = new InOutput (display,
				 gc,
				 this,
				 fontstruct,
				 70, 10,
				 OUTPUTMINSIZE, OUTPUTMINSIZE,
				 text);

  DrawMoveableIcon ();

};

InputIcon::~InputIcon ()
{
  delete InOutputwindow;
}

void 
InputIcon::Doubleclicked ()
{
  InOutputwindow->MapMe ();
}

void 
InputIcon::Connectin (Connection * acuteconnection)
{
  if (In == NULL) {
    switch (acuteconnection->GetStartIcon ()->GetTyp ()) {
    case MOVEABLEICONTYP_RULE:
      if (InOutputwindow->GetDataset ()->Lese_OutputDim () == 0) {
	Mainwindow->printError ("Input-Objects dataset has no outputs!");
	break;
      }
    case MOVEABLEICONTYP_CLASSIFY:
      acuteconnection->ErasetempArrow ();
      acuteconnection->SetEnd ((int) (x + width * .5), (int) (y + height * .5));
      acuteconnection->DrawArrow ();

      Mainwindow->SetStatus (MAINNORMALSTATUS);
      acuteconnection->SetStatus (CONNECTSTATUS_VISIBLE);
      if (acuteconnection->GetEndIcon () != this) {
	if (acuteconnection->GetEndIcon () != NULL)
	  acuteconnection->GetEndIcon ()->Action (ICON_REMOVECONNECTIN, (int) (Mainwindow->GetAcuteconnection ()));

	acuteconnection->SetEndIcon (this);
	In = acuteconnection;
      }
      ChangeConnectin (acuteconnection);
      XDefineCursor (display, Mainwindow->GetWindow (), maincursor);
      break;

    case MOVEABLEICONTYP_INPUT:
      Mainwindow->printError ("No connection from Input-Object to Input-Object!");
      break;

    case MOVEABLEICONTYP_CLUSTER:
      Mainwindow->printError ("No connection from Cluster-Object to Input-Object!");
      break;
    }
  } else
    Mainwindow->printError ("Only one input to Input-Object!");
}

void 
InputIcon::Connectout (Connection * acuteconnection)
{
  Mainwindow->SetStatus (MAINENDCONNECTSTATUS);
  acuteconnection->SetStartIcon (this);
  *Outputs += acuteconnection;
  XDefineCursor (display, Mainwindow->GetWindow (), createendcursor);
}

void 
InputIcon::ChangeConnectin (Connection * acuteconnection)
{
  InOutputwindow->Action (OUTPUT_ICONLOSTINPUTCONNECTION, 0);
  switch (Mainwindow->GetAcuteconnection ()->GetStartIcon ()->GetTyp ()) {
  case MOVEABLEICONTYP_RULE:
    {
      RuleResultInfo *tempinfo;
      InOutputwindow->Action (OUTPUT_ICONREQUESTFORRULERESULTINFO, (int) &tempinfo);
      In->SendActiontoStart (ICON_REQUESTFORRULERESULTS, (int) tempinfo);
      InOutputwindow->Action (OUTPUT_REDRAW, 0);
      InOutputwindow->UpdateOutput ();
    }
    break;

  case MOVEABLEICONTYP_CLASSIFY:
    In->SendActiontoStart (ICON_REQUESTFORCLASSIFYRESULTS, (int) InOutputwindow->GetDataset ());
    InOutputwindow->Action (OUTPUT_REDRAW, 0);
    InOutputwindow->UpdateOutput ();
    break;

  default:
    break;
  }
}

void 
InputIcon::ChangeConnectout (Connection * acuteconnection)
{

  if (acuteconnection->GetEndIcon ()->GetTyp () != MOVEABLEICONTYP_INPUT) {
    char bothdimsequal = TRUE;
    if (acuteconnection->GetEndIcon ()->GetTyp () == MOVEABLEICONTYP_CLUSTER) {
      Clustering *targetclustering;
      Datensatz *Dataset = InOutputwindow->GetDataset ();
      acuteconnection->SendActiontoEnd (ICON_REQUESTFORCLUSTERING, (int) (&targetclustering));

      if (targetclustering->Lese_Clusteranzahl () > 0)
	if (targetclustering->Lese_Clusterdimension () != Dataset->Lese_Daten ().Lese_Dim ())
	  bothdimsequal = FALSE;
    }
    if (bothdimsequal) {
      acuteconnection->GetStartIcon ()->Action (ICON_REMOVECONNECTOUT, (int) (Mainwindow->GetAcuteconnection ()));
      acuteconnection->ErasetempArrow ();
      acuteconnection->SetStart ((int) (x + width * .5), (int) (y + height * .5));
      acuteconnection->DrawArrow ();
      Mainwindow->SetStatus (MAINNORMALSTATUS);
      acuteconnection->SetStartIcon (this);
      acuteconnection->SetStatus (CONNECTSTATUS_VISIBLE);
      *Outputs += acuteconnection;

      Outputs->SendActiontoallEnds (ICON_SENDINGDATASET,
				    (int) (InOutputwindow->GetDataset ()));
      Outputs->SendActiontoallEnds (ICON_SENDINGDVECTOR,
			   (int) InOutputwindow->GetSelectedinputvector ());

      short counter = 1;
      char *newtext = (char *) malloc (sizeof (char) * 3);	// 3 has to be enough

      Connectlist *temp = Outputs;
      Connection *connection = temp->GetConnection ();
      while (connection != NULL) {
	sprintf (newtext, "%d", counter);
	connection->SendActiontoEnd (ICON_SENDINGINDEX, (int) newtext);
	temp = temp->GetNext ();
	if (temp != NULL)
	  connection = temp->GetConnection ();
	else
	  connection = NULL;
	counter++;
      }

      XDefineCursor (display, Mainwindow->GetWindow (), maincursor);
    } else
      Mainwindow->printError ("Cluster dimension not equal to data dimension!");
  } else
    Mainwindow->printError ("No connection from Input-Object to Input-Object!");
}

void 
InputIcon::DrawMoveableIcon ()
{
  short textx, texty, depth;
  width = XTextWidth (fontstruct, text, strlen (text)) + 4;
  width = MAX (width, MOVEABLEICON_WIDTH);
  height = width;

  SizeMe (width, height);

  textx = (int) ((width - XTextWidth (fontstruct, text, strlen (text))) * .5);
  texty = (int) ((height + fontstruct->max_bounds.ascent +
     fontstruct->max_bounds.descent) * .5 - fontstruct->max_bounds.descent);

  if (pixmap_up != 0)
    XFreePixmap (display, pixmap_up);
  if (pixmap_down != 0)
    XFreePixmap (display, pixmap_down);
  depth = DefaultDepth (display, DefaultScreen (display));
  pixmap_up = XCreatePixmap (display, window, width, height, depth);
  pixmap_down = XCreatePixmap (display, window, width, height, depth);

  XSetForeground (display, gc, yellowbuttonpix);
  XSetBackground (display, gc, yellowbuttonpix);
  XFillRectangle (display, pixmap_up, gc, 0, 0, width, height);
  XFillRectangle (display, pixmap_down, gc, 0, 0, width, height);

  XSegment linesrightlow[4] =
  {
    {width - 1, 1, width - 1, height - 1},
    {width - 2, 2, width - 2, height - 1},
    {1, height - 1, width - 1, height - 1},
    {2, height - 2, width - 1, height - 2},
  }, lineslefthigh[6] =
  {
    {0, 0, width - 2, 0},
    {0, 1, width - 3, 1},
    {0, 0, 0, (height / 4) - 1},
    {1, 0, 1, (height / 4) - 2},
    {0, (3 * height / 4), 0, height - 2},
    {1, (3 * height / 4), 1, height - 3}
  }, shadowarrowlines[7] =
  {
    {1, (height / 4), (width / 2), (height / 4)},
    {2, (height / 4) - 1, (width / 2) + 1, (height / 4) - 1},
    {(width / 2), (height / 4), (width / 2), 2},
    {(width / 2) + 1, (height / 4), (width / 2) + 1, 2},
    {(width / 2), 2, width - 3, (height / 2) - 1},
    {(width / 2), (3 * height / 4), (width / 2), height - 3},
    {(width / 2) + 1, (3 * height / 4), (width / 2) + 1, height - 3}
  }, lightarrowlines[4] =
  {
    {1, (3 * height / 4), (width / 2) + 1, (3 * height / 4)},
    {2, (3 * height / 4) + 1, (width / 2), (3 * height / 4) + 1},
    {(width / 2) + 2, height - 4, width - 3, (height / 2)},
    {(width / 2) + 2, height - 5, width - 4, (height / 2)}
  };

  XSetForeground (display, gc, yellowshadowpix);
  XDrawSegments (display, pixmap_up, gc, linesrightlow, 4);
  XDrawSegments (display, pixmap_down, gc, lineslefthigh, 6);
  XDrawSegments (display, pixmap_up, gc, shadowarrowlines, 7);
  XDrawSegments (display, pixmap_down, gc, lightarrowlines, 4);

  XSetForeground (display, gc, yellowlightpix);
  XDrawSegments (display, pixmap_up, gc, lineslefthigh, 6);
  XDrawSegments (display, pixmap_down, gc, linesrightlow, 4);
  XDrawSegments (display, pixmap_down, gc, shadowarrowlines, 7);
  XDrawSegments (display, pixmap_up, gc, lightarrowlines, 4);

  XSetForeground (display, gc, blackpix);
  XDrawString (display, pixmap_up, gc, textx, texty, text, strlen (text));
  XDrawString (display, pixmap_down, gc, textx, texty, text, strlen (text));
}

void 
InputIcon::Action (int actiontyp, int value)
{
  switch (actiontyp) {
  case ICON_DATASETCLASSIFIED:
    InOutputwindow->Action (OUTPUT_CLASSIFIED, value);
    break;

  case ICON_RULERESULTSCHANGED:
  case ICON_CLASSIFYRESULTSCHANGED:
    InOutputwindow->Action (OUTPUT_REDRAW, 0);
    InOutputwindow->UpdateOutput ();
    break;

  case ICON_REMOVECONNECTOUT:
    *Outputs -= (Connection *) (value);
    break;

  case ICON_REMOVECONNECTIN:
    In = NULL;
    InOutputwindow->Action (OUTPUT_ICONLOSTINPUTCONNECTION, 0);
    break;

  case ICON_DATASETMODIFIED:
    Outputs->SendActiontoallEnds (ICON_DATASETMODIFIED, 0);
    if (In != NULL) {
      switch (In->GetStartIcon ()->GetTyp ()) {
      case MOVEABLEICONTYP_RULE:
	In->SendActiontoStart (ICON_REQUESTFORRULERESULTS, value);
	break;

      case MOVEABLEICONTYP_CLASSIFY:
	In->SendActiontoStart (ICON_REQUESTFORCLASSIFYRESULTS, (int) InOutputwindow->GetDataset ());
//                InOutputwindow->Action(OUTPUT_CLASSIFIED,0);
	break;

      default:
	break;
      }
    }
    break;


  case ICON_REQUESTFORDVECTOR:
    *((DVektor **) value) = InOutputwindow->GetSelectedinputvector ();
    break;

  case ICON_SENDINGDVECTOR:
    Outputs->SendActiontoallEnds (ICON_SENDINGDVECTOR, value);
    break;


  case ICON_REQUESTFORDATASET:
    *((Datensatz **) value) = InOutputwindow->GetDataset ();
    break;

  case ICON_REQUESTFORINDEX:
    {
      short counter = 1;
      char *newtext = (char *) malloc (sizeof (char) * 3);
      Connectlist *temp = Outputs;
      Connection *connection = temp->GetConnection ();
      while ((connection == NULL) || (connection != (Connection *) value)) {
	counter++;
	temp = temp->GetNext ();
	if (temp != NULL)
	  connection = temp->GetConnection ();
	else
	  temp = NULL;
      }
      sprintf (newtext, "%d", counter);
      if (connection != NULL)
	connection->SendActiontoEnd (ICON_SENDINGINDEX, (int) newtext);
    }
    break;
  }
}


ClusterIcon::ClusterIcon (Display * initdisplay,
	     GC initgc,
	     XFontStruct * fontstruct,
	     MainWindow * initMainwindow,
	     int initx, int inity,
	     int initwidth, int initheight,
	     char *text,
	     ClusterInfo * values):
//                         Clustering *theclustering):
MoveableIcon (initdisplay,
	      initgc,
	      fontstruct,
	      initMainwindow,
	      initx, inity, initwidth, initheight,
	      text)
{
  pixmap_up = 0;
  pixmap_down = 0;
  indextext = "";

  ClusterOutputwindow = new ClusterOutput (display,
					   gc,
					   this,
					   fontstruct,
					   70, 10,
					   OUTPUTMINSIZE, OUTPUTMINSIZE,
					   text);
  if (values != NULL)
    ClusterOutputwindow->SetClustertyp (values);
  else {
    ClusterInfo defaultvalues;
    defaultvalues.clusternumber = MAXINT;
    defaultvalues.repeatnumber = MAXINT;
    defaultvalues.clustertyp = FCM;
    defaultvalues.validitytyp = NONE;
    ClusterOutputwindow->SetClustertyp (&defaultvalues);
//  if(theclustering!=NULL)
    //    ClusterOutputwindow->Action(OUTPUT_LOADOK, (int) theclustering);
  }
};

ClusterIcon::~ClusterIcon ()
{
  delete ClusterOutputwindow;
}

void 
ClusterIcon::Doubleclicked ()
{
  ClusterOutputwindow->MapMe ();
}

void 
ClusterIcon::Pressedbutton3 ()
{
  ClusterOutputwindow->Action (OUTPUT_DISPLAYCLUSTEROPTIONS_ACTION, 0);
}

void 
ClusterIcon::Connectin (Connection * acuteconnection)
{
  if (acuteconnection->GetStartIcon ()->GetTyp () != MOVEABLEICONTYP_RULE) {
    if (acuteconnection->GetStartIcon ()->GetTyp () != MOVEABLEICONTYP_CLASSIFY) {
      if (In == NULL) {
	if (acuteconnection->GetStartIcon () != this) {
	  char bothdimsequal = TRUE;
	  Datensatz *tempdataset;
	  Clustering *clustering = ClusterOutputwindow->Getclustering ();
	  acuteconnection->SendActiontoStart (ICON_REQUESTFORDATASET, (int) &tempdataset);
	  if (tempdataset != NULL)
	    if ((clustering->Lese_Clusteranzahl () > 0) &&
		(tempdataset->Lese_Daten ().Lese_Dim ()) != clustering->Lese_Clusterdimension ())
	      bothdimsequal = FALSE;

	  if (bothdimsequal) {
	    acuteconnection->GetStartIcon ()->AddStatus (ICON_CHECKED);
	    status |= ICON_CHECKED;
	    char circlefound = FALSE;
	    Outputs->SendActiontoallEnds (ICON_SEARCHFORCIRCLE, (int) &circlefound);
	    status &= ~ICON_CHECKED;
	    acuteconnection->GetStartIcon ()->DelStatus (ICON_CHECKED);
	    if (!circlefound) {
	      acuteconnection->ErasetempArrow ();
	      acuteconnection->SetEnd ((int) (x + width * .5), (int) (y + height * .5));
	      acuteconnection->DrawArrow ();

	      Mainwindow->SetStatus (MAINNORMALSTATUS);
	      acuteconnection->SetStatus (CONNECTSTATUS_VISIBLE);
	      if (acuteconnection->GetEndIcon () != this) {
		if (acuteconnection->GetEndIcon () != NULL)
		  acuteconnection->GetEndIcon ()->Action (ICON_REMOVECONNECTIN, (int) acuteconnection);

		acuteconnection->SetEndIcon (this);
		In = acuteconnection;
		Datensatz *tempdataset;
		In->SendActiontoStart (ICON_REQUESTFORDATASET, (int) &tempdataset);
		ClusterOutputwindow->SetDataset (tempdataset);
		Outputs->SendActiontoallEnds (ICON_SENDINGDATASET, (int) tempdataset);

		DVektor *tempinput;
		In->SendActiontoStart (ICON_REQUESTFORDVECTOR, (int) &tempinput);
		ClusterOutputwindow->SetSelectedinputvector (tempinput);
		Outputs->SendActiontoallEnds (ICON_SENDINGDVECTOR, (int) tempinput);

		In->SendActiontoStart (ICON_REQUESTFORINDEX, (int) In);
	      }
	      XDefineCursor (display, Mainwindow->GetWindow (), maincursor);
	    } else
	      Mainwindow->printError ("Circle not allowed!");
	  } else
	    Mainwindow->printError ("Cluster dimension not equal to data dimension!");
	} else
	  Mainwindow->printError ("Cluster-Object cannot be connected to itself!");
      } else
	Mainwindow->printError ("Only one input to Cluster-Object!");
    } else
      Mainwindow->printError ("No connection from Classify-Object to Cluster-Object!");
  } else
    Mainwindow->printError ("No connection from Rule-Object to Cluster-Object!");
}

void 
ClusterIcon::Connectout (Connection * acuteconnection)
{
  Mainwindow->SetStatus (MAINENDCONNECTSTATUS);
  acuteconnection->SetStartIcon (this);
  *Outputs += acuteconnection;
  XDefineCursor (display, Mainwindow->GetWindow (), createendcursor);
}

void 
ClusterIcon::ChangeConnectin (Connection * acuteconnection)
{
}

void 
ClusterIcon::ChangeConnectout (Connection * acuteconnection)
{
  switch (acuteconnection->GetEndIcon ()->GetTyp ()) {
  case MOVEABLEICONTYP_CLUSTER:
    if (acuteconnection->GetEndIcon () != this) {
      char bothdimsequal = TRUE;

      Clustering *targetclustering, *clustering = ClusterOutputwindow->Getclustering ();
      Datensatz *Dataset = ClusterOutputwindow->GetDataset ();

      acuteconnection->SendActiontoEnd (ICON_REQUESTFORCLUSTERING, (int) (&targetclustering));
      if (targetclustering->Lese_Clusteranzahl () > 0) {
	if (Dataset != NULL)
	  if (Dataset->Lese_Daten ().Lese_Dim () != targetclustering->Lese_Clusterdimension ()) {
	    Mainwindow->printError ("Cluster dimension not equal to data dimension!");
	    bothdimsequal = FALSE;
	  }
	if ((clustering->Lese_Clusteranzahl () > 0) &&
	    (clustering->Lese_Clusterdimension () != targetclustering->Lese_Clusterdimension ())) {
	  Mainwindow->printError ("Cluster dimensions of both objects have to be equal!");
	  bothdimsequal = FALSE;
	}
      }
      if (bothdimsequal) {
	status |= ICON_CHECKED;
	char circlefound = FALSE;
	acuteconnection->SendActiontoEnd (ICON_SEARCHFORCIRCLE, (int) &circlefound);
	status &= ~ICON_CHECKED;
	if (!circlefound) {
	  acuteconnection->ErasetempArrow ();
	  acuteconnection->SetStart ((int) (x + width * .5), (int) (y + height * .5));
	  acuteconnection->DrawArrow ();
	  acuteconnection->GetStartIcon ()->Action (ICON_REMOVECONNECTOUT, (int) (Mainwindow->GetAcuteconnection ()));
	  acuteconnection->SetStartIcon (this);
	  acuteconnection->SetStatus (CONNECTSTATUS_VISIBLE);

	  *Outputs += acuteconnection;
	  Outputs->SendActiontoallEnds (ICON_SENDINGDATASET,
				(int) (ClusterOutputwindow->GetDataset ()));
	  Outputs->SendActiontoallEnds (ICON_SENDINGDVECTOR,
			     (int) GetOutput ()->GetSelectedinputvector ());

	  short counter = 1;
	  char *newtext = (char *) malloc (sizeof (char) * (strlen (indextext) + 4));	// has to

	  Connectlist *temp = Outputs;
	  Connection *connection = temp->GetConnection ();
	  while (temp != NULL) {
	    sprintf (newtext, "%s-%d", indextext, counter);
	    if (connection != NULL)
	      connection->SendActiontoEnd (ICON_SENDINGINDEX, (int) newtext);
	    temp = temp->GetNext ();
	    if (temp != NULL)
	      connection = temp->GetConnection ();
	    else
	      connection = NULL;
	    counter++;
	  }

	  Mainwindow->SetStatus (MAINNORMALSTATUS);
	  XDefineCursor (display, Mainwindow->GetWindow (), maincursor);
	} else
	  Mainwindow->printError ("Circle not allowed!");
      }
    } else
      Mainwindow->printError ("Cluster-Object cannot be connected to itself!");
    break;

  case MOVEABLEICONTYP_INPUT:
    Mainwindow->printError ("No connection from Cluster-Object to Input-Object!");
    break;

  case MOVEABLEICONTYP_CLASSIFY:
    Mainwindow->printError ("No connection from Cluster-Object to Classify-Object!");
    break;

  case MOVEABLEICONTYP_RULE:
    Mainwindow->printError ("No connection from Cluster-Object to Rule-Object!");
    break;
  }
}

void 
ClusterIcon::SetIconname (char *newtext)
{
  short length = strlen (newtext), counter, textstart = 0;

  for (counter = length - 1; counter >= 0; counter--)
    if (newtext[counter] == '/')	// found last "/"
     {
      textstart = counter + 1;
      break;
    }
//text+=textstart;

  text = newtext + textstart;
  DrawMoveableIcon ();
}

void 
ClusterIcon::DrawMoveableIcon ()
{
  char *clustertext = ClusterOutputwindow->Getclustering ()->Lese_Name ();

  short clustertextwidth = XTextWidth (fontstruct, clustertext, strlen (clustertext));
  short indextextwidth = XTextWidth (fontstruct, indextext, strlen (indextext));
  short textwidth = XTextWidth (fontstruct, text, strlen (text));
  short textheight = fontstruct->max_bounds.ascent +
  fontstruct->max_bounds.descent;
  width = MAX (clustertextwidth, textwidth);
  width = MAX (width, indextextwidth);
  height = 5 * textheight;
  width += 4;

  if (pixmap_up != 0) {
    char *space = " ";
    char *windowtext = (char *) malloc (sizeof (char) *
	   (4 + strlen (text) + strlen (indextext) + strlen (clustertext)));
    strcpy (windowtext, "");
    strcat (windowtext, text);
    strcat (windowtext, space);
    strcat (windowtext, indextext);
    strcat (windowtext, space);
    strcat (windowtext, clustertext);
    ClusterOutputwindow->SetWindowName (windowtext);
    ClusterOutputwindow->Getmemberoutput ()->SetWindowName (windowtext);

    XFreePixmap (display, pixmap_up);
    XFreePixmap (display, pixmap_down);
  }
  short depth;
  width = height = MAX (width, height);
  SizeMe (width, height);

  depth = DefaultDepth (display, DefaultScreen (display));
  pixmap_up = XCreatePixmap (display, window, width, height, depth);
  pixmap_down = XCreatePixmap (display, window, width, height, depth);

  XSetForeground (display, gc, greenbuttonpix);
  XSetBackground (display, gc, greenbuttonpix);
  XFillRectangle (display, pixmap_up, gc, 0, 0, width, height);
  XFillRectangle (display, pixmap_down, gc, 0, 0, width, height);

  XSegment linesrightlow[4] =
  {
    {width - 1, height / 4, width - 1, height / 2},
    {width - 2, height / 4 + 1, width - 2, height / 2},
    {1, height - 1, width / 2, height - 1},
    {2, height - 2, width / 2 - 1, height - 2},
  }, lineslefthigh[6] =
  {
    {0, 0, width / 2, 0},
    {0, 1, width / 2 - 1, 1},
    {0, 0, 0, height / 2 - 1},
    {1, 0, 1, height / 2 - 4},
    {0, 3 * height / 4, 0, height},
    {1, 3 * height / 4, 1, height - 1}
  }, shadowobjectlines[4] =
  {
    {width / 2, 0, width / 2, height / 4},
    {1 + width / 2, 0, 1 + width / 2, 1 + height / 4},
    {0, height / 2, width / 8, height / 4},
    {0, height / 2 - 1, width / 8, height / 4 - 1}
  }, lightobjectlines[5] =
  {
    {1, 3 * height / 4, 3 * width / 8 + 1, 3 * height / 4},
    {2, 1 + 3 * height / 4, 3 * width / 8, 3 * height / 4 + 1},
    {1 + width / 2, height / 4, width - 1, height / 4},
    {width / 2, 1 + height / 4, width - 2, 1 + height / 4},
    {width / 8, height / 4 - 1, 3 * width / 8, 3 * height / 4}
  };


  XSetForeground (display, gc, greenshadowpix);
  XDrawArc (display, pixmap_up, gc, width / 2, height / 2, width, height, 90 * 64, 90 * 64);
  XDrawArc (display, pixmap_up, gc, width / 2, 1 + height / 2, width, height, 90 * 64, 90 * 64);
  XDrawArc (display, pixmap_up, gc, 1 + width / 2, height / 2, width, height, 90 * 64, 90 * 64);

  XSetForeground (display, gc, greenlightpix);
  XDrawArc (display, pixmap_down, gc, width / 2, height / 2, width, height, 90 * 64, 90 * 64);
  XDrawArc (display, pixmap_down, gc, width / 2, 1 + height / 2, width, height, 90 * 64, 90 * 64);
  XDrawArc (display, pixmap_down, gc, 1 + width / 2, height / 2, width, height, 90 * 64, 90 * 64);

  XDrawSegments (display, pixmap_up, gc, lineslefthigh, 6);
  XDrawSegments (display, pixmap_down, gc, linesrightlow, 4);
  XDrawSegments (display, pixmap_down, gc, shadowobjectlines, 4);
  XDrawSegments (display, pixmap_up, gc, lightobjectlines, 5);

  XSetForeground (display, gc, greenshadowpix);
  XDrawSegments (display, pixmap_up, gc, linesrightlow, 4);
  XDrawSegments (display, pixmap_down, gc, lineslefthigh, 6);
  XDrawSegments (display, pixmap_up, gc, shadowobjectlines, 4);
  XDrawSegments (display, pixmap_down, gc, lightobjectlines, 5);

  short textx, texty;
  XSetForeground (display, gc, blackpix);
  textx = (int) ((width - XTextWidth (fontstruct, text, strlen (text))) * .5);
  texty = (int) ((height + fontstruct->max_bounds.ascent +
     fontstruct->max_bounds.descent) * .5 - fontstruct->max_bounds.descent);
  XDrawString (display, pixmap_up, gc, textx, texty - textheight, text, strlen (text));
  XDrawString (display, pixmap_down, gc, textx, texty - textheight, text, strlen (text));

  textx = (int) ((width - XTextWidth (fontstruct, clustertext, strlen (clustertext))) * .5);
  XDrawString (display, pixmap_up, gc, textx, texty + textheight, clustertext, strlen (clustertext));
  XDrawString (display, pixmap_down, gc, textx, texty + textheight, clustertext, strlen (clustertext));

  textx = (int) ((width - XTextWidth (fontstruct, indextext, strlen (indextext))) * .5);
  XDrawString (display, pixmap_up, gc, textx, texty, indextext, strlen (indextext));
  XDrawString (display, pixmap_down, gc, textx, texty, indextext, strlen (indextext));

  DrawIcon (~ICON_SELECTED);
  CenterConnections ();
}

void 
ClusterIcon::Action (int actiontyp, int value)
{
  switch (actiontyp) {
  case ICON_SEARCHFORCIRCLE:
    if (status & ICON_CHECKED) {
      *((char *) value) = TRUE;
      break;
    }
    status |= ICON_CHECKED;
    Outputs->SendActiontoallEnds (ICON_SEARCHFORCIRCLE, value);
    status &= ~ICON_CHECKED;
    break;

  case ICON_SENDINGDVECTOR:
    GetOutput ()->SetSelectedinputvector ((DVektor *) value);
    Outputs->SendActiontoallEnds (ICON_SENDINGDVECTOR, value);
    break;

  case ICON_REQUESTFORDVECTOR:
    *((DVektor **) value) = ClusterOutputwindow->GetSelectedinputvector ();
    break;



  case ICON_REMOVECONNECTIN:
    In = NULL;
    ClusterOutputwindow->SetDataset ((Datensatz *) NULL);
    text = "unplugged";
    indextext = " ";
    DrawMoveableIcon ();
    if (Outputs != NULL) {
      Outputs->SendActiontoallEnds (ICON_SENDINGDVECTOR,
				    (int) NULL);
      Outputs->SendActiontoallEnds (ICON_SENDINGDATASET, (int) NULL);
    }
    break;

  case ICON_REMOVECONNECTOUT:
    *Outputs -= (Connection *) (value);
    break;

  case ICON_DATASETMODIFIED:
//       ClusterOutputwindow->Action(OUTPUT_REDRAW,0);
    //       ClusterOutputwindow->UpdateOutput();
    ClusterOutputwindow->Action (OUTPUT_DATASETMODIFIED, 0);
    Outputs->SendActiontoallEnds (ICON_DATASETMODIFIED, 0);
    break;

  case ICON_SENDINGDATASET:
    if ((Datensatz *) value != ClusterOutputwindow->GetDataset ()) {
      if ((Datensatz *) value == NULL) {
	text = "unplugged";
	indextext = " ";
	DrawMoveableIcon ();
      }
      ClusterOutputwindow->SetDataset ((Datensatz *) value);
      Outputs->SendActiontoallEnds (ICON_SENDINGDATASET, value);
    }
    break;

  case ICON_REQUESTFORDATASET:
    *((Datensatz **) value) = ClusterOutputwindow->GetDataset ();
    break;

  case ICON_REQUESTFORCLUSTERING:
    *((Clustering **) value) = ClusterOutputwindow->Getclustering ();
    break;

  case ICON_SENDINGDATASETNAME:
    text = (char *) value;
    DrawMoveableIcon ();
    break;

  case ICON_REQUESTFORINDEX:
    {
      short counter = 1;
      char *newtext = (char *) malloc (sizeof (char) * (4 + strlen (indextext)));
      char *temptext = (char *) malloc (sizeof (char) * 4);
      Connectlist *temp = Outputs;
      while ((temp != NULL) && (temp->GetConnection () != (Connection *) value)) {
	counter++;
	temp = temp->GetNext ();
      }
      sprintf (temptext, "-%d", counter);
      strcpy (newtext, indextext);
      strcat (newtext, temptext);
      Connection *connection = temp->GetConnection ();
      if ((connection != NULL) && strcmp (indextext, ""))
	connection->SendActiontoEnd (ICON_SENDINGINDEX, (int) newtext);
    }
    break;

  case ICON_SENDINGINDEX:
    indextext = (char *) value;
    DrawMoveableIcon ();
    {
      Connectlist *temp = Outputs;
      while (temp != NULL) {
	Action (ICON_REQUESTFORINDEX, (int) temp->GetConnection ());
	temp = temp->GetNext ();
      }
    }
    break;
  }
}


RuleIcon::RuleIcon (Display * initdisplay,
	  GC initgc,
	  XFontStruct * fontstruct,
	  MainWindow * initMainwindow,
	  int initx, int inity,
	  int initwidth, int initheight,
	  char *inittext,
	  RuleInfo * ruleinfo):
MoveableIcon (initdisplay,
	      initgc,
	      fontstruct,
	      initMainwindow,
	      initx, inity,
	      initwidth, initheight,
	      inittext)
{
  index = 0;
  pixmap_up = 0;
  RuleOutputwindow = new RuleOutput (display,
				     gc,
				     this,
				     fontstruct,
				     70, 10,
				     OUTPUTMINSIZE, OUTPUTMINSIZE,
				     text,
				     ruleinfo);

  DrawMoveableIcon ();
};

RuleIcon::~RuleIcon ()
{
  delete RuleOutputwindow;
}

void 
RuleIcon::Doubleclicked ()
{
  RuleOutputwindow->MapMe ();
}

void 
RuleIcon::Connectin (Connection * acuteconnection)
{
  switch (acuteconnection->GetStartIcon ()->GetTyp ()) {
  case MOVEABLEICONTYP_INPUT:
    if (In == NULL) {
      if (Outputs->GetConnection () == NULL) {
	acuteconnection->ErasetempArrow ();
	acuteconnection->SetEnd ((int) (x + width * .5), (int) (y + height * .5));
	acuteconnection->DrawArrow ();

	Mainwindow->SetStatus (MAINNORMALSTATUS);
	acuteconnection->SetStatus (CONNECTSTATUS_VISIBLE);
	if (acuteconnection->GetEndIcon () != this) {
	  if (acuteconnection->GetEndIcon () != NULL)
	    acuteconnection->GetEndIcon ()->Action (ICON_REMOVECONNECTIN, (int) (acuteconnection));

	  acuteconnection->SetEndIcon (this);
	  In = acuteconnection;
	  Datensatz *tempdataset;
	  In->SendActiontoStart (ICON_REQUESTFORDATASET, (int) &tempdataset);
	  RuleOutputwindow->SetDataset (tempdataset);
	}
	XDefineCursor (display, Mainwindow->GetWindow (), maincursor);
      } else
	Mainwindow->printError ("Only input if no output exists!");
    } else
      Mainwindow->printError ("Only one input to Rule-Object!");
    break;

  case MOVEABLEICONTYP_CLUSTER:
    Mainwindow->printError ("No connection from Cluster-Object to Rule-Object!");
    break;

  case MOVEABLEICONTYP_RULE:
    Mainwindow->printError ("No connection from Rule-Object to Rule-Object!");
    break;

  case MOVEABLEICONTYP_CLASSIFY:
    Mainwindow->printError ("No connection from Classify-Object to Rule-Object!");
    break;

  default:
    break;
  }
}

void 
RuleIcon::Connectout (Connection * acuteconnection)
{
  if (In == NULL) {
    Mainwindow->SetStatus (MAINENDCONNECTSTATUS);
    acuteconnection->SetStartIcon (this);
    *Outputs += acuteconnection;
    XDefineCursor (display, Mainwindow->GetWindow (), createendcursor);
  } else
    Mainwindow->printError ("Only output, if no input exists!");
}

void 
RuleIcon::ChangeConnectin (Connection * acuteconnection)
{
}

void 
RuleIcon::ChangeConnectout (Connection * acuteconnection)
{
  if (Outputs->GetConnection () == NULL) {
    if (In == NULL) {
      switch (acuteconnection->GetEndIcon ()->GetTyp ()) {
      case MOVEABLEICONTYP_INPUT:
	acuteconnection->GetStartIcon ()->Action (ICON_REMOVECONNECTOUT, (int) (Mainwindow->GetAcuteconnection ()));
	acuteconnection->ErasetempArrow ();
	acuteconnection->SetStart ((int) (x + width * .5), (int) (y + height * .5));
	acuteconnection->DrawArrow ();
	acuteconnection->GetStartIcon ()->Action (ICON_REMOVECONNECTOUT, (int) (Mainwindow->GetAcuteconnection ()));
	acuteconnection->SetStartIcon (this);
	acuteconnection->SetStatus (CONNECTSTATUS_VISIBLE);
	Mainwindow->SetStatus (MAINNORMALSTATUS);
	*Outputs += Mainwindow->GetAcuteconnection ();
	XDefineCursor (display, Mainwindow->GetWindow (), maincursor);
	acuteconnection->GetEndIcon ()->ChangeConnectin (acuteconnection);
	break;

      case MOVEABLEICONTYP_CLASSIFY:
	Mainwindow->printError ("No connection from Rule-Object to Classify-Object!");
	break;

      case MOVEABLEICONTYP_CLUSTER:
	Mainwindow->printError ("No connection from Rule-Object to Cluster-Object!");
	break;

      case MOVEABLEICONTYP_RULE:
	Mainwindow->printError ("No connection from Rule-Object to Rule-Object!");
	break;
      }
    } else
      Mainwindow->printError ("Only output, if no input exists!");
  } else
    Mainwindow->printError ("Only output, if no output exists!");
}

void 
RuleIcon::DrawMoveableIcon ()
{
  short textx, texty, depth;
  char *indextext = (char *) malloc (sizeof (char) * 5);	// has to be enough

  sprintf (indextext, "%d", index);
  short textheight = fontstruct->max_bounds.ascent +
  fontstruct->max_bounds.descent;
  width = MAX (XTextWidth (fontstruct, text, strlen (indextext)),
	       XTextWidth (fontstruct, text, strlen (text)));
  width += 4;
  height = 5 * textheight;
  width = height = (((width) > (height)) ? (width) : (height));
  SizeMe (width, height);

  if (pixmap_up != 0) {
    XFreePixmap (display, pixmap_up);
    XFreePixmap (display, pixmap_down);
    char *space = " ";
    char *windowtext = (char *) malloc (sizeof (char) *
				  (4 + strlen (text) + strlen (indextext)));
    strcpy (windowtext, "");
    strcat (windowtext, text);
    strcat (windowtext, space);
    strcat (windowtext, indextext);
    RuleOutputwindow->SetWindowName (windowtext);
  }
  depth = DefaultDepth (display, DefaultScreen (display));
  pixmap_up = XCreatePixmap (display, window, width, height, depth);
  pixmap_down = XCreatePixmap (display, window, width, height, depth);

  XSetForeground (display, gc, redbuttonpix);
  XFillRectangle (display, pixmap_up, gc, 0, 0, width, height);
  XFillRectangle (display, pixmap_down, gc, 0, 0, width, height);

  XSegment shadowlines[12] =
  {
    {1, height - 1, width - 1, height - 1},
    {2, height - 2, width - 1, height - 2},
    {width - 1, 1, width - 1, height - 1},
    {width - 2, 1, width - 2, height - 2},
    {width / 8, 3 * height / 8, 3 * width / 8, height / 8},
    {1 + width / 8, 3 * height / 8, 3 * width / 8, 1 + height / 8},
    {width / 8, 1 + 3 * height / 8, 5 * width / 8, 5 * height / 8},
  {1 + width / 8, 1 + 3 * height / 8, 5 * width / 8 - 1, 5 * height / 8 - 1},
    {3 * width / 4, 5 * height / 8, 5 * width / 8, 3 * height / 4},
    {3 * width / 4, 5 * height / 8 + 1, 5 * width / 8 + 1, 3 * height / 4},
    {3 * width / 4, 7 * height / 8, 5 * width / 8, 3 * height / 4},
    {3 * width / 4, 7 * height / 8 - 1, 5 * width / 8 + 1, 3 * height / 4}
  };

  XSegment lightlines[10] =
  {
    {0, 0, 0, height - 2},
    {1, 0, 1, height - 3},
    {0, 0, width - 2, 0},
    {0, 1, width - 3, 1},
    {1 + 3 * width / 8, height / 8, 5 * width / 8, 5 * height / 8},
  {1 + 3 * width / 8, 1 + height / 8, 5 * width / 8 - 1, 5 * height / 8 - 1},
    {3 * width / 4, 5 * height / 8, 7 * width / 8, 3 * height / 4},
    {3 * width / 4, 5 * height / 8 + 1, 7 * width / 8 - 1, 3 * height / 4},
    {3 * width / 4, 7 * height / 8, 7 * width / 8, 3 * height / 4},
    {3 * width / 4, 7 * height / 8 - 1, 7 * width / 8 - 1, 3 * height / 4}
  };

  XSetForeground (display, gc, redshadowpix);
  XDrawSegments (display, pixmap_up, gc, shadowlines, 12);
  XDrawSegments (display, pixmap_down, gc, lightlines, 10);

  XSetForeground (display, gc, redlightpix);
  XDrawSegments (display, pixmap_down, gc, shadowlines, 12);
  XDrawSegments (display, pixmap_up, gc, lightlines, 10);

  XSetForeground (display, gc, blackpix);
  texty = (int) ((height + fontstruct->max_bounds.ascent +
		  fontstruct->max_bounds.descent) * .5 - fontstruct->max_bounds.descent) - textheight;
  textx = (int) ((width - XTextWidth (fontstruct, text, strlen (text))) * .5);
//texty=2*textheight;
  XDrawString (display, pixmap_up, gc, textx, texty, text, strlen (text));
  XDrawString (display, pixmap_down, gc, textx, texty, text, strlen (text));

  textx = (int) ((width - XTextWidth (fontstruct, indextext, strlen (indextext))) * .5);
  texty += 2 * textheight;
  XDrawString (display, pixmap_up, gc, textx, texty, indextext, strlen (indextext));
  XDrawString (display, pixmap_down, gc, textx, texty, indextext, strlen (indextext));
}

void 
RuleIcon::Action (int actiontyp, int value)
{
  switch (actiontyp) {
  case ICON_REQUESTFORRULERESULTS:
    RuleOutputwindow->CalculateResults ((RuleResultInfo *) value);
    break;

  case ICON_RULERESULTSCHANGED:
    Outputs->SendActiontoallEnds (ICON_RULERESULTSCHANGED, 0);
    break;

  case ICON_REMOVECONNECTIN:
    RuleOutputwindow->SetDataset ((Datensatz *) NULL);
    In = NULL;
    break;

  case ICON_REMOVECONNECTOUT:
    *Outputs -= (Connection *) (value);
    RuleOutputwindow->SetDataset ((Datensatz *) NULL);
    break;

  case ICON_SENDINGDATASET:
    RuleOutputwindow->SetDataset ((Datensatz *) value);
    break;

  case ICON_REQUESTFORDVECTOR:
    if (In != NULL)
      In->SendActiontoStart (ICON_REQUESTFORDVECTOR, value);
    break;

  case ICON_SENDINGDVECTOR:
    RuleOutputwindow->SetSelectedinputvector ((DVektor *) value);
    break;

  }
}

ClassifyIcon::ClassifyIcon (Display * initdisplay,
	      GC initgc,
	      XFontStruct * fontstruct,
	      MainWindow * initMainwindow,
	      int initx, int inity,
	      int initwidth, int initheight,
	      char *inittext,
	      RuleInfo * ruleinfo):
MoveableIcon (initdisplay,
	      initgc,
	      fontstruct,
	      initMainwindow,
	      initx, inity,
	      initwidth, initheight,
	      inittext)
{
  index = 0;
  pixmap_up = 0;
  ClassifyOutputwindow = new ClassifyOutput (display,
					     gc,
					     this,
					     fontstruct,
					     70, 10,
					     OUTPUTMINSIZE, OUTPUTMINSIZE,
					     text,
					     ruleinfo);

  DrawMoveableIcon ();
};

ClassifyIcon::~ClassifyIcon ()
{
  delete ClassifyOutputwindow;
}

void 
ClassifyIcon::Doubleclicked ()
{
  ClassifyOutputwindow->MapMe ();
}

void 
ClassifyIcon::Connectin (Connection * acuteconnection)
{
  switch (acuteconnection->GetStartIcon ()->GetTyp ()) {
  case MOVEABLEICONTYP_INPUT:
    if (In == NULL) {
      if (Outputs->GetConnection () == NULL) {
	acuteconnection->ErasetempArrow ();
	acuteconnection->SetEnd ((int) (x + width * .5), (int) (y + height * .5));
	acuteconnection->DrawArrow ();

	Mainwindow->SetStatus (MAINNORMALSTATUS);
	acuteconnection->SetStatus (CONNECTSTATUS_VISIBLE);
	if (acuteconnection->GetEndIcon () != this) {
	  if (acuteconnection->GetEndIcon () != NULL)
	    acuteconnection->GetEndIcon ()->Action (ICON_REMOVECONNECTIN, (int) (acuteconnection));

	  acuteconnection->SetEndIcon (this);
	  In = acuteconnection;
	  Datensatz *tempdataset;
	  In->SendActiontoStart (ICON_REQUESTFORDATASET, (int) &tempdataset);
	  ClassifyOutputwindow->SetDataset (tempdataset);
	}
	XDefineCursor (display, Mainwindow->GetWindow (), maincursor);
      } else
	Mainwindow->printError ("Only input if no output exists!");
    } else
      Mainwindow->printError ("Only one input to Classify-Object!");
    break;

  case MOVEABLEICONTYP_CLUSTER:
    Mainwindow->printError ("No connection from Cluster-Object to Classify-Object!");
    break;

  case MOVEABLEICONTYP_RULE:
    Mainwindow->printError ("No connection from Rule-Object to Classify-Object!");
    break;

  case MOVEABLEICONTYP_CLASSIFY:
    Mainwindow->printError ("No connection from Classify-Object to Classify-Object!");
    break;
  }
}

void 
ClassifyIcon::Connectout (Connection * acuteconnection)
{
  if (In == NULL) {
    Mainwindow->SetStatus (MAINENDCONNECTSTATUS);
    acuteconnection->SetStartIcon (this);
    *Outputs += acuteconnection;
    XDefineCursor (display, Mainwindow->GetWindow (), createendcursor);
  } else
    Mainwindow->printError ("Only output, if no input exists!");
}

void 
ClassifyIcon::ChangeConnectin (Connection * acuteconnection)
{
}

void 
ClassifyIcon::ChangeConnectout (Connection * acuteconnection)
{
  if (Outputs->GetConnection () == NULL) {
    if (In == NULL) {
      switch (acuteconnection->GetEndIcon ()->GetTyp ()) {
      case MOVEABLEICONTYP_INPUT:
	acuteconnection->GetStartIcon ()->Action (ICON_REMOVECONNECTOUT, (int) (Mainwindow->GetAcuteconnection ()));
	acuteconnection->ErasetempArrow ();
	acuteconnection->SetStart ((int) (x + width * .5), (int) (y + height * .5));
	acuteconnection->DrawArrow ();
	acuteconnection->GetStartIcon ()->Action (ICON_REMOVECONNECTOUT, (int) (Mainwindow->GetAcuteconnection ()));
	acuteconnection->SetStartIcon (this);
	acuteconnection->SetStatus (CONNECTSTATUS_VISIBLE);
	Mainwindow->SetStatus (MAINNORMALSTATUS);
	*Outputs += Mainwindow->GetAcuteconnection ();
	XDefineCursor (display, Mainwindow->GetWindow (), maincursor);
	acuteconnection->GetEndIcon ()->ChangeConnectin (acuteconnection);
	break;

      case MOVEABLEICONTYP_CLASSIFY:
	Mainwindow->printError ("No connection from Classify-Object to Classify-Object!");
	break;

      case MOVEABLEICONTYP_CLUSTER:
	Mainwindow->printError ("No connection from Classify-Object to Cluster-Object!");
	break;

      case MOVEABLEICONTYP_RULE:
	Mainwindow->printError ("No connection from Classify-Object to Rule-Object!");
	break;
      }
    } else
      Mainwindow->printError ("Only output, if no input exists!");
  } else
    Mainwindow->printError ("Only output, if no output exists!");
}

void 
ClassifyIcon::DrawMoveableIcon ()
{
  short textx, texty, depth;
  char *indextext = (char *) malloc (sizeof (char) * 5);	// has to be enough

  sprintf (indextext, "%d", index);
  short textheight = fontstruct->max_bounds.ascent +
  fontstruct->max_bounds.descent;
  width = MAX (XTextWidth (fontstruct, text, strlen (indextext)),
	       XTextWidth (fontstruct, text, strlen (text)));
  width += 4;
  height = 5 * textheight;
  width = height = (((width) > (height)) ? (width) : (height));
  SizeMe (width, height);

  if (pixmap_up != 0) {
    XFreePixmap (display, pixmap_up);
    XFreePixmap (display, pixmap_down);
    char *space = " ";
    char *windowtext = (char *) malloc (sizeof (char) *
				  (4 + strlen (text) + strlen (indextext)));
    strcpy (windowtext, "");
    strcat (windowtext, text);
    strcat (windowtext, space);
    strcat (windowtext, indextext);
    ClassifyOutputwindow->SetWindowName (windowtext);
  }
  depth = DefaultDepth (display, DefaultScreen (display));
  pixmap_up = XCreatePixmap (display, window, width, height, depth);
  pixmap_down = XCreatePixmap (display, window, width, height, depth);

  XSetForeground (display, gc, bluebuttonpix);
  XFillRectangle (display, pixmap_up, gc, 0, 0, width, height);
  XFillRectangle (display, pixmap_down, gc, 0, 0, width, height);

  XSegment shadowlines[28] =
  {
    {1, height - 1, width / 2, height - 1},
    {2, height - 2, width / 2, height - 2},
    {5 * width / 8, height - 1, width - 1, height - 1},
    {5 * width / 8 + 1, height - 2, width - 1, height - 2},
    {width - 1, 1, width - 1, height / 4},
    {width - 2, 1, width - 2, height / 4 - 1},
    {3 * width / 4, 3 * height / 8, 3 * width / 4, height / 2},
  {3 * width / 4 + 1, 3 * height / 8 - 1, 3 * width / 4 + 1, height / 2 - 1},
    {7 * width / 8, height / 4, width - 2, height / 4},
    {7 * width / 8 + 1, height / 4 - 1, width - 1, height / 4 - 1},
    {3 * width / 4 + 1, height / 8, 7 * width / 8, height / 8},
    {3 * width / 4, height / 8 - 1, 7 * width / 8 + 1, height / 8 - 1},
    {3 * width / 4, height / 8, 3 * width / 4, height / 4 - 1},
    {3 * width / 4 + 1, height / 8 - 1, 3 * width / 4 + 1, height / 4},
    {5 * width / 8, height / 4, 3 * width / 4 + 1, height / 4},
    {5 * width / 8 + 1, height / 4 - 1, 3 * width / 4, height / 4 - 1},
    {width / 2, height / 4, width / 2, 3 * height / 8},
    {width / 2 + 1, height / 4, width / 2 + 1, 3 * height / 8 - 1},
    {width - 1, 3 * height / 8, width - 1, height - 1},
    {width - 2, 3 * height / 8 + 1, width - 2, height - 2},
    {width / 4, 5 * height / 8, width / 4, 3 * height / 4},
    {width / 4 + 1, 5 * height / 8 + 1, width / 4 + 1, 3 * height / 4 - 1},
    {width / 4, 5 * height / 8, 3 * width / 4, 5 * height / 8},
  {width / 4 + 1, 5 * height / 8 + 1, 3 * width / 4 - 1, 5 * height / 8 + 1},
    {5 * width / 8, 0, width / 2, height / 4},
    {5 * width / 8 + 1, 0, width / 2 + 1, height / 4},
    {5 * width / 8, 3 * height / 4, width / 2, height - 1},
    {5 * width / 8 + 1, 3 * height / 4 - 1, width / 2 + 1, height - 1}
  };

  XSegment lightlines[26] =
  {
    {0, 0, 0, 3 * height / 8},
    {1, 0, 1, 3 * height / 8 - 1},
    {0, height / 2 + 2, 0, height - 2},
    {1, height / 2 + 2, 1, height - 3},
    {0, 0, 5 * width / 8, 0},
    {0, 1, 5 * width / 8 - 1, 1},
    {3 * height / 4, 0, width - 2, 0},
    {3 * height / 4, 1, width - 3, 1},
    {7 * width / 8, 3 * height / 8, width - 1, 3 * height / 8},
    {7 * width / 8 + 1, 3 * height / 8 + 1, width - 2, 3 * height / 8 + 1},
    {5 * width / 8, height - 2, 3 * width / 4, 3 * height / 4},
    {5 * width / 8 - 1, height - 1, 3 * width / 4 - 1, 3 * height / 4 - 1},
    {width / 2, 3 * height / 8, 3 * width / 4, 3 * height / 8},
  {width / 2 + 1, 3 * height / 8 - 1, 3 * width / 4 + 1, 3 * height / 8 - 1},
    {7 * width / 8, height / 8, 7 * width / 8, height / 4},
    {7 * width / 8 + 1, height / 8 - 1, 7 * width / 8 + 1, height / 4 - 1},
    {3 * width / 4, 5 * height / 8, 3 * width / 4, 3 * height / 4},
    {3 * width / 4 - 1, 5 * height / 8 + 1, 3 * width / 4 - 1, 3 * height / 4 - 1},
    {3 * width / 4, height / 2, 7 * width / 8, height / 2},
    {3 * width / 4 + 1, height / 2 - 1, 7 * width / 8 + 1, height / 2 - 1},
    {7 * width / 8, height / 2 - 1, 7 * width / 8, 3 * height / 8},
    {7 * width / 8 + 1, height / 2, 7 * width / 8 + 1, 3 * height / 8 + 1},
    {5 * width / 8, height / 4, 3 * width / 4 - 1, 0},
    {5 * width / 8 + 1, height / 4 - 1, 3 * width / 4, 1},
    {width / 4, 3 * height / 4, 5 * width / 8, 3 * height / 4},
  {width / 4 + 1, 3 * height / 4 - 1, 5 * width / 8 + 1, 3 * height / 4 - 1}
  };

  XSetForeground (display, gc, blueshadowpix);
  XDrawArc (display, pixmap_up, gc, -width / 8, -3 * height / 8, width / 4, 3 * height / 4, 270 * 64, 360 * 64);
  XDrawArc (display, pixmap_up, gc, -width / 8, -3 * height / 8 + 1, width / 4, 3 * height / 4, 270 * 64, 360 * 64);
  XDrawArc (display, pixmap_up, gc, -width / 8 + 1, -3 * height / 8, width / 4, 3 * height / 4, 270 * 64, 360 * 64);
  XDrawArc (display, pixmap_down, gc, -width / 4, -height / 2, width / 2, height, 270 * 64, 360 * 64);
  XDrawArc (display, pixmap_down, gc, -width / 4, -height / 2 + 1, width / 2, height, 270 * 64, 360 * 64);
  XDrawArc (display, pixmap_down, gc, -width / 4 + 1, -height / 2, width / 2, height, 270 * 64, 360 * 64);

  XSetForeground (display, gc, bluelightpix);
  XDrawArc (display, pixmap_up, gc, -width / 4, -height / 2, width / 2, height, 270 * 64, 360 * 64);
  XDrawArc (display, pixmap_up, gc, -width / 4, -height / 2 + 1, width / 2, height, 270 * 64, 360 * 64);
  XDrawArc (display, pixmap_up, gc, -width / 4 + 1, -height / 2, width / 2, height, 270 * 64, 360 * 64);
  XDrawArc (display, pixmap_down, gc, -width / 8, -3 * height / 8, width / 4, 3 * height / 4, 270 * 64, 360 * 64);
  XDrawArc (display, pixmap_down, gc, -width / 8, -3 * height / 8 + 1, width / 4, 3 * height / 4, 270 * 64, 360 * 64);
  XDrawArc (display, pixmap_down, gc, -width / 8 + 1, -3 * height / 8, width / 4, 3 * height / 4, 270 * 64, 360 * 64);

  XSetForeground (display, gc, blueshadowpix);
  XDrawSegments (display, pixmap_up, gc, shadowlines, 28);
  XDrawSegments (display, pixmap_down, gc, lightlines, 26);

  XSetForeground (display, gc, bluelightpix);
  XDrawSegments (display, pixmap_down, gc, shadowlines, 28);
  XDrawSegments (display, pixmap_up, gc, lightlines, 26);

  XSetForeground (display, gc, blackpix);
  texty = (int) ((height + fontstruct->max_bounds.ascent +
		  fontstruct->max_bounds.descent) * .5 - fontstruct->max_bounds.descent) - textheight;
  textx = (int) ((width - XTextWidth (fontstruct, text, strlen (text))) * .5);
  XDrawString (display, pixmap_up, gc, textx, texty, text, strlen (text));
  XDrawString (display, pixmap_down, gc, textx, texty, text, strlen (text));

  textx = (int) ((width - XTextWidth (fontstruct, indextext, strlen (indextext))) * .5);
  texty += 2 * textheight;
  XDrawString (display, pixmap_up, gc, textx, texty, indextext, strlen (indextext));
  XDrawString (display, pixmap_down, gc, textx, texty, indextext, strlen (indextext));
}

void 
ClassifyIcon::Action (int actiontyp, int value)
{
  switch (actiontyp) {
  case ICON_REQUESTFORCLASSIFYRESULTS:
    ClassifyOutputwindow->CalculateResults ((Datensatz *) value);
    break;

  case ICON_CLASSIFYRESULTSCHANGED:
    Outputs->SendActiontoallEnds (ICON_CLASSIFYRESULTSCHANGED, 0);
    break;

  case ICON_REMOVECONNECTIN:
    ClassifyOutputwindow->SetDataset ((Datensatz *) NULL);
    In = NULL;
    break;

  case ICON_REMOVECONNECTOUT:
    *Outputs -= (Connection *) (value);
    ClassifyOutputwindow->SetDataset ((Datensatz *) NULL);
    break;

  case ICON_REQUESTFORDVECTOR:
    if (In != NULL)
      In->SendActiontoStart (ICON_REQUESTFORDVECTOR, value);
    break;

  case ICON_SENDINGDVECTOR:
    ClassifyOutputwindow->SetSelectedinputvector ((DVektor *) value);
    break;

  }
}
