/********************************************************/
/*   filename: regel.h                                  */
/*                                                      */
/********************************************************/
/* programmed by: Oliver Wagner                         */
/* last change:  (XXX: not updated)                     */
/********************************************************/
#ifndef _REGEL_H
#define _REGEL_H

class Parameter;
class Trapez_Parameter;
class Konvex_Parameter;
class Regel;
class Regelbasis;
class Regel_Ausgabe;
class Regelbasis_Ausgabe;
class Klasse;
class Klassifizierung;
class Klassen_Ausgabe;
class Klassifizierungs_Ausgabe;

enum Findungs_Art {
  Dim_einzeln, Dim_insgesamt
};
enum Parameter_Typ {
  Trapez, Konvex
};
enum Regel_Typ {
  Dummy, Typ_Regel, Typ_Klasse
};


#define PARAMETER_F_PROZENT .02
#define PARAMETER_SCHRITTE 40

#define _REGEL_DATA
#include "fehler.h"
#include "matrix.h"
#include "daten.h"
#include "norm.h"
#include "vektor.h"
#include "cluster.h"
#undef _REGEL_DATA
#endif

#ifndef _REGEL_DATA
#define _REGEL_DATA

class Parameter
{
#if 0
  protected:
#else
  public:
#endif
  Parameter_Typ So;
  public:
  Parameter () {
  };
  Parameter (const Parameter & Das) {
  };
  virtual ~ Parameter () {
  };
  virtual void print () = 0;

  virtual Parameter_Typ Lese_Parameter_Typ () = 0;

  virtual void Setze_Parameter (Parameter & Damit) = 0;
  virtual void Setze_xParameter (int Index, double Damit) = 0;
  virtual int Lese_Parameter_Anzahl () = 0;
  virtual double &Lese_xParameter (int i) = 0;
  virtual double Lese_yParameter (int i) = 0;

  virtual char Finde_Parameter (DVektorArray & Daraus,
				double Z_Grenze) = 0;

  virtual void Speichern (FILE * File) = 0;

  virtual double Berechne_Schwerpunkt (double Hoehe, DVektor & x, DVektor & y) = 0;
  virtual double Berechne_Hoehe (double Eingabe) = 0;
  virtual void Berechne_Form (double Hoehe, DVektor & x, DVektor & y) = 0;

  virtual Parameter & operator = (Parameter & Das) = 0;
protected:
};

class Trapez_Parameter: public Parameter
{
  protected:
  double Daten[4];
  double Steigung[2];
    public:
    Trapez_Parameter ();
    Trapez_Parameter (const Trapez_Parameter & Das);
    Trapez_Parameter (const Parameter & Das);
  void print ();

  inline Parameter_Typ Lese_Parameter_Typ () {
    return (So);
  };
  void Setze_Parameter (Parameter & Damit);
  void Setze_xParameter (int Index, double Damit);
  inline int Lese_Parameter_Anzahl () {
    return (4);
  };
  inline double &Lese_xParameter (int i) {
    return (Daten[i]);
  };
  double Lese_yParameter (int i);

  char Finde_Parameter (DVektorArray & Daraus, double Z_Grenze);

  void Speichern (FILE * File);
  friend Trapez_Parameter Trapez_einlesen (FILE * File);

  double Berechne_Schwerpunkt (double Hoehe, DVektor & x, DVektor & y);
  double Berechne_Hoehe (double Eingabe);
  void Berechne_Form (double Hoehe, DVektor & x, DVektor & y);

    Parameter & operator = (Parameter & Das);
    protected:
  double Parameter_Fehler (DVektorArray & Damit);
};

class Konvex_Parameter: public Parameter
{
  protected:
  DVektorArray Daten;
  public:
  Konvex_Parameter ();
  Konvex_Parameter (const Konvex_Parameter & Das);
    Konvex_Parameter (const Parameter & Das);
  void print ();

  inline Parameter_Typ Lese_Parameter_Typ () {
    return (So);
  };
  void Setze_Parameter (Parameter & Damit);
  void Setze_xParameter (int Index, double Damit);
  void Setze_yParameter (int Index, double Damit);
  inline int Lese_Parameter_Anzahl () {
    return (Daten.Lese_Dim ());
  };
  inline double &Lese_xParameter (int i) {
    return (Daten[0][i]);
  };
  inline double Lese_yParameter (int i) {
    return (Daten[1][i]);
  };

  char Finde_Parameter (DVektorArray & Daraus, double Z_Grenze);

  void Speichern (FILE * File);
  friend Konvex_Parameter Konvex_einlesen (FILE * File);

  double Berechne_Schwerpunkt (double Hoehe, DVektor & x, DVektor & y);
  double Berechne_Hoehe (double Eingabe);
  void Berechne_Form (double Hoehe, DVektor & x, DVektor & y);

    Parameter & operator = (Parameter & Das);
};

/**************************************************************************/
class Dummy_Regel
{
  protected:
  Parameter ** Daten;
  int xDim, yDim;
  char **Namen;
  int Nummer;
    public:
    Dummy_Regel (int xDimension = 0, int yDimension = 0, Parameter_Typ PT = Trapez);
    Dummy_Regel (const Dummy_Regel & Das);
    virtual ~ Dummy_Regel ();

  virtual Regel_Typ Lese_Typ () {
    return (Dummy);
  };
  virtual void print () = 0;
  inline int Lese_Dimension () {
    return (xDim + yDim);
  };
  inline int &Lese_Nummer () {
    return (Nummer);
  };
  void Setze_Nummer (int i) {
    Nummer = i;
  };
  inline char *Lese_Name (int i) {
    return (Namen[i]);
  };
  void Setze_Name (int i, char *Damit) {
    strcpy (Namen[i], Damit);
  };

  void Speichern (FILE * File);

  void Setze_Parameter (int Index, Parameter & Damit);
  inline Parameter & Lese_Parameter (int Index) {
    return (*(Daten[Index]));
  };

    Dummy_Regel & operator = (const Dummy_Regel & Das);
    protected:
  void File_einlesen (FILE * File, int x, int y);
};

class Regel:public Dummy_Regel
{
  public:
  Regel (int xDimension = 0, int yDimension = 0, Parameter_Typ PT = Trapez);
    Regel (const Regel & Das);
   ~Regel ();

  inline Regel_Typ Lese_Typ () {
    return (Typ_Regel);
  };
  void print ();
  inline int Lese_xDimension () {
    return (xDim);
  };
  inline int Lese_yDimension () {
    return (yDim);
  };
  void Setze_Dimensionen (int In, int Out);
  void Setze_xDimension (int i);
  void Setze_yDimension (int i);

  friend Regel Regel_einlesen (FILE * File, int x, int y);

  void Berechne (DVektor & Eingabe, Regel_Ausgabe & Ausgabe);
  void Berechne (DVektor & Eingabe, BVektor & Diese_Eingaben,
		 BVektor & Diese_Ausgaben, Regel_Ausgabe & Ausgabe);

    Regel & operator = (Regel & Das);
  inline Parameter & operator[] (int i) {
    return (*Daten[i]);
  };
};

/**********************************************/
class Regelbasis
{
  Regel **Regeln;
  int Anzahl, xDim, yDim;
    public:
    Regelbasis (int Regelanzahl = 0, int xDimension = 0, int yDimension = 0);
   ~Regelbasis ();
  void print ();
  inline int Lese_Dimension () {
    return (xDim + yDim);
  };
  inline int Lese_xDimension () {
    return (xDim);
  };
  inline int Lese_yDimension () {
    return (yDim);
  };
  void Setze_xDimension (int i);
  void Setze_yDimension (int i);
  void Setze_Dimensionen (int In, int Out);

  inline char *Lese_Namen (int i) {
    return (Regeln[0]->Lese_Name (i));
  };

  void Setze_Regelanzahl (int Damit);
  inline int Lese_Regelanzahl () {
    return (Anzahl);
  };
  inline Regel & Lese_Regel (int Nummer) {
    return (*(Regeln[Nummer]));
  };

  void Speichern (char *Name);
  friend Regelbasis Regelbasis_einlesen (char *Name);

  void Berechne (DVektor & Eingabe, Regelbasis_Ausgabe & Ausgabe);
  void Berechne (DVektor & Eingabe, BVektor & Diese_Regeln,
		 BVektor & Diese_Eingaben, BVektor & Diese_Ausgaben,
		 Regelbasis_Ausgabe & Ausgabe);

    Regelbasis & operator = (const Regelbasis & Das);
  void operator += (Regel & Diese);
  void operator += (Regelbasis & Das);
  inline Regel & operator[] (int i) {
    return (*(Regeln[i]));
  };

/* Ermitteln einfach die Regeln: */
  friend Regelbasis Ermittle_Regeln (
				      Clustering & Cluster,
				      double Z_Grenze,
				      Datensatz & Daten,
				      int InputDim,
				      BVektor & relevante_Dims,
				      Parameter_Typ PT);
  Regelbasis Ermittle_Regeln (
			       Clustering & Cluster,
			       double Z_Grenze,
			       Datensatz & Daten,
			       int Input_Dim,
			       BVektor & relevante_Dims,
			       Parameter_Typ PT,
			       BVektor & Diese_Cluster);
  Regelbasis Ermittle_Regeln (
			       Clustering & Cluster,
			       double Z_Grenze,
			       Datensatz & Daten,
			       int Input_Dim,
			       BVektor & relevante_Dims,
			       Parameter_Typ PT,
			       int ClusterNr);

/* Ermitteln die Regeln, ueberpruefen aber die Konvexitaet der  */
/* Zugehoerigkeitkurven, und teilen sie falls noetig:           */
  friend Regelbasis Finde_Regeln (double Huegel_Grenze,
				  Findungs_Art Art, Clustering & Cluster,
				  double Z_Grenze,
				  Datensatz & Daten,
				  int InputDim,
				  BVektor & relevante_Dims,
				  Parameter_Typ PT);
  Regelbasis Finde_Regeln (double Huegel_Grenze,
			   Findungs_Art Art, Clustering & Cluster,
			   double Z_Grenze,
			   Datensatz & Daten,
			   int Input_Dim,
			   BVektor & relevante_Dims,
			   Parameter_Typ PT,
			   BVektor & Diese_Cluster);
  Regelbasis Finde_Regeln (double Huegel_Grenze,
			   Findungs_Art Art, Clustering & Cluster,
			   double Z_Grenze,
			   Datensatz & Daten,
			   int InputDim,
			   BVektor & relevante_Dims,
			   Parameter_Typ PT,
			   int ClusterNr);

};

/* Friend Funktionen: */

Regelbasis Ermittle_Regeln (Clustering & Cluster,
			    double Z_Grenze,
			    Datensatz & Daten,
			    int InputDim,
			    BVektor & relevante_Dims,
			    Parameter_Typ PT);
Regelbasis Ermittle_Regeln (Clustering & Cluster,
			    double Z_Grenze,
			    Datensatz & Daten,
			    int Input_Dim,
			    BVektor & relevante_Dims,
			    Parameter_Typ PT,
			    BVektor & Diese_Cluster);
Regelbasis Ermittle_Regeln (Clustering & Cluster,
			    double Z_Grenze,
			    Datensatz & Daten,
			    int Input_Dim,
			    BVektor & relevante_Dims,
			    Parameter_Typ PT,
			    int ClusterNr);

Regelbasis Finde_Regeln (double Huegel_Grenze,
			 Findungs_Art Art, Clustering & Cluster,
			 double Z_Grenze,
			 Datensatz & Daten,
			 int InputDim,
			 BVektor & relevante_Dims,
			 Parameter_Typ PT);
Regelbasis Finde_Regeln (double Huegel_Grenze,
			 Findungs_Art Art, Clustering & Cluster,
			 double Z_Grenze,
			 Datensatz & Daten,
			 int Input_Dim,
			 BVektor & relevante_Dims,
			 Parameter_Typ PT,
			 BVektor & Diese_Cluster);
Regelbasis Finde_Regeln (double Huegel_Grenze,
			 Findungs_Art Art, Clustering & Cluster,
			 double Z_Grenze,
			 Datensatz & Daten,
			 int Input_Dim,
			 BVektor & relevante_Dims,
			 Parameter_Typ PT,
			 int ClusterNr);

/*************************************************/
double Berechne_RC (DVektorArray & GruppeAInput,
		    DVektorArray & GruppeBInput,
		    DVektorArray & OutputA,
		    DVektorArray & OutputB,
		    Regelbasis & RegelnA, Regelbasis & RegelnB);

/***********************************************************/
class Regel_Ausgabe
{
  friend Regel;
    protected:
  double Uebereinstimmung;	/* Minimale Hoehe       */
  DVektor InputHoehen;		/* Uebereinst. pro Input */
  DVektor *X_OutputKurve;	/* x-Werte der Ausgabef. */
  DVektor *Y_OutputKurve;	/* y-Werte der Ausgabef. */
  DVektor Ergebnisse;		/* i.A. die Schwerpunkte */
  int Input_Dim, Output_Dim;	/* kommt von der Regel  */
    public:
    Regel_Ausgabe (int InputDim = 0, int OutputDim = 0);
   ~Regel_Ausgabe ();

  void Setze_Dim (int xDim, int yDim);
  inline double Lese_Uebereinstimmung () {
    return (Uebereinstimmung);
  };
  inline DVektor & Lese_InputHoehen () {
    return (InputHoehen);
  };
  inline DVektor & Lese_X_Output (int i) {
    return (X_OutputKurve[i]);
  };
  inline DVektor & Lese_Y_Output (int i) {
    return (Y_OutputKurve[i]);
  };
  inline DVektor & Lese_Ergebnisse () {
    return (Ergebnisse);
  };
  inline int Lese_InputDim () {
    return (Input_Dim);
  };
  inline int Lese_OutputDim () {
    return (Output_Dim);
  };

    Regel_Ausgabe & operator = (Regel_Ausgabe & Das);
};

class Regelbasis_Ausgabe
{
  friend Regelbasis;
    protected:
    Regel_Ausgabe ** Ausgaben;	/* aus den einzelnen Regel      */
  DVektor Ergebnisse;		/* berechnete Endergebnisse     */
  int Anzahl;
    public:
    Regelbasis_Ausgabe (int Groesse = 0, int InputDim = 0, int OutputDim = 0);
   ~Regelbasis_Ausgabe ();

  void Setze_Dim (int xDim, int yDim);
  void Setze_Groesse (int So);

  inline DVektor & Lese_Ergebnisse () {
    return (Ergebnisse);
  } inline Regel_Ausgabe & operator[] (int i) {
    return (*(Ausgaben[i]));
  };
    Regelbasis_Ausgabe & operator = (Regelbasis_Ausgabe & Das);
};

/****************************************************************/
/*              Klassifizierungs-Stuff                          */
/****************************************************************/
class Klasse: public Dummy_Regel
{
  protected:
  public:
  Klasse (int Dimension = 0, Parameter_Typ PT = Trapez);
  Klasse (const Klasse & Das);
   ~Klasse () {
  };

  inline Regel_Typ Lese_Typ () {
    return (Typ_Klasse);
  };
  void print ();
  inline int Lese_Dimension () {
    return (xDim);
  };
  void Setze_Dimension (int So);

  friend Klasse Klasse_einlesen (FILE * File, int Dim);

  void Berechne (DVektor & Eingabe, Klassen_Ausgabe & Ausgabe);
  void Berechne (DVektor & Eingabe, BVektor & Diese_Eingaben,
		 Klassen_Ausgabe & Ausgabe);
  Klasse & operator = (Klasse & Das);
  inline Parameter & operator[] (int i) {
    return (*Daten[i]);
  };
};

class Klassifizierung
{
  Klasse **Klassen;
  int Anzahl, Dim;
  IVektor Klassen_Nummern;
    public:
    Klassifizierung (int Klassenanzahl = 0, int Dimension = 0);
   ~Klassifizierung ();
  void print ();
  inline int Lese_Dimension () {
    return (Dim);
  };
  void Setze_Dimension (int So);

  void Setze_Klassenanzahl (int So);
  inline int Lese_Klassenanzahl () const{
    return (Anzahl);
  };
  inline Klasse & Lese_Klasse (int Nummer) {
    return (*(Klassen[Nummer]));
  };
  inline IVektor & Lese_Klassen_Nummern () {
    return (Klassen_Nummern);
  };

  void Speichern (char *Name);
  friend Klassifizierung Klassifizierung_einlesen (char *Name);

  void Berechne (DVektor & Eingabe, Klassifizierungs_Ausgabe & Ausgabe);
  void Berechne (DVektor & Eingabe, BVektor & Diese_Regeln,
		 BVektor & Diese_Eingaben,
		 Klassifizierungs_Ausgabe & Ausgabe);

    Klassifizierung & operator = (const Klassifizierung & Das);
  void operator += (Klasse & Das);
  void operator += (const Klassifizierung & Diese);
  inline Klasse & operator[] (int i) const {
    return (*(Klassen[i]));
  };

  friend Klassifizierung Ermittle_Klassifizierung (
						    Clustering & Cluster,
						    double Z_Grenze,
						    Datensatz & Daten,
						    int InputDim,
						    BVektor & relevante_Dims,
						    Parameter_Typ PT);
  friend Klassifizierung Ermittle_Klassifizierung (
						    Clustering & Cluster,
						    double Z_Grenze,
						    Datensatz & Daten,
						    int Input_Dim,
						    BVektor & relevante_Dims,
						    Parameter_Typ PT,
						    BVektor & Diese_Cluster);
  friend Klassifizierung Ermittle_Klassifizierung (
						    Clustering & Cluster,
						    double Z_Grenze,
						    Datensatz & Daten,
						    int Input_Dim,
						    BVektor & relevante_Dims,
						    Parameter_Typ PT,
						    int ClusterNr);
};

class Klassen_Ausgabe
{
  friend Klasse;
    protected:
  double Uebereinstimmung;	/* Minimale Hoehe       */
  DVektor InputHoehen;		/* Uebereinst. pro Input */
    public:
    Klassen_Ausgabe (int Dimension = 0);

  inline void Setze_Dim (int Dimension) {
    InputHoehen.Setze_Dim (Dimension);
  };
  inline int Lese_Dim () {
    return (InputHoehen.Lese_Dim ());
  };

  inline double Lese_Uebereinstimmung () {
    return (Uebereinstimmung);
  };
  inline DVektor & Lese_InputHoehen () {
    return (InputHoehen);
  };
    Klassen_Ausgabe & operator = (Klassen_Ausgabe & Das);
};

class Klassifizierungs_Ausgabe
{
  friend Klassifizierung;
    protected:
    Klassen_Ausgabe ** Ausgaben;	/* aus den einzelnen Klassen    */
  int Ergebnis;			/* zu der Klassse gehoert Eing. */
  int Aus_dieser_Regel;
  DVektor Klassen_Zugehoerigkeiten;
  int Anzahl;
  char geklappt;
  char eindeutig;
    public:
    Klassifizierungs_Ausgabe (int Groesse = 0, int Dim = 0);
   ~Klassifizierungs_Ausgabe ();

  void Setze_Dim (int Dim);
  void Setze_Groesse (int So);

  inline int Lese_Ergebnis () {
    return (Ergebnis);
  } inline int Lese_Aus_dieser_Regel () {
    return (Aus_dieser_Regel);
  } inline DVektor & Lese_Zugehoerigkeiten () {
    return (Klassen_Zugehoerigkeiten);
  };
  inline char &Lese_geklappt () {
    return (geklappt);
  };
  inline char &Lese_eindeutig () {
    return (eindeutig);
  };
  inline Klassen_Ausgabe & operator[] (int i) {
    return (*(Ausgaben[i]));
  };
    Klassifizierungs_Ausgabe & operator = (Klassifizierungs_Ausgabe & Das);
};

#endif
