//**************************************************************
//*   filename: selectwindow.cc                                *
//*                                                            *
//**************************************************************
//* programmed by: Thomas Wagner                               *
//* last change: 24-05-95                                      *
//* last change: 10-10-98 HT                                     *
//**************************************************************
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>

#include "selectwindow.h"
#include "mainwindow.h"

#include "include_bmps/littlearrowup.bmp"
#include "include_bmps/littlearrowdown.bmp"

extern unsigned long buttonpix, blackpix, shadowpix, lightpix;
extern Cursor maincursor, waitcursor;
extern MainWindow *Mainwindow;

static SelectWindowInfo Infos[CLUSTERTYPNUMBER] =
{
  {"FCM", TEXTOPTION_SELECTABLE, FCM},
  {"GK", TEXTOPTION_SELECTABLE, GK},
  {"GK parallel", TEXTOPTION_SELECTABLE, GK_parallel},
  {"GG", TEXTOPTION_SELECTABLE, GG},
  {"GG parallel", TEXTOPTION_SELECTABLE, GG_parallel},
  {"not yet", 0, 10},
  {"not yet", 0, 10},
  {"not yet", 0, 10},
  {"to be continued ..", 0, 10}
};

static SelectWindowInfo ValidityInfos[VALIDYTYPNUMBER] =
{
  {"S-Validity", TEXTOPTION_SELECTABLE, S},
  {"PK-Validity", TEXTOPTION_SELECTABLE, PK},
  {"CS-Validity", TEXTOPTION_SELECTABLE, CS},
  {"D1-Validity", TEXTOPTION_SELECTABLE, D1},
  {"FHV-Validity", TEXTOPTION_SELECTABLE, FHV},
  {"DPA-Validity", TEXTOPTION_SELECTABLE, DPA},
  {"PD-Validity", TEXTOPTION_SELECTABLE, PD},
  {"Classify-Validity", TEXTOPTION_SELECTABLE, CLASSIFY}
};

static SelectWindowInfo RuletypInfos[RULETYPNUMBER] =
{
  {"Rule type 1", TEXTOPTION_SELECTABLE, RULETYP1},
  {"Rule type 2 a", TEXTOPTION_SELECTABLE, RULETYP2A},
  {"Rule type 2 b", TEXTOPTION_SELECTABLE, RULETYP2B}
};

static SelectWindowInfo ParametertypInfos[PARAMETERTYPNUMBER] =
{
  {"trapezium parameter", TEXTOPTION_SELECTABLE, Trapez},
  {"convex parameter", TEXTOPTION_SELECTABLE, Konvex},
  {"not yet", 0, 0},
  {"not yet", 0, 0},
  {"not yet", 0, 0}
};

RuleInfo::RuleInfo (		//DVektorArray *Cluster,
		   //DMatrix  *Membership,
	   Clustering * initclustering,
	   int Inputdim,
	   BVektor * Selecteddims,
	   BVektor * Selectedcluster,
	   Datensatz * Dataset)
{
  clustering = initclustering;
  inputdim = Inputdim;
  selecteddims = Selecteddims;
  selectedcluster = Selectedcluster;
  dataset = Dataset;
}

RuleInfo::~RuleInfo ()
{
}

void 
RuleInfo::operator= (RuleInfo & source)
{
  clustering = source.clustering;
  inputdim = source.inputdim;
  selecteddims = source.selecteddims;
  selectedcluster = source.selectedcluster;
  dataset = source.dataset;
}


ClusterInfo::ClusterInfo (Datensatz * Dataset,
	     char Ispossib,
	     int Clustertyp,
	     int Clusternumber,
	     int Repeatnumber,
	     double initM,
	     double initK,
	     int initexactness,
	     double initdelta_u,
	     Guete_Typ Validitytyp,
	     BVektor * Bvector)
{
  dataset = Dataset;
  ispossib = Ispossib;
  clustertyp = Clustertyp;
  clusternumber = Clusternumber;
  repeatnumber = Repeatnumber;
  M = initM;
  K = initK;
  exactness = initexactness;
  delta_u = initdelta_u;
  bvector = Bvector;
  validitytyp = Validitytyp;
}

void 
ClusterInfo::operator= (ClusterInfo & source)
{
  dataset = source.dataset;
  ispossib = source.ispossib;
  clustertyp = source.clustertyp;
  clusternumber = source.clusternumber;
  repeatnumber = source.repeatnumber;
  M = source.M;
  K = source.K;
  delta_u = source.delta_u;
  bvector = source.bvector;
}

SelectWindow::SelectWindow (Display * initdisplay,
	      GC initgc,
	      BigWindow * initParent,
	      XFontStruct * fontstruct,
	      int initx, int inity, int initwidth, int initheight,
	      SelectWindowInfo * Infos,
	      unsigned short Infonumber,
	      int Changeactionnumber,
	      int Doubleclickactionnumber,
	      unsigned short initoptions):
BigWindow (initdisplay,
	   initgc,
	   initParent->GetWindow (),
	   fontstruct,
	   initx, inity,
	   initwidth, initheight,
	   NULL, 0)
{
  Parent = initParent;
  options = initoptions;
  infos = Infos;
  infonumber = Infonumber;
  offset = 0;
  doubleclickactionnumber = Doubleclickactionnumber;
  changeactionnumber = Changeactionnumber;
  unsigned short counter;
  scrollwindow = NULL;

  short textwidth = 0, maxheight;
  short depth, textheight = fontstruct->max_bounds.ascent +
  fontstruct->max_bounds.descent;

// first add options, to place text left in textboxes
  for (counter = 0; counter < infonumber; counter++)
    infos[counter].options |= TEXTOPTION_GIVENXPOS;

  textnumber = (int) (height / textheight);
  if (textnumber > infonumber)
    textnumber = infonumber;

  textwidth = width - 4;

  if (textnumber < infonumber)	// scrollwindow is needed
   {
    textwidth -= (SCROLL_MINWIDTH + SELECTWINDOW_INSIDESPACE);
    scrollwindow = new ScrollWindow (display,
				     gc,
				     this,
				     fontstruct,
				     textwidth + SELECTWINDOW_INSIDESPACE, 2,
			       SCROLL_MINWIDTH, textnumber * textheight + 1,
				     0, infonumber - textnumber,
				     0,
				     SELECTWINDOW_SCROLLED_ACTION);
    scrollwindow->MapMe ();
    scrollwindow->SetValue (infonumber - textnumber);
  }
  textboxes = (TextBox **) malloc (sizeof (TextBox *) * (textnumber));
  for (counter = 0; counter < textnumber; counter++) {
    if (counter < infonumber)
      textboxes[counter] = new TextBox (display, gc, this, fontstruct,
					2, 2 + textheight * counter,
					textwidth, textheight,
				    (int) SELECTWINDOW_BORDERBESIDETEXT / 2,
					infos[counter].Selecttext, 0,
					infos[counter].options,
					CHANGETEXTSELECTION_ACTION,
					TEXT_DOUBLECLICKED);
    else
      textboxes[counter] = new TextBox (display, gc, this, fontstruct,
					2, 2 + textheight * counter,
					textwidth, textheight,
				    (int) SELECTWINDOW_BORDERBESIDETEXT / 2,
					"", 0,
					TEXTOPTION_DEFAULT,
					CHANGETEXTSELECTION_ACTION,
					TEXT_DOUBLECLICKED);
  }

  if (scrollwindow != NULL)
    maxheight = MAX (textnumber * textheight + 4, scrollwindow->Getheight ());
  else
    maxheight = textnumber * textheight + 4;

  SizeMe (width, maxheight);

  depth = DefaultDepth (display, DefaultScreen (display));
  pixmap = XCreatePixmap (display, window, width, height, depth);
  XSetForeground (display, gc, buttonpix);
  XFillRectangle (display, pixmap, gc, 0, 0, width, height);
  DrawBorder (pixmap, 0, 0, textwidth + 4, textnumber * textheight + 4, 0, 0, "", LOW);

  SetSelectedInput (ExposureMask);
  FixMySize ();
  MapMe ();
}

SelectWindow::~SelectWindow ()
{
}

void 
SelectWindow::ChangeInfo (SelectWindowInfo * Infos,
			  unsigned short Infonumber)
{
  free (infos);
  infos = Infos;
  infonumber = Infonumber;
  unsigned short counter;
  for (counter = 0; counter < textnumber; counter++) {
    if (counter < infonumber) {
      textboxes[counter]->SetStatus (infos[counter].options);
      textboxes[counter]->ChangeText (infos[counter].Selecttext);
    } else {
      textboxes[counter]->SetStatus (TEXTOPTION_GIVENXPOS);
      textboxes[counter]->ChangeText ("");
    }
  }

  if (scrollwindow != NULL) {
    if (infonumber > textnumber) {
      scrollwindow->SetMinMax (0, (infonumber - textnumber));
      scrollwindow->SetValue (infonumber - textnumber);
    } else {
      scrollwindow->SetMinMax (0, 0);
      scrollwindow->SetValue (0);
    }
  } else {
    if (infonumber > textnumber) {
      short textheight = fontstruct->max_bounds.ascent +
      fontstruct->max_bounds.descent;
      scrollwindow = new ScrollWindow (display,
				       gc,
				       this,
				       fontstruct,
				       0, 2,
			       SCROLL_MINWIDTH, textnumber * textheight + 1,
				       0, infonumber - textnumber,
				       0,
				       SELECTWINDOW_SCROLLED_ACTION);

      for (counter = 0; counter < textnumber; counter++)
	textboxes[counter]->ResizeMe (-scrollwindow->Getwidth (), 0);

      short textwidth = textboxes[0]->Getwidth ();
      scrollwindow->MoveMe (textwidth + SELECTWINDOW_INSIDESPACE, 0);
      scrollwindow->SetValue (infonumber - textnumber);
      scrollwindow->MapMe ();
      XSetForeground (display, gc, buttonpix);
      XFillRectangle (display, pixmap, gc, 0, 0, width, height);
      DrawBorder (pixmap, 0, 0, textwidth + 4, height, 0, 0, "", LOW);
      XCopyArea (display, pixmap, window, gc, 0, 0, width, height, 0, 0);
    }
  }
}

void 
SelectWindow::Action (int Actionnumber, int value)
{
  unsigned short counter;
  switch (Actionnumber) {
  case CHANGETEXTSELECTION_ACTION:
    if (options & SELECTWINDOWOPTION_EXCLUSIVESELECTION) {
      for (counter = 0; counter < infonumber; counter++)
	infos[counter].options &= ~TEXTOPTION_SELECTED;
      for (counter = 0; counter < textnumber; counter++) {
	if (value != (int) (textboxes[counter]))
	  textboxes[counter]->UnselectMe ();
	else {
	  textboxes[counter]->SelectMe ();
	  infos[counter + offset].options |= TEXTOPTION_SELECTED;
	}
      }
    } else {
      for (counter = 0; counter < textnumber; counter++) {
	if (value == (int) (textboxes[counter])) {
	  if (textboxes[counter]->IsSelected ()) {
	    textboxes[counter]->UnselectMe ();
	    infos[counter + offset].options &= ~TEXTOPTION_SELECTED;
	  } else {
	    textboxes[counter]->SelectMe ();
	    infos[counter + offset].options |= TEXTOPTION_SELECTED;
	  }
	}
      }
    }
    Parent->Action (changeactionnumber, value);
    break;

  case TEXT_DOUBLECLICKED:
    Parent->Action (doubleclickactionnumber, value);
    break;

  case SELECTWINDOW_SCROLLED_ACTION:
    if (infonumber > textnumber) {
      offset = (infonumber - textnumber) - value;
      for (counter = 0; counter < textnumber; counter++) {
	textboxes[counter]->SetStatus (infos[counter + offset].options);
	textboxes[counter]->ChangeText (infos[counter + offset].Selecttext);
      }
    } else {
      offset = 0;
      for (counter = 0; counter < infonumber; counter++) {
	textboxes[counter]->SetStatus (infos[counter].options);
	textboxes[counter]->ChangeText (infos[counter].Selecttext);
      }
      for (counter = infonumber; counter < textnumber; counter++) {
	textboxes[counter]->SetStatus (0);
	textboxes[counter]->ChangeText ("");
      }
    }
    break;
  }
}

int 
SelectWindow::GetAcuteInfo ()
{
  unsigned short counter, foundindex = 0;
  for (counter = 0; counter < infonumber; counter++) {
    if (infos[counter].options & TEXTOPTION_SELECTED)
      foundindex = counter;
  }
  return infos[foundindex].info;
}

BVektor *
SelectWindow::Getbvector ()
{
  unsigned short counter;
  BVektor *bvector = new BVektor (infonumber, NULL);
  for (counter = 0; counter < infonumber; counter++) {
    if (infos[counter].options & TEXTOPTION_SELECTED)
      bvector->Setze_i (counter, TRUE);
  }
  return /***/ bvector;
}

void 
SelectWindow::HandleEvent (XEvent * Event)
{
  switch (Event->type) {
  case Expose:
    XCopyArea (display, pixmap, window, gc,
	       Event->xgraphicsexpose.x, Event->xgraphicsexpose.y,
	       Event->xgraphicsexpose.width, Event->xgraphicsexpose.height,
	       Event->xgraphicsexpose.x, Event->xgraphicsexpose.y);
    break;
  }
}


ClusterOptionWindow::ClusterOptionWindow (Display * initdisplay,
		     GC initgc,
		     BigWindow * initParent,
		     XFontStruct * fontstruct,
		     int initx, int inity,
		     char *text,
		     int initactionnumber,
		     ClusterInfo * clusterinfo):
BigWindow (initdisplay,
	   initgc,
	   DefaultRootWindow (initdisplay),
	   fontstruct,
	   initx, inity, 350, 200,
	   text,
	   DEFAULT_BDWIDTH)
{
  char *possitext = "possibilistic";
  actionnumber = initactionnumber;
  SelectWindowInfo *datainfo;
  exactnessselect = NULL;
  dataselectwindow = NULL;
  myclusterinfo = *(clusterinfo);
  Parent = initParent;
  unsigned short counter;
  short leftheight = 0, rightheight = 0;
  short intselectypos;		// pos. of numberoutput

  short clusterselectwidth, clusterselectheight;	// of cluster-select-window

  // first the right elements
  clusternumberselect =
    new IntegerSelectWindow (display,
			     gc,
			     this,
			     fontstruct,
			     0, 0,
			     "Clusterno.",
			     "auto",
			     2, 100, 2, DISPLAY_LIGHTICON);
  if (clusterinfo->clusternumber < MAXINT)
    clusternumberselect->SetValue (clusterinfo->clusternumber);

  repeatnumberselect =
    new IntegerSelectWindow (display,
			     gc,
			     this,
			     fontstruct,
			     0, 0,
			     "repeats",
			     "opt.",
			     1, 1000, 1, DISPLAY_LIGHTICON);
  if (clusterinfo->repeatnumber < MAXINT)
    repeatnumberselect->SetValue (clusterinfo->repeatnumber);
/* //+
  Mselect =
    new IntegerSelectWindow (display,
			     gc,
			     this,
			     fontstruct,
			     0, 0,
			     "M",
			     "100",
			     2.0, 100, clusterinfo->M, 0);

  Kselect =  
    new IntegerSelectWindow (display,
			     gc,
			     this,
			     fontstruct,
			     0, 0,
			     "K",
			     "100",
			     1.0, 100, clusterinfo->K, 0);
			     
  Mselect->SetValue (clusterinfo->M);
  Kselect->SetValue (clusterinfo->K);
*/

  if (clusterinfo->dataset != NULL)
    if (clusterinfo->dataset->Lese_Daten_Typ () == Klassifiziert) {
      exactnessselect =
	new IntegerSelectWindow (display,
				 gc,
				 this,
				 fontstruct,
				 0, 0,
				 "exactness",
				 "100",
				 0, 100, clusterinfo->exactness, 0);
      exactnessselect->SetValue (clusterinfo->exactness);
    }
  intselectypos = SELECTWINDOW_BORDERBESIDETEXT;
  clusternumberselect->PlaceMe (width - clusternumberselect->Getwidth () - SELECTWINDOW_INSIDESPACE, intselectypos);


  intselectypos = clusternumberselect->Gety () +
    clusternumberselect->Getheight () +
    SELECTWINDOW_INSIDESPACE;
  repeatnumberselect->PlaceMe (width - repeatnumberselect->Getwidth () - SELECTWINDOW_INSIDESPACE, intselectypos);

  intselectypos = repeatnumberselect->Gety () +
    repeatnumberselect->Getheight () +
    SELECTWINDOW_INSIDESPACE;

  //+  Mselect->PlaceMe (width - Mselect->Getwidth () - SELECTWINDOW_INSIDESPACE, intselectypos);
  //+  Kselect->PlaceMe (Mselect->Getx () - Kselect->Getwidth () - SELECTWINDOW_INSIDESPACE, intselectypos);

  //+ rightheight = Kselect->Gety () + Kselect->Getheight () + SELECTWINDOW_INSIDESPACE;
  rightheight =SELECTWINDOW_INSIDESPACE;

  if (exactnessselect != NULL) {
/* //+    intselectypos = Mselect->Gety () +
      Mselect->Getheight () +
      SELECTWINDOW_INSIDESPACE;
*/
intselectypos = repeatnumberselect->Gety()+repeatnumberselect->Getheight()+
                SELECTWINDOW_INSIDESPACE;

    exactnessselect->PlaceMe (width - exactnessselect->Getwidth () - SELECTWINDOW_INSIDESPACE, intselectypos);
    rightheight = exactnessselect->Gety () +
      exactnessselect->Getheight () + SELECTWINDOW_INSIDESPACE;
  }
// now the left elements
  clusterselectwidth = width - clusternumberselect->Getwidth () - 2 * SELECTWINDOW_BORDERBESIDETEXT;
  clusterselectheight = 60;

  for (counter = 0; counter < CLUSTERTYPNUMBER; counter++) {
    if ((unsigned short) clusterinfo->clustertyp == Infos[counter].info)
      Infos[counter].options |= TEXTOPTION_SELECTED;
    else
      Infos[counter].options &= ~TEXTOPTION_SELECTED;
  }

  clusterselectwindow
    = new SelectWindow (display,
			gc,
			this,
			fontstruct,
			SELECTWINDOW_BORDERBESIDETEXT,
			SELECTWINDOW_BORDERBESIDETEXT,
			clusterselectwidth, clusterselectheight,
			Infos,
			CLUSTERTYPNUMBER, CHANGESELECTED_ACTION, 0,
			SELECTWINDOWOPTION_EXCLUSIVESELECTION);
  leftheight = clusterselectwindow->Gety () +
    clusterselectwindow->Getheight () +
    SELECTWINDOW_INSIDESPACE;


  for (counter = 0; counter < VALIDYTYPNUMBER; counter++) {
    if ((unsigned short) clusterinfo->validitytyp == ValidityInfos[counter].info)
      ValidityInfos[counter].options |= TEXTOPTION_SELECTED;
    else
      ValidityInfos[counter].options &= ~TEXTOPTION_SELECTED;
  }

  ValidityInfos[counter - 1].options &= ~TEXTOPTION_SELECTABLE;
  if (exactnessselect != NULL)
    ValidityInfos[counter - 1].options |= TEXTOPTION_SELECTABLE;

  validityselectwindow
    = new SelectWindow (display,
			gc,
			this,
			fontstruct,
			SELECTWINDOW_BORDERBESIDETEXT,
			leftheight,
			clusterselectwidth, clusterselectheight,
			ValidityInfos,
			VALIDYTYPNUMBER,
			CHANGESELECTED_ACTION, 0,
			SELECTWINDOWOPTION_EXCLUSIVESELECTION);
  leftheight = validityselectwindow->Gety () +
    validityselectwindow->Getheight () +
    SELECTWINDOW_INSIDESPACE;

  if (clusterinfo->dataset != NULL) {
    datainfo = (SelectWindowInfo *) malloc (sizeof (SelectWindowInfo) * clusterinfo->dataset->Lese_DatenDim ());

    for (counter = 0; counter < clusterinfo->dataset->Lese_DatenDim (); counter++) {
      datainfo[counter].Selecttext = clusterinfo->dataset->Lese_Name_Pointer (counter);
      datainfo[counter].options = TEXTOPTION_SELECTABLE;
      if (clusterinfo->bvector->Lese_i (counter))
	datainfo[counter].options |= TEXTOPTION_SELECTED;
      datainfo[counter].info = counter;
    }

    dataselectwindow
      = new SelectWindow (display,
			  gc,
			  this,
			  fontstruct,
			  SELECTWINDOW_BORDERBESIDETEXT,
			  leftheight,
			  clusterselectwidth, clusterselectheight,
			  datainfo,
			  clusterinfo->dataset->Lese_DatenDim (),
			  CHANGESELECTED_ACTION, 0, 0);

    leftheight += dataselectwindow->Getheight () +
      SELECTWINDOW_INSIDESPACE;
  }
  possilighticon = new OnOffIcon (display,
				  gc,
				  this,
				  fontstruct,
				  SELECTWINDOW_BORDERBESIDETEXT,
				  leftheight,
				  DEFAULTLIGHTSIZE, DEFAULTLIGHTSIZE,
				  NULL, NULL, 0, 0,
				  0,
				  ICON_SELECTABLE | ICON_LIGHTICON);
  possilighticon->MapMe ();
  if (clusterinfo->ispossib)
    possilighticon->AddStatus (ICON_SELECTED);

  leftheight += possilighticon->Getheight () + SELECTWINDOW_INSIDESPACE;

  char *newtext;
//delta_u=clusterinfo->delta_u;
  newtext = (char *) malloc (sizeof (char) * DOUBLELENGTH + 2);
  sprintf (newtext, "%lf", clusterinfo->delta_u);
  char *delta_u_text = "delta_U:";
  short textwidthtextwidth = XTextWidth (fontstruct, delta_u_text, strlen (delta_u_text));
  short textwidth = XTextWidth (fontstruct, newtext, strlen (newtext)) + SELECTWINDOW_INSIDESPACE;
  short textheight = fontstruct->max_bounds.ascent +
  fontstruct->max_bounds.descent;

  delta_u_select = new TextBox (display, gc, this, fontstruct,
		     textwidthtextwidth + 2 * SELECTWINDOW_BORDERBESIDETEXT,
				leftheight,
				textwidth, textheight,
				0, newtext, 0,
				TEXTOPTION_DOUBLEINPUT | TEXTOPTION_DARK,
				DELTA_U_CHANGE_ACTION, 0);

  leftheight += delta_u_select->Getheight () + SELECTWINDOW_INSIDESPACE;

//+ Setzen wir die M-Box neu
  char *M_newtext;
  M_newtext = (char *) malloc (sizeof (char) * DOUBLELENGTH + 2);
  sprintf (M_newtext, "%lf", clusterinfo->M);
  char *M_text = "M:";
  short M_textwidthtextwidth = XTextWidth (fontstruct, M_text, strlen (M_text));
  short M_textwidth = XTextWidth (fontstruct, M_newtext, strlen (M_newtext)) + SELECTWINDOW_INSIDESPACE;
  short M_textheight = fontstruct->max_bounds.ascent +
  fontstruct->max_bounds.descent;

  M_select = new TextBox (display, gc, this, fontstruct,
		     M_textwidthtextwidth + 2 * SELECTWINDOW_BORDERBESIDETEXT,
				leftheight,
				M_textwidth, M_textheight,
				0, M_newtext, 0,
				TEXTOPTION_DOUBLEINPUT | TEXTOPTION_DARK,
				M_CHANGE_ACTION, 0);

  leftheight += M_select->Getheight () + SELECTWINDOW_INSIDESPACE;

//+ Setzen wir die K-Box neu
  char *K_newtext;
  K_newtext = (char *) malloc (sizeof (char) * DOUBLELENGTH + 2);
  sprintf (K_newtext, "%lf", clusterinfo->K);
  char *K_text = "K:";
  short K_textwidthtextwidth = XTextWidth (fontstruct, K_text, strlen (K_text));
  short K_textwidth = XTextWidth (fontstruct, K_newtext, strlen (K_newtext)) + SELECTWINDOW_INSIDESPACE;
  short K_textheight = fontstruct->max_bounds.ascent +
  fontstruct->max_bounds.descent;

  K_select = new TextBox (display, gc, this, fontstruct,
		     K_textwidthtextwidth + 2 * SELECTWINDOW_BORDERBESIDETEXT,
				leftheight,
				K_textwidth, K_textheight,
				0, K_newtext, 0,
				TEXTOPTION_DOUBLEINPUT | TEXTOPTION_DARK,
				K_CHANGE_ACTION, 0);

  leftheight += K_select->Getheight () + SELECTWINDOW_INSIDESPACE;



// now the actionicons at the bottom
  short actioniconypos = MAX (leftheight, rightheight);
  OK = new ActionIcon (display,
		       gc,
		       this,
		       fontstruct, 35, actioniconypos, 40, 30,
		       "Ok", NULL, 0, 0, SELECTED_OK, ICON_SELECTABLE);
  Cancel = new ActionIcon (display,
			   gc,
			   this,
			   fontstruct, 105, actioniconypos, 40, 30,
		    "Cancel", NULL, 0, 0, SELECTED_CANCEL, ICON_SELECTABLE);

// place icons in middle
  {
    short freewidthhalf = (width - 3 * OK->Getwidth ()) / 2;
    OK->PlaceMe (freewidthhalf, OK->Gety ());
    Cancel->PlaceMe (freewidthhalf + 2 * OK->Getwidth (), OK->Gety ());
  }
// now resize window to fit
  short depth;
  SizeMe (width, OK->Gety () + OK->Getheight () + SELECTWINDOW_BORDERBESIDETEXT);


  depth = DefaultDepth (display, DefaultScreen (display));
  pixmap = XCreatePixmap (display, window, width, height, depth);


  XSetForeground (display, gc, buttonpix);
  XFillRectangle (display, pixmap, gc, 0, 0, width, height);
  DrawBorder (pixmap, 0, 0, width, height, 0, 0, "", HIGH);


  XSetForeground (display, gc, blackpix);
  {
    short possx = possilighticon->Getwidth () + possilighticon->Getx () + SELECTWINDOW_INSIDESPACE,
      possy = possilighticon->Gety () + possilighticon->Getheight ();
    XDrawString (display,
		 pixmap,
		 gc,
		 possx, possy,
//            possilighticon->Getwidth()+possilighticon->Getx()+SELECTWINDOW_INSIDESPACE,
    //            possilighticon->Gety()+possilighticon->Getheight(), 
		 possitext, strlen (possitext));

    possarea_left = possx;
    possarea_right = possx + XTextWidth (fontstruct, possitext, strlen (possitext));
    possarea_bottom = possy;
    possarea_top = possy - textheight;
  }
  
  XDrawString (display, pixmap, gc,
	       SELECTWINDOW_BORDERBESIDETEXT,
	       delta_u_select->Gety () + delta_u_select->Getheight () - 3,
	       delta_u_text, strlen (delta_u_text));
    DrawBorder (pixmap, delta_u_select->Getx () - 2, delta_u_select->Gety () - 2,
    delta_u_select->Getwidth () + 2, delta_u_select->Getheight () + 2, 0, 0,
	      "", LOW);

  //+ f"ur M-Box
  XDrawString (display, pixmap, gc,
	       SELECTWINDOW_BORDERBESIDETEXT,
	       + M_select->Gety () + M_select->Getheight () - 3,
	       M_text, strlen (M_text));
    DrawBorder (pixmap, M_select->Getx () - 2, M_select->Gety () - 2,
    M_select->Getwidth () + 2, M_select->Getheight () + 2, 0, 0,
	      "", LOW);

  //+ f"ur K-Box
  XDrawString (display, pixmap, gc,
	       SELECTWINDOW_BORDERBESIDETEXT,
	       + K_select->Gety () + K_select->Getheight () - 3,
	       K_text, strlen (K_text));
    DrawBorder (pixmap, K_select->Getx () - 2, K_select->Gety () - 2,
    K_select->Getwidth () + 2, K_select->Getheight () + 2, 0, 0,
	      "", LOW);

	           
  SetSelectedInput (ExposureMask
		    | EnterWindowMask
		    | LeaveWindowMask
		    | ButtonPressMask
		    | ButtonReleaseMask);

  FixMySize ();
  MapMe ();
  XDefineCursor (display, window, maincursor);
}

ClusterOptionWindow::~ClusterOptionWindow ()
{
  UnmapMe ();
  XFreePixmap (display, pixmap);
  delete OK;
  delete Cancel;
  delete clusternumberselect;
  delete repeatnumberselect;
  //+  delete Mselect;
  //+ delete Kselect;
  if (exactnessselect != NULL)
    delete exactnessselect;
  delete delta_u_select;
  delete M_select; //+
  delete K_select; //+
}

void 
ClusterOptionWindow::Action (int Actionnumber, int value)
{
  switch (Actionnumber) {
  case DELTA_U_CHANGE_ACTION:
    myclusterinfo.delta_u = *((double *) value);
    break;
  case M_CHANGE_ACTION: //+ M-Box
    myclusterinfo.M = *((double *) value);
    break;
  case K_CHANGE_ACTION: //+ K-Box
    myclusterinfo.K = *((double *) value);
    break;

  case SELECTED_OK:
    myclusterinfo.repeatnumber = repeatnumberselect->GetValue ();
    myclusterinfo.clusternumber = clusternumberselect->GetValue ();
//+    myclusterinfo.M = Mselect->GetValue ();
//+    myclusterinfo.K = Kselect->GetValue ();
    if (exactnessselect != NULL)
      myclusterinfo.exactness = exactnessselect->GetValue ();
    myclusterinfo.validitytyp = (Guete_Typ) (validityselectwindow->GetAcuteInfo ());
    myclusterinfo.clustertyp = clusterselectwindow->GetAcuteInfo ();
    myclusterinfo.delta_u = MAX (MIN (myclusterinfo.delta_u, 1.0), 0.0);
    myclusterinfo.M = MAX (MIN (myclusterinfo.M, 100.0), 1.0001); //+
    myclusterinfo.K = MAX (MIN (myclusterinfo.K, 100.0), 1.0); //+
    if (possilighticon->IsSelected ())
      myclusterinfo.ispossib = 1;
    else
      myclusterinfo.ispossib = 0;
    if (myclusterinfo.dataset != NULL)
      myclusterinfo.resultbvector = *dataselectwindow->Getbvector ();
    UnmapMe ();
    XUnmapWindow (display, window);
    Parent->Action (actionnumber, (int) &myclusterinfo);
#if 0
    /* XXX: would you believe this ??? */
    delete this;
#endif
    break;

  case SELECTED_CANCEL:
    UnmapMe ();
    Parent->Action (actionnumber, (int) NULL);
#if 0
    /* XXX: would you believe this ??? */
    delete this;
#endif
    break;
  }
}

void 
ClusterOptionWindow::HandleEvent (XEvent * Event)
{
  switch (Event->type) {
  case Expose:
    XCopyArea (display, pixmap, window, gc,
	       Event->xgraphicsexpose.x, Event->xgraphicsexpose.y,
	       Event->xgraphicsexpose.width, Event->xgraphicsexpose.height,
	       Event->xgraphicsexpose.x, Event->xgraphicsexpose.y);
    break;

  case ButtonPress:
    if ((possarea_left <= Event->xbutton.x) &&
	(possarea_right >= Event->xbutton.x) &&
	(possarea_top <= Event->xbutton.y) &&
	(possarea_bottom >= Event->xbutton.y)) {
      if (possilighticon->IsSelected ())
	possilighticon->UnselectMe ();
      else
	possilighticon->SelectMe ();
    }
    break;
  }
}

RuleOptionWindow::RuleOptionWindow (Display * initdisplay,
		  GC initgc,
		  BigWindow * initParent,
		  XFontStruct * fontstruct,
		  int initx, int inity,
		  char *text,
		  int Actionnumber,
		  RuleInfo * Ruleinfo,
		  RULEOPTIONSTATUS initoption):
BigWindow (initdisplay,
	   initgc,
	   DefaultRootWindow (initdisplay),
	   fontstruct,
	   initx, inity,
	   200, 200,
	   text,
	   DEFAULT_BDWIDTH)
{
  threshold = .01;
  hillheight = DEFAULT_HUEGEL_GRENZE;
  Parent = initParent;
  actionnumber = Actionnumber;
  ruleselectwindow = NULL;
  ruleinfo = Ruleinfo;
  short xpos = SELECTWINDOW_BORDERBESIDETEXT, ypos = SELECTWINDOW_BORDERBESIDETEXT,
    windowwidth = 150, counter, textheight = fontstruct->max_bounds.ascent +
  fontstruct->max_bounds.descent;

  SelectWindowInfo *datainfo;
  if (ruleinfo->dataset != NULL) {
    datainfo = (SelectWindowInfo *) malloc (sizeof (SelectWindowInfo) * ruleinfo->dataset->Lese_DatenDim ());

    for (counter = 0; counter < ruleinfo->dataset->Lese_DatenDim (); counter++) {
      datainfo[counter].Selecttext = ruleinfo->dataset->Lese_Name_Pointer (counter);
      datainfo[counter].options = TEXTOPTION_SELECTABLE;
      if (ruleinfo->selecteddims->Lese_i (counter))
	datainfo[counter].options |= TEXTOPTION_SELECTED;
      datainfo[counter].info = counter;
    }

    dataselectwindow
      = new SelectWindow (display,
			  gc,
			  this,
			  fontstruct,
			  xpos,
			  ypos,
			  windowwidth, 200,
			  datainfo,
			  ruleinfo->dataset->Lese_DatenDim (),
			  CHANGESELECTED_ACTION, 0, 0);

  }
  ypos = dataselectwindow->Gety () +
    dataselectwindow->Getheight () +
    SELECTWINDOW_INSIDESPACE;

  if (initoption == RULEOPTION) {
    RuletypInfos[0].options |= TEXTOPTION_SELECTED;
    for (counter = 1; counter < RULETYPNUMBER; counter++)
      RuletypInfos[counter].options &= ~TEXTOPTION_SELECTED;

    ruleselectwindow
      = new SelectWindow (display,
			  gc,
			  this,
			  fontstruct,
			  SELECTWINDOW_BORDERBESIDETEXT, ypos,
			  windowwidth, 40,
			  RuletypInfos,
			  RULETYPNUMBER,
			  CHANGESELECTED_ACTION, 0,
			  SELECTWINDOWOPTION_EXCLUSIVESELECTION);
    ypos = ruleselectwindow->Gety () +
      ruleselectwindow->Getheight () +
      SELECTWINDOW_INSIDESPACE;
  }
  ParametertypInfos[0].options |= TEXTOPTION_SELECTED;
  for (counter = 1; counter < PARAMETERTYPNUMBER; counter++)
    ParametertypInfos[counter].options &= ~TEXTOPTION_SELECTED;

  parameterselectwindow
    = new SelectWindow (display,
			gc,
			this,
			fontstruct,
			SELECTWINDOW_BORDERBESIDETEXT, ypos,
			windowwidth, 60,
			ParametertypInfos,
			PARAMETERTYPNUMBER,
			CHANGESELECTED_ACTION, 0,
			SELECTWINDOWOPTION_EXCLUSIVESELECTION);

  ypos = parameterselectwindow->Gety () +
    parameterselectwindow->Getheight () +
    SELECTWINDOW_INSIDESPACE;


  char *newtext;
  newtext = (char *) malloc (sizeof (char) * DOUBLELENGTH + 2);
  sprintf (newtext, "%lf", threshold);
  char *thresholdtext = "threshold:";
  short textwidthtextwidth = XTextWidth (fontstruct, thresholdtext, strlen (thresholdtext));
  short textwidth = XTextWidth (fontstruct, newtext, strlen (newtext)) + SELECTWINDOW_INSIDESPACE;
  lowerbound = new TextBox (display, gc, this, fontstruct,
	       textwidthtextwidth + 2 * SELECTWINDOW_BORDERBESIDETEXT, ypos,
			    textwidth, textheight,
			    0, newtext, 0,
			    TEXTOPTION_DOUBLEINPUT | TEXTOPTION_DARK,
			    THRESHOLDCHANGE_ACTION, 0);

  ypos = lowerbound->Gety () +
    lowerbound->Getheight () +
    SELECTWINDOW_INSIDESPACE;

  char *hillvaluetext;
  hillvaluetext = (char *) malloc (sizeof (char) * DOUBLELENGTH + 2);
  sprintf (hillvaluetext, "%lf", hillheight);

  char *hilltext = "hillheight:";
  textwidthtextwidth = XTextWidth (fontstruct, thresholdtext, strlen (hilltext));
  textwidth = XTextWidth (fontstruct, newtext, strlen (newtext)) + SELECTWINDOW_INSIDESPACE;

  hillheightbox = new TextBox (display, gc, this, fontstruct,
	       textwidthtextwidth + 2 * SELECTWINDOW_BORDERBESIDETEXT, ypos,
			       textwidth, textheight,
			       0, hillvaluetext, 0,
			       TEXTOPTION_DOUBLEINPUT | TEXTOPTION_DARK,
			       HILLHEIGHTCHANGE_ACTION, 0);

  ypos = hillheightbox->Gety () +
    hillheightbox->Getheight () +
    SELECTWINDOW_INSIDESPACE;


  OK = new ActionIcon (display,
		       gc,
		       this,
		       fontstruct,
		       35, ypos, 40, 30,
		       "Ok", NULL, 0, 0, SELECTED_OK, ICON_SELECTABLE);
  Cancel = new ActionIcon (display,
			   gc,
			   this,
			   fontstruct,
			   105, ypos, 40, 30,
		    "Cancel", NULL, 0, 0, SELECTED_CANCEL, ICON_SELECTABLE);

// now resize window to fit
  short depth;
  SizeMe (windowwidth + 2 * SELECTWINDOW_BORDERBESIDETEXT,
	  OK->Gety () + OK->Getheight () + SELECTWINDOW_BORDERBESIDETEXT);

  depth = DefaultDepth (display, DefaultScreen (display));
  pixmap = XCreatePixmap (display, window, width, height, depth);


  XSetForeground (display, gc, buttonpix);
  XFillRectangle (display, pixmap, gc, 0, 0, width, height);
  DrawBorder (pixmap, 0, 0, width, height, 0, 0, "", HIGH);
  DrawBorder (pixmap, lowerbound->Getx () - 2, lowerbound->Gety () - 2,
	    lowerbound->Getwidth () + 2, lowerbound->Getheight () + 2, 0, 0,
	      "", LOW);
  DrawBorder (pixmap, hillheightbox->Getx () - 2, hillheightbox->Gety () - 2,
      hillheightbox->Getwidth () + 2, hillheightbox->Getheight () + 2, 0, 0,
	      "", LOW);

  XSetForeground (display, gc, blackpix);
  XDrawString (display, pixmap, gc,
	       SELECTWINDOW_BORDERBESIDETEXT,
	       lowerbound->Gety () + lowerbound->Getheight () - 3,
	       thresholdtext, strlen (thresholdtext));
  XDrawString (display, pixmap, gc,
	       SELECTWINDOW_BORDERBESIDETEXT,
	       hillheightbox->Gety () + hillheightbox->Getheight () - 3,
	       hilltext, strlen (hilltext));

  SetSelectedInput (ExposureMask
		    | EnterWindowMask
		    | LeaveWindowMask
		    | ButtonPressMask
		    | ButtonReleaseMask);

  FixMySize ();
  MapMe ();
}

RuleOptionWindow::~RuleOptionWindow ()
{
  XFreePixmap (display, pixmap);
  delete OK;
  delete Cancel;
  delete hillheightbox;
  delete lowerbound;
  delete dataselectwindow;
  if (ruleselectwindow != NULL)
    delete ruleselectwindow;
  delete parameterselectwindow;
}

void 
RuleOptionWindow::Action (int Actionnumber, int value)
{
  switch (Actionnumber) {
  case THRESHOLDCHANGE_ACTION:
    threshold = *((double *) value);
    threshold = MAX (MIN (threshold, 1.0), 0.0);
    lowerbound->ChangeText (threshold);
    break;

  case HILLHEIGHTCHANGE_ACTION:
    hillheight = *((double *) value);
    hillheight = MAX (MIN (hillheight, 1.0), 0.0);
    hillheightbox->ChangeText (hillheight);
    break;

  case SELECTED_OK:
    if (ruleinfo->selecteddims->Anzahl_gesetzt () > 0) {
      XDefineCursor (display, window, waitcursor);
      XFlush (display);
      if (ruleselectwindow != NULL)	// initoptions==RULEOPTION
       {
	switch (ruleselectwindow->GetAcuteInfo ()) {
	case RULETYP1:
	  ruleinfo->rules = Ermittle_Regeln (*(ruleinfo->clustering),
					     threshold,
					     *(ruleinfo->dataset),
					     ruleinfo->inputdim,
					 *(dataselectwindow->Getbvector ()),
		     (Parameter_Typ) parameterselectwindow->GetAcuteInfo (),
					     *(ruleinfo->selectedcluster));
	  break;

	case RULETYP2A:
	  ruleinfo->rules = Finde_Regeln (hillheight,
					  Dim_insgesamt,
					  *(ruleinfo->clustering),
					  threshold,
					  *(ruleinfo->dataset),
					  ruleinfo->inputdim,
					  *(dataselectwindow->Getbvector ()),
		     (Parameter_Typ) parameterselectwindow->GetAcuteInfo (),
					  *(ruleinfo->selectedcluster));
	  break;

	case RULETYP2B:
	  ruleinfo->rules = Finde_Regeln (hillheight,
					  Dim_einzeln,
					  *(ruleinfo->clustering),
					  threshold,
					  *(ruleinfo->dataset),
					  ruleinfo->inputdim,
					  *(dataselectwindow->Getbvector ()),
		     (Parameter_Typ) parameterselectwindow->GetAcuteInfo (),
					  *(ruleinfo->selectedcluster));
	  break;
	}
      } else {
	ruleinfo->classify = Ermittle_Klassifizierung (*(ruleinfo->clustering),
						       threshold,
						       *(ruleinfo->dataset),
						       ruleinfo->inputdim,
					 *(dataselectwindow->Getbvector ()),
		     (Parameter_Typ) parameterselectwindow->GetAcuteInfo (),
					      *(ruleinfo->selectedcluster));
      }
      XUngrabPointer (display, CurrentTime);
      short rulesfound = FALSE;

      if (ruleselectwindow != NULL)	// initoptions==RULEOPTION
       {
	if (ruleinfo->rules.Lese_Regelanzahl () > 0)
	  rulesfound = TRUE;
      } else {
	if (ruleinfo->classify.Lese_Klassenanzahl () > 0)
	  rulesfound = TRUE;
      }
      if (rulesfound)
	Parent->Action (actionnumber, (int) ruleinfo);
      else {
	Mainwindow->printError ("No rules found!");
	Parent->Action (actionnumber, (int) NULL);
      }
    } else
      Parent->Action (actionnumber, (int) NULL);
    UnmapMe ();
#if 0
    /* XXX: would you believe this ??? */
    delete this;
#endif
    break;

  case SELECTED_CANCEL:
    UnmapMe ();
    Parent->Action (actionnumber, (int) NULL);
#if 0
    /* XXX: would you believe this ??? */
    delete this;
#endif
    break;
  }
}

void 
RuleOptionWindow::HandleEvent (XEvent * Event)
{
  switch (Event->type) {
  case Expose:
    XCopyArea (display, pixmap, window, gc,
	       Event->xgraphicsexpose.x, Event->xgraphicsexpose.y,
	       Event->xgraphicsexpose.width, Event->xgraphicsexpose.height,
	       Event->xgraphicsexpose.x, Event->xgraphicsexpose.y);
    break;
  }
}


IntegerSelectWindow::IntegerSelectWindow (Display * initdisplay,
		     GC initgc,
		     BigWindow * Parent,
		     XFontStruct * fontstruct,
		     int initx, int inity,
		     char *inittext,
		     char *initdefaulttext,
		     int Minvalue, int Maxvalue, int initvalue,
		     unsigned short options):
BigWindow (initdisplay,
	   initgc,
	   Parent->GetWindow (),
	   fontstruct,
	   initx, inity,
	   10, 10,
	   inittext, 0)
{
  minvalue = Minvalue;
  maxvalue = Maxvalue;
  value = initvalue;
  defaulttext = initdefaulttext;
  lighticon = NULL;
  short integerwidth = XTextWidth (fontstruct, defaulttext, strlen (defaulttext)) + 10;
  short integerboxxpos, integerboxypos;

  short textwidth = XTextWidth (fontstruct, text, strlen (text));
  short textxpos, textypos;
  short depth, integerheight = fontstruct->max_bounds.ascent +
  fontstruct->max_bounds.descent;
  short arrowsize = 11, arrowupxpos = 10, arrowupypos = 0, arrowdownypos;
  short lightxpos = 0, lightypos;

  integerboxypos = arrowupypos + arrowsize + 2 * DEFAULT_BDWIDTH + 2;
  lightypos = integerboxypos +
    (int) (.5 * (integerheight - DEFAULTLIGHTSIZE)) +
    DEFAULT_BDWIDTH;

  textxpos = lightxpos + DEFAULTLIGHTSIZE + SELECTWINDOW_INSIDESPACE;
  textypos = integerboxypos + integerheight - fontstruct->max_bounds.descent + DEFAULT_BDWIDTH;
  integerboxxpos = textxpos + textwidth + SELECTWINDOW_INSIDESPACE;

  integerbox = new TextBox (display, gc, this, fontstruct,
			    integerboxxpos, integerboxypos,
			    integerwidth, integerheight,
			    0, defaulttext, 1,
			    TEXTOPTION_DARK,
			    INTEGERSELECTWINDOW_VALUECHANGE,
			    INTEGERSELECTWINDOW_NOOP);


  arrowdownypos = integerboxypos + integerheight + 4;
  arrowupxpos = integerboxxpos + (int) (.5 * (integerwidth - arrowsize));

  arrowup = new ActionIcon (display,
			    gc,
			    this,
			    fontstruct,
			    arrowupxpos, arrowupypos, arrowsize, arrowsize,
			    "", littlearrowup_bits,
			    littlearrowup_width, littlearrowup_height,
			    INTEGERSELECTWINDOW_COUNTUP,
			    ICON_SELECTABLE | ICON_REPEATEDCLICK);

  arrowdown = new ActionIcon (display,
			      gc,
			      this,
			      fontstruct,
			   arrowupxpos, arrowdownypos, arrowsize, arrowsize,
			      "", littlearrowdown_bits,
			      littlearrowdown_width, littlearrowdown_height,
			      INTEGERSELECTWINDOW_COUNTDOWN,
			      ICON_SELECTABLE | ICON_REPEATEDCLICK);

  lighticon = new OnOffIcon (display,
			     gc,
			     this,
			     fontstruct,
			     lightxpos, lightypos,
			     DEFAULTLIGHTSIZE, DEFAULTLIGHTSIZE,
			     NULL, NULL, 0, 0,
			     INTEGERSELECTWINDOW_LIGHTCHANGE,
			     ICON_SELECTABLE | ICON_LIGHTICON);

  if (options & DISPLAY_LIGHTICON)
    lighticon->MapMe ();

  width = integerboxxpos + integerwidth + 4;
  height = arrowdownypos + arrowsize + 2 * DEFAULT_BDWIDTH;
  XResizeWindow (display, window, width, height);

  depth = DefaultDepth (display, DefaultScreen (display));
  pixmap = XCreatePixmap (display, window, width, height, depth);
  XSetForeground (display, gc, buttonpix);
  XFillRectangle (display, pixmap, gc, 0, 0, width, height);

  XSetForeground (display, gc, blackpix);
  XDrawString (display, pixmap, gc,
	       textxpos, textypos, text, strlen (text));

  clickarea_left = textxpos;
  clickarea_right = textxpos + XTextWidth (fontstruct, text, strlen (text));
  clickarea_bottom = textypos;
  clickarea_top = textypos - integerheight;

  DrawBorder (pixmap, integerboxxpos - 2, integerboxypos - 2,
	      integerbox->Getwidth () + 6, integerbox->Getheight () + 6,
	      0, 0, "", LOW);

  if (options & DISPLAY_LIGHTICON)
    SetSelectedInput (ExposureMask | ButtonPressMask);
  else
    SetSelectedInput (ExposureMask);

  MapMe ();
}

void 
IntegerSelectWindow::SetDefault ()
{
  lighticon->DelStatus (ICON_SELECTED);
  integerbox->ChangeText (defaulttext);
}

void 
IntegerSelectWindow::SetValue (int newvalue)
{
  value = newvalue;
  lighticon->AddStatus (ICON_SELECTED);

  integerbox->SetStatus (integerbox->GetStatus () | TEXTOPTION_INTINPUT);
  integerbox->ChangeText (value);
}

int 
IntegerSelectWindow::GetValue ()
{
  if (lighticon->IsSelected ())
    return (MAX (MIN (value, maxvalue), minvalue));
  else
    return (MAXINT);
}

void 
IntegerSelectWindow::Action (int Actionnumber, int Value)
{
  switch (Actionnumber) {
  case INTEGERSELECTWINDOW_COUNTUP:
    if (lighticon->IsSelected () && (value < maxvalue))
      integerbox->ChangeText (++value);
    break;

  case INTEGERSELECTWINDOW_COUNTDOWN:
    if (lighticon->IsSelected () && (value > minvalue))
      integerbox->ChangeText (--value);
    break;
  case INTEGERSELECTWINDOW_LIGHTCHANGE:
    if (Value == SWITCHON) {
      integerbox->SetStatus (integerbox->GetStatus () | TEXTOPTION_INTINPUT);
      integerbox->ChangeText (value);
    } else {
      integerbox->SetStatus (integerbox->GetStatus () & ~TEXTOPTION_INTINPUT);
      integerbox->ChangeText (defaulttext);
    }
    break;
  case INTEGERSELECTWINDOW_VALUECHANGE:
    value = *((int *) Value);
    break;
  }
}

void 
IntegerSelectWindow::HandleEvent (XEvent * Event)
{
  switch (Event->type) {
  case Expose:
    XCopyArea (display, pixmap, window, gc,
	       Event->xgraphicsexpose.x, Event->xgraphicsexpose.y,
	       Event->xgraphicsexpose.width, Event->xgraphicsexpose.height,
	       Event->xgraphicsexpose.x, Event->xgraphicsexpose.y);
    break;
  case ButtonPress:
    if (lighticon->IsMapped () &&
	(clickarea_left <= Event->xbutton.x) &&
	(clickarea_right >= Event->xbutton.x) &&
	(clickarea_top <= Event->xbutton.y) &&
	(clickarea_bottom >= Event->xbutton.y)) {
      if (lighticon->IsSelected ()) {
	lighticon->UnselectMe ();
	Action (INTEGERSELECTWINDOW_LIGHTCHANGE, SWITCHOFF);
      } else {
	lighticon->SelectMe ();
	Action (INTEGERSELECTWINDOW_LIGHTCHANGE, SWITCHON);
      }
    }
    break;
  }
}
