/********************************************************/
/*   filename: vektor.cc                                */
/*                                                      */
/********************************************************/
/* programmed by: Oliver Wagner                         */
/* last change: 26-04-95                                */
/********************************************************/
#include "vektor.h"

/********************************************************/
/*              Alles zur Klasse BVektor                */
/********************************************************/

BVektor::BVektor (int Groesse, char *Inhalt)
{
  int i;

  n = Groesse;
  gesetzt = 0;
  if (n != 0) {
    if ((Daten = (char *) malloc (sizeof (char) * ((n / 8) + 1))) == NULL)
        Fehlermeldung ("BVektor::BVektor()", SPEICHERFEHLER);
    if ((gesetzte = (int *) malloc (sizeof (int) * n)) == NULL)
        Fehlermeldung ("BVektor::BVektor()", SPEICHERFEHLER);

    if (Inhalt != NULL)
      for (i = 0; (i < n) && (Inhalt != NULL); i++)
	Setze_i (i, Inhalt[i]);
    else
      for (i = 0; i < ((n / 8) + 1); i++)
	Daten[i] = 0;
  } else {
    Daten = NULL;
    gesetzte = 0;
  }
};

BVektor::BVektor (const BVektor & Das)
{
  int i;

  n = Das.n;
  gesetzt = Das.gesetzt;

  if (n != 0) {
    if ((Daten = (char *) malloc (sizeof (char) * ((n / 8) + 1))) == NULL)
        Fehlermeldung ("BVektor::BVektor(const BVektor& Das)", SPEICHERFEHLER);
    if ((gesetzte = (int *) malloc (sizeof (int) * n)) == NULL)
        Fehlermeldung ("BVektor::BVektor(const BVektor& Das)", SPEICHERFEHLER);
    for (i = 0; i < (n / 8) + 1; i++) {
      Daten[i] = Das.Daten[i];
      gesetzte[i] = Das.gesetzte[i];
    }
    for (; i < n; i++)		/* Der Rest */
      gesetzte[i] = Das.gesetzte[i];
  } else {
    Daten = NULL;
    gesetzte = NULL;
  }

};

BVektor::~BVektor ()
{
  if (n != 0) {
    n = 0;
    free (Daten);
    free (gesetzte);
    Daten = NULL;
    gesetzte = NULL;
  }
};

void 
BVektor::Setze_Dim (int Damit)
{
  int i;

  if (n != Damit) {
    if (n != 0) {
      free (Daten);
      free (gesetzte);
    }
    if (Damit != 0) {
      if ((Daten = (char *) malloc (sizeof (char) * ((Damit / 8) + 1))) == NULL)
	  Fehlermeldung ("BVektor::Setze_Dim()", SPEICHERFEHLER);
      if ((gesetzte = (int *) malloc (sizeof (int) * Damit)) == NULL)
	  Fehlermeldung ("BVektor::Setze_Dim()", SPEICHERFEHLER);
      for (i = 0; i < (Damit / 8) + 1; i++)
	Daten[i] = 0;
    } else {
      Daten = NULL;
      gesetzte = NULL;
    }
    n = Damit;
    gesetzt = 0;
  }
};

void 
BVektor::print ()
{
  int i;

  printf ("\nBVektor :");
  for (i = 0; i < n; i++)
    if (Lese_i (i))
      printf ("1 ");
    else
      printf ("0 ");
  printf ("gesetze:");
  for (i = 0; i < gesetzt; i++)
    printf ("%d ", gesetzte[i]);
  printf ("\n");
};

void 
BVektor::Setze_i (int Pos, char x)
{
  int i;

  if (x) {
    if (!(Daten[Pos / 8] & (1 << (Pos % 8)))) {
      gesetzt++;		/* neu gesetzt */
      Daten[Pos / 8] |= (1 << (Pos % 8));

      /* jetzt den Index an die passende Stelle */
      if (gesetzt == 1)		/* Der erste liegt immer am Anfang */
	gesetzte[0] = Pos;
      else if (Pos < gesetzte[0]) {
	for (i = gesetzt - 1; i > 0; i--)
	  gesetzte[i] = gesetzte[i - 1];
	gesetzte[0] = Pos;
      } else			/* nicht am Anfang */
	for (i = gesetzt - 1; i > 0; i--) {
	  if (Pos < gesetzte[i - 1]) {	/* Positionen tauschen */
	    gesetzte[i] = gesetzte[i - 1];
	  } else {		/* jetzt setzen */
	    gesetzte[i] = Pos;
	    break;
	  }
	}
    }
  } else {
    if (Daten[Pos / 8] & (1 << (Pos % 8))) {
      gesetzt--;		/* war vorher gesetzt */
      Daten[Pos / 8] &= ((1 << (Pos % 8)) ^ 255);
      /* jetzt den Index wieder rauspulen */
      if (gesetzte != 0) {	/* sonst nichts zu tun */
	for (i = 0; gesetzte[i] != Pos; i++);
	Pos = i;		/* Dort ist Pos */
	for (i = Pos; i < gesetzt; i++) {	/* Positionen tauschen */
	  gesetzte[i] = gesetzte[i + 1];
	}
      }
    }
  }
};

char 
BVektor::Lese_i (int Pos)
{
  return ((TRUE) && ((Daten[(Pos / 8)]) & (1 << (Pos % 8))));
};

void 
BVektor::Speichern (FILE * File)
{
  int i;

  fprintf (File, "%d %d ", n, gesetzt);
  for (i = 0; i <= (n / 8); i++)
    fprintf (File, "%d ", ((int) Daten[i]));
  for (i = 0; i < gesetzt; i++)
    fprintf (File, "%d ", gesetzte[i]);
  fprintf (File, "\n");
};

int 
BVektor::Laden (FILE * File)
{
  int i, Buffer;

  if (!(fscanf (File, "%d %d", &i, &Buffer))) {
#if 0
    this. ~ BVektor ();
#endif
    return (0);
  }
  Setze_Dim (i);
  gesetzt = Buffer;
  for (i = 0; i <= (n / 8); i++) {
    if (!(fscanf (File, "%d", &Buffer))) {
#if 0
      this. ~ BVektor ();
#endif
      return (0);
    }
    Daten[i] = (char) Buffer;
  }
  for (i = 0; i < gesetzt; i++)
    if (!(fscanf (File, "%d ", &(gesetzte[i])))) {
#if 0
      this. ~ BVektor ();
#endif
      return (0);
    }
  return (1);
};

BVektor 
BVektor::concat (BVektor & Dahinter)
return Result;
{
  int i, j = 0;

  Result.n = n + Dahinter.n;

  if (Result.n != 0) {
    if ((Result.Daten = (char *) malloc (sizeof (char) * ((Result.n / 8) + 1))) == NULL)
        Fehlermeldung ("BVektor::concat()", SPEICHERFEHLER);
    if ((Result.gesetzte = (int *) malloc (sizeof (int) * Result.n)) == NULL)
        Fehlermeldung ("BVektor::concat()", SPEICHERFEHLER);
    for (i = 0; i < n; i++)
      Result.Setze_i (j++, (*this)[i]);
    for (i = 0; i < Dahinter.n; i++)
      Result.Setze_i (j++, Dahinter[i]);
    /* Jetzt die Indizes aneinanderhaengen */
    j = 0;
    for (i = 0; i < gesetzt; i++)
      Result.gesetzte[j++] = gesetzte[i];
    for (i = 0; i < Dahinter.gesetzt; i++)
      Result.gesetzte[j++] = Dahinter.gesetzte[i] + n;
  } else {
    Result.Daten = NULL;
    Result.gesetzt = 0;
    Result.gesetzte = NULL;
  }
};

BVektor 
BVektor::operator& (BVektor & A)
return Result (*this);
{
  int i, j, l = 0;
  char Maske;

  Result.gesetzt = 0;
  if (Result.n != 0) {
    for (i = 0; i < (n / 8); i++) {
      Result.Daten[i] &= A.Daten[i];
      Maske = 1;
      for (j = 0; j < 8; j++, l++) {
	if (Maske & Result.Daten[i])
	  Result.gesetzte[Result.gesetzt++] = l;
	Maske <<= 1;
      }
    }
    Maske = 1;
    Result.Daten[i] &= A.Daten[i];
    for (j = 0; j < (n % 8); j++, l++) {
      if (Maske & Result.Daten[i])
	Result.gesetzte[Result.gesetzt++] = l;
      Maske <<= 1;
    }
  }
};

void 
BVektor::operator= (const BVektor & Das)
{
  int i;

  Setze_Dim (Das.n);
  if (n != 0) {
    for (i = 0; i < (n / 8) + 1; i++)
      Daten[i] = Das.Daten[i];
    for (i = 0; i < n; i++)
      gesetzte[i] = Das.gesetzte[i];
  }
  gesetzt = Das.gesetzt;
};

void 
BVektor::operator= (char Damit)
{
  int i;

  if (n != 0) {
    if (Damit == 0) {
      gesetzt = 0;
      for (i = 0; i < (n / 8) + 1; i++)
	Daten[i] = (char) 0;
    } else {
      gesetzt = n;
      for (i = 0; i < (n / 8) + 1; i++)
	Daten[i] = (char) 255;
      for (i = 0; i < n; i++)
	gesetzte[i] = i;
    }
  }
};

/********************************************************/
/*              Alles zur Klasse IVektor                */
/********************************************************/
IVektor::IVektor (int Groesse, int *Inhalt)
{
  int i;

  n = Groesse;
  if (n != 0) {
    if ((Daten = (int *) malloc (sizeof (int) * n)) == NULL)
        Fehlermeldung ("IVektor::IVektor()", SPEICHERFEHLER);

    if (Inhalt != NULL)
      for (i = 0; i < n; i++)
	Daten[i] = Inhalt[i];
    else
      for (i = 0; i < n; i++)
	Daten[i] = 0;
  } else
    Daten = NULL;
};

IVektor::IVektor (const IVektor & Das)
{
  int i;

  n = Das.n;
  if (n != 0) {
    if ((Daten = (int *) malloc (sizeof (int) * n)) == NULL)
        Fehlermeldung ("IVektor::IVektor(const IVektor& Das)", SPEICHERFEHLER);

    for (i = 0; i < n; i++)
      Daten[i] = Das.Daten[i];
  } else
    Daten = NULL;
};

IVektor::~IVektor ()
{
  if (n != 0) {
    n = 0;
    free (Daten);
    Daten = NULL;
  }
};

void 
IVektor::Setze_Dim (int Damit)
{
  int i;

  if (n != Damit) {
    if (n != 0)
      free (Daten);
    n = Damit;
    if (n != 0) {
      if ((Daten = (int *) malloc (sizeof (int) * n)) == NULL)
	  Fehlermeldung ("IVektor::Setze_Dim(int Damit)", SPEICHERFEHLER);
      else {
	for (i = 0; i < n; i++)
	  Daten[i] = 0;
      }
    } else
      Daten = NULL;
  }
};

char 
IVektor::Ist_drinnen (int Das)
{
  int i;

  for (i = 0; i < n; i++)
    if (Daten[i] == Das)
      return (TRUE);
  return (FALSE);
};

int 
IVektor::Lese_Index (int Davon)
{
  int i;

  for (i = 0; i < n; i++)
    if (Daten[i] == Davon)
      return (i);
  return (-1);
};

void 
IVektor::print ()
{
  int i;

  printf ("\nIVektor :");
  for (i = 0; i < n; i++)
    printf ("%d ", Daten[i]);
  printf ("\n");
};

double 
IVektor::Vektorlaenge ()
{
  int i;
  double Result = 0;

  for (i = 0; i < n; i++) {
    Result += Daten[i] * Daten[i];
  }
  return (sqrt (Result));
};

void 
IVektor::Tausche (int den, int mitdem)
{
  int Buffer;

  Buffer = Daten[den];
  Daten[den] = Daten[mitdem];
  Daten[mitdem] = Buffer;
};

IVektor 
IVektor::Projiziere (BVektor & Diese)
return Result (Diese.Anzahl_gesetzt (), NULL);
{
  int i, j;

  for (i = 0, j = 0; i < n; i++)
    if (Diese.Lese_i (i) == TRUE)
      Result.Daten[j++] = Daten[i];
};

IVektor 
IVektor::Projiziere (IVektor & Diese)
return Result (Diese.Lese_Dim (), NULL);
{
  int i;

  for (i = 0; i < n; i++)
    Result[i] = Daten[Diese[i]];
};

IVektor 
IVektor::Projiziere (int ab)
  return Result (n - ab, NULL);
{
  int i;

  for (i = 0; i + ab < n; i++)
    Result.Daten[i] = Daten[i + ab];
};

void 
IVektor::Speichern (FILE * File)
{
  int i;

  fprintf (File, "%d ", n);
  for (i = 0; i < n; i++)
    fprintf (File, "%d ", Daten[i]);
  fprintf (File, "\n");
};

int 
IVektor::Laden (FILE * File)
{
  int Dim;

  if (!(fscanf (File, "%d", &Dim))) {
#if 0
    this. ~ IVektor ();
#endif
    return (0);
  }
  Setze_Dim (Dim);
  for (Dim = 0; Dim < n; Dim++) {
    if (!(fscanf (File, "%d", &Daten[Dim]))) {
#if 0
      this. ~ IVektor ();
#endif
      return (0);
    }
  }
  return (1);
};

void 
IVektor::Tausche (IVektor & Damit)
{
  int *iBuffer;
  int i;

  iBuffer = Damit.Daten;
  Damit.Daten = Daten;
  Daten = iBuffer;
  i = Damit.n;
  Damit.n = n;
  n = i;
};

void 
IVektor::Plus_Sortiert (int Das)
{
  int *Buffer, i, j = 0;

  Buffer = Daten;
  if ((Daten = (int *) malloc (sizeof (int) * (n + 1))) == NULL)
      Fehlermeldung ("IVektor::Plus_Sortiert(int Das)", SPEICHERFEHLER);

  for (i = 0; i < n; i++) {
    if (Das > Buffer[j])
      Daten[i] = Buffer[j++];
    else {
      Daten[i++] = Das;
      for (; i <= n; i++)
	Daten[i] = Buffer[j++];
      break;
    }
  }
  n++;
  if (i < n)
    Daten[i] = Das;

  free (Buffer);
};

void 
IVektor::operator-= (IVektor & Minus)
{
  int i;

  for (i = 0; i < n; i++)
    Daten[i] = Daten[i] - Minus.Daten[i];
};

void 
IVektor::operator+= (IVektor & Plus)
{
  int i;

  for (i = 0; i < n; i++)
    Daten[i] = Daten[i] + Plus.Daten[i];
};

void 
IVektor::operator+= (int Das)
{
  int *Buffer, i;

  Buffer = Daten;
  if ((Daten = (int *) malloc (sizeof (int) * (n + 1))) == NULL)
      Fehlermeldung ("IVektor::operator+=(int Das)", SPEICHERFEHLER);
  for (i = 0; i < n; i++)
    Daten[i] = Buffer[i];
  Daten[i] = Das;
  free (Buffer);
  n++;
};

int 
IVektor::operator* (IVektor & Damit)
{
  int Result = 0, i;

  for (i = 0; i < n; i++)
    Result += Daten[i] * Damit.Daten[i];
  return (Result);
};

IVektor 
IVektor::operator- (IVektor & Das) return Result;
{
  int i;

  Result.n = Das.n;
  if (Das.n != 0) {
    if ((Result.Daten = (int *) malloc (sizeof (int) * Das.n)) == NULL)
        Fehlermeldung ("IVektor::operator -", SPEICHERFEHLER);
    for (i = 0; i < Das.n; i++)
      Result.Daten[i] = Daten[i] - Das.Daten[i];
  } else
    Result.Daten = NULL;
};

void 
IVektor::operator= (const IVektor & Das)
{
  int i;

  Setze_Dim (Das.n);
  for (i = 0; i < n; i++)
    Daten[i] = Das.Daten[i];
};

void 
IVektor::operator= (int Das)
{
  int i;

  for (i = 0; i < n; i++)
    Daten[i] = Das;

};
/********************************************************/
/*              Alles zur Klasse DVektor                */
/********************************************************/
DVektor::DVektor (int Groesse, int Nr, double *Inhalt)
{
  int i;

  n = Groesse;
  Nummer = Nr;

  if (n != 0) {
    if ((Daten = (double *) malloc (sizeof (double) * n)) == NULL)
        Fehlermeldung ("DVektor::DVektor()", SPEICHERFEHLER);

    if (Inhalt != NULL)
      for (i = 0; i < n; i++)
	Daten[i] = Inhalt[i];
    else
      for (i = 0; i < n; i++)
	Daten[i] = 0;
  } else
    Daten = NULL;
};

DVektor::~DVektor ()
{
  if (n != 0) {
    n = 0;
    free (Daten);
    Daten = NULL;
  }
};

DVektor::DVektor (const DVektor & Daraus)
{
  int i;

  n = Daraus.n;
  Nummer = Daraus.Nummer;

  if (n != 0) {
    if ((Daten = (double *) malloc (sizeof (double) * n)) == NULL)
        Fehlermeldung ("DVektor::DVektor(DVektor& )", SPEICHERFEHLER);

    for (i = 0; i < n; i++)
      Daten[i] = Daraus.Daten[i];
  } else
    Daten = NULL;
};

void 
DVektor::Setze_Dim (int Damit)
{
  int i;

  if (n != Damit) {
    if (n != 0)
      free (Daten);
    n = Damit;
    if (Damit != 0) {
      if ((Daten = (double *) malloc (sizeof (double) * Damit)) == NULL)
	  Fehlermeldung ("DVektor::Setze_Dim()", SPEICHERFEHLER);
      for (i = 0; i < Damit; i++)
	Daten[i] = 0;
    } else
      Daten = NULL;
  }
};

void 
DVektor::NullVektor ()
{
  int i;

  for (i = 0; i < n; i++)
    Daten[i] = 0;
};

void 
DVektor::print ()
{
  int i;

  printf ("\nDVektor %d:", Nummer);
  for (i = 0; i < n; i++)
    printf ("%lf ", Daten[i]);
  printf ("\n");
};

double 
DVektor::Vektorlaenge ()
{
  int i;
  double Result = 0;

  for (i = 0; i < n; i++) {
    Result += Daten[i] * Daten[i];
  }
  return (sqrt (Result));
};

void 
DVektor::Speichern (FILE * File)
{
  int i;

  fprintf (File, "%d %d ", Nummer, n);
  for (i = 0; i < n; i++)
    fprintf (File, "%lf ", Daten[i]);
  fprintf (File, "\n");
};

int 
DVektor::Laden (FILE * File)
{
  int Dim;

  if (!(fscanf (File, "%d %d ", &Nummer, &Dim))) {
#if 0
    this. ~ DVektor ();
#endif
    return (0);
  }
  Setze_Dim (Dim);
  for (Dim = 0; Dim < n; Dim++) {
    if (!(fscanf (File, "%lf ", &Daten[Dim]))) {
#if 0
      this. ~ DVektor ();
#endif
      return (0);
    }
  }
  return (1);
};

DVektor 
DVektor::Deprojiziere (BVektor & Diese)
return Result (Diese.Lese_Dim (), Nummer, NULL);
{
  int i, k = 0;

  for (i = 0; i < Diese.Lese_Dim (); i++) {
    if (Diese.Lese_i (i))
      Result.Daten[i] = Daten[k++];
    else
      Result.Daten[i] = 0;
  }
  Result.Nummer = Nummer;
};

DVektor 
DVektor::Projiziere (BVektor & Diese)
return Result (Diese.Anzahl_gesetzt (), Nummer, NULL);
{
  int i, j;

  for (i = 0, j = 0; i < n; i++) {
    if (Diese.Lese_i (i) == TRUE)
      Result.Daten[j++] = Daten[i];
  }
  Result.Nummer = Nummer;
};

DVektor 
DVektor::Projiziere (IVektor & Diese)
return Result (Diese.Lese_Dim (), Nummer, NULL);
{
  int i;

  for (i = 0; i < n; i++)
    Result[i] = Daten[Diese[i]];
  Result.Nummer = Nummer;
};

DVektor 
DVektor::Projiziere (int ab)
  return Result (n - ab, Nummer, NULL);
{
  int i;

  for (i = 0; i + ab < n; i++)
    Result.Daten[i] = Daten[i + ab];
  Result.Nummer = Nummer;
};

DVektor 
DVektor::Projiziere (int ab, int bis)
  return Result ((bis - ab) + 1, Nummer, NULL);
{
  int i;

  for (i = 0; i + ab <= bis; i++)
    Result.Daten[i] = Daten[i + ab];
  Result.Nummer = Nummer;
};

void 
DVektor::VektorT_Mult (DVektor & Dieser, DMatrix & Damit)
{
  int i, j;
  double *Buffer;

  if (Damit.Lese_Dim_n () != 0) {
    if ((Buffer = (double *) malloc (sizeof (double) * Damit.Lese_Dim_n ())) == NULL)
        Fehlermeldung ("void DVektor::VektorT_Mult()", SPEICHERFEHLER);
  } else
    Buffer = NULL;

  n = Damit.Lese_Dim_n ();
  if (n != 0)
    free (Daten);
  Daten = Buffer;

  for (j = 0; j < n; j++) {
    Buffer[j] = 0;
    for (i = 0; i < Damit.Lese_Dim_m (); i++)
      Buffer[j] += Damit.Lese_i_j (i, j) * Dieser.Lese_i (i);
  }

};

DVektor 
DVektor::x (DVektor & B)
return Result (n, 0, NULL);
{
  int i;

  for (i = 0; i < n; i++)
    Result[i] = Daten[(i + 1) % n] * B[(i + 2) % n] - Daten[(i - 1) % n] * B[(i - 2) % n];
};

DMatrix 
DVektor::Produkt (DVektor & B)
return Result (n, B.n);
{
  int i, j;

  for (i = 0; i < n; i++)
    for (j = 0; j < B.n; j++)
      Result.Setze_i_j (i, j, Daten[i] * B.Daten[j]);
};

DMatrix 
DVektor::Produkt ()
return Result (n, n);
{
  int i, j;

  for (i = 0; i < n; i++) {
    for (j = 0; j < i; j++) {
      Result[i][j] = Result[j][i] = Daten[i] * Daten[j];
    }
    Result[i][i] = Daten[i] * Daten[i];
  }
};

void 
DVektor::Normiere ()
{
  int i;
  double Laenge = 0;

  for (i = 0; i < n; i++)
    Laenge += Daten[i] * Daten[i];
  Laenge = sqrt (Laenge);
  for (i = 0; i < n; i++)
    Daten[i] = Daten[i] / Laenge;
}

double 
DVektor::Abstand (DVektor & Davon)
{
  int i;
  double Result = 0;

  for (i = 0; i < n; i++)
    Result += (Davon.Daten[i] - Daten[i]) * (Davon.Daten[i] - Daten[i]);

  return (sqrt (Result));
};

double 
DVektor::Abstand (DVektor & Davon, BVektor & relevant)
{
  int i;
  double Result = 0;

  for (i = 0; i < relevant.Anzahl_gesetzt (); i++)
    Result += (Davon.Daten[relevant.Lese_gesetzte (i)] - Daten[relevant.Lese_gesetzte (i)]) *
      (Davon.Daten[relevant.Lese_gesetzte (i)] - Daten[relevant.Lese_gesetzte (i)]);

  return (sqrt (Result));
};

double 
DVektor::cos_Winkel (DVektor & Dazu)
{
  double Result = 0, LaengeA = 0, LaengeB = 0;
  int i;

  if (n > 1) {
    for (i = 0; i < n; i++) {	/* Laengen bestimmen */
      LaengeA += Daten[i] * Daten[i];
      LaengeB += Dazu.Daten[i] * Dazu.Daten[i];
    }
    LaengeA = sqrt (LaengeA * LaengeB);

    for (i = 0; i < n; i++)
      Result += Daten[i] * Dazu.Daten[i];
    Result /= LaengeA;
    return (Result);
  } else {
    if (((Dazu.Daten[0] >= 0) && (Daten[0] >= 0)) ||
	((Dazu.Daten[0] < 0) && (Daten[0] < 0)))	/* gleiches vorzeichen */
      return (1);
    else
      return (0);
  }
};


void 
DVektor::Quicksort (int a, int b, IVektor & Indizes)
{
  int r, i, j, iBuffer;
  double x, DBuffer;

  if (b > a) {
    i = a;
    j = b;
    /* Waehle r einmal in der Mitte */
    r = (a + b) / 2;
    x = Daten[r];
    do {
      while (Daten[i] < x)
	i++;
      while (Daten[j] > x)
	j--;
      if (i <= j) {
	DBuffer = Daten[i];
	iBuffer = Indizes[i];

	Daten[i] = Daten[j];
	Indizes[i] = Indizes[j];

	Daten[j] = DBuffer;
	Indizes[j] = iBuffer;

	i++;
	j--;
      }
    } while (j >= i);
    Quicksort (a, j, Indizes);
    Quicksort (i, b, Indizes);
  }
};


void 
DVektor::Quicksort (int a, int b, int Index)
{
  int r, i, j, NrBuffer;
  double x, *DBuffer;

  if (b > a) {
    i = a;
    j = b;
    /* Waehle r einmal in der Mitte */
    r = (a + b) / 2;
    x = this[r].Daten[Index];
    do {
      while (this[i].Daten[Index] < x)
	i++;
      while (this[j].Daten[Index] > x)
	j--;
      if (i <= j) {
	DBuffer = this[i].Daten;
	NrBuffer = this[i].Nummer;

	this[i].Daten = this[j].Daten;
	this[i].Nummer = this[j].Nummer;

	this[j].Daten = DBuffer;
	this[j].Nummer = NrBuffer;
	i++;
	j--;
      }
    } while (j >= i);
    Quicksort (a, j, Index);
    Quicksort (i, b, Index);
  }
};

void 
DVektor::Tausche (DVektor & Damit)
{
  double *dBuffer;
  int i;

  dBuffer = Damit.Daten;
  Damit.Daten = Daten;
  Daten = dBuffer;
  i = Damit.Nummer;
  Damit.Nummer = Nummer;
  Nummer = i;
  i = Damit.n;
  Damit.n = n;
  n = i;
};

double DVektor::operator () (double x) {
  double Result = 0;
  int i;

  if (n != 0)
    Result = Daten[n - 1];
  for (i = n - 2; i >= 0; i--)
    Result = Result * x + Daten[i];
  return (Result);
};

void 
DVektor::operator= (const DVektor & A)
{
  int i;

  Setze_Dim (A.n);
  for (i = 0; i < n; i++)
    Daten[i] = A.Daten[i];
  Nummer = A.Nummer;
};

void 
DVektor::operator= (double Das)
{
  int i;

  for (i = 0; i < n; i++)
    Daten[i] = Das;
};

char 
DVektor::operator== (DVektor & Der)
{
  int i;

  if ((Der.n != n) || (Der.Nummer != Nummer))
    return (FALSE);

  for (i = 0; i < n; i++) {
    if (Daten[i] != Der.Daten[i])
      return (FALSE);
  }
  return (TRUE);
};

char 
DVektor::operator!= (DVektor & Der)
{
  int i;

  if ((Der.n != n) || (Der.Nummer != Nummer))
    return (TRUE);

  for (i = 0; i < n; i++) {
    if (Daten[i] != Der.Daten[i])
      return (TRUE);
  }
  return (FALSE);
};

void 
DVektor::operator*= (double Damit)
{
  int i;

  for (i = 0; i < n; i++)
    Daten[i] *= Damit;
};

void 
DVektor::operator+= (DVektor & Plus)
{
  int i;

  Setze_Dim (Plus.n);
  for (i = 0; i < n; i++)
    Daten[i] += Plus.Daten[i];
};

void 
DVektor::operator-= (DVektor & Minus)
{
  int i;

  Setze_Dim (Minus.n);
  for (i = 0; i < n; i++)
    Daten[i] -= Minus.Daten[i];
};

void 
DVektor::operator<<= (int soviel)
{
  *this >>= (-soviel);
};

void 
DVektor::operator>>= (int soviel)
{
  int i, j;
  double Buffer;

  if (soviel < 0)
    soviel = n + (soviel % n);

  Buffer = Daten[0];
  for (i = 0, j = soviel % n; j != 0; i = j, j = (i + soviel) % n)
    Daten[i] = Daten[j];
  Daten[i] = Buffer;
};

void 
DVektor::operator/= (double Divisor)
{
  int i;

  for (i = 0; i < n; i++)
    Daten[i] /= Divisor;
};

DVektor 
DVektor::operator* (double Skalar)
  return Result (n, 0, NULL);
{
  int i;

  for (i = 0; i < n; i++)
    Result.Daten[i] = Daten[i] * Skalar;
};

DVektor 
DVektor::operator* (DMatrix & Die)
return Result;
{
  int i, j;

  if (Die.Lese_Dim_n () != 0) {
    if ((Result.Daten = (double *) malloc (sizeof (double) * Die.Lese_Dim_n ())) == NULL)
        Fehlermeldung ("void DVektor::operator*(DMatrix& Die)", SPEICHERFEHLER);
  } else
    Result.Daten = NULL;

  Result.n = Die.Lese_Dim_n ();
  Result.Nummer = 0;

  for (j = 0; j < Result.n; j++) {
    Result.Daten[j] = 0;
    for (i = 0; i < Die.Lese_Dim_m (); i++)
      Result.Daten[j] += Die.Lese_i_j (i, j) * Lese_i (i);
  }
};

double 
DVektor::operator* (DVektor & Das)
{
  double Result = 0;
  int i;

  for (i = 0; i < n; i++)
    Result += Daten[i] * Das.Daten[i];

  return (Result);
};

DVektor 
DVektor::operator+ (DVektor & Das)
return Result (n, 0, NULL);
{
  int i;

  for (i = 0; i < n; i++)
    Result.Daten[i] = Daten[i] + Das.Daten[i];
};

DVektor 
DVektor::operator- (DVektor & Das)
return Result (n, 0, NULL);
{
  int i;

  for (i = 0; i < n; i++)
    Result.Daten[i] = Daten[i] - Das.Daten[i];
};

/********************************************************/
/*              Alles zur Klasse BVektorArray           */
/********************************************************/
BVektorArray::BVektorArray (int VektorDim, int ArrayGroesse,
	      BVektor * Diese)
{
  int i;

  Groesse = ArrayGroesse;
  VektorDimension = VektorDim;

  if (Groesse != 0) {
    if ((Daten = (BVektor **) malloc (sizeof (BVektor *) * Groesse)) == NULL)
      Fehlermeldung ("BVektorArray::BVektorArray()", SPEICHERFEHLER);

    for (i = 0; i < Groesse; i++)
      Daten[i] = new BVektor (VektorDimension, NULL);

    if (Diese != NULL)
      for (i = 0; i < Groesse; i++)
	*(Daten[i]) = Diese[i];
  } else
    Daten = NULL;
};

BVektorArray::BVektorArray (const BVektorArray & Daraus)
{
  int i;

  Groesse = Daraus.Groesse;
  VektorDimension = Daraus.VektorDimension;

  if (Groesse != 0) {
    if ((Daten = (BVektor **) malloc (sizeof (BVektor *) * Groesse)) == NULL)
      Fehlermeldung ("BVektorArray::BVektorArray(const BVektorArray& Daraus)", SPEICHERFEHLER);

    for (i = 0; i < Groesse; i++)
      Daten[i] = new BVektor (*(Daraus.Daten[i]));
  } else
    Daten = NULL;
};

BVektorArray::~BVektorArray ()
{
  VektorDimension = 0;

  int i;

  if (Groesse != 0) {
    for (i = 0; i < Groesse; i++) {
      delete Daten[i];
    }
    free (Daten);
    Daten = NULL;
    Groesse = 0;
  }
};

void 
BVektorArray::print ()
{
  int i;

  for (i = 0; i < Groesse; i++)
    Daten[i]->print ();
};

void 
BVektorArray::Setze_Dim (int Damit)
{
  int i;

  if (Damit != VektorDimension) {
    VektorDimension = Damit;
    for (i = 0; i < Groesse; i++)
      Daten[i]->Setze_Dim (Damit);
  }
};

void 
BVektorArray::Setze_Groesse (int Damit)
{
  int i;
  BVektor **Buffer;

  if (Damit != Groesse) {
    for (i = Damit; i < Groesse; i++)
      delete Daten[i];		/* Ueberfl"ussige L"oschen */
    Buffer = Daten;
    if (Damit != 0) {
      if ((Daten = (BVektor **) malloc (sizeof (BVektor *) * Damit)) == NULL)
	Fehlermeldung ("BVektorArray:Setze_Groesse()", SPEICHERFEHLER);
      for (i = 0; (i < Groesse) && (i < Damit); i++) {	/* Alte "ubernehmen */
	Daten[i] = Buffer[i];
	*(Daten[i]) = 0;;
      }
      for (; i < Damit; i++)
	Daten[i] = new BVektor (VektorDimension, NULL);
    } else
      Daten = NULL;
    Groesse = Damit;
    free (Buffer);
  }
}

void 
BVektorArray::Loesche (int von, int bis)
{
  if ((bis >= von) && (von >= 0) && (bis < Groesse)) {
    int i, j;
    BVektor **DatenBuffer;

    if ((Groesse - (von - bis + 1)) != 0) {
      if ((DatenBuffer = (BVektor **) malloc (sizeof (BVektor *) * (Groesse - (von - bis + 1)))) == NULL)
	Fehlermeldung ("void BVektorArray::Loesche...", SPEICHERFEHLER);
    } else
      DatenBuffer = 0;
    for (j = 0, i = 0; i < von; i++)
      DatenBuffer[j++] = Daten[i];
    for (i = von; i <= bis; i++)
      delete Daten[i];
    for (; i < Groesse; i++)
      DatenBuffer[j++] = Daten[i];

    Groesse = Groesse - (von - bis + 1);
    free (Daten);
    Daten = DatenBuffer;
  }
};

void 
BVektorArray::operator+= (BVektor & Damit)
{
  int i;
  BVektor **DatenBuffer;

  if ((VektorDimension != Damit.Lese_Dim ()) &&
      (Groesse != 0))
    Fehlermeldung ("BVektorArray::operator+=", DIMFEHLER);

  VektorDimension = Damit.Lese_Dim ();

  if ((DatenBuffer = (BVektor **) malloc (sizeof (BVektor *) * (Groesse + 1))) == NULL)
    Fehlermeldung ("void BVektorArray::operator+=(BVektor& Damit)", SPEICHERFEHLER);

  for (i = 0; i < Groesse; i++)
    DatenBuffer[i] = Daten[i];
  DatenBuffer[Groesse] = new BVektor (Damit);

  Groesse++;
  free (Daten);
  Daten = DatenBuffer;
};

void 
BVektorArray::operator-= (BVektor & Den)
{
  int i, j, der = -1;
  BVektor **DatenBuffer;

  for (i = 0; i < Groesse; i++) {	/* den passenden Suchen */
    if (Daten[i] == &Den) {
      der = i;
      delete Daten[i];
      break;
    }
  }
  if ((Groesse != 0) && (der != -1)) {
    if (Groesse != 1) {
      if ((DatenBuffer = (BVektor **) malloc (sizeof (BVektor *) * (Groesse - 1))) == NULL)
	Fehlermeldung ("void BVektorArray::operator-=(BVektor& Damit)", SPEICHERFEHLER);
    } else
      DatenBuffer = 0;
    for (i = 0, j = 0; i < Groesse; i++) {
      if (i != der)
	DatenBuffer[j] = Daten[i];
    }
    Groesse--;
    free (Daten);
    Daten = DatenBuffer;
  }
};

BVektorArray & BVektorArray::operator++ (int) {
  int i;
  BVektor **DatenBuffer;

  if ((DatenBuffer = (BVektor **) malloc (sizeof (BVektor *) * (Groesse + 1))) == NULL)
    Fehlermeldung ("void BVektorArray::operator+=(BVektor& Damit)", SPEICHERFEHLER);

  for (i = 0; i < Groesse; i++)
    DatenBuffer[i] = Daten[i];
  DatenBuffer[Groesse] = new BVektor (VektorDimension, NULL);

  Groesse++;
  free (Daten);
  Daten = DatenBuffer;
};


void 
BVektorArray::operator= (char Das)
{
  int i;

  for (i = 0; i < Groesse; i++)
    *(Daten[i]) = Das;
};

void 
BVektorArray::operator= (const BVektorArray & Das)
{
  int i;

  if ((Groesse != Das.Groesse) || (VektorDimension != Das.VektorDimension)) {
#if 0
    this. ~ BVektorArray ();
#else
    this-> ~ BVektorArray ();
#endif
    Groesse = Das.Groesse;
    VektorDimension = Das.VektorDimension;
    if (Groesse != 0) {
      if ((Daten = (BVektor **) malloc (sizeof (BVektor *) * Groesse)) == NULL)
	Fehlermeldung ("void IBektorArray::operator=(BVektorArray *Das)", SPEICHERFEHLER);
    } else
      Daten = NULL;

    for (i = 0; i < Groesse; i++)
      Daten[i] = new BVektor (*(Das.Daten[i]));
  } else {
    for (i = 0; i < Groesse; i++)
      (*(Daten[i])) = *(Das.Daten[i]);
  }
};

/********************************************************/
/*              Alles zur Klasse IVektorArray           */
/********************************************************/
IVektorArray::IVektorArray (int VektorDim, int ArrayGroesse,
	      IVektor * Diese)
{
  int i;

  Groesse = ArrayGroesse;
  VektorDimension = VektorDim;

  if (Groesse != 0) {
    if ((Daten = (IVektor **) malloc (sizeof (IVektor *) * Groesse)) == NULL)
      Fehlermeldung ("IVektorArray::IVektorArray()", SPEICHERFEHLER);

    for (i = 0; i < Groesse; i++)
      Daten[i] = new IVektor (VektorDimension, NULL);

    if (Diese != NULL)
      for (i = 0; i < Groesse; i++)
	Setze_Vektor (i, Diese[i]);
  } else
    Daten = NULL;
};

IVektorArray::IVektorArray (const IVektorArray & Daraus)
{
  int i;

  Groesse = Daraus.Groesse;
  VektorDimension = Daraus.VektorDimension;

  if (Groesse != 0) {
    if ((Daten = (IVektor **) malloc (sizeof (IVektor *) * Groesse)) == NULL)
      Fehlermeldung ("IVektorArray::IVektorArray(const IVektorArray& Daraus)", SPEICHERFEHLER);

    for (i = 0; i < Groesse; i++)
      Daten[i] = new IVektor (*(Daraus.Daten[i]));
  } else
    Daten = NULL;
};

IVektorArray::~IVektorArray ()
{
  VektorDimension = 0;

  int i;

  if (Groesse != 0) {
    for (i = 0; i < Groesse; i++) {
      delete Daten[i];
    }
    free (Daten);
    Daten = NULL;
    Groesse = 0;
  }
};

void 
IVektorArray::Setze_Dim (int Damit)
{
  int i;

  if (Damit != VektorDimension) {
    VektorDimension = Damit;
    for (i = 0; i < Groesse; i++)
      Daten[i]->Setze_Dim (Damit);
  }
};

void 
IVektorArray::Setze_Groesse (int Damit)
{
  int i;
  IVektor **Buffer;

  if (Damit != Groesse) {
    for (i = Damit; i < Groesse; i++)
      delete Daten[i];		/* Ueberfl"ussige L"oschen */
    Buffer = Daten;
    if (Damit != 0) {
      if ((Daten = (IVektor **) malloc (sizeof (IVektor *) * Damit)) == NULL)
	Fehlermeldung ("IVktorArray:Setze_Groesse()", SPEICHERFEHLER);
      for (i = 0; (i < Groesse) && (i < Damit); i++) {	/* Alte "ubernehmen */
	Daten[i] = Buffer[i];
	*(Daten[i]) = 0;;
      }
      for (; i < Damit; i++)
	Daten[i] = new IVektor (VektorDimension, NULL);
    } else
      Daten = NULL;
    Groesse = Damit;
    free (Buffer);
  }
}

void 
IVektorArray::operator+= (IVektor & Damit)
{
  int i;
  IVektor **DatenBuffer;

  if ((VektorDimension != Damit.Lese_Dim ()) &&
      (Groesse != 0))
    Fehlermeldung ("IVektorArray::operator+=", DIMFEHLER);

  VektorDimension = Damit.Lese_Dim ();

  if ((DatenBuffer = (IVektor **) malloc (sizeof (IVektor *) * (Groesse + 1))) == NULL)
    Fehlermeldung ("void IVektorArray::operator+=(IVektor& Damit)", SPEICHERFEHLER);

  for (i = 0; i < Groesse; i++)
    DatenBuffer[i] = Daten[i];
  DatenBuffer[Groesse] = new IVektor (Damit);

  Groesse++;
  free (Daten);
  Daten = DatenBuffer;
};

void 
IVektorArray::operator-= (IVektor & Den)
{
  int i, j, der = -1;
  IVektor **DatenBuffer;

  for (i = 0; i < Groesse; i++) {	/* den passenden Suchen */
    if (Daten[i] == &Den) {
      der = i;
      delete Daten[i];
      break;
    }
  }
  if ((Groesse != 0) && (der != -1)) {
    if (Groesse != 1) {
      if ((DatenBuffer = (IVektor **) malloc (sizeof (IVektor *) * (Groesse - 1))) == NULL)
	Fehlermeldung ("void IVektorArray::operator+=(IVektor& Damit)", SPEICHERFEHLER);
    } else
      DatenBuffer = 0;
    for (i = 0, j = 0; i < Groesse; i++) {
      if (i != der)
	DatenBuffer[j] = Daten[i];
    }
    Groesse--;
    free (Daten);
    Daten = DatenBuffer;
  }
};

char 
IVektorArray::Ist_vorhanden (IVektor & Soeiner)
{
  int i, j;
  char Result;

  for (i = 0; i < Groesse; i++) {
    Result = TRUE;
    for (j = 0; j < VektorDimension; j++) {
      if (Soeiner.Lese_i (j) != Daten[i]->Lese_i (j)) {		/* nicht gleich */
	Result = FALSE;
	break;
      }
    }
    if (Result == TRUE)
      return (TRUE);
  }
  return (FALSE);
};

void 
IVektorArray::operator= (const IVektorArray & Das)
{
  int i;

  if ((Groesse != Das.Groesse) || (VektorDimension != Das.VektorDimension)) {
#if 0
    this. ~ IVektorArray ();
#else
    this-> ~ IVektorArray ();
#endif
    Groesse = Das.Groesse;
    VektorDimension = Das.VektorDimension;
    if (Groesse != 0) {
      if ((Daten = (IVektor **) malloc (sizeof (IVektor *) * Groesse)) == NULL)
	Fehlermeldung ("void IVektorArray::operator=(IVektorArray *Das)", SPEICHERFEHLER);
    } else
      Daten = NULL;

    for (i = 0; i < Groesse; i++)
      Daten[i] = new IVektor (*(Das.Daten[i]));
  } else {
    for (i = 0; i < Groesse; i++)
      (*(Daten[i])) = *(Das.Daten[i]);
  }
};

/********************************************************/
/*              Alles zur Klasse DVektorArray           */
/********************************************************/
DVektorArray::DVektorArray (int VektorDim, int ArrayGroesse,
	      DVektor * Diese)
{
  int i;

  Groesse = ArrayGroesse;
  VektorDimension = VektorDim;

  Durchschnitt.DVektor (VektorDim, 0, NULL);
  Max.DVektor (VektorDim, 0, NULL);
  Min.DVektor (VektorDim, 0, NULL);
  MaxNummer.IVektor (VektorDim, NULL);
  MinNummer.IVektor (VektorDim, NULL);

  for (i = 0; i < VektorDim; i++) {
    Max[i] = -MAXDOUBLE;
    Min[i] = MAXDOUBLE;
  }


  if (Groesse != 0) {
    if ((Daten = (DVektor **) malloc (sizeof (DVektor *) * Groesse)) == NULL)
      Fehlermeldung ("DVektorArray::DVektorArray()", SPEICHERFEHLER);

    for (i = 0; i < Groesse; i++)
      Daten[i] = new DVektor (VektorDimension, i, NULL);

    if (Diese != NULL)
      for (i = 0; i < Groesse; i++)
	Setze_Vektor (i, Diese[i].Lese_Daten ());
  } else
    Daten = NULL;
};

DVektorArray::DVektorArray (const DVektorArray & Daraus)
{
  int i;

  Groesse = Daraus.Groesse;
  VektorDimension = Daraus.VektorDimension;

  if (Groesse != 0) {
    if ((Daten = (DVektor **) malloc (sizeof (DVektor *) * Groesse)) == NULL)
      Fehlermeldung ("DVektorArray::DVektorArray(const DVektorArray& Daraus)", SPEICHERFEHLER);
    for (i = 0; i < Groesse; i++)
      Daten[i] = new DVektor (*(Daraus.Daten[i]));
  } else
    Daten = NULL;

  Min = (DVektor &) Daraus.Min;
  Max = (DVektor &) Daraus.Max;
  MinNummer = (IVektor &) Daraus.MinNummer;
  MaxNummer = (IVektor &) Daraus.MaxNummer;
  Durchschnitt = (DVektor &) Daraus.Durchschnitt;

};

DVektorArray::~DVektorArray ()
{
  VektorDimension = 0;

  int i;

  if (Groesse != 0) {
    for (i = 0; i < Groesse; i++) {
      delete Daten[i];
    }
    free (Daten);
    Daten = NULL;
    Groesse = 0;
  }
};

void 
DVektorArray::Setze_Dim (int Damit)
{
  int i;

  if (Damit != VektorDimension) {
    VektorDimension = Damit;
    for (i = 0; i < Groesse; i++)
      Daten[i]->Setze_Dim (Damit);
    Durchschnitt.Setze_Dim (Damit);
    Min.Setze_Dim (Damit);
    Max.Setze_Dim (Damit);
    MinNummer.Setze_Dim (Damit);
    MaxNummer.Setze_Dim (Damit);
  }
};

void 
DVektorArray::Setze_Groesse (int Damit)
{
  int i;
  DVektor **Buffer;

  if (Damit != Groesse) {
    for (i = Damit; i < Groesse; i++)
      delete Daten[i];		/* Ueberfl"ussige L"oschen */
    Buffer = Daten;
    if (Damit != 0) {
      if ((Daten = (DVektor **) malloc (sizeof (DVektor *) * Damit)) == NULL)
	Fehlermeldung ("DVktorArray:Setze_Groesse()", SPEICHERFEHLER);
      for (i = 0; (i < Groesse) && (i < Damit); i++) {	/* Alte "ubernehmen */
	Daten[i] = Buffer[i];
	Daten[i]->NullVektor ();
      }
      for (; i < Damit; i++)
	Daten[i] = new DVektor (VektorDimension, i, NULL);
    } else
      Daten = NULL;
    Groesse = Damit;
    Durchschnitt = 0;
    Min = 0;
    Max = 0;
    MinNummer = 0;
    MaxNummer = 0;
    free (Buffer);
  }
}

void 
DVektorArray::NullVektor ()
{
  int i;

  for (i = 0; i < Groesse; i++)
    Daten[i]->NullVektor ();
};

DVektor 
DVektorArray::Zufallsvektor ()
return Result (VektorDimension, 0, NULL);
{
  int i;

  if (Groesse != 0) {
    for (i = 0; i < VektorDimension; i++)
      Result[i] = (double) (Min[i] +
		  ((double) rand () / (double) MAXINT) * (Max[i] - Min[i]));
  }
};

int 
DVektorArray::Lese_Index (DVektor & Der)
{
  int i;

  for (i = 0; i < Groesse; i++) {
    if (Daten[i] == &Der)
      return (i);
  }

  return (-1);
};

void 
DVektorArray::Speichern (FILE * File)
{
  int i;

  fprintf (File, "%d %d\n", Groesse, VektorDimension);
  Durchschnitt.Speichern (File);
  Min.Speichern (File);
  Max.Speichern (File);
  MinNummer.Speichern (File);
  MaxNummer.Speichern (File);
  for (i = 0; i < Groesse; i++)
    Daten[i]->Speichern (File);
};

int 
DVektorArray::Laden (FILE * File)
{
  int x, y;

  if (!(fscanf (File, "%d %d ", &x, &y)))
    return (0);
  Setze_Groesse (x);
  Setze_Dim (y);
  if ((!Durchschnitt.Laden (File)) || (!Min.Laden (File)) ||
      (!Max.Laden (File)) || (!MinNummer.Laden (File)) ||
      (!MaxNummer.Laden (File))) {
#if 0
    this. ~ DVektorArray ();
#endif
    return (0);
  }
  for (x = 0; x < Groesse; x++) {
    if (!(Daten[x]->Laden (File))) {
#if 0
      this. ~ DVektorArray ();
#endif
      return (0);
    }
  }
  return (1);
};

DVektorArray 
DVektorArray::Deprojiziere (BVektor & Diese)
return Result (Diese.Lese_Dim (), Groesse, NULL);
{
  int i;

  for (i = 0; i < Groesse; i++) {
    *(Result.Daten[i]) = Daten[i]->Deprojiziere (Diese);
  }
};

DVektorArray 
DVektorArray::Projiziere (BVektor & Diese)
return Result;
{
  int i;

  if (Groesse != 0) {
    if ((Result.Daten = (DVektor **) malloc (sizeof (DVektor *) * Groesse)) == NULL)
      Fehlermeldung ("DVektorArray::Projiziere", SPEICHERFEHLER);
  } else
    Result.Daten = NULL;

  Result.Groesse = Groesse;
  Result.VektorDimension = Diese.Anzahl_gesetzt ();
  for (i = 0; i < Groesse; i++) {
    Result.Daten[i] = new DVektor (Diese.Anzahl_gesetzt (),
				   Daten[i]->Lese_Nummer (), NULL);
    (*(Result.Daten[i])) = Daten[i]->Projiziere (Diese);
    Result.Daten[i]->Setze_Nummer (Daten[i]->Lese_Nummer ());
  }

  Result.Min = Min.Projiziere (Diese);
  Result.Max = Max.Projiziere (Diese);
  Result.MinNummer = MinNummer.Projiziere (Diese);
  Result.MaxNummer = MaxNummer.Projiziere (Diese);
  Result.Durchschnitt = Durchschnitt.Projiziere (Diese);
};

DVektorArray 
DVektorArray::Projiziere (IVektor & Diese)
return Result;
{
  int i;

  if (Groesse != 0) {
    if ((Result.Daten = (DVektor **) malloc (sizeof (DVektor *) * Groesse)) == NULL)
      Fehlermeldung ("DVektorArray::Projiziere", SPEICHERFEHLER);
  } else
    Result.Daten = NULL;

  Result.Groesse = Groesse;
  Result.VektorDimension = Diese.Lese_Dim ();
  for (i = 0; i < Groesse; i++) {
    Result.Daten[i] = new DVektor (Diese.Lese_Dim (),
				   Daten[i]->Lese_Nummer (), NULL);
    (*(Result.Daten[i])) = Daten[i]->Projiziere (Diese);
    Result.Daten[i]->Setze_Nummer (Daten[i]->Lese_Nummer ());
  }

  Result.Min = Min.Projiziere (Diese);
  Result.Max = Max.Projiziere (Diese);
  Result.MinNummer = MinNummer.Projiziere (Diese);
  Result.MaxNummer = MaxNummer.Projiziere (Diese);
  Result.Durchschnitt = Durchschnitt.Projiziere (Diese);
};

DVektorArray 
DVektorArray::Projiziere (int ab)
  return Result;
{
  int i;

  if (Groesse != 0) {
    if ((Result.Daten = (DVektor **) malloc (sizeof (DVektor *) * Groesse)) == NULL)
      Fehlermeldung ("DVektorArray::Projiziere", SPEICHERFEHLER);
  } else
    Result.Daten = NULL;

  for (i = 0; i < Groesse; i++) {
    Result.Daten[i] = new DVektor (0, Daten[i]->Lese_Nummer (), NULL);
    (*(Result.Daten[i])) = Daten[i]->Projiziere (ab);
    Result.Daten[i]->Setze_Nummer (Daten[i]->Lese_Nummer ());
  }

  Result.Groesse = Groesse;
  Result.VektorDimension = Daten[i]->Lese_Dim ();

  Result.Min = Min.Projiziere (ab);
  Result.Max = Max.Projiziere (ab);
  Result.MinNummer = MinNummer.Projiziere (ab);
  Result.MaxNummer = MaxNummer.Projiziere (ab);
  Result.Durchschnitt = Durchschnitt.Projiziere (ab);
};

DVektorArray 
DVektorArray::Quicksort (int Index)
  return Result;
{
  int i, j, MinIndex = 0, MaxIndex = 0;
  double Minwert, Maxwert;

  Result = *this;
  Result.Quicksort (0, Groesse - 1, Index);

/* Neue Maxima und Minima bestimmen */
  for (i = 0; i < VektorDimension; i++) {
    Minwert = MAXDOUBLE;
    Maxwert = -MAXDOUBLE;
    for (j = 0; j < Groesse; j++) {
      if (Result.Daten[j]->Lese_i (i) < Minwert) {
	MinIndex = j;
	Minwert = Result.Daten[j]->Lese_i (i);
      }
      if (Result.Daten[j]->Lese_i (i) > Maxwert) {
	MaxIndex = j;
	Maxwert = Daten[j]->Lese_i (i);
      }
    }
    Result.Max.Setze_i (i, Maxwert);
    Result.MaxNummer.Setze_i (i, MaxIndex);
    Result.Min.Setze_i (i, Minwert);
    Result.MinNummer.Setze_i (i, MinIndex);
  }
};

void 
DVektorArray::Setze_Vektor (int i, DVektor & Damit)
{
  Setze_Vektor (i, Damit.Lese_Daten ());
  Daten[i]->Setze_Nummer (Damit.Lese_Nummer ());
};

void 
DVektorArray::Setze_Vektor (int i, double *Damit)
{
  int x;

  for (x = 0; x < VektorDimension; x++)
    Setze_Vektor (i, x, Damit[x]);
};

void 
DVektorArray::Setze_Vektor (int Vektor, int Index, double x)
{
  int i;

/*****************************************/
  /*     neuen Durchschnitt feststellen    */
/*****************************************/
  Durchschnitt[Index] = ((Lese_Durchschnitt (Index) * Lese_Groesse ()) +
		       x - Daten[Vektor]->Lese_i (Index)) / Lese_Groesse ();

/*****************************************/
  /* neues Maximum und Minimum feststellen */
/*****************************************/
  if ((Max.Lese_i (Index) == -MAXDOUBLE) && (Min.Lese_i (Index) == MAXDOUBLE)) {
    /* noch uninitialisiert */
    Max[Index] = x;
    Min[Index] = x;
  } else if (x > Max[Index]) {
    MaxNummer[Index] = Vektor;
    Max[Index] = x;
  } else if (x < Min[Index]) {
    MinNummer[Index] = Vektor;
    Min[Index] = x;
  } else if ((*Daten[Vektor])[Index] == Max[Index]) {	/* War das Max ? */
    if (x < Max[Index]) {	/* neues Max aus allen suchen */
      Max[Index] = -MAXDOUBLE;
      for (i = 0; i < Groesse; i++)
	if ((*Daten[i])[Index] > Max[Index]) {
	  MaxNummer[Index] = i;
	  Max[Index] = (*Daten[i])[Index];
	}
    }
  } else if ((*Daten[Vektor])[Index] == Min[Index]) {	/* War das Min ? */
    if (x > Min[Index]) {	/* neues Max aus allen suchen */
      Min[Index] = MAXDOUBLE;
      for (i = 0; i < Groesse; i++)
	if ((*Daten[i])[Index] < Min[Index]) {
	  MinNummer[Index] = i;
	  Min[Index] = (*Daten[i])[Index];
	}
    }
  }
/*****************************************/
  /*            neuen Wert einsetzen       */
/*****************************************/
  (*Daten[Vektor])[Index] = x;
};

void 
DVektorArray::Aktualisiere_Min_etc ()
{
  int i, DatenNr;

  Durchschnitt.Setze_Dim (VektorDimension);
  Min.Setze_Dim (VektorDimension);
  Max.Setze_Dim (VektorDimension);
  MinNummer.Setze_Dim (VektorDimension);
  MaxNummer.Setze_Dim (VektorDimension);
  MinNummer = 0;
  MaxNummer = 0;

  if (Groesse != 0) {		/* es gibt was zu tun */
    Durchschnitt = (*Daten[0]);
    Max = *Daten[0];
    Min = *Daten[0];
    for (DatenNr = 1; DatenNr < Groesse; DatenNr++) {
      Durchschnitt += (*Daten[DatenNr]);
      for (i = 0; i < VektorDimension; i++) {
	if ((*Daten[DatenNr])[i] > Max[i]) {
	  Max[i] = (*Daten[DatenNr])[i];
	  MaxNummer[i] = DatenNr;
	}
	if ((*Daten[DatenNr])[i] < Min[i]) {
	  Min[i] = (*Daten[DatenNr])[i];
	  MinNummer[i] = DatenNr;
	}
      }
    }
    Durchschnitt /= Groesse;
  }
};

void 
DVektorArray::Quicksort (int a, int b, int Index)
{
  int r, i, j;
  double x;
  DVektor *VBuffer;

  if (b > a) {
    i = a;
    j = b;
    /* Waehle r einmal in der Mitte */
    r = (a + b) / 2;
    x = Daten[r]->Lese_i (Index);
    do {
      while (Daten[i]->Lese_i (Index) < x)
	i++;
      while (Daten[j]->Lese_i (Index) > x)
	j--;
      if (i <= j) {
	VBuffer = Daten[i];
	Daten[i] = Daten[j];
	Daten[j] = VBuffer;
	i++;
	j--;
      }
    } while (j >= i);
    Quicksort (a, j, Index);
    Quicksort (i, b, Index);
  }
};

void 
DVektorArray::print ()
{
  int i;

  for (i = 0; i < Groesse; i++)
#if 0
    Daten[i].print ();
#else
    Daten[i]->print ();
#endif
};

void 
DVektorArray::print (int Anfang, int Ende)
{
  int i;

  if ((Anfang <= Ende) && (Anfang >= 0) && (Ende < Groesse))
    for (i = Anfang; i <= Ende; i++)
      Daten[i]->print ();
};

void 
DVektorArray::Tausche (DVektorArray & Damit)
{
  DVektor **dBuffer;
  int i;

  dBuffer = Damit.Daten;
  Damit.Daten = Daten;
  Daten = dBuffer;
  i = Damit.VektorDimension;
  Damit.VektorDimension = VektorDimension;
  VektorDimension = i;
  i = Damit.Groesse;
  Damit.Groesse = Groesse;
  Groesse = i;

  Durchschnitt.Tausche (Damit.Durchschnitt);
  Min.Tausche (Damit.Min);
  Max.Tausche (Damit.Max);
  MinNummer.Tausche (Damit.MinNummer);
  MaxNummer.Tausche (Damit.MaxNummer);
};

void 
DVektorArray::neuer_Zufallspunkt ()
{
  int i;
  DVektor Buffer (VektorDimension, 0, NULL);

  for (i = 0; i < VektorDimension; i++) {
    Buffer[i] = Lese_Min (i) +
      ((double) rand () / (double) MAXINT) * (Lese_Max (i) - Lese_Min (i));
  }
  (*this) += Buffer;
};

DVektorArray 
DVektorArray::Mache_Konvex ()
return Result;
{
  Result = this.Mache_Konvex (1, 0);
};

DVektorArray 
DVektorArray::Mache_Konvex (int Index, double Untergrenze)
return Result;
{
  int i, j, k, Mitte;
  double Hoehe;
  BVektor Diese (Groesse, NULL), Loescher (Groesse, NULL);

  /* von links bis zum Maximum */
  Hoehe = 0;
  Mitte = 0;

  for (i = 0; i < Groesse; i++) {
    if ((Daten[i]->Lese_i (Index) >= Hoehe) &&
	(Daten[i]->Lese_i (Index) >= Untergrenze)) {
      Diese.Setze_i (i, TRUE);
      Hoehe = Daten[i]->Lese_i (Index);
      Mitte = i;
    }
  }
  /* von rechts bis zum Maximum */
  Hoehe = 0;
  for (i = Groesse - 1; i > Mitte; i--) {
    if ((Daten[i]->Lese_i (Index) >= Hoehe) &&
	(Daten[i]->Lese_i (Index) >= Untergrenze)) {
      Diese.Setze_i (i, TRUE);
      Hoehe = Daten[i]->Lese_i (Index);
    }
  }

  /* doppelte x-Positionen loeschen */
  for (i = 0; i < Diese.Anzahl_gesetzt () - 1;) {
    if (Daten[Diese.Lese_gesetzte (i)]->Lese_i (0) ==
	Daten[Diese.Lese_gesetzte (i + 1)]->Lese_i (0))
      Diese.Setze_i (Diese.Lese_gesetzte (i + 1), FALSE);
    else
      i++;
  }

  /* 3-fache gleiche y-Werte loeschen */
  Loescher = TRUE;
  for (i = 0; i < Diese.Anzahl_gesetzt (); i = j) {
    Hoehe = Daten[Diese.Lese_gesetzte (i)]->Lese_i (1);
    for (j = i + 1; j < Diese.Anzahl_gesetzt (); j++) {
      if (Hoehe != Daten[Diese.Lese_gesetzte (j)]->Lese_i (1))
	break;
    }
    /* Die dazwischen loeschen */
    for (k = i + 1; k < j - 1; k++)
      Loescher.Setze_i (Diese.Lese_gesetzte (k), FALSE);
  }
  Diese = Diese & Loescher;

  /* jetzt die gefundenen Daten uebernehmen */
  Result.DVektorArray (2, Diese.Anzahl_gesetzt (), NULL);
  for (i = 0; i < Diese.Anzahl_gesetzt (); i++) {
    Result.Setze_Vektor (i, 0, Daten[Diese.Lese_gesetzte (i)]->Lese_i (0));
    Result.Setze_Vektor (i, 1, Daten[Diese.Lese_gesetzte (i)]->Lese_i (Index));
  }
};

IVektor *
DVektorArray::Finde_Huegel (DVektor & Y, int &soviel,
			    double Huegel_Grenze,
			    int Naeher_erlaubt,
			    double Winkel_Cosinus)
{
  IVektor in_Huegel (Groesse, NULL);
  IVektor *Result;
  DVektor Abstand (Groesse, 0, NULL), Gerade1 (VektorDimension, 0, NULL),
    Gerade2 (VektorDimension, 0, NULL);
  int i = 0, j, k, Gruppe, aktuelleGruppe, Buffer, naeher, zugeordnet,
    Der = 0;
  char vorhanden, groessere;
  double Hoehe, DBuffer;

/********************************************************/
/* Zunaechst Zugehoerigkeiten oberhalb Grenze bestimmen */
/********************************************************/
  Gruppe = 1;
  for (i = 0; i < Groesse; i++) {	/* alle Punkte untersuchen */
    if (Y[i] >= Huegel_Grenze) {	/* ueber Grenze, dann alle in anderen Gruppen suchen */
      /* zunaechst Abstaende berechnen */
      for (j = 0; j < Groesse; j++)
	Abstand[j] = Daten[i]->Abstand (*(Daten[j]));

      if (in_Huegel.Lese_i (i) != 0)
	aktuelleGruppe = in_Huegel[i];
      else
	aktuelleGruppe = Gruppe++;
      for (j = 0; j < Groesse; j++) {	/* mit jedem anderen testen */
	if ((in_Huegel.Lese_i (j) != aktuelleGruppe) && (j != i) &&
	    (Y[j] >= Huegel_Grenze)) {	/* mit allen anderen testen, die nicht schon in selber Gruppe */
	  Gerade1 = (*(Daten[i])) - (*(Daten[j]));
	  naeher = 0;
	  for (k = 0; (k < Groesse) && (naeher <= Naeher_erlaubt); k++) {	/* pruefen, ob unter den uebrigen naehere */
	    if ((k != i) && (k != j) && (in_Huegel[k] != aktuelleGruppe) &&
		(Abstand[k] < Abstand[j])) {
	      Gerade2 = (*(Daten[i])) - (*(Daten[k]));
	      if (Gerade1.cos_Winkel (Gerade2) >= Winkel_Cosinus) {	/* naeher und in aehnlicher Richtung */
		if (Y[k] >= Huegel_Grenze) {
		  naeher = MAXINT;
		  break;
		} else {
		  naeher++;
		}
	      }
	    }
	  }
	  if (naeher <= Naeher_erlaubt) {
	    if (in_Huegel[j] == 0)
	      in_Huegel[j] = aktuelleGruppe;
	    else {		/* alle der gefundenen Gruppen umnennen */
	      Buffer = in_Huegel[j];
	      for (k = 0; k < Groesse; k++)
		if (in_Huegel[k] == aktuelleGruppe)
		  in_Huegel[k] = Buffer;
	      aktuelleGruppe = Buffer;
	    }
	  }
	}
      }
      in_Huegel[i] = aktuelleGruppe;
    }				/* if(Y[i] >= Huegel_Grenze) */
  }
/***************************************************/
/*** testen wieviele Huegel und durchnummerieren ***/
  for (i = 0, soviel = 1; i < Groesse; i++) {
    k = 0;
    vorhanden = groessere = FALSE;
    for (j = 0; j < Groesse; j++) {
      if (in_Huegel.Lese_i (j) == soviel)
	vorhanden = TRUE;
      if (in_Huegel.Lese_i (j) > soviel) {
	groessere = TRUE;
	k = in_Huegel.Lese_i (j);
      }
    }
    /* irgendeinen hoeheren umbenennen */
    if (vorhanden)
      soviel += 1;
    else {
      if (groessere) {		/* Groesseren umbenennen */
	for (j = 0; j < Groesse; j++) {
	  if (in_Huegel[j] == k)
	    in_Huegel[j] = soviel;
	}
	soviel += 1;
      } else {
	soviel -= 1;
	break;			/* fertig */
      }
    }
  }
/*********************************************/
/* jetzt noch die unter der Grenzen zuordnen */
/*********************************************/
  /* zu welchem Huegel am naechsten */
  zugeordnet = 0;
  for (i = 0; i < Groesse; i++)
    if (in_Huegel[i] != 0)
      zugeordnet += 1;

  while (zugeordnet != Groesse) {
    Hoehe = 0;
    /* zunaechst den hoechsten davon finden */
    for (i = 0; i < Groesse; i++)
      if ((Y[i] > Hoehe) && (in_Huegel[i] == 0)) {
	Der = i;		/* Der hoechste */
	Hoehe = Y[i];
      }
    /* worein damit ? */
    Hoehe = MAXDOUBLE;
    for (i = 0; i < Groesse; i++) {
      if (in_Huegel[i] != 0) {
#if 0
	DBuffer = Daten[Der].Abstand (*(Daten[i]));
#else
	DBuffer = Daten[Der]->Abstand (*(Daten[i]));
#endif
	if (DBuffer < Hoehe) {
	  Gruppe = in_Huegel[i];
	  Hoehe = DBuffer;
	}
      }
    }
    in_Huegel.Setze_i (Der, Gruppe);
    zugeordnet++;
  }

/**********************************/
/* zum Schluss die Huegel basteln */
/**********************************/
  if (soviel != 0) {
    if ((Result = (IVektor *) malloc (sizeof (IVektor) * soviel)) == NULL)
      Fehlermeldung ("IVektor *DVektorArray::Finde_Huegel(...)", SPEICHERFEHLER);
  } else
    Result = NULL;

  for (i = 0; i < soviel; i++) {
    zugeordnet = 0;
    for (j = 0; j < Groesse; j++) {
      if (in_Huegel[j] == (i + 1))
	zugeordnet += 1;
    }
    Result[i].IVektor (zugeordnet, NULL);
    Der = 0;
    for (k = 0; k < Groesse; k++) {
      if (in_Huegel[k] == (i + 1)) {
	Result[i][Der] = k;
	Der += 1;
      }
    }
  }

  return (Result);
};

/**************************************************************/
/* Separates Suchen in den einzelnen Dimensionen nacheinander */
/**************************************************************/
IVektor *
DVektorArray::Finde_Huegel (DVektor & Y, int &soviele,
			    double Huegel_Grenze, int Naeher_erlaubt)
{
  int i = 0, j, k, Dim, Der = 0, aktueller_Huegel, dazwischen, rechts,
    links, Max_Index = 0;
  double Max;
  BVektor Huegel (Groesse, NULL);
  BVektorArray in_Huegel (Groesse, 1, NULL), Huegel_Ergebnis (Groesse, 1, NULL);
  IVektor *Result;
  DVektorArray Arbeits_Y (VektorDimension + 1, Groesse, NULL);

/* Initialisierungen */
  Huegel_Ergebnis = TRUE;
  for (i = 0; i < Groesse; i++) {
    for (j = 0; j < VektorDimension; j++)
      Arbeits_Y[i][j] = (*(Daten[i]))[j];
    Arbeits_Y[i][VektorDimension] = Y[i];
    Arbeits_Y[i].Setze_Nummer (i);	/* sozusagen der Index */
  }
/* Los geht's */
  for (Dim = 0; Dim < VektorDimension; Dim++) {
    links = 0;
    rechts = 0;
    Arbeits_Y = Arbeits_Y.Quicksort (Dim);
    aktueller_Huegel = 0;
    dazwischen = 0;
    in_Huegel.Setze_Groesse (1);
    in_Huegel[0] = FALSE;

    /* zunaechst vor dem ersten Huegel suchen */
    for (i = 0; i < Groesse; i++) {	/* Der Anfang */
      in_Huegel[0].Setze_i (Arbeits_Y[i].Lese_Nummer (), TRUE);
      if (Arbeits_Y[i][VektorDimension] >= Huegel_Grenze) {
	i++;
	break;
      }
    }

    for (; i < Groesse; i++) {
      if (Arbeits_Y[i][VektorDimension] >= Huegel_Grenze) {
	if (dazwischen < Naeher_erlaubt) {	/* bis hierher Luecke in Huegel fuellen */
	  for (; dazwischen != -1; dazwischen--)
	    in_Huegel[aktueller_Huegel].Setze_i (Arbeits_Y[i - dazwischen].Lese_Nummer (), TRUE);
	  links = i;
	} else {		/* neuen Huegel beginnen */
	  in_Huegel++;
	  in_Huegel[aktueller_Huegel + 1].Setze_i (Arbeits_Y[i].Lese_Nummer (), TRUE);
	  rechts = i;
	  /* jetzt dazwischen entscheiden */
	  dazwischen = rechts - links - 1;
	  while (dazwischen) {
	    Max = -1.0;
	    /* Zunaechst den hoechsten ermitteln */
	    for (j = links + 1; j < rechts; j++) {
	      if ((Arbeits_Y[j][VektorDimension] >= Max) &&
	      (!in_Huegel[aktueller_Huegel][Arbeits_Y[j].Lese_Nummer ()]) &&
		  (!in_Huegel[aktueller_Huegel + 1][Arbeits_Y[j].Lese_Nummer ()])) {	/* ddie groesste Zugehoerigkeit finden */
		Max = Arbeits_Y[j][VektorDimension];
		Max_Index = j;
	      }
	    }
	    /* und zuordnen */

	    if (Arbeits_Y[Max_Index][Dim] == Arbeits_Y[links][Dim]) {
	      for (k = links + 1; k <= Max_Index; k++)
		in_Huegel[aktueller_Huegel].Setze_i (Arbeits_Y[k].Lese_Nummer (), TRUE);
	      dazwischen -= Max_Index - links;
	      links = Max_Index;
	    } else if (Arbeits_Y[Max_Index][Dim] == Arbeits_Y[rechts][Dim]) {
	      /* zum rechten Huegel zuordnen */
	      for (k = Max_Index; k < rechts; k++)
		in_Huegel[aktueller_Huegel + 1].Setze_i (Arbeits_Y[k].Lese_Nummer (), TRUE);
	      dazwischen -= rechts - Max_Index;
	      rechts = Max_Index;
	    } else if (((Arbeits_Y[links][VektorDimension] - Arbeits_Y[Max_Index][VektorDimension]) /
		     (Arbeits_Y[Max_Index][Dim] - Arbeits_Y[links][Dim])) >=
		       ((Arbeits_Y[rechts][VektorDimension] - Arbeits_Y[Max_Index][VektorDimension]) /
		    (Arbeits_Y[rechts][Dim] - Arbeits_Y[Max_Index][Dim]))) {
	      /* bis zum Rand dem linken Huegel zuordnen */
	      for (k = links + 1; k <= Max_Index; k++)
		in_Huegel[aktueller_Huegel].Setze_i (Arbeits_Y[k].Lese_Nummer (), TRUE);
	      dazwischen -= Max_Index - links;
	      links = Max_Index;
	    } else {
	      /* zum rechten Huegel zuordnen */
	      for (k = Max_Index; k < rechts; k++)
		in_Huegel[aktueller_Huegel + 1].Setze_i (Arbeits_Y[k].Lese_Nummer (), TRUE);
	      dazwischen -= rechts - Max_Index;
	      rechts = Max_Index;
	    }
	  }
	  aktueller_Huegel++;
	}
	dazwischen = 0;
      } else
	dazwischen++;
    }
    /* jetzt rechts vom letzten Huegel auffuellen */
    if (dazwischen != 0) {
      for (i = Groesse - dazwischen - 1; i < Groesse; i++) {
	in_Huegel[aktueller_Huegel].Setze_i (Arbeits_Y[i].Lese_Nummer (), TRUE);
      }
    }
    /* jetzt mit den bisherigen Ergebnissen schneiden */
    int Anzahl;

    Anzahl = Huegel_Ergebnis.Lese_Groesse ();
    for (i = 0; i < Anzahl; i++) {
      for (j = 0; j < in_Huegel.Lese_Groesse (); j++) {
	Huegel = Huegel_Ergebnis[i] & in_Huegel[j];
	for (k = 0; k < in_Huegel.Lese_Dim (); k++) {
	  if (Huegel[Arbeits_Y[k].Lese_Nummer ()]) {
	    if (Arbeits_Y[k][VektorDimension] >= Huegel_Grenze) {
	      Huegel_Ergebnis += Huegel;
	      break;
	    }
	  }
	}
      }
    }
    /* jetzt die unterteilten loeschen */
    Huegel_Ergebnis.Loesche (0, Anzahl - 1);
  }

/**********************************/
/* zum Schluss die Huegel basteln */
/**********************************/
  soviele = Huegel_Ergebnis.Lese_Groesse ();
  if (Huegel_Ergebnis.Lese_Groesse () != 0) {
    if ((Result = (IVektor *) malloc (sizeof (IVektor) * Huegel_Ergebnis.Lese_Groesse ())) == NULL)
      Fehlermeldung ("IVektor *DVektorArray::Finde_Huegel(...)", SPEICHERFEHLER);
  } else
    Result = NULL;

  for (i = 0; i < Huegel_Ergebnis.Lese_Groesse (); i++) {
    Result[i].IVektor (Huegel_Ergebnis[i].Anzahl_gesetzt (), NULL);
    Der = 0;
    for (j = 0; j < Groesse; j++) {
      if (Huegel_Ergebnis[i][j]) {
	Result[i][Der] = j;
	Der += 1;
      }
    }
  }
  return (Result);
}


void 
DVektorArray::operator= (const DVektorArray & Das)
{
  int i;

  if ((Groesse != Das.Groesse) || (VektorDimension != Das.VektorDimension)) {
#if 0
    this. ~ DVektorArray ();
#else
    this-> ~ DVektorArray ();
#endif
    Groesse = Das.Groesse;
    VektorDimension = Das.VektorDimension;
    if (Groesse != 0) {
      if ((Daten = (DVektor **) malloc (sizeof (DVektor *) * Groesse)) == NULL)
	Fehlermeldung ("void DVektorArray::operator=(DVektorArray *Das)", SPEICHERFEHLER);
      for (i = 0; i < Groesse; i++)
	Daten[i] = new DVektor (*(Das.Daten[i]));
    } else
      Daten = NULL;
  } else {
    for (i = 0; i < Groesse; i++)
      (*(Daten[i])) = *(Das.Daten[i]);
  }

  Min = Das.Min;
  Max = Das.Max;
  MinNummer = Das.MinNummer;
  MaxNummer = Das.MaxNummer;
  Durchschnitt = Das.Durchschnitt;
};

void 
DVektorArray::operator+= (const DVektor & Damit)
{
  int i;
  DVektor **DatenBuffer;

  if ((VektorDimension != Damit.Lese_Dim ()) &&
      (Groesse != 0))
    Fehlermeldung ("DVektorArray::operator+=", DIMFEHLER);

  VektorDimension = Damit.Lese_Dim ();

  if ((DatenBuffer = (DVektor **) malloc (sizeof (DVektor *) * (Groesse + 1))) == NULL)
    Fehlermeldung ("DVektorArray::operator+=", SPEICHERFEHLER);

  for (i = 0; i < Groesse; i++)
    DatenBuffer[i] = Daten[i];
  free (Daten);
  DatenBuffer[Groesse] = new DVektor (Damit);

  if (Groesse == 0) {
    Max.DVektor (Damit);
    Min.DVektor (Damit);
    Durchschnitt.DVektor (Damit);
    MaxNummer.IVektor (VektorDimension, NULL);
    MinNummer.IVektor (VektorDimension, NULL);
  } else {
    for (i = 0; i < VektorDimension; i++) {
      if (Damit[i] > Max[i]) {	/* neues Maximum */
	Max[i] = Damit[i];
	MaxNummer[i] = Groesse;
      }
      if (Damit[i] < Min[i]) {	/* neues Minimum */
	Min[i] = Damit[i];
	MinNummer[i] = Groesse;
      }
      Durchschnitt[i] = (Durchschnitt[i] * Groesse + Damit[i]) / (Groesse + 1);
    }
  }
  Groesse++;
  Daten = DatenBuffer;
};

void 
DVektorArray::operator-= (DVektor & Damit)
{
  int i, j, Das = 0;
  DVektor **DatenBuffer;

  if (VektorDimension != Damit.Lese_Dim ())
    Fehlermeldung ("DVektorArray::operator-=", DIMFEHLER);

  if ((DatenBuffer = (DVektor **) malloc (sizeof (DVektor *) * (Groesse - 1))) == NULL)
    Fehlermeldung ("DVektorArray::operator+=", SPEICHERFEHLER);

  j = 0;
  for (i = 0; i < Groesse; i++) {
    if (Daten[i] != &Damit) {	/* Der is es nicht */
      DatenBuffer[j++] = Daten[i];
    } else
      Das = i;			/* Merken */
  }

  Groesse--;
  /* Jetzt Durchschnitt etc. neu berechnen */
  if (Groesse == 0) {
    for (i = 0; i < VektorDimension; i++) {
      Max[i] = -MAXDOUBLE;
      MaxNummer[i] = 0;
      Min[i] = MAXDOUBLE;
      MinNummer[i] = 0;
      Durchschnitt[i] = 0;
    }
  } else {			/* Groesse != 0 */
    for (i = 0; i < VektorDimension; i++) {
      if (Damit[i] == Max[i]) {	/* war das alte Maximum -> neues Maximum finden  */
	Max[i] = -MAXDOUBLE;
	for (j = 0; j < Groesse; j++) {
	  if ((*Daten[j])[i] > Max[i]) {
	    Max[i] = Damit[i];
	    MaxNummer[i] = j;
	  }
	}
      }
      if (Damit[i] == Min[i]) {	/* war das alte Minimum -> neues Minimum finden  */
	Min[i] = MAXDOUBLE;
	for (j = 0; j < Groesse; j++) {
	  if ((*Daten[j])[i] < Min[i]) {
	    Min[i] = Damit[i];
	    MinNummer[i] = j;
	  }
	}
      }
      Durchschnitt[i] = (Durchschnitt[i] * (Groesse + 1) - (*Daten[Das])[i]) / Groesse;
    }
  }

  delete Daten[Das];
  free (Daten);
  Daten = DatenBuffer;
};
